/* arbiter.c

Functions implementing Arbiter related common ioctls
for the Linux character device driver
module for the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2003-2005 & 2010-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

arbiter.c $Id: 7faff9cff996ebfc57daa5b1368c1d3381ea044c $
*/

#include <linux/version.h>
#include <linux/module.h>
#include <asm/io.h>

#include "pciaerhw.h"
#include "internal.h"

/*****************************************************************************/

/*
 * SetArbiterConfig
 * Called by Mon_ioctl & Map_ioctl to implement PCIAER_IOC_SET_ARB_CONFIG. The
 * parameter ulConfig should be one of the PCIAER_IOC_ARB_* constants in which
 * case the function returns 0; otherwise the function returns -EINVAL.
 */

int SetArbiterConfig(struct pciaer_board *pBoard, unsigned long ulConfig)
{
	u32 u32ArbCfgBits;
	void __iomem *pCfgRegB1;
	unsigned long ulHwLockFlags;

	switch (ulConfig) {
	/* case PCIAER_IOC_ARB_DISABLED:
		u32ArbCfgBits = CRB1_ARB_DISABLE;
		break;
	*/
	case PCIAER_IOC_ARB_0_16:
		u32ArbCfgBits = CRB1_ARB_1_16;
		break;
	case PCIAER_IOC_ARB_1_15:
		u32ArbCfgBits = CRB1_ARB_2_15;
		break;
	case PCIAER_IOC_ARB_2_14:
		u32ArbCfgBits = CRB1_ARB_4_14;
		break;
	default:
		return -EINVAL;
		break;
	}

	pCfgRegB1 = XilinxRegAddress(pBoard, CFG_REG_B1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	iowrite32((ioread32(pCfgRegB1) & ~CRB1_ARB_CFG_MASK) | u32ArbCfgBits,
		  pCfgRegB1);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return 0;
}

/*****************************************************************************/

/*
 * GetArbiterConfig
 * Called by Mon_ioctl & Map_ioctl to implement PCIAER_IOC_GET_ARB_CONFIG.
 * The return value is one of the PCIAER_IOC_ARB_* constants.
 */

int GetArbiterConfig(struct pciaer_board *pBoard)
{
	void __iomem *pCfgRegB1;
	int nConfig;

	pCfgRegB1 = XilinxRegAddress(pBoard, CFG_REG_B1);

	switch (ioread32(pCfgRegB1) & CRB1_ARB_CFG_MASK) {
	case CRB1_ARB_DISABLE:
		nConfig = PCIAER_IOC_ARB_DISABLED;
		break;
	case CRB1_ARB_1_16:
		nConfig = PCIAER_IOC_ARB_0_16;
		break;
	case CRB1_ARB_2_15:
		nConfig = PCIAER_IOC_ARB_1_15;
		break;
	case CRB1_ARB_4_14:
		nConfig = PCIAER_IOC_ARB_2_14;
		break;
	default:
		nConfig = -1;	/* The default should never be reached if the */
		break;		/* CRB1_ARB constants are properly defined.   */
	}
	return nConfig;
}

/*****************************************************************************/

/*
 * SetSequencerArbiterChannel
 * Called by Seq_ioctl to implement PCIAER_IOC_SET_SEQ_ARB_CHANNEL. The
 * parameter ulChannel can be 0, 1, 2 or 3, in which case the function returns
 * 0; otherwise the function returns -EINVAL.
 */

int SetSequencerArbiterChannel(struct pciaer_board *pBoard, unsigned long ulChannel)
{
	u32 u32SeqChSelBits;
	void __iomem *pCfgRegA1;
	unsigned long ulHwLockFlags;

	switch (ulChannel) {
	case 0:
		u32SeqChSelBits = CRA1_SEQ_CH_SEL_0;
		break;
	case 1:
		u32SeqChSelBits = CRA1_SEQ_CH_SEL_1;
		break;
	case 2:
		u32SeqChSelBits = CRA1_SEQ_CH_SEL_2;
		break;
	case 3:
		u32SeqChSelBits = CRA1_SEQ_CH_SEL_3;
		break;
	default:
		return -EINVAL;
		break;
	}

	pCfgRegA1 = XilinxRegAddress(pBoard, CFG_REG_A1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	iowrite32((ioread32(pCfgRegA1) & ~CRA1_SEQ_CH_SEL_MASK) |
		  u32SeqChSelBits, pCfgRegA1);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return 0;
}

/*****************************************************************************/

/*
 * GetSequencerArbiterChannel
 * Called by Seq_ioctl to implement PCIAER_IOC_GET_SEQ_ARB_CHANNEL. The return
 * value is an integer in the range 0..3 repesenting the channel number.
 */

int GetSequencerArbiterChannel(struct pciaer_board *pBoard)
{
	void __iomem *pCfgRegA1;
	int nChannel;

	pCfgRegA1 = XilinxRegAddress(pBoard, CFG_REG_A1);

	switch (ioread32(pCfgRegA1) & CRA1_SEQ_CH_SEL_MASK) {
	case CRA1_SEQ_CH_SEL_0:
		nChannel = 0;
		break;
	case CRA1_SEQ_CH_SEL_1:
		nChannel = 1;
		break;
	case CRA1_SEQ_CH_SEL_2:
		nChannel = 2;
		break;
	case CRA1_SEQ_CH_SEL_3:
		nChannel = 3;
		break;
	default:
		nChannel = -1;	/* The default should never be reached if the */
		break;		/* CRA1_SEQ_CH constants are properly defined.*/
	}
	return nChannel;
}

/*****************************************************************************/

/* EOF */
