/* counter.c

Functions implementing Counter related common ioctls
for the Linux character device driver
module for the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2003-2005 & 2008-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

counter.c $Id: 11e4a101a42675dda2be420e89e06fc8a0668350 $
*/

#include <linux/version.h>
#include <linux/module.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "pciaerhw.h"
#include "internal.h"

/*****************************************************************************/

/*
 * SetCounterPeriod
 * Called by Seq_ioctl & Mon_ioctl to implement PCIAER_IOC_SET_CNTR_PERIOD.
 * The parameter ulMicroseconds can be 1, 10, 50 or 100, in which
 * case the function returns 0; otherwise the function returns -EINVAL.
 */

int SetCounterPeriod(struct pciaer_board *pBoard, unsigned long ulMicroseconds)
{
	u32 u32AerClockBits;
	void __iomem *pCfgRegB1;
	unsigned long ulHwLockFlags;

	switch (ulMicroseconds) {
	case 1:
		u32AerClockBits = CRB1_AER_CLOCK_1US;
		break;
	case 10:
		u32AerClockBits = CRB1_AER_CLOCK_10US;
		break;
	case 50:
		u32AerClockBits = CRB1_AER_CLOCK_50US;
		break;
	case 100:
		u32AerClockBits = CRB1_AER_CLOCK_100US;
		break;
	default:
		return -EINVAL;
		break;
	}

	pCfgRegB1 = XilinxRegAddress(pBoard, CFG_REG_B1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	iowrite32((ioread32(pCfgRegB1) & ~CRB1_AER_CLOCK_MASK) |
		  u32AerClockBits, pCfgRegB1);
	FlushPciWrites(pBoard);
	atomic_set(&pBoard->CounterPeriod_us, ulMicroseconds);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return 0;
}

/*****************************************************************************/

/*
 * ResetCounter
 * Resets the counter and stores the time at which it did so in the board struct.
 * Called by ResetCounterGetTime below, and also from ResetXilinx in driver.c
 */

int ResetCounter(struct pciaer_board *pBoard)
{
	void __iomem *pCfgRegB1;
	unsigned long ulHwLockFlags;
	u32 u32RegContents;

	pCfgRegB1 = XilinxRegAddress(pBoard, CFG_REG_B1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	u32RegContents = ioread32(pCfgRegB1);
	iowrite32(u32RegContents & ~CRB1_TIME_RESET_, pCfgRegB1);
	FlushPciWrites(pBoard);
	/* No delay needed here. Confirmed in phone call to Vittorio, 3.7.03 */
	iowrite32(u32RegContents | CRB1_TIME_RESET_, pCfgRegB1);
	FlushPciWrites(pBoard);

	do_gettimeofday(&pBoard->tvLastCounterResetTime);

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return 0;
}

/*****************************************************************************/

/*
 * GetLastResetTime
 * Safely reads the reset time stored in the board struct into the timeval
 * pointed to by the ptvReset parameter.
 * Called by ResetCounterGetTime and GetLastCounterResetTime.
 */

static void ReadLastResetTime(struct pciaer_board *pBoard, struct timeval *ptvReset)
{
	unsigned long ulHwLockFlags;

	read_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	ptvReset->tv_sec = pBoard->tvLastCounterResetTime.tv_sec;
	ptvReset->tv_usec = pBoard->tvLastCounterResetTime.tv_usec;

	read_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);
}

/*****************************************************************************/

/*
 * ResetCounterGetTime
 * Called by Seq_ioctl & Mon_ioctl to implement PCIAER_IOC_RST_CNTR_GET_TIME
 */

int ResetCounterGetTime(struct pciaer_board *pBoard, __user struct timeval *ptvUser,
			fmode_t FileMode)
{
	int nError;

	if (!(FileMode & FMODE_WRITE))
		nError = -EACCES;
	else if (ptvUser == NULL)
		nError = ResetCounter(pBoard);
	else if (!(FileMode & FMODE_READ))
		nError = -EACCES;
	else if (!access_ok(VERIFY_WRITE, ptvUser, sizeof(struct timeval)))
		nError = -EFAULT;
	else {
		nError = ResetCounter(pBoard);
		if (nError == 0) {
			struct timeval tvReset;

			ReadLastResetTime(pBoard, &tvReset);
			if (__copy_to_user
			    (ptvUser, &tvReset, sizeof(struct timeval)) != 0)
				nError = -EFAULT;
		}
	}

	return nError;
}

/*****************************************************************************/

/*
 * GetLastCounterResetTime
 * Called by Seq_ioctl & Mon_ioctl to implement PCIAER_IOC_GET_CNTR_LAST_RST_TIME
 */

int GetLastCounterResetTime(struct pciaer_board *pBoard,
			    __user struct timeval *ptvUser)
{
	struct timeval tvReset;

	ReadLastResetTime(pBoard, &tvReset);
	return (copy_to_user(ptvUser, &tvReset, sizeof(struct timeval)) ==
		0) ? 0 : -EFAULT;
}

/*****************************************************************************/

/* EOF */
