/* debugfs.c

Support for reading regs via debugfs. This is at least a partial replacement for
the code that was in debugioctl.c for use by the testcode.

Copyright (C) 2009-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institute of Neuroinformatics (www.ini.uzh.ch)
University of Zurich and ETH Zurich,
CH-8057 Zurich, Switzerland

debugfs.c $Id: 4c4f09bb3728d3317ddbd1591b15f9106329a892 $
*/

#include <linux/version.h>
#include <linux/module.h>
#include <linux/debugfs.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "s5920.h"
#include "pciaerhw.h"
#include "internal.h"

/*****************************************************************************/

static struct dentry *DebugFSRoot;

/*****************************************************************************/

static ssize_t dfs_read_reg(struct file *fp, __user char *pchUserBuffer,
			    size_t BytesRequested, loff_t *pOffset)
{
	void __iomem *pReg;
	u32 value;
	char buf[10];		/* 8 hex digits + \n + \0 */

	if (*pOffset != 0L)
		return 0;

	if (unlikely(BytesRequested < sizeof(buf)))
		return -EINVAL;

	pReg = (void __iomem *)fp->f_dentry->d_inode->i_private;
	value = ioread32(pReg);
	snprintf(buf, sizeof(buf), "%08X\n", value);
	if (unlikely(__copy_to_user(pchUserBuffer, buf, sizeof(buf)) != 0L))
		return -EFAULT;

	*pOffset += sizeof(buf);
	return sizeof(buf);
}

/*****************************************************************************/

static ssize_t dfs_write_reg(struct file *fp, const __user char *pchUserBuffer,
			     size_t BytesToWrite, loff_t *pOffset)
{
	int err;
	char buf[10];		/* 8 hex digits + optional \n + \0 */
	u32 value;
	void __iomem *pReg;

	err = TestSysAdminCapability();
	if (err != 0)
		return err;

	if (unlikely(BytesToWrite >= sizeof(buf) || *pOffset != 0L))
		return -EINVAL;

	if (unlikely(__copy_from_user(buf, pchUserBuffer, BytesToWrite) != 0))
		return -EFAULT;

	if (buf[BytesToWrite - 1] == '\n')
		buf[BytesToWrite - 1] = '\0';
	else {
		if (BytesToWrite > 8)
			return -EINVAL;
		buf[BytesToWrite] = '\0';
	}

	if (kstrtou32(buf, 16, &value) != 0)
		return -EINVAL;

	pReg = (void __iomem *)fp->f_dentry->d_inode->i_private;

	iowrite32(value, pReg);
	mb();
	ioread32(pReg);

	return BytesToWrite;
}

/*****************************************************************************/

static ssize_t dfs_xilinx_reset(struct file *fp, const __user char *pchUserBuffer,
				size_t BytesToWrite, loff_t *pOffset)
{
	int err;
	struct pciaer_board *b;

	err = TestSysAdminCapability();
	if (err != 0)
		return err;

	b = fp->f_dentry->d_inode->i_private;

	err = ResetXilinx(b);
	if (err != 0)
		return err;

	/* Re-enable default interrupts */

	EnableInt(b,
		  CRC1_DIS_INT_FH_MON | CRC1_DIS_INT_FF_MON |
		  CRC1_DIS_INT_FF_MAP);

	return BytesToWrite;
}

/*****************************************************************************/

static ssize_t dfs_read_stats(__user char *pchUserBuffer, size_t BytesRequested,
			      loff_t *pOffset, struct pciaer_board *b,
			      int (*statsfn)(struct pciaer_board *,
					     struct internal_stats *, bool reset))
{
	struct internal_stats s;
	int error;
	long long time_ns;
	struct timeval timestamp;
	u32 remainder;
	char *buf;
	int nchars;
	unsigned long bytes_not_copied;

	if (*pOffset != 0L)
		return 0;

	if (unlikely(BytesRequested < sizeof(buf)))
		return -EINVAL;

	memset(&s, 0, sizeof(s));
	error = statsfn(b, &s, false);
	time_ns = sched_clock();

	if (unlikely(error != 0))
		return error;

	buf = devm_kzalloc(&b->pPciDev->dev, BytesRequested, GFP_KERNEL);
	if (unlikely(!buf))
		return -ENOMEM;

	timestamp.tv_sec = div_u64_rem(time_ns, 1000000000, &remainder);
	timestamp.tv_usec = (remainder + 500) / 1000;

	nchars = snprintf(buf, BytesRequested, "%ld words_transferred\n"
					       "%ld user_transfers\n"
					       "%ld total_interrupts\n"
					       "%ld overflows_underflows\n"
					       "%ld timeouts\n"
					       "%ld memory_usage\n"
					       "%ld.%06ld timestamp\n",
					       s.words_transferred,
					       s.user_transfers,
					       s.total_interrupts,
					       s.overflows_underflows,
					       s.timeouts,
					       s.memory_usage,
					       (long)timestamp.tv_sec,
						(long)timestamp.tv_usec
			 );

	if (unlikely(nchars > BytesRequested))
		nchars = BytesRequested;

	bytes_not_copied = __copy_to_user(pchUserBuffer, buf, nchars);

	devm_kfree(&b->pPciDev->dev, buf);

	if (unlikely(bytes_not_copied != 0L))
		return -EFAULT;

	*pOffset += nchars;
	return nchars;
}

/*****************************************************************************/

static ssize_t dfs_reset_stats(size_t BytesToWrite, struct pciaer_board *b,
			       int (*statsfn)(struct pciaer_board *,
					      struct internal_stats *, bool reset))
{
	int error;

	error = statsfn(b, NULL, true);
	if (unlikely(error != 0))
		return error;
	return BytesToWrite;
}

/*****************************************************************************/

static ssize_t dfs_read_mon_stats(struct file *fp, __user char *pchUserBuffer,
				  size_t BytesRequested, loff_t *pOffset)
{
	return dfs_read_stats(pchUserBuffer, BytesRequested, pOffset,
			      fp->f_dentry->d_inode->i_private,
			      GetMonitorStats);
}

/*****************************************************************************/

static ssize_t dfs_reset_mon_stats(struct file *fp,
				   const __user char *pchUserBuffer,
				   size_t BytesToWrite, loff_t *pOffset)
{
	return dfs_reset_stats(BytesToWrite, fp->f_dentry->d_inode->i_private,
			       GetMonitorStats);
}

/*****************************************************************************/

static ssize_t dfs_read_seq_stats(struct file *fp, __user char *pchUserBuffer,
				  size_t BytesRequested, loff_t *pOffset)
{
	return dfs_read_stats(pchUserBuffer, BytesRequested, pOffset,
			      fp->f_dentry->d_inode->i_private,
			      GetSequencerStats);
}

/*****************************************************************************/

static ssize_t dfs_reset_seq_stats(struct file *fp,
				   const __user char *pchUserBuffer,
				   size_t BytesToWrite, loff_t *pOffset)
{
	return dfs_reset_stats(BytesToWrite, fp->f_dentry->d_inode->i_private,
			       GetSequencerStats);
}

/*****************************************************************************/

static ssize_t dfs_read_map_stats(struct file *fp, __user char *pchUserBuffer,
				  size_t BytesRequested, loff_t *pOffset)
{
	return dfs_read_stats(pchUserBuffer, BytesRequested, pOffset,
			      fp->f_dentry->d_inode->i_private,
			      GetMapperStats);
}

/*****************************************************************************/

static ssize_t dfs_reset_map_stats(struct file *fp,
				   const __user char *pchUserBuffer,
				   size_t BytesToWrite, loff_t *pOffset)
{
	return dfs_reset_stats(BytesToWrite, fp->f_dentry->d_inode->i_private,
			       GetMapperStats);
}

/*****************************************************************************/

static const struct file_operations roregfops = {.read = dfs_read_reg };
static const struct file_operations rwregfops = {.read = dfs_read_reg,
						 .write = dfs_write_reg };
static const struct file_operations xilinxrstfops = {.write = dfs_xilinx_reset};
static const struct file_operations monstatsfops = {.read = dfs_read_mon_stats,
						    .write = dfs_reset_mon_stats };
static const struct file_operations seqstatsfops = {.read = dfs_read_seq_stats,
						    .write = dfs_reset_seq_stats };
static const struct file_operations mapstatsfops = {.read = dfs_read_map_stats,
						    .write = dfs_reset_map_stats };

/*****************************************************************************/

/*
 * InitDebugFs
 */

void InitDebugFS(void)
{
	DebugFSRoot = debugfs_create_dir(DriverName, NULL);

	if (DebugFSRoot == NULL)
		pr_warning("%s: Something went wrong creating debugfs directory\n",
		DriverName);

	if (PTR_ERR(DebugFSRoot) == -ENODEV) {
		pr_info("%s: No debugfs\n", DriverName);
		DebugFSRoot = NULL;
	}
}

/*****************************************************************************/

static void CreateRegsFile(struct dentry *parent, char *name, mode_t mode,
		       void __iomem *reg)
{
	debugfs_create_file(name, mode, parent, (void *)reg,
			    (mode & S_IWUSR) ? &rwregfops : &roregfops);
}

/*****************************************************************************/

/*
 * AddBoardToDebugFS
 */

void AddBoardToDebugFS(struct pciaer_board *pBoard)
{
	struct dentry *d;
	struct dentry *ds;
	struct dentry *dx;

	if (DebugFSRoot != NULL) {
		d = debugfs_create_dir(pBoard->DeviceName, DebugFSRoot);
		pBoard->dDebugFSEntry = d;

		ds = debugfs_create_dir("s5920", d);
		CreateRegsFile(ds, "omb", S_IRUSR | S_IWUSR,
			   S5920OpRegAddress(pBoard, S5920_OMB));
		CreateRegsFile(ds, "imb", S_IRUSR,
			   S5920OpRegAddress(pBoard, S5920_IMB));
		CreateRegsFile(ds, "mbef", S_IRUSR,
			   S5920OpRegAddress(pBoard, S5920_MBEF));
		CreateRegsFile(ds, "intcsr", S_IRUSR | S_IWUSR,
			   S5920OpRegAddress(pBoard, S5920_INTCSR));
		CreateRegsFile(ds, "rcr", S_IRUSR | S_IWUSR,
			   S5920OpRegAddress(pBoard, S5920_RCR));
		CreateRegsFile(ds, "ptcr", S_IRUSR | S_IWUSR,
			   S5920OpRegAddress(pBoard, S5920_PTCR));

		dx = debugfs_create_dir("xilinx", d);
		CreateRegsFile(dx, "sra1", S_IRUGO | S_IWUSR,
			   XilinxRegAddress(pBoard, STAT_REG_A1));
		CreateRegsFile(dx, "cra1", S_IRUGO | S_IWUSR,
			   XilinxRegAddress(pBoard, CFG_REG_A1));
		CreateRegsFile(dx, "crb1", S_IRUGO | S_IWUSR,
			   XilinxRegAddress(pBoard, CFG_REG_B1));
		CreateRegsFile(dx, "crc1", S_IRUGO | S_IWUSR,
			   XilinxRegAddress(pBoard, CFG_REG_C1));
		CreateRegsFile(dx, "rr1", S_IRUGO,
			   XilinxRegAddress(pBoard, REL_REG_1));
		CreateRegsFile(dx, "sra2", S_IRUGO,
			   XilinxRegAddress(pBoard, STAT_REG_A2));
		CreateRegsFile(dx, "cra2", S_IRUGO | S_IWUSR,
			   XilinxRegAddress(pBoard, CFG_REG_A2));
		CreateRegsFile(dx, "rr2", S_IRUGO,
			   XilinxRegAddress(pBoard, REL_REG_2));
			   
		debugfs_create_file("reset", S_IWUSR, dx, pBoard,
				    &xilinxrstfops);

		debugfs_create_file("monstats", S_IRUGO | S_IWUGO, d, pBoard,
				    &monstatsfops);
		debugfs_create_file("seqstats", S_IRUGO | S_IWUGO, d, pBoard,
				    &seqstatsfops);
		debugfs_create_file("mapstats", S_IRUGO | S_IWUGO, d, pBoard,
				    &mapstatsfops);
	}
}

/*****************************************************************************/

/*
 * CleanUpDebugFS
 */

void CleanUpDebugFS(void)
{
	if (DebugFSRoot != NULL) {
		debugfs_remove_recursive(DebugFSRoot);
		DebugFSRoot = NULL;
	}
}

/*****************************************************************************/

/* EOF */
