/*
driver.c

This file contains code for the Linux character device driver
module for the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2000 & 2003-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

driver.c $Id: 2249031a1e8ee53995136be533d0c1eb47a1a204 $
*/

#include <linux/version.h>
#include <linux/module.h>

#include <linux/delay.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "s5920.h"
#include "pciaerhw.h"
#include "internal.h"
#include "version.h"

/*****************************************************************************/

/*
 * By default, the driver will use dynamic allocation to obtain a major device
 * number from the kernel by calling register_chrdev with a major number of
 * zero. If this is not desired, a specific major number can be given on the
 * insmod command line.
 */

INTEGER_MODULE_PARAMETER(major, 0,
			 "Major device number to use. Defaults to 0 for dynamic allocation.");

/*****************************************************************************/

/* Global data */

/*
 * DriverName is the string used to identify our module in printk error
 * messages, and also in registering our driver with the kernel.
 */

const char DriverName[] = "pciaer";

/*
 * MinorDeviceFileOps
 * Table of pointers to the four minor device dispatch tables, indexed by minor
 * device number.
 */

extern const struct file_operations MonFileOps;	/* in monitor.c */
extern const struct file_operations SeqFileOps;	/* in sequencer.c */
extern const struct file_operations MapFileOps;	/* in mapper.c */
extern const struct file_operations NvrFileOps;	/* in s5920nvram.c */
static const struct file_operations *MinorDeviceFileOps[] = {
	&NvrFileOps,
	&MapFileOps,
	&MonFileOps,
	&SeqFileOps
};

/*****************************************************************************/

/*
 * PciAer_open
 * Check that the board indicated by the NUM field in the minor number is
 * present. If the minor number is OK, put a pointer to the relevant struct
 * pciaer_board into the file's private_data member, change fp->f_op depending
 * on the minor number and call the minor device's open method.
 */

static int PciAerMain_open(struct inode *INode, struct file *fp)
{
	int iBoard;

	iBoard = iminor(INode) >> PCIAER_MINOR_NUM_SHIFT;

	fp->private_data = GetBoard(iBoard);

	if (fp->private_data == NULL)
		return -ENODEV;

	fp->f_op = MinorDeviceFileOps[iminor(INode) & PCIAER_MINOR_TYPE_MASK];

	return (fp->f_op->open != NULL) ? fp->f_op->open(INode, fp) : -ENODEV;
}

/*****************************************************************************/

/*
 * MainFileOps
 * The dispatch table for the driver. We support only open, since open will
 * always change the f_op pointer to one of the minor devices' dispatch tables.
 * PciAer_release must be called from minor device release methods.
 */

static const struct file_operations MainFileOps = {
	.owner = THIS_MODULE,
	.open = PciAerMain_open
};

/*****************************************************************************/

/*
 * CheckMajor
 * Zero means allocate dynamically. Apart from that, according to
 * linux/Documentation/devices.tex, acceptable major device numbers are 60-63,
 * 120-127 and 240-254, but I read somewhere else that >128 is not supported,
 * and this is born out by experience. Nonetheless, I allow the range documented
 * in devices.tex here; the kernel can complain at register_chrdev time if it
 * objects to 8-bit numbers.
 * Returns 0 for acceptable major device number and -EINVAL otherwise.
 */

static int CheckMajor(unsigned int m)
{
	switch (m) {
	case 0:
	case 60:
	case 61:
	case 62:
	case 63:
	case 120:
	case 121:
	case 122:
	case 123:
	case 124:
	case 125:
	case 126:
	case 127:
	case 240:
	case 241:
	case 242:
	case 243:
	case 244:
	case 245:
	case 246:
	case 247:
	case 248:
	case 249:
	case 250:
	case 251:
	case 252:
	case 253:
	case 254:
		return 0;
		break;
	default:
		return -EINVAL;
		break;
	}
}

/*****************************************************************************/

/*
 * RegisterDriver
 */

int RegisterDriver(void)
{
	int nResult;

	pr_info("%s: driver version %d.%02d initializing\n",
	       DriverName, VER_MAJOR, VER_MINOR);

	/*
	 * Check major device number, because it may have been changed using
	 * insmod
	 */

	if (CheckMajor(major) != 0) {
		pr_err("%s: invalid major device number %u\n", DriverName, major);
		return -EINVAL;
	}

	nResult = register_chrdev(major, DriverName, &MainFileOps);
	if (nResult < 0) {
		pr_err("%s: registration with major number %u failed with error %d\n",
		       DriverName, major, -nResult);
		return nResult;
	}

	/*
	 * If major was zero, we were doing dynamic major number allocation, so
	 * update major number from result.
	 */

	if (major == 0)
		major = nResult;

	pr_info("%s: registered with major number %u\n", DriverName, major);

	return 0;
}

/*****************************************************************************/

/*
 * ResetXilinx								@xilinx
 * Called by InitHardware during initialization, and possibly as a result of
 * writing to a debugfs xilinx/reset file.
 */

int ResetXilinx(struct pciaer_board *pBoard)
{
	unsigned long ulHwLockFlags;
	void __iomem *pXilinxBase;
	u16 rr1;
	u16 rr2;

	/*
	 * Reset add-on (SYSRST#) i.e. the Xilinx chips & ensure state
	 * information in board struct is correct.
	 */

	/* ABORT_INIT_POINT(25); */
	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	WriteS5920OpReg(pBoard, S5920_RCR, S5920_RCR_ADDON_RESET);
	pBoard->fSeqEnabled = false;
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	/*
	 * The rest of the function brings the Xilinx chips out of reset again
	 * and sets the Xilinx registers to sensible default values. This cannot
	 * and should not be attempted if the registers are not accessible
	 * because the stored base address is zero. If this is the case, we just
	 * return at this point.
	 */

	pXilinxBase = XilinxRegAddress(pBoard, 0);
	if (pXilinxBase == NULL)
		return 0;

	msleep(XILINX_RESET_DELAY_MS);

	/* ABORT_INIT_POINT(26); */

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	WriteS5920OpReg(pBoard, S5920_RCR, 0L);
	udelay(S5920_RESET_DELAY_US);

	/*
	 * Test that Xilinx chips are sucessfully reset by testing the FPGA
	 * version regs.
	 */

	iowrite32(~0, pXilinxBase + REL_REG_1);
	mb();
	rr1 = ioread32(pXilinxBase + REL_REG_1);
	iowrite32(~0, pXilinxBase + REL_REG_2);
	mb();
	rr2 = ioread32(pXilinxBase + REL_REG_2);

	if (rr1 == (u16)~0 || rr1 == 0 || rr2 == (u16)~0 || rr2 == 0) {
		write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);
		return -ENODEV;
	}

	pBoard->fpga1_release = rr1;
	pBoard->fpga2_release = rr2;

	/* Initialize Xilinx Configuration Register A1. */

	/* ABORT_INIT_POINT(27); */
	iowrite32(0			 /* Run with monitor enabled, */
		  | CRA1_MON_IGN_CH_MASK /* but all channels ignored, */
		  | 0			 /* and time labels enabled */
		  | CRA1_MON_FIFO_MR_	 /* Monitor FIFO not in reset */
		  | CRA1_MON_FIFO_RT_	 /* Monitor FIFO not in re-transmit mode */
		  | CRA1_SEQ_DISABLE	 /* Sequencer disabled */
		  | CRA1_SEQ_CH_SEL_3	 /* Sequencer connected to arbiter channel 3 */
		  | 0			 /* Sequencer FIFO in reset */
		  | CRA1_SEQ_FIFO_RT_	 /* Sequencer FIFO not in re-transmit mode */
		  | 0			 /* Mapper input enabled, otherwise no ACK's are generated */
		  | CRA1_MAP_IN_FIFO_MR_ /* Mapper FIFO not in reset */
		  , pXilinxBase + CFG_REG_A1);
	FlushPciWrites(pBoard);

	/* Initialize Xilinx Configuration Register B1. */

	/* ABORT_INIT_POINT(28); */
	iowrite32(0			/* Time counter enabled */
		  | 0			/* Time counter in reset */
		  | CRB1_AER_CLOCK_1US	/* Clock period = 1 microsecond */
		  | CRB1_ARB_4_14	/* Four sender chips, 14 address bits */
		  | 0			/* Mapper ignores none of the 4 channels */
		  , pXilinxBase + CFG_REG_B1);
	FlushPciWrites(pBoard);

	/* Initialize Xilinx Configuration Register C1. */

	/* ABORT_INIT_POINT(29); */
	iowrite32(CRC1_DIS_INT_FH_MON | CRC1_DIS_INT_FF_MON	/* Disable monitor interrupts */
		  | CRC1_DIS_INT_FE_SEQ | CRC1_DIS_INT_FH_SEQ	/* Disable sequencer interrupts */
		  | CRC1_DIS_INT_FF_MAP	/* Disable mapper interrupts */
		  | CRC1_EN_ANTI_META	/* Enable anti-metastability flip-flops */
		  | CRC1_DIS_SM_DEBUG	/* Disable LED flashing FPGA 1 debug feature */
		  , pXilinxBase + CFG_REG_C1);
	FlushPciWrites(pBoard);

	/* Initialize Xilinx Configuration Register A2. */

	/* ABORT_INIT_POINT(30); */
	iowrite32(CRA2_MAP_OUT_CFG_PASSTHRU	/* Mapper operated in pass-through mode */
		  | CRA2_AER_OUT_CFG_1_16	/* One receiver chip, 16 address bits */
		  | 0				/* Mapper output enabled */
		  | CRA2_AER_PROTOCOL_P2P	/* Point-to-point protocol */
		  , pXilinxBase + CFG_REG_A2);
	FlushPciWrites(pBoard);

	/* Clear interrupt flags */

	/* ABORT_INIT_POINT(31); */
	iowrite32(SRA1_INT_FLAGS_MASK, pXilinxBase + STAT_REG_A1);
	FlushPciWrites(pBoard);

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	SetCounterPeriod(pBoard, GetCounterPeriod(pBoard));
	ResetCounter(pBoard);

	return 0;
}

/*****************************************************************************/

/*
 * InitHardware								@s5920
 * Called by PciProbeCallback to initialize each board.
 */

int InitHardware(struct pciaer_board *pBoard)
{
	u8 u8PassThruCfgRegion[MAX_NUM_BADR];

	/*
	 * u8PassThruCfgRegion[ 0 ] is not used because there is no control over
	 * Region 0 via the PTCR. (The PTCR itself appears in Region 0.)
	 */

	int iRegion;
	unsigned long ulPtcr;
	unsigned long ulHwLockFlags;
	int nError;

	if (pBoard->pPciDev == NULL)
		return -ENODEV;

	/* Prepare S5920 Pass-Thru configuration */

	u8PassThruCfgRegion[XILINXREGS_BADR] =
	    S5920_PTCR_PTADR_DRIVEN | XILINXREGS_WS;
	u8PassThruCfgRegion[MONSEQFIFO_BADR] = MONSEQFIFO_WS;
	u8PassThruCfgRegion[MAPPERSRAM_BADR] =
	    S5920_PTCR_PTADR_DRIVEN | MAPPERSRAM_WS;
	u8PassThruCfgRegion[UNUSED_BADR] = S5920_PTCR_PTADR_DRIVEN;	/* power-up default */
	ulPtcr = 0L;
	for (iRegion = 1; iRegion < MAX_NUM_BADR; iRegion++)
		ulPtcr |=
		    ((unsigned long)u8PassThruCfgRegion[iRegion]) <<
		    ((iRegion - 1) * 8);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	/*
	 * Disable add-on interrupts (ADDINT#) and clear any pending mailbox
	 * interrupts.
	 */

	/* ABORT_INIT_POINT(21); */
	WriteS5920OpReg(pBoard, S5920_INTCSR,
			S5920_INTCSR_IMB_INT | S5920_INTCSR_OMB_INT);
	/* ABORT_INIT_POINT(22); */

	/* Setup Pass-Thru configuration on S5920 */

	/* ABORT_INIT_POINT(23); */
	WriteS5920OpReg(pBoard, S5920_PTCR, ulPtcr);
	/* ABORT_INIT_POINT(24); */

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	/* Reset and initialize Xilinx FPGAs */

	nError = ResetXilinx(pBoard);
	if (nError != 0)
		return nError;

	return 0;
}

/*****************************************************************************/

/*
 * DisableHardware						@s5920/@xilinx
 * This function is called from ReleaseBoard which itself is normally called
 * from PciRemoveCallback, but may also be called from PciProbeCallback if an
 * error occurs during initialisation. In the latter case, the pciaer_board
 * structure, and in particular the array of pointers to the base addresses may
 * not have been filled in completely, so we need to be particularly careful
 * about possible NULL pointers.
 */

void DisableHardware(struct pciaer_board *pBoard)
{
	unsigned long ulHwLockFlags;
	void __iomem *pXilinxBase;

	/*
	 * If the board struct doesn't refer to a board which exists, do nothing
	 * and return.
	 */

	if (pBoard->pPciDev == NULL)
		return;

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	/*
	 * Disable add-on interrupts (ADDINT#) and clear any pending mailbox
	 * interrupts.
	 */

	WriteS5920OpReg(pBoard, S5920_INTCSR,
			S5920_INTCSR_IMB_INT | S5920_INTCSR_OMB_INT);

	/* The rest of the function deals with the Xilinx registers which are
	 * only accessible if the relevant stored base address is valid, so if
	 * it isn't, we only unlock the hardware lock and return.
	 */

	pXilinxBase = XilinxRegAddress(pBoard, 0);
	if (pXilinxBase != NULL) {
		/* Switch off everything else */

		iowrite32(CRA1_MON_DISABLE	 /* Monitor disabled */
			  | CRA1_MON_IGN_CH_MASK /* Monitor ignores all channels */
			  | 0			 /* Time labels enabled */
			  | 0			 /* Monitor FIFO in reset */
			  | CRA1_MON_FIFO_RT_	 /* Monitor FIFO not in re-transmit mode */
			  | CRA1_SEQ_DISABLE	 /* Sequencer disabled */
			  | CRA1_SEQ_CH_SEL_3	 /* Sequencer connected to arbiter channel 3 */
			  | 0			 /* Sequencer FIFO in reset */
			  | CRA1_SEQ_FIFO_RT_	 /* Sequencer FIFO not in re-transmit mode */
			  | CRA1_MAP_IN_DISABLE	 /* Mapper input disabled */
			  | 0			 /* Mapper FIFO in reset */
			  , pXilinxBase + CFG_REG_A1);
		FlushPciWrites(pBoard);

		iowrite32(CRB1_TIME_DISABLE	 /* Time counter disabled */
			  | 0			 /* Time counter in reset */
			  | CRB1_AER_CLOCK_1US	 /* Clock period = 1 microsecond */
			  | CRB1_ARB_4_14	 /* Four sender chips, 14 address bits */
			  | CRB1_MAP_IGN_CH_MASK /* Mapper ignores all channels */
			  , pXilinxBase + CFG_REG_B1);
		FlushPciWrites(pBoard);

		iowrite32(CRC1_DIS_INT_FH_MON | CRC1_DIS_INT_FF_MON	/* Disable monitor interrupts */
			  | CRC1_DIS_INT_FE_SEQ | CRC1_DIS_INT_FH_SEQ	/* Disable sequencer interrupts */
			  | CRC1_DIS_INT_FF_MAP	/* Disable mapper interrupts */
			  | CRC1_EN_ANTI_META	/* Keep anti-metastability flip-flops enabled */
			  | CRC1_DIS_SM_DEBUG	/* Keep LED flashing FPGA 1 debug feature disabled */
			  , pXilinxBase + CFG_REG_C1);
		FlushPciWrites(pBoard);

		iowrite32(CRA2_MAP_OUT_CFG_PASSTHRU	/* Mapper operated in pass-through mode */
			  | CRA2_AER_OUT_CFG_1_16	/* One receiver chip, 16 address bits */
			  | CRA2_MAP_OUT_DISABLE	/* Mapper output disabled */
			  | CRA2_AER_PROTOCOL_P2P	/* Point-to-point protocol */
			  , pXilinxBase + CFG_REG_A2);
		FlushPciWrites(pBoard);
	}

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);
}

/*****************************************************************************/

int IncSingleUserLock(struct pciaer_board *pBoard)
{
	spin_lock(&pBoard->OpenFlagsLock);

	if (pBoard->nOpens != 0 && current_euid() != pBoard->uidCurrentUser) {
		spin_unlock(&pBoard->OpenFlagsLock);
		return -EBUSY;
	}

	if (pBoard->nOpens++ == 0)
		pBoard->uidCurrentUser = current_euid();

	spin_unlock(&pBoard->OpenFlagsLock);

	return 0;
}

/*****************************************************************************/

void DecSingleUserLock(struct pciaer_board *pBoard)
{
	spin_lock(&pBoard->OpenFlagsLock);
	if (--pBoard->nOpens == 0)
		pBoard->uidCurrentUser = -1;
	spin_unlock(&pBoard->OpenFlagsLock);
}

/*****************************************************************************/

/*
 * PciAer_release
 * Clears the file's private_data field.
 */

int PciAer_release(struct inode *INode, struct file *fp)
{
	fp->private_data = NULL;
	return 0;
}

/*****************************************************************************/

/*
 * CheckIoctlPerm
 * This function is called from the minor device ioctl functions to check for
 * the correct permissions. The first parameter is a pointer to a file struct,
 * and the second parameter should be an ioctl number in which there are
 * 'direction' bits extractable using the _IOC_DIR macro. The assumption is that
 * the four possible 'directions' require modes as follows:
 * _IOC_NONE			FMODE_WRITE
 * _IOC_READ			FMODE_READ
 * _IOC_WRITE			FMODE_WRITE
 * _IOC_READ | _IOC_WRITE	FMODE_READ
 * Exceptions to this pattern must be dealt with by the caller.
 * The function returns 0 if the current ioctl is OK on the current file, and
 * -EACCES otherwise.
 */

int CheckIoctlPerm(struct file *fp, unsigned int uCmd)
{
	fmode_t ModeRequired =
	    (_IOC_DIR(uCmd) & _IOC_READ) ? FMODE_READ : FMODE_WRITE;
	return (ModeRequired & fp->f_mode) ? 0 : -EACCES;
}

/*****************************************************************************/

/*
 * GetVersionInfo
 * Called by the individual minor device *_ioctl functions to implement
 * PCIAER_IOC_GET_VERSION_INFO. The second parameter should be a pointer to a
 * pciaer_version_info struct in user space to be filled in.
 * The function returns 0 for success and -errno otherwise.
 */

int GetVersionInfo(struct pciaer_board *pBoard,
		   __user struct pciaer_version_info *pUserVersionInfo)
{
	struct pciaer_version_info VersionInfo;

	VersionInfo.driver_version = (VER_MAJOR << 8) | VER_MINOR;
	VersionInfo.fpga1_release = pBoard->fpga1_release;
	VersionInfo.fpga2_release = pBoard->fpga2_release;
	VersionInfo.s5920_revision_id = pBoard->pPciDev->revision;

	return (copy_to_user
		(pUserVersionInfo, &VersionInfo,
		 sizeof(struct pciaer_version_info)) == 0) ? 0 : -EFAULT;
}

/*****************************************************************************/

/*
 * ResetFifo							   @xilinx ?
 * Called by Mon_ioctl, Seq_ioctl and Map_ioctl to implement PCIAER_IOC_RST_FIFO.
 */

void ResetFifo(struct pciaer_board *pBoard, u32 u32CRA1FifoResetBit)
{
	void __iomem *pCfgRegA1;
	unsigned long ulHwLockFlags;
	u32 u32RegContents;

	pCfgRegA1 = XilinxRegAddress(pBoard, CFG_REG_A1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	u32RegContents = ioread32(pCfgRegA1);
	iowrite32(u32RegContents & ~u32CRA1FifoResetBit, pCfgRegA1);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	msleep(FIFO_RESET_DELAY_MS);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	u32RegContents = ioread32(pCfgRegA1);
	iowrite32(u32RegContents | u32CRA1FifoResetBit, pCfgRegA1);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);
}

/*****************************************************************************/

void UnregisterDriver(void)
{
	unregister_chrdev(major, DriverName);

	pr_info("%s: unregistered, bye\n", DriverName);
}

/*****************************************************************************/

/* EOF */
