/* internal.h */
#ifndef PCIAER_INTERNAL_H
#define PCIAER_INTERNAL_H

/*
Header file containing definitions and declarations for use only within the
source of the pciaer.ko Linux character device driver
module for the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2000 & 2003-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

internal.h $Id: facd07f897d6e23e8ea47f078fdce83ba5acd657 $
*/

#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <asm/atomic.h>

#include "pciaer.h"

#ifndef __iomem
#define __iomem
#endif

#define INTEGER_MODULE_PARAMETER(X, I, D) static unsigned int X = (I); module_param(X, uint, S_IRUGO); MODULE_PARM_DESC(X, D)
#define BOOLEAN_MODULE_PARAMETER(X, D) static bool X = false; module_param(X, bool, S_IRUGO); MODULE_PARM_DESC(X, D)
#define STRING_MODULE_PARAMETER(X, D) static char *X; module_param(X, charp, S_IRUGO); MODULE_PARM_DESC(X, D)

#define HIBYTE(W)	(((W)>>8) & 0xFF)
#define LOBYTE(W)	((W) & 0xFF)

#define MAX_DEVNAME_LEN	8

/*****************************************************************************/

/* Type used to hold details of a board */

struct pciaer_board {
	struct pci_dev *pPciDev;		/* Pointer to the kernel's
						   struct pci_dev for the board, or NULL
						   if a board is absent. */

	unsigned int iBoard;			/* Board index */

	char DeviceName[MAX_DEVNAME_LEN];	/* Nul terminated name of this device,
						   e.g. pciaer0\0 */

	void __iomem *pBaseAddr[NUM_USED_BADR];	/* Array of virtual addresses obtained
						   from ioremap, or NULL if a region is
						   unused, is in IO space, or is currently
						   not mapped into virtual memory address
						   space. */

	void __iomem *pReadToFlush;		/* Address from which to read in order to
						   flush writes to the board */

	rwlock_t HardwareLock;			/* R/W spinlock used to guarantee serialization
						   of hardware accesses where needed.
						   Also protects tvLastCounterResetTime, see
						   below! */

	u16 fpga1_release;			/* Release number of FPGA1, or 0xFFFF if unknown */
	u16 fpga2_release;			/* Release number of FPGA2, or 0xFFFF if unknown */

	unsigned int uMonFifoDepth;		/* Depth of Monitor FIFO in words (not bytes) */
	unsigned int uSeqFifoDepth;		/* Depth of Sequencer FIFO in words (not bytes) */
	unsigned int uMapFifoDepth;		/* Depth of Mapper FIFO in words (not bytes) */

	bool fIrqHandlerInstalled;		/* True iff an interrupt handler is installed */
/* ------------------------- */
	spinlock_t OpenFlagsLock;		/* Spinlock to protect access to the following
						   variables: nOpens and uidCurrentUser */

	unsigned int nOpens;			/* Count of the number of times the Monitor,
						   Sequencer and Mapper have been opened and not
						   released */

	uid_t uidCurrentUser;			/* The effective user ID of the user who first
						   opened one of the Monitor, Sequencer and
						   Mapper, or -1 if no device is open */
/* ------------------------- */
	atomic_t CounterPeriod_us;		/* The current counter period in microseconds */

	struct timeval tvLastCounterResetTime;	/* The time at which the counter was last reset.
						   Readers & Writers use HardwareLock (see above)
						   to guarantee atomicity! */
/* ------------------------- */
	struct mutex mtxMonitor;		/* Mutex to protect access to nMonitorOpens,
						   pMonitorBuffer and the buffer it points to,
						   ulMonWordsRead, ulMonReads, and the Monitor
						   FIFO and its hardware flags */

	unsigned int nMonitorOpens;		/* Count of the number of times the Monitor
						   has been opened and not released. */

	void *pMonitorBuffer;			/* Buffer allocated in Mon_open and freed in
						   Mon_release through which Monitor data are
						   copied */

	unsigned long ulMonWordsRead;		/* Statistic: Number of words read so far */

	unsigned long ulMonReads;		/* Statistic: Number of reads so far */

	atomic_t MonFifoHalfFullIntCount;	/* Number of Monitor FIFO Half Full interrupts */

	atomic_t MonFifoFullIntCount;		/* Number of Monitor FIFO Full interrupts */
/* ------------------------- */
	struct mutex mtxSequencer;		/* Mutex to protect access to nSequencerOpens,
						   pSequencerBuffer and the buffer it points to,
						   fSeqEnabled, ulSeqWordsWritten, ulSeqWrites
						   and the Sequencer FIFO and its hardware flags */

	unsigned int nSequencerOpens;		/* Count of the number of times the Sequencer
						   has been opened and not released. */

	void *pSequencerBuffer;			/* Buffer allocated in Seq_open and freed in
						   Seq_release through which Sequencer data are
						   copied */

	bool fSeqEnabled;			/* True iff the sequencer hardware is enabled */

	unsigned long ulSeqWordsWritten;	/* Statistic: Number of words written so far */

	unsigned long ulSeqWrites;		/* Statistic: Number of writes so far */

	atomic_t SeqFifoHalfEmptyIntCount;	/* Number of Sequencer FIFO Half Empty interrupts */

	atomic_t SeqFifoEmptyIntCount;		/* Number of Sequencer FIFO Empty interrupts */
/* ------------------------- */
	struct rw_semaphore rwsMapMem;		/* Reader-writer semaphore to protect access to
						   Mapper SRAM and related data, @@@ currently
						   only u32MapperNextFreeSpikeTblEntry and
						   puhMapperDestListBuffer */

	u32 u32MapperNextFreeSpikeTblEntry;	/* Offset in SRAM of next free spike table entry
						   @@@ This is a kludge to get some mapper
						   functionality going. In future we'll store a
						   list of free areas and their lengths and do
						   proper memory management. */

	unsigned short *puhMapperDestListBuffer;/* Buffer allocated in SetMapping and freed in
						   Map_release used for temporary storage of the
						   list of destination addresses in SetMapping */

	unsigned short uhMapperDestListBufferLen;/* Length of above buffer in unsigned shorts */

	atomic_t MapFifoFullIntCount;		/* Number of times the Mapper FIFO has filled */
/* ------------------------- */
	struct mutex mtxS5920NvrCache;		/* Mutex to protect access to uS5920NvrSize,
						   S5920NvrCacheBytes, S5920NvrCacheFlags and the
						   contents of the cache. */

	unsigned int uS5920NvrSize;		/* See s5920nvram.c */
	unsigned char *S5920NvrCacheBytes;	/* See s5920nvram.c */
	unsigned char *S5920NvrCacheFlags;	/* See s5920nvram.c */
/* ------------------------- */
	struct dentry *dDebugFSEntry;		/* Our directory in debugfs, or NULL if we don't
						   have one. */
};

/*****************************************************************************/

/*
 * Internal statistics structure common to monitor, sequencer and mapper.
 */

struct internal_stats {
	unsigned long words_transferred;
	unsigned long user_transfers;
	unsigned long total_interrupts;
	unsigned long overflows_underflows;
	unsigned long timeouts;
	unsigned long memory_usage;
};

/*****************************************************************************/

/*
 * Inline functions for getting register/base addresses given a pointer to a
 * struct pciaer_board
 */

static inline void __iomem *S5920OpRegAddress(const struct pciaer_board *b,
					      unsigned int reg)
{
	return b->pBaseAddr[S5920OPREGS_BADR] + reg;
}

static inline void __iomem *XilinxRegAddress(const struct pciaer_board *b,
					     unsigned int reg)
{
	return b->pBaseAddr[XILINXREGS_BADR] + reg;
}

static inline void __iomem *FifoAddress(const struct pciaer_board *b)
{
	return b->pBaseAddr[MONSEQFIFO_BADR];
}

static inline void __iomem *SramAddress(const struct pciaer_board *b)
{
	return b->pBaseAddr[MAPPERSRAM_BADR];
}

/*****************************************************************************/

extern const char DriverName[];	/* in pciaer.c */

extern const struct file_operations MonFileOps;
extern const struct file_operations SeqFileOps;
extern const struct file_operations MapFileOps;
extern const struct file_operations NvrFileOps;

/*****************************************************************************/

/* Functions in pcimodule.c */

extern struct pciaer_board *GetBoard(unsigned int iBoard);

/* Functions in driver.c */

extern int ResetXilinx(struct pciaer_board *pBoard);
extern int RegisterDriver(void);
extern void UnregisterDriver(void);
extern int InitHardware(struct pciaer_board *pBoard);
extern void DisableHardware(struct pciaer_board *pBoard);
extern int PciAer_release(struct inode *INode, struct file *fp);
extern int IncSingleUserLock(struct pciaer_board *pBoard);
extern void DecSingleUserLock(struct pciaer_board *pBoard);
extern int CheckIoctlPerm(struct file *fp, unsigned int uCmd);
extern int GetVersionInfo(struct pciaer_board *pBoard,
			  __user struct pciaer_version_info *pVersionInfo);
extern void ResetFifo(struct pciaer_board *pBoard, u32 u32CRA1FifoResetBit);

/* Functions in irq.c */

extern int SetupInterrupts(struct pciaer_board *pBoard);
extern bool EnableInt(struct pciaer_board *pBoard, u32 u32IntToEnable);
extern void DisableInt(struct pciaer_board *pBoard, u32 u32IntToDisable);

/* ISRs in monitor.c, sequencer.c and mapper.c */

extern void MonFifoHalfFullISR(struct pciaer_board *pBoard);
extern void MonFifoFullISR(struct pciaer_board *pBoard);
extern void SeqFifoHalfEmptyISR(struct pciaer_board *pBoard);
extern void SeqFifoEmptyISR(struct pciaer_board *pBoard);
extern void MapFifoFullISR(struct pciaer_board *pBoard);
extern unsigned long MapperOomHandler(struct pciaer_board *pBoard);

/* Functions in counter.c */

extern int SetCounterPeriod(struct pciaer_board *pBoard,
			    unsigned long ulMicroseconds);
extern int ResetCounter(struct pciaer_board *pBoard);
extern int ResetCounterGetTime(struct pciaer_board *pBoard,
			       __user struct timeval *ptvUser,
			       fmode_t FileMode);
extern int GetLastCounterResetTime(struct pciaer_board *pBoard,
				   __user struct timeval *ptvReset);

/* Functions in arbiter.c */

extern int SetArbiterConfig(struct pciaer_board *pBoard, unsigned long ulConfig);
extern int GetArbiterConfig(struct pciaer_board *pBoard);
extern int SetSequencerArbiterChannel(struct pciaer_board *pBoard,
				      unsigned long ulChannel);
extern int GetSequencerArbiterChannel(struct pciaer_board *pBoard);

/* Functions in s5920nvram.c */

extern unsigned int ReadS5920OpReg(struct pciaer_board *pBoard, unsigned int uReg);
extern void WriteS5920OpReg(struct pciaer_board *pBoard, unsigned int uReg,
			    unsigned int uValue);
extern void S5920NvrCleanup(struct pciaer_board *pBoard);
extern unsigned long S5920OomHandler(struct pciaer_board *pBoard);

/* Functions in debugfs.c */

extern void InitDebugFS(void);
extern void AddBoardToDebugFS(struct pciaer_board *pBoard);
extern void CleanUpDebugFS(void);

/* GetXxxStatistics functions */
extern int GetMonitorStats(struct pciaer_board *pBoard,
			   struct internal_stats *pStats, bool fReset);
extern int GetSequencerStats(struct pciaer_board *pBoard,
			     struct internal_stats *pStats, bool fReset);
extern int GetMapperStats(struct pciaer_board *pBoard,
			  struct internal_stats *pStats, bool fReset);

/*****************************************************************************/

/*
 * FlushPciWrites
 * Reads from the board to force the PCI bridge to flush the preceding write to
 * the board.
 */

static inline void FlushPciWrites(struct pciaer_board *pBoard)
{
	mb();
	ioread32(pBoard->pReadToFlush);
}

/*****************************************************************************/

/*
 * GetCounterPeriod
 * Called by Seq_ioctl & Mon_ioctl to implement PCIAER_IOC_GET_CNTR_PERIOD.
 * Also called from DisableMapperOut.
 * The return value is the current counter period in microseconds.
 */

static inline int GetCounterPeriod(struct pciaer_board *pBoard)
{
	return atomic_read(&pBoard->CounterPeriod_us);
}

/*****************************************************************************/

static inline int TestSysAdminCapability(void)
{
	return capable(CAP_SYS_ADMIN) ? 0 : -EPERM;
}

/*****************************************************************************/

#endif

/* EOF */
