/*

irq.c

This file contains code for the Linux 2.6.* character device driver
module for the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2000, 2003, 2005, 2006 & 2008-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

irq.c $Id: c40b56f588bf2b9e3a7dd1ba3de365015b2ce50c $
*/

#include <linux/version.h>

#include <linux/module.h>
#include <linux/interrupt.h>

#include "s5920.h"
#include "pciaerhw.h"
#include "internal.h"

/*****************************************************************************/

#define MIN_FPGA1_RELEASE_FOR_IRQ	0x4202

/*****************************************************************************/

/*
 * InterruptHandler
 * This is the master interrupt handler for all interrupts generated by any
 * PCI-AER board. PCI interrupts are shared, so first of all we need to
 * determine if the incoming interrupt was for us at all by looking to see
 * whether our S5920 has it's Interrupt Asserted bit set. If not we just exit.
 * If the interrupt is for us, we determine which of the various FPGA1 interrupt
 * flags were set including clearing the interrupt until none of the flags are
 * set and hence the S5920 no longer sees an Add-On interrupt. See p2-58 onwards
 * in the S5920 data book.
 */

static irqreturn_t InterruptHandler(int iIrq, void *pvBoardStruct)
{
	struct pciaer_board *pBoard;
	void __iomem *pINTCSR;
	u32 u32INTCSRContents;
	void __iomem *pStatRegA1;
	unsigned long ulHwLockFlags;
	u32 u32SRA1Contents;
	u32 u32IntFlagsToClear;

	pBoard = (struct pciaer_board *) pvBoardStruct;
	pINTCSR = S5920OpRegAddress(pBoard, S5920_INTCSR);

	/* If the incoming interrupt was not generated by our hardware, exit */

	u32INTCSRContents = ioread32(pINTCSR);

	if (!(u32INTCSRContents & S5920_INTCSR_INT_ASSERTED))
		return IRQ_NONE;

	/*
	 * Test for & clear spurious mailbox interrupts. This is very defensive.
	 * Mailbox interrupts should never be enabled, but if they ever became
	 * so and one was to spuriously fire, no amount of clearing our add-on
	 * interrupts would ever clear INTA# from the S5920. Therefore we make
	 * sure to definitely clear such interrupts here if they occur.
	 */

	if (u32INTCSRContents & (S5920_INTCSR_OMB_INT | S5920_INTCSR_IMB_INT)) {
		write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
		iowrite32(u32INTCSRContents &
			  ~(S5920_INTCSR_OMB_INT_ENABLE |
			    S5920_INTCSR_IMB_INT_ENABLE), pINTCSR);
		FlushPciWrites(pBoard);
		write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

		dev_err(&pBoard->pPciDev->dev, "Spurious mailbox interrupt\n");
	}

	/*
	 * Now loop around handling and clearing each of our possible
	 * FPGA1-based interrupts until the S5920 no longer sees the add-on
	 * interrupt line asserted.
	 */

	pStatRegA1 = XilinxRegAddress(pBoard, STAT_REG_A1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	while (ioread32(pINTCSR) & S5920_INTCSR_ADDINT_INT) {
		u32SRA1Contents = ioread32(pStatRegA1);
		u32IntFlagsToClear = 0;

		if (u32SRA1Contents & SRA1_INT_FH_MON) {
			u32IntFlagsToClear |= SRA1_INT_FH_MON;
			MonFifoHalfFullISR(pBoard);
		}

		if (u32SRA1Contents & SRA1_INT_FF_MON) {
			u32IntFlagsToClear |= SRA1_INT_FF_MON;
			MonFifoFullISR(pBoard);
		}

		if (u32SRA1Contents & SRA1_INT_FE_SEQ) {
			u32IntFlagsToClear |= SRA1_INT_FE_SEQ;
			SeqFifoEmptyISR(pBoard);
		}

		if (u32SRA1Contents & SRA1_INT_FH_SEQ) {
			u32IntFlagsToClear |= SRA1_INT_FH_SEQ;
			SeqFifoHalfEmptyISR(pBoard);
		}

		if (u32SRA1Contents & SRA1_INT_FF_MAP) {
			u32IntFlagsToClear |= SRA1_INT_FF_MAP;
			MapFifoFullISR(pBoard);
		}

		iowrite32(u32IntFlagsToClear, pStatRegA1);
		/*
		 * No extra read is necessary to flush, since the next operation
		 * will be the ioread32 in the while condition.
		 */
	}
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return IRQ_HANDLED;
}

/*****************************************************************************/

/*
 * EnableInt
 * Called to enable specific FPGA1 interrupts. The u32IntToEnable parameter
 * must be a combination of CRC1_DIS_INT_F?_??? bits.
 * No interrupts are enabled if there is no handler installed.
 * Returns true if interrupt handler is installed, and false otherwise.
 */

bool EnableInt(struct pciaer_board *pBoard, u32 u32IntToEnable)
{
	void __iomem *pCfgRegC1;
	unsigned long ulHwLockFlags;

	if (!pBoard->fIrqHandlerInstalled)
		return false;

	pCfgRegC1 = XilinxRegAddress(pBoard, CFG_REG_C1);
	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	iowrite32(ioread32(pCfgRegC1) & ~u32IntToEnable, pCfgRegC1);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);
	return true;
}

/*****************************************************************************/

/*
 * DisableInt
 * Called to disable specific FPGA1 interrupts. The u32IntToDisable parameter
 * must be a combination of CRC1_DIS_INT_F?_??? bits.
 */

void DisableInt(struct pciaer_board *pBoard, u32 u32IntToDisable)
{
	void __iomem *pCfgRegC1;
	unsigned long ulHwLockFlags;

	pCfgRegC1 = XilinxRegAddress(pBoard, CFG_REG_C1);
	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	iowrite32(ioread32(pCfgRegC1) | u32IntToDisable, pCfgRegC1);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);
}

/*****************************************************************************/

/*
 * SetupInterrupts
 * Called once per successfully initialized board from PciProbeCallback.
 * If FPGA1 is new enough to incorporate the expected interrupt flag mechanism,
 * SetupInterrupts tries to install an interrupt handler by calling
 * devm_request_irq. (The interrupt handler is uninstalled using devm_free_irq
 * from ReleaseBoards.)
 * Iff an interrupt handler is successfully installed, S5920 add-on interrupts
 * are enabled. (They are disabled in DisableHardware which is called from
 * ReleaseBoards BEFORE devm_free_irq.)
 */

int SetupInterrupts(struct pciaer_board *pBoard)
{
	struct pci_dev *pPciDevice;
	int nError;
	unsigned long ulHwLockFlags;

	/*
	 * If we cannot access the Xilinx regs, this is probably because the
	 * driver has loaded on a board on which the NVRAM is not yet programmed
	 * or corrupt.
	 */

	if (XilinxRegAddress(pBoard, 0) == NULL)
		return -ENXIO;

	pPciDevice = pBoard->pPciDev;

	/*
	 * We don't support the use of interrupts if the FPGA1 code is too old,
	 * since there was no reasonable interrupt clear mechanism.
	 */

	if (pBoard->fpga1_release == 0xFFFF ||
	    pBoard->fpga1_release < MIN_FPGA1_RELEASE_FOR_IRQ) {
		dev_warn(&pPciDevice->dev,
			 "Interrupts not supported on FPGA1 release < 0x%04x\n",
			 MIN_FPGA1_RELEASE_FOR_IRQ);
		return -EOPNOTSUPP;
	}

	/*
	 * Install interrupt handler. We use devm_request_threaded_irq, but
	 * we don't (yet) actually request a thread, hence the NULL parameter.
	 */

	nError = devm_request_threaded_irq(&pPciDevice->dev, pPciDevice->irq,
					   InterruptHandler, NULL, IRQF_SHARED,
					   pBoard->DeviceName, pBoard);
	if (nError != 0) {
		dev_warn(&pPciDevice->dev,
			 "Failed to install interrupt handler for board %u on IRQ %u: error %d\n",
			 pBoard->iBoard, pBoard->pPciDev->irq, -nError);
		return nError;
	}
	pBoard->fIrqHandlerInstalled = true;
	dev_info(&pPciDevice->dev, "Board %u using IRQ %u\n", pBoard->iBoard,
		 pPciDevice->irq);

	/* Enable interrupts at the PCI interface chip level */

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	WriteS5920OpReg(pBoard, S5920_INTCSR, S5920_INTCSR_ADDINT_ENABLE);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return 0;
}

/*****************************************************************************/

/* EOF */
