/* mapper.c

Mapper minor device code for the Linux character device driver
module for the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2000 & 2003-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

mapper.c $Id: a6f34449b472ac92df1108042019595679daa195 $
*/

#ifdef CONFIG_PCIAER_DEBUG
#define DEBUG
#endif

#include <linux/compat.h>
#include <linux/device.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/personality.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>

#include "pciaerhw.h"
#include "internal.h"

#ifndef CONFIG_PCIAER_DEBUG
#define INLINE inline
#else
#define INLINE
#endif

/*****************************************************************************/

/*
 * Some current boards are fitted with only 2 out of the possible 4 banks of
 * SRAM. The banks fitted are (probably) non-contiguous, being the first &
 * third(?)! For now, we assume only the first bank of 512KW is useable.
 *
 * @@@ In the longer term, the SRAM size and layout should be determined
 * dynamically (and without destroying the contents) for each board.
 *
 * NB. The length is given in terms of addressable words, each addressable word
 * being on a 4 byte boundary.
 */

#define MAPPER_SRAM_LENGTH	(512 * 1024)

/*
 * The maximum time we're prepared to wait within DisableMapperOut for the
 * mapper to go idle is defined in terms of counter periods.
 */

#define MAX_DIS_MAP_OUT_CNTR_PERIODS	128

/*
 * The threshhold for the number of destination addresses above which a kernel
 * warning message is emitted in the function CountDestinationAddresses (128 is
 * the longest list of destinations that stressmap will currently create).
 */

#define DEST_COUNT_WARN_THRESHHOLD	128

/*****************************************************************************/

/*
 * The equivalent structure to a pciaer_mapping struct in user land, with an
 * untranslated dest_ae pointer still pointing out there into user space.
 */

struct user_mapping {
	unsigned short source_ae;
	unsigned short dest_count;
	unsigned short __user *dest_ae;
};

/*
 * Same thing for use with compat_ioctl - i.e. what this struct looks like when
 * passed into a 64-bit kernel from a 32-bit user space program.
 */
#ifdef CONFIG_COMPAT
struct compat_mapping {
	u16 source_ae;
	u16 dest_count;
	compat_uptr_t dest_ae_ptr;
} __packed;
#endif

/*****************************************************************************/

/*
 * Module parameter allow_mmap_write can be set true to allow user-space
 * apps to use mprotect to gain write access to write directly to the mapper
 * SRAM using mmap, bypassing the driver's efforts to enforce SRAM integrity by
 * flushing PCI writes etc. This is primarily intended only for use by the
 * boardtest program, and possibly by other debugging code.
 */

BOOLEAN_MODULE_PARAMETER(allow_mmap_write,
			 "Allow unreliable direct writes to mmapped Mapper SRAM");

/*****************************************************************************/

static inline int read_implies_exec(void)
{
	return current->personality & READ_IMPLIES_EXEC;
}

/*****************************************************************************/

void MapFifoFullISR(struct pciaer_board *pBoard)
{
	atomic_inc(&pBoard->MapFifoFullIntCount);
}

/*****************************************************************************/

/*
 * SetMapperChannelSelect
 * Called by Map_ioctl to implement PCIAER_IOC_SET_MAP_CH_SEL. The parameter
 * ulMask should be a 4-bit mask containing a 1 for each channel which should be
 * processed by the mapper and a 0 for each channel which should be ignored. If
 * any other bits are set, the function returns -EINVAL, otherwise the function
 * returns 0.
 */

static int SetMapperChannelSelect(struct pciaer_board *pBoard, unsigned long ulMask)
{
	void __iomem *pCfgRegB1;
	unsigned long ulHwLockFlags;
	u32 u32RegContents;

	if (ulMask & ~0xF)
		return -EINVAL;

	pCfgRegB1 = XilinxRegAddress(pBoard, CFG_REG_B1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	u32RegContents = ioread32(pCfgRegB1);
	u32RegContents &= ~CRB1_MAP_IGN_CH_MASK;
	u32RegContents |=
	    (ulMask << CRB1_MAP_IGN_CH_SHIFT) ^ CRB1_MAP_IGN_CH_MASK;
	iowrite32(u32RegContents, pCfgRegB1);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return 0;
}

/*****************************************************************************/

/*
 * GetMapperChannelSelect
 * Called by Map_ioctl to implement PCIAER_IOC_GET_MAP_CH_SEL. The return value
 * is a 4-bit mask containing a 1 for each channel which is being processed by
 * the mapper and a 0 for each channel which is being ignored.
 */

static int GetMapperChannelSelect(struct pciaer_board *pBoard)
{
	u32 u32RegContents;

	u32RegContents = ioread32(XilinxRegAddress(pBoard, CFG_REG_B1));
	return ((u32RegContents & CRB1_MAP_IGN_CH_MASK) ^ CRB1_MAP_IGN_CH_MASK)
	    >> CRB1_MAP_IGN_CH_SHIFT;
}

/*****************************************************************************/

/*
 * SetMapperOutConfig
 * Called by Map_ioctl to implement PCIAER_IOC_SET_MAP_OUT_CONFIG. The
 * parameter ulConfig should be one of the PCIAER_IOC_MAP_OUT_* constants in
 * which case the function returns 0; otherwise the function returns -EINVAL.
 */

static int SetMapperOutConfig(struct pciaer_board *pBoard, unsigned long ulConfig)
{
	u32 u32MapOutCfgBits;
	void __iomem *pCfgRegA2;
	unsigned long ulHwLockFlags;

	switch (ulConfig) {
	case PCIAER_IOC_MAP_OUT_PASS_THRU:
		u32MapOutCfgBits = CRA2_MAP_OUT_CFG_PASSTHRU;
		break;
	/*
	 * Switching into 1 to 1 mode is not currently supported, because it
	 * would cause any 1 to many mapping pointers to be treated as
	 * addresses. In the longer term, requesting 1 to 1 mode should be
	 * permitted iff there are no 1 to many mappings in SRAM, and creating 1
	 * to many mappings should not be permitted in 1 to 1 mode.
	 */

	/* case PCIAER_IOC_MAP_OUT_1_TO_1:
		u32MapOutCfgBits = CRA2_MAP_OUT_CFG_1_TO_1;
		break;
	*/
	case PCIAER_IOC_MAP_OUT_1_TO_MANY:
		u32MapOutCfgBits = CRA2_MAP_OUT_CFG_1_TO_MANY;
		break;
	default:
		return -EINVAL;
		break;
	}

	pCfgRegA2 = XilinxRegAddress(pBoard, CFG_REG_A2);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	iowrite32((ioread32(pCfgRegA2) & ~CRA2_MAP_OUT_CFG_MASK) |
		  u32MapOutCfgBits, pCfgRegA2);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return 0;
}

/*****************************************************************************/

/*
 * GetMapperOutConfig
 * Called by Map_ioctl to implement PCIAER_IOC_GET_MAP_OUT_CONFIG.
 * The return value is one of the PCIAER_IOC_ARB_* constants, or -EIO if the
 * contents of CRA2 indicate the 'not used' value, previously labelled
 * 'Disabled'.
 */

static int GetMapperOutConfig(struct pciaer_board *pBoard)
{
	int nResult;

	switch (ioread32(XilinxRegAddress(pBoard, CFG_REG_A2)) &
		CRA2_MAP_OUT_CFG_MASK) {
	case CRA2_MAP_OUT_CFG_DISABLE:
		dev_err(&pBoard->pPciDev->dev,
			"CfgMapperOut[1:0] = 00 'Not Used'\n");
		nResult = -EIO;
		break;
	case CRA2_MAP_OUT_CFG_PASSTHRU:
		nResult = PCIAER_IOC_MAP_OUT_PASS_THRU;
		break;
	case CRA2_MAP_OUT_CFG_1_TO_1:
		nResult = PCIAER_IOC_MAP_OUT_1_TO_1;
		break;
	case CRA2_MAP_OUT_CFG_1_TO_MANY:
		nResult = PCIAER_IOC_MAP_OUT_1_TO_MANY;
		break;
	default:
		/*
		 * The default should never be reached if the CRA2_MAP_OUT_CFG
		 * constants are properly defined.
		 */
		dev_err(&pBoard->pPciDev->dev, __FILE__ "(%d) internal error\n",
			__LINE__);
		nResult = -EIO;
		break;
	}
	return nResult;
}

/*****************************************************************************/

/*
 * SetMapperDemuxConfig
 * Called by Map_ioctl to implement PCIAER_IOC_SET_MAP_DEMUX_CONFIG. The
 * parameter ulConfig should be one of the PCIAER_IOC_MAP_DEMUX_* constants in
 * which case the function returns 0; otherwise the function returns -EINVAL.
 */

static int SetMapperDemuxConfig(struct pciaer_board *pBoard, unsigned long ulConfig)
{
	u32 u32MapDemuxBits;
	void __iomem *pCfgRegA2;
	unsigned long ulHwLockFlags;

	switch (ulConfig) {
	/* case PCIAER_IOC_MAP_DEMUX_DISABLED:
		u32MapDemuxBits = CRA2_AER_OUT_CFG_DISABLE;
		break;
	*/
	case PCIAER_IOC_MAP_DEMUX_0_16:
		u32MapDemuxBits = CRA2_AER_OUT_CFG_1_16;
		break;
	case PCIAER_IOC_MAP_DEMUX_1_15:
		u32MapDemuxBits = CRA2_AER_OUT_CFG_2_15;
		break;
	case PCIAER_IOC_MAP_DEMUX_2_14:
		u32MapDemuxBits = CRA2_AER_OUT_CFG_4_14;
		break;
	default:
		return -EINVAL;
		break;
	}

	pCfgRegA2 = XilinxRegAddress(pBoard, CFG_REG_A2);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	iowrite32((ioread32(pCfgRegA2) & ~CRA2_AER_OUT_CFG_MASK) |
		  u32MapDemuxBits, pCfgRegA2);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return 0;
}

/*****************************************************************************/

/*
 * GetMapperDemuxConfig
 * Called by Map_ioctl to implement PCIAER_IOC_GET_MAP_DEMUX_CONFIG.
 * The return value is one of the PCIAER_IOC_MAP_DEMUX_* constants.
 */

static int GetMapperDemuxConfig(struct pciaer_board *pBoard)
{
	int nConfig;

	switch (ioread32(XilinxRegAddress(pBoard, CFG_REG_A2)) &
		CRA2_AER_OUT_CFG_MASK) {
	case CRA2_AER_OUT_CFG_DISABLE:
		nConfig = PCIAER_IOC_MAP_DEMUX_DISABLED;
		break;
	case CRA2_AER_OUT_CFG_1_16:
		nConfig = PCIAER_IOC_MAP_DEMUX_0_16;
		break;
	case CRA2_AER_OUT_CFG_2_15:
		nConfig = PCIAER_IOC_MAP_DEMUX_1_15;
		break;
	case CRA2_AER_OUT_CFG_4_14:
		nConfig = PCIAER_IOC_MAP_DEMUX_2_14;
		break;
	default:
		nConfig = -1;	/* The default should never be reached if the CRA2_AER_OUT_CFG */
		break;		/* constants are properly defined. */
	}
	return nConfig;
}

/*****************************************************************************/

/*
 * SetMapperAerProtocol
 * Called by Map_ioctl to implement PCIAER_IOC_SET_MAP_AER_PROTOCOL. The
 * parameter ulConfig should be one of the PCIAER_IOC_MAP_AER_* constants in
 * which case the function returns 0; otherwise the function returns -EINVAL.
 */

static int SetMapperAerProtocol(struct pciaer_board *pBoard,
				unsigned long ulProtocol)
{
	u32 u32AerProtocolBit;
	void __iomem *pCfgRegA2;
	unsigned long ulHwLockFlags;

	switch (ulProtocol) {
	case PCIAER_IOC_MAP_AER_P2P:
		u32AerProtocolBit = CRA2_AER_PROTOCOL_P2P;
		break;
	case PCIAER_IOC_MAP_AER_SCX:
		u32AerProtocolBit = CRA2_AER_PROTOCOL_MULTI;
		break;
	default:
		return -EINVAL;
		break;
	}

	pCfgRegA2 = XilinxRegAddress(pBoard, CFG_REG_A2);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	iowrite32((ioread32(pCfgRegA2) & ~CRA2_AER_PROTOCOL_MASK) |
		  u32AerProtocolBit, pCfgRegA2);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return 0;
}

/*****************************************************************************/

/*
 * GetMapperAerProtocol
 * Called by Map_ioctl to implement PCIAER_IOC_GET_MAP_AER_PROTOCOL.
 * The return value is one of the PCIAER_IOC_MAP_AER_* constants.
 */

static int GetMapperAerProtocol(struct pciaer_board *pBoard)
{
	int nProtocol;

	switch (ioread32(XilinxRegAddress(pBoard, CFG_REG_A2)) &
		CRA2_AER_PROTOCOL_MASK) {
	case CRA2_AER_PROTOCOL_P2P:
		nProtocol = PCIAER_IOC_MAP_AER_P2P;
		break;
	case CRA2_AER_PROTOCOL_MULTI:
		nProtocol = PCIAER_IOC_MAP_AER_SCX;
		break;
	default:
		nProtocol = -1;	/* The default should never be reached if the CRA2_AER_PROTOCOL */
		break;		/* constants are properly defined. */
	}
	return nProtocol;
}

/*****************************************************************************/

/*
 * EnableMapperOut
 * Called by Map_ioctl to implement PCIAER_IOC_ENABLE_MAP_OUT, and also from
 * RestoreMapperOut and Map_release.
 */

static void EnableMapperOut(struct pciaer_board *pBoard)
{
	void __iomem *pCfgRegA2;
	unsigned long ulHwLockFlags;

	pCfgRegA2 = XilinxRegAddress(pBoard, CFG_REG_A2);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	iowrite32(ioread32(pCfgRegA2) & ~CRA2_MAP_OUT_DISABLE, pCfgRegA2);
	FlushPciWrites(pBoard);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);
}

/*****************************************************************************/

/*
 * DisableMapperOut
 * Called by Map_ioctl to implement PCIAER_IOC_ENABLE_MAP_OUT, and by various
 * other mapper ioctl functions that need to suspend mapping to access the SRAM.
 * If the parameter ulWait is non-zero, the function waits for the mapper to
 * become idle before returning. If the 3rd parameter is non-NULL, then the
 * contents of CRA2 before the call will be stored at the given location. The
 * caller can determine from this whether the mapper output should be enabled
 * again when it has finished with the SRAM.
 * The return value should normally be zero, but may be -EBUSY if for some
 * reason the mapper hardware remains indicating that it is busy for a certain
 * number of counter periods after we try to make it go idle.
 */

static int DisableMapperOut(struct pciaer_board *pBoard, unsigned long ulWait,
			    u32 *pu32CRA2Value)
{
	void __iomem *pCfgRegA2;
	unsigned long ulHwLockFlags;
	u32 u32InitialCRA2Value;

	pCfgRegA2 = XilinxRegAddress(pBoard, CFG_REG_A2);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	u32InitialCRA2Value = ioread32(pCfgRegA2);

	if (pu32CRA2Value != NULL)
		*pu32CRA2Value = u32InitialCRA2Value;

	iowrite32(u32InitialCRA2Value | CRA2_MAP_OUT_DISABLE, pCfgRegA2);
	FlushPciWrites(pBoard);

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	if (ulWait != 0) {
		/*
		 * Have to do busy waiting on the MMU busy/idle bit.
		 * There's no hardware interrupt support for this.
		 */

		void __iomem *pStatRegA2;
		int nCounterPeriod_us;
		unsigned int uBusyWaitCycles;

		pStatRegA2 = XilinxRegAddress(pBoard, STAT_REG_A2);
		nCounterPeriod_us = GetCounterPeriod(pBoard);
		uBusyWaitCycles = 0;

		/*
		 * Wait for the MMU to be idle, delaying one counter period
		 * between each test of the idle bit, so as not to saturate the
		 * PCI bus.
		 */

		mb();
		while (!(ioread32(pStatRegA2) & SRA2_MMU_IDLE)) {
			if (++uBusyWaitCycles >= MAX_DIS_MAP_OUT_CNTR_PERIODS) {
				write_lock_irqsave(&pBoard->HardwareLock,
						   ulHwLockFlags);
				iowrite32(u32InitialCRA2Value, pCfgRegA2);
				FlushPciWrites(pBoard);
				write_unlock_irqrestore(&pBoard->HardwareLock,
							ulHwLockFlags);
				return -EBUSY;
			}
			usleep_range(nCounterPeriod_us, nCounterPeriod_us + 1);
		}
	}

	return 0;
}

/*****************************************************************************/

/*
 * RestoreMapperOut
 * Called as a counterpart to DisableMapperOut by the various other mapper ioctl
 * functions which need to suspend mapping to access the SRAM.
 * The u32PreviousCRA2Value parameter should be the value returned by
 * DisableMapperOut. If this value shows that Mapper Out was previously
 * disabled, it is left diabled by this function, otherwise it is enabled.
 */

static inline void RestoreMapperOut(struct pciaer_board *pBoard,
				    u32 u32PreviousCRA2Value)
{
	if ((u32PreviousCRA2Value & CRA2_MAP_OUT_DISABLE) == 0)
		EnableMapperOut(pBoard);
}

/*****************************************************************************/

/*
 * GetMapperOutState
 * Called by Map_ioctl to implement PCIAER_IOC_GET_MAP_OUT_STATE, and also by
 * Map_release.
 */

static inline int GetMapperOutState(struct pciaer_board *pBoard)
{
	void __iomem *pXilinxBaseAddr;
	unsigned long ulHwLockFlags;
	u32 u32StatRegA2Contents;
	u32 u32CfgRegA2Contents;

	pXilinxBaseAddr = XilinxRegAddress(pBoard, 0);

	read_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	u32StatRegA2Contents = ioread32(pXilinxBaseAddr + STAT_REG_A2);
	u32CfgRegA2Contents = ioread32(pXilinxBaseAddr + CFG_REG_A2);
	read_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	/*
	 * Check that the flag bit positions defined for the IOCTL match those
	 * defined for the hardware
	 */

	BUILD_BUG_ON(PCIAER_IOC_MAP_OUT_BUSY != SRA2_MMU_IDLE);
	BUILD_BUG_ON(PCIAER_IOC_MAP_OUT_ENABLED != CRA2_MAP_OUT_DISABLE);

	return (~u32StatRegA2Contents & SRA2_MMU_IDLE) | (~u32CfgRegA2Contents &
							  CRA2_MAP_OUT_DISABLE);
}

/*****************************************************************************/

/*
 * SRamRead
 * One central place where different strategies for reliably reading from SRAM
 * can be tried out.
 */

static INLINE u32 SRamRead(struct pciaer_board *pBoard, void __iomem *pAddress)
{
#ifdef CONFIG_PCIAER_DEBUG
	/* dev_dbg( &pBoard->pPciDev->dev, "SRR %lx %lx\n", SramAddress(pBoard),
		(unsigned long)pAddress ); */
	if (pAddress < SramAddress(pBoard)
	    || pAddress >=
	    SramAddress(pBoard) + MAPPER_SRAM_LENGTH * sizeof(unsigned int)) {
		dev_err(&pBoard->pPciDev->dev, "SRamRead with bad address %p\n", pAddress);
		BUG();
		return 0xFFFFFFFF;
	}
#endif
	return ioread32(pAddress);
}

/*****************************************************************************/

/*
 * SRamWrite
 * One central place where different strategies for reliably writing to SRAM can
 * be tried out.
 */

static INLINE void SRamWrite(struct pciaer_board *pBoard, u32 u32Value,
			     void __iomem *pAddress)
{
#ifdef CONFIG_PCIAER_DEBUG
	/* dev_dbg( &pBoard->pPciDev->dev, "SRW %lx %x %lx\n",
		SramAddress(pBoard), u32Value, (unsigned long)pAddress ); */
	if (pAddress < SramAddress(pBoard)
	    || pAddress >=
	    SramAddress(pBoard) + MAPPER_SRAM_LENGTH * sizeof(unsigned int)) {
		dev_err(&pBoard->pPciDev->dev, "SRamWrite with bad address %p\n", pAddress);
		BUG();
		return;
	}
#endif
	iowrite32(u32Value, pAddress);
	FlushPciWrites(pBoard);
}

/*****************************************************************************/

/*
 * MapperClear
 * Implements PCIAER_IOC_MAP_CLEAR.
 */

static int MapperClear(struct pciaer_board *pBoard)
{
	u32 u32InitialCRA2Value;
	int nError;
	struct pciaer_mapper_sram __iomem *pSRam;
	unsigned long ulHwLockFlags;
	int i;

	down_write(&pBoard->rwsMapMem);

	/*
	 * Disable mapper out and wait for idle.
	 * Store previous state of mapper out.
	 */

	nError = DisableMapperOut(pBoard, 1L, &u32InitialCRA2Value);
	if (nError != 0) {
		up_write(&pBoard->rwsMapMem);
		return nError;
	}

	/* Fill pointer table with MAPPER_PTR_NOT_USEDs. */

	pSRam = (struct pciaer_mapper_sram __iomem *) SramAddress(pBoard);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	for (i = 0; i < PCIAER_MAPPER_PTR_TBL_LENGTH; i++)
		SRamWrite(pBoard, MAPPER_PTR_TYPE_NOT_USED,
			  &pSRam->pointer_table[i]);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	pBoard->u32MapperNextFreeSpikeTblEntry = PCIAER_MAPPER_PTR_TBL_LENGTH;	/* @@@ Kludge - see struct pciaer_board comment */

	/* Restore state of mapper out. */

	RestoreMapperOut(pBoard, u32InitialCRA2Value);

	up_write(&pBoard->rwsMapMem);

	return 0;
}

/*****************************************************************************/

/*
 * CopyUserMappingStruct
 * Helper function that copies a pciaer_mapping struct from user space to kernel
 * space. fVerify should be VERIFY_READ or VERIFY_WRITE if verification for
 * writing is required. Returns 0 if successful or -EFAULT otherwise.
 * Additionally checks that source_ae is valid, and returns -EINVAL otherwise.
 * NB. pKern->dest_ae is still a user space pointer after calling this function.
 */

static int CopyUserMappingStruct(struct user_mapping *pKern,
				 __user struct user_mapping *pUser, int fVerify)
{
	if (!access_ok(fVerify, pUser, sizeof(struct user_mapping)))
		return -EFAULT;

	if (__copy_from_user(pKern, pUser, sizeof(struct user_mapping)) != 0)
		return -EFAULT;

	if (pKern->source_ae == MAPPER_END_LABEL)
		return -EINVAL;

	return 0;
}

/*****************************************************************************/

/*
 * CopyCompatMappingStruct
 * Helper function that copies a pciaer_mapping struct from 32-bit user space to
 * kernel space. fVerify should be VERIFY_READ or VERIFY_WRITE if verification
 * for writing is required. Returns 0 if successful or -EFAULT otherwise.
 * Additionally checks that source_ae is valid, and returns -EINVAL otherwise.
 * NB. pKern->dest_ae is still a user space pointer after calling this function.
 */
#ifdef CONFIG_COMPAT
static int CopyCompatMappingStruct(struct user_mapping *pKern,
				 __user struct compat_mapping *pUser32, int fVerify)
{
	struct compat_mapping Intermediate;

	if (!access_ok(fVerify, pUser32, sizeof(struct compat_mapping)))
		return -EFAULT;

	if (__copy_from_user(&Intermediate, pUser32, sizeof(struct compat_mapping)) != 0)
		return -EFAULT;

	if (Intermediate.source_ae == MAPPER_END_LABEL)
		return -EINVAL;

	pKern->source_ae = Intermediate.source_ae;
	pKern->dest_count = Intermediate.dest_count;
	pKern->dest_ae = compat_ptr(Intermediate.dest_ae_ptr);

	return 0;
}
#endif

/*****************************************************************************/

/*
 * SetMapping
 * Implements PCIAER_IOC_MAP_SET_MAPPING.
 */

static int SetMapping(struct pciaer_board *pBoard, struct user_mapping *pMapping)
{
	int nError;
	struct device *pDev;
	int i;
	u32 u32InitialCRA2Value;
	u32 __iomem *pu32SRam;
	unsigned long ulHwLockFlags;
	u32 u32PointerTableEntry;
	u32 __iomem *pu32SpikeTableAddress;

	/* Parameter validation */

	pDev = &pBoard->pPciDev->dev;

	down_write(&pBoard->rwsMapMem);

	if (pMapping->dest_count > pBoard->uhMapperDestListBufferLen) {
		if (pBoard->puhMapperDestListBuffer != NULL)
			devm_kfree(pDev, pBoard->puhMapperDestListBuffer);
		pBoard->puhMapperDestListBuffer =
		    devm_kzalloc(pDev, pMapping->dest_count * sizeof(unsigned short),
			    GFP_KERNEL);
		if (pBoard->puhMapperDestListBuffer != NULL) {
			pBoard->uhMapperDestListBufferLen = pMapping->dest_count;
		} else {
			up_write(&pBoard->rwsMapMem);
			return -ENOMEM;
		}
	}

	if (pMapping->dest_count != 0) {
		/* Copy destination list to buffer */

		if (copy_from_user
		    (pBoard->puhMapperDestListBuffer, pMapping->dest_ae,
		     pMapping->dest_count * sizeof(unsigned short)) != 0) {
			up_write(&pBoard->rwsMapMem);
			return -EFAULT;
		}
		/* Check destination list for invalid addresses */

		for (i = 0; i < pMapping->dest_count; i++)
			if (pBoard->puhMapperDestListBuffer[i] ==
			    MAPPER_END_LABEL) {
				up_write(&pBoard->rwsMapMem);
				return -EINVAL;
			}
	}

	/* Is there space available? */

	if (pMapping->dest_count > 1 &&
	    pBoard->u32MapperNextFreeSpikeTblEntry + pMapping->dest_count + 1 >
	    MAPPER_SRAM_LENGTH) {
		up_write(&pBoard->rwsMapMem);
		return -ENOSPC;
	}

	/*
	 * Disable mapper out and wait for idle.
	 * Store previous state of mapper out.
	 */

	nError = DisableMapperOut(pBoard, 1L, &u32InitialCRA2Value);
	if (nError != 0) {
		up_write(&pBoard->rwsMapMem);
		return nError;
	}

	/* Make appropriate entry/ies in SRAM */

	pu32SRam = (u32 __iomem *) SramAddress(pBoard);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	switch (pMapping->dest_count) {
	case 0:
		u32PointerTableEntry = MAPPER_PTR_TYPE_NOT_USED;
		break;

	case 1:
		u32PointerTableEntry =
		    MAPPER_PTR_TYPE_DIRECT | pBoard->puhMapperDestListBuffer[0];
		break;

	default:
		/* Copy destination list to SRAM */

		pu32SpikeTableAddress =
		    &pu32SRam[pBoard->u32MapperNextFreeSpikeTblEntry];
		u32PointerTableEntry =
		    MAPPER_PTR_TYPE_INDIRECT | pBoard->
		    u32MapperNextFreeSpikeTblEntry;
		pBoard->u32MapperNextFreeSpikeTblEntry +=
		    pMapping->dest_count + 1;
		for (i = 0; i < pMapping->dest_count; i++) {
			SRamWrite(pBoard, pBoard->puhMapperDestListBuffer[i],
				  pu32SpikeTableAddress++);
			/* dev_dbg( &pBoard->pPciDev->dev, "w %p %04x\n",
				&pu32SRam[ (u32PointerTableEntry & MAPPER_PTR_INDIRECT_MASK) + i ],
				uhDestAE ); */
		}
		SRamWrite(pBoard, MAPPER_END_LABEL, pu32SpikeTableAddress);
		break;
	}

	/* Set pointer in pointer table. */

	SRamWrite(pBoard, u32PointerTableEntry, &pu32SRam[pMapping->source_ae]);

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	/* Restore state of mapper out. */

	RestoreMapperOut(pBoard, u32InitialCRA2Value);

	up_write(&pBoard->rwsMapMem);

	return 0;
}

/*****************************************************************************/

/*
 * CountDestinationAddresses
 * Returns the number of destination addresses for a given AE, or -EIO if there
 * is bad data in the SRAM. Must be called with MapperOut disabled and rwsMapMem
 * held for reading. Called from GetMappingCount and GetMapping.
 */

static int CountDestinationAddresses(struct pciaer_board *pBoard,
				     unsigned short uhAE)
{
	u32 __iomem *pu32SRam;
	u32 u32PointerTableEntry;
	u32 u32SRamOffset;
	int nReturnValue;
	unsigned short uhDestAddr;

	pu32SRam = (u32 __iomem *) SramAddress(pBoard);

	u32PointerTableEntry = SRamRead(pBoard, &pu32SRam[uhAE]);

	switch (u32PointerTableEntry & MAPPER_PTR_TYPE_MASK) {
	case MAPPER_PTR_TYPE_NOT_USED:
		nReturnValue = 0;
		break;

	case MAPPER_PTR_TYPE_DIRECT:
		nReturnValue = 1;
		break;

	case MAPPER_PTR_TYPE_INDIRECT:
		nReturnValue = 0;
		u32SRamOffset = u32PointerTableEntry & MAPPER_PTR_INDIRECT_MASK;
		do {
			if (u32SRamOffset >= MAPPER_SRAM_LENGTH) {
				dev_err(&pBoard->pPciDev->dev,
					"SRAM offset 0x%08x outside valid SRAM while counting destination addresses for source AE 0x%04x\n",
					u32SRamOffset, (unsigned int)uhAE);
				nReturnValue = -EIO;
				break;
			}
			uhDestAddr =
			    SRamRead(pBoard,
				     &pu32SRam[u32SRamOffset++]) &
			    MAPPER_PTR_DIRECT_MASK;
			if (uhDestAddr != MAPPER_END_LABEL)
				nReturnValue++;
		} while (uhDestAddr != MAPPER_END_LABEL);

		/*
		 * Emit a warning if we come across an apparently over-large
		 * list of destinations. (Perhaps we missed the MAPPER_END_LABEL
		 * due to some hardware or addressing problem.)
		 */

		if (nReturnValue > DEST_COUNT_WARN_THRESHHOLD)
			dev_warn(&pBoard->pPciDev->dev,
				 "Counted %d destinations for source AE 0x%04x\n",
				 nReturnValue, (unsigned int)uhAE);

		break;

	default:
		dev_err(&pBoard->pPciDev->dev,
			"Bad pointer table entry 0x%08x for source AE 0x%04x\n",
			u32PointerTableEntry, (unsigned int)uhAE);
		nReturnValue = -EIO;
		break;
	}

	return nReturnValue;
}

/*****************************************************************************/

/*
 * GetMappingCount
 * Implements PCIAER_IOC_GET_MAP_MAPPING_COUNT
 */

static int GetMappingCount(struct pciaer_board *pBoard,
			   struct user_mapping *pMapping,
			   unsigned short __user *pUserDestCount)
{
	int nError;
	u32 u32InitialCRA2Value;
	int nDestCount;

	down_read(&pBoard->rwsMapMem);

	/*
	 * Disable mapper out and wait for idle.
	 * Store previous state of mapper out.
	 */

	nError = DisableMapperOut(pBoard, 1L, &u32InitialCRA2Value);
	if (nError != 0) {
		up_read(&pBoard->rwsMapMem);
		return nError;
	}

	/* Count destination addresses in SRAM */

	nDestCount = CountDestinationAddresses(pBoard, pMapping->source_ae);

	/* Restore state of mapper out. */

	RestoreMapperOut(pBoard, u32InitialCRA2Value);

	up_read(&pBoard->rwsMapMem);

	/*
	 * If there was an error, report it, otherwise put result in user's
	 * buffer.
	 */

	if (nDestCount < 0)
		return nDestCount;	/* error number */
	else
		return __put_user((unsigned short)nDestCount, pUserDestCount);
}

/*****************************************************************************/

/*
 * GetMapping
 * Implements PCIAER_IOC_GET_MAP_MAPPING
 */

static int GetMapping(struct pciaer_board *pBoard,
		      struct user_mapping *pMapping,
		      unsigned short __user *pUserDestCount)
{
	int nError;
	u32 u32InitialCRA2Value;
	u32 __iomem *pu32SRam;
	u32 u32PointerTableEntry;
	u32 u32SRamOffset;
	int nDestCount;
	int i;

	if (pMapping->dest_count != 0
	    && !access_ok(VERIFY_WRITE, pMapping->dest_ae,
			  pMapping->dest_count * sizeof(unsigned short)))
		return -EFAULT;

	down_read(&pBoard->rwsMapMem);

	/*
	 * Disable mapper out and wait for idle.
	 * Store previous state of mapper out.
	 */

	nError = DisableMapperOut(pBoard, 1L, &u32InitialCRA2Value);
	if (nError != 0) {
		up_read(&pBoard->rwsMapMem);
		return nError;
	}

	/* Count destination addresses in SRAM */

	nDestCount = CountDestinationAddresses(pBoard, pMapping->source_ae);
	if (nDestCount < 0)	/* Error occurred */
		nError = nDestCount;	/* Negative error number */
	if (nDestCount > (int)pMapping->dest_count) {
		/* dev_dbg( &pBoard->pPciDev->dev, "%s(%d):%d>%hu\n",
			__FILE__, __LINE__ - 2, nDestCount, pMapping->dest_count ); */
		/* dev_dbg( &pBoard->pPciDev->dev, "second count: %d\n",
			CountDestinationAddresses( pBoard, pMapping->source_ae ) ); */
		nError = -EINVAL;
	}

	if (nError == 0) {
		nError =
		    __put_user((unsigned short)nDestCount, pUserDestCount);

		if (nError == 0 && nDestCount != 0) {
			pu32SRam = (u32 __iomem *) SramAddress(pBoard);
			u32PointerTableEntry =
			    SRamRead(pBoard, &pu32SRam[pMapping->source_ae]);
			switch (u32PointerTableEntry & MAPPER_PTR_TYPE_MASK) {
			case MAPPER_PTR_TYPE_DIRECT:
				nError =
				    __put_user((unsigned
						short)(u32PointerTableEntry &
						       MAPPER_PTR_DIRECT_MASK),
					       &pMapping->dest_ae[0]);
				break;
			case MAPPER_PTR_TYPE_INDIRECT:
				u32SRamOffset =
				    u32PointerTableEntry &
				    MAPPER_PTR_INDIRECT_MASK;
				if (u32SRamOffset + nDestCount >
				    MAPPER_SRAM_LENGTH) {
					dev_err(&pBoard->pPciDev->dev,
						"SRAM offset 0x%08x outside valid SRAM while counting destination addresses for source AE 0x%04x\n",
						u32SRamOffset,
						(unsigned int)pMapping->
						source_ae);
					nError = -EIO;
				} else
					for (i = 0; i < nDestCount; i++) {
						nError =
						    __put_user((unsigned short)
							       SRamRead(pBoard,
									&pu32SRam
									[u32SRamOffset++]),
							       &pMapping->
							       dest_ae[i]);
						if (nError != 0)
							break;
					}
				break;
			default:
				dev_err(&pBoard->pPciDev->dev,
					"Bad pointer table entry 0x%08x for source AE 0x%04x\n",
					u32PointerTableEntry,
					(unsigned int)pMapping->source_ae);
				nError = -EIO;
				break;
			}
		}
	}

	/* Restore state of mapper out. */

	RestoreMapperOut(pBoard, u32InitialCRA2Value);

	up_read(&pBoard->rwsMapMem);

	return nError;
}

/*****************************************************************************/

/*
 * FindNextMapping
 * Implements PCIAER_IOC_FIND_NEXT_MAP_MAPPING
 */

static int FindNextMapping(struct pciaer_board *pBoard, __user int *pnArg)
{
	int nArg;
	int nError;
	unsigned short uhSourceAE;
	u32 __iomem *pu32SRam;
	u32 u32InitialCRA2Value;
	u32 u32PointerType;

	if (!access_ok(VERIFY_WRITE, pnArg, sizeof(int)))
		return -EFAULT;

	nError = __get_user(nArg, pnArg);
	if (nError != 0)
		return nError;

	if (nArg < -1 || nArg >= MAPPER_END_LABEL)
		return -EINVAL;

	uhSourceAE = (unsigned short)(nArg + 1);

	pu32SRam = (u32 __iomem *) SramAddress(pBoard);

	down_read(&pBoard->rwsMapMem);

	/*
	 * Disable mapper out and wait for idle.
	 * Store previous state of mapper out.
	 */

	nError = DisableMapperOut(pBoard, 1L, &u32InitialCRA2Value);
	if (nError != 0) {
		up_read(&pBoard->rwsMapMem);
		return nError;
	}

	while (uhSourceAE != MAPPER_END_LABEL) {
		u32PointerType =
		    SRamRead(pBoard,
			     &pu32SRam[uhSourceAE]) & MAPPER_PTR_TYPE_MASK;
		if (u32PointerType == MAPPER_PTR_TYPE_INDIRECT
		    || u32PointerType == MAPPER_PTR_TYPE_DIRECT)
			break;
		else
			uhSourceAE++;
	}

	/* Restore state of mapper out. */

	RestoreMapperOut(pBoard, u32InitialCRA2Value);

	up_read(&pBoard->rwsMapMem);

	nArg =
	    (uhSourceAE !=
	     MAPPER_END_LABEL) ? (int)(unsigned int)uhSourceAE : -1;
	return __put_user(nArg, pnArg);
}

/*****************************************************************************/

/*
 * GetMappedMap
 * Implements PCIAER_IOC_GET_MAP_MAPPED_MAP.
 */

static int GetMappedMap(struct pciaer_board *pBoard, __user u32 *pu32BitVector)
{
	u32 __iomem *pu32SRam;
	u32 u32InitialCRA2Value;
	int iSourceAE;
	int iU32;
	u32 u32Bits;
	u32 u32Bit;
	u32 u32PointerType;
	int nError;

	if (!access_ok(VERIFY_WRITE, pu32BitVector, 64 * 1024 / 8)) /* 64K bits */
		return -EFAULT;

	pu32SRam = (u32 __iomem *) SramAddress(pBoard);

	down_read(&pBoard->rwsMapMem);

	/*
	 * Disable mapper out and wait for idle.
	 * Store previous state of mapper out.
	 */

	nError = DisableMapperOut(pBoard, 1L, &u32InitialCRA2Value);
	if (nError != 0) {
		up_read(&pBoard->rwsMapMem);
		return nError;
	}

	/* Fill user's buffer a u32 at a time for efficiency */

	iSourceAE = 0;
	for (iU32 = 0; iU32 < 64 * 1024 / 32; iU32++) {
		u32Bits = 0;
		for (u32Bit = 0x00000001; u32Bit != 0x00000000; u32Bit <<= 1) {
			u32PointerType =
			    SRamRead(pBoard,
				     &pu32SRam[iSourceAE++]) &
			    MAPPER_PTR_TYPE_MASK;
			if (u32PointerType == MAPPER_PTR_TYPE_INDIRECT
			    || u32PointerType == MAPPER_PTR_TYPE_DIRECT)
				u32Bits |= u32Bit;
		}
		nError = __put_user(u32Bits, &pu32BitVector[iU32]);
		if (nError != 0)
			break;
	}

	/* Restore state of mapper out. */

	RestoreMapperOut(pBoard, u32InitialCRA2Value);

	up_read(&pBoard->rwsMapMem);

	/* Last bit (representing source AE 0xFFFF) should always be zero. */

	if (nError == 0)
		nError =
		    __put_user(u32Bits & 0x7fffffff,
			       &pu32BitVector[64 * 1024 / 32 - 1]);

	return nError;
}

/*****************************************************************************/

/*
 * GetMapperStats
 * Fills in an internal_stats struct on behalf of debugfs.c and resets the
 * statistics if required.
 */

int GetMapperStats(struct pciaer_board *pBoard, struct internal_stats *pStats,
		   bool fReset)
{
	int cFifoFills;

	cFifoFills = atomic_read(&pBoard->MapFifoFullIntCount);
	if (fReset)
		atomic_sub(cFifoFills, &pBoard->MapFifoFullIntCount);

	if (pStats != NULL) {
		pStats->total_interrupts = cFifoFills;
		pStats->overflows_underflows = cFifoFills;
	}

	return 0;
}

/*****************************************************************************/

/*
 * MapperOomHandler
 * Frees any pages being used to cache the board's SRAM.
 * Returns the number of pages so freed.
 */

unsigned long MapperOomHandler(struct pciaer_board *pBoard)
{
	return 0UL;
}

/*****************************************************************************/

/*
 * Map_ioctl
 * This function handles all ioctls on the Mapper minor device.
 */

static long Map_ioctl(struct file *fp, unsigned int uCmd, unsigned long ulArgs)
{
	struct pciaer_board *pBoard;
	int nError;
	__user struct user_mapping *pUserMapping;
	struct user_mapping KernMapping;
	extern int ModuleOomHandler(struct notifier_block *nb, unsigned long dummy, void *pvFreedPages);
	unsigned long ulTest;

	nError = CheckIoctlPerm(fp, uCmd);
	if (nError != 0 && uCmd != PCIAER_IOC_GET_MAP_MAPPED_MAP)
		return nError;

	pBoard = (struct pciaer_board *) (fp->private_data);

	nError = -ENOTTY;

	switch (uCmd) {
	case PCIAER_IOC_GET_VERSION_INFO:
		nError =
		    GetVersionInfo(pBoard,
				   (__user struct pciaer_version_info *) ulArgs);
		break;

	case PCIAER_IOC_SET_SEQ_ARB_CHANNEL:
		nError = SetSequencerArbiterChannel(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_SEQ_ARB_CHANNEL:
		nError =
		    put_user(GetSequencerArbiterChannel(pBoard),
			     (__user int *)ulArgs);
		break;

	case PCIAER_IOC_SET_ARB_CONFIG:
		nError = SetArbiterConfig(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_ARB_CONFIG:
		nError =
		    put_user(GetArbiterConfig(pBoard), (__user int *)ulArgs);
		break;

	case PCIAER_IOC_SET_MAP_CH_SEL:
		nError = SetMapperChannelSelect(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_MAP_CH_SEL:
		nError =
		    put_user(GetMapperChannelSelect(pBoard),
			     (__user int *)ulArgs);
		break;

	case PCIAER_IOC_SET_MAP_MAPPING:
		pUserMapping = (__user struct user_mapping *)ulArgs;
		nError = CopyUserMappingStruct(&KernMapping,
					       pUserMapping,
					       VERIFY_READ);
		if (nError == 0)
			nError = SetMapping(pBoard, &KernMapping);
		break;

	case PCIAER_IOC_GET_MAP_MAPPING_COUNT:
		pUserMapping = (__user struct user_mapping *)ulArgs;
		nError = CopyUserMappingStruct(&KernMapping,
					       pUserMapping,
					       VERIFY_WRITE);
		if (nError == 0)
			nError = GetMappingCount(pBoard, &KernMapping, &pUserMapping->dest_count);
		break;

	case PCIAER_IOC_GET_MAP_MAPPING:
		pUserMapping = (__user struct user_mapping *)ulArgs;
		nError = CopyUserMappingStruct(&KernMapping,
					       pUserMapping,
					       VERIFY_WRITE);
		if (nError == 0)
			nError = GetMapping(pBoard, &KernMapping, &pUserMapping->dest_count);
		break;

	case PCIAER_IOC_MAP_ADD_TO_MAPPING:
		/* @@@ */
		break;

	case PCIAER_IOC_MAP_DEL_FROM_MAPPING:
		/* @@@ */
		break;

	case PCIAER_IOC_FIND_NEXT_MAP_MAPPING:
		nError = FindNextMapping(pBoard, (__user int *)ulArgs);
		break;

	case PCIAER_IOC_GET_MAP_MAPPED_MAP:
		if (fp->f_mode & FMODE_READ)
			nError = GetMappedMap(pBoard, (__user u32 *) ulArgs);
		else
			nError = -EACCES;
		break;

	case PCIAER_IOC_MAP_CLEAR:
		nError = MapperClear(pBoard);
		break;

	case PCIAER_IOC_MAP_COMPACT:
	    /* XXX Temporarily subvert this ioctl for testing OOM handler! */
	    ModuleOomHandler(NULL, 0UL, &ulTest);
		/*
		 * @@@ At the moment, memory is always maximally compacted, so
		 * this ioctl does nothing. In the future, this will be a big,
		 * tricky function! */
		nError = 0;
		break;

	case PCIAER_IOC_GET_MAP_FREE_SPACE:
		nError =
		    put_user(MAPPER_SRAM_LENGTH -
			     pBoard->u32MapperNextFreeSpikeTblEntry,
			     (__user unsigned int *)ulArgs);
		break;

	case PCIAER_IOC_GET_MAP_FIFO_FLAGS:
		/*
		 * Check that the flag bit positions defined for the IOCTL match
		 * those defined for the hardware
		 */
		BUILD_BUG_ON(PCIAER_IOC_MAP_EMPTY != SRA1_MAP_FIFO_EF_);
		BUILD_BUG_ON(PCIAER_IOC_MAP_FULL != SRA1_MAP_FIFO_FF_);

		nError = put_user(~ioread32(XilinxRegAddress(pBoard, STAT_REG_A1)), (__user int *)ulArgs);	/* Inverts bits for +ve logic result */
		break;

	case PCIAER_IOC_RST_FIFO:
		ResetFifo(pBoard, CRA1_MAP_IN_FIFO_MR_);
		nError = 0;
		break;

	case PCIAER_IOC_ENABLE_MAP_OUT:
		EnableMapperOut(pBoard);
		nError = 0;
		break;

	case PCIAER_IOC_DISABLE_MAP_OUT:
		nError = DisableMapperOut(pBoard, ulArgs, NULL);
		break;

	case PCIAER_IOC_GET_MAP_OUT_STATE:
		nError =
		    put_user(GetMapperOutState(pBoard), (__user int *)ulArgs);
		break;

	case PCIAER_IOC_SET_MAP_OUT_CONFIG:
		nError = SetMapperOutConfig(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_MAP_OUT_CONFIG:
		{
			int nResult;

			nResult = GetMapperOutConfig(pBoard);
			if (nResult >= 0)
				nError =
				    put_user(nResult, (__user int *)ulArgs);
			else
				nError = nResult;
		}
		break;

	case PCIAER_IOC_SET_MAP_DEMUX_CONFIG:
		nError = SetMapperDemuxConfig(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_MAP_DEMUX_CONFIG:
		nError =
		    put_user(GetMapperDemuxConfig(pBoard),
			     (__user int *)ulArgs);
		break;

	case PCIAER_IOC_SET_MAP_AER_PROTOCOL:
		nError = SetMapperAerProtocol(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_MAP_AER_PROTOCOL:
		nError =
		    put_user(GetMapperAerProtocol(pBoard),
			     (__user int *)ulArgs);
		break;

	default:
		nError = -ENOTTY;
		break;
	}
	return nError;
}

/*****************************************************************************/

/*
 * Map_compat_ioctl
 * Implements the compat_ioctl call for the Mapper minor device.
 * The ioctls PCIAER_IOC_SET_MAP_MAPPING, PCIAER_IOC_GET_MAP_MAPPING_COUNT,
 * PCIAER_IOC_GET_MAP_MAPPING, PCIAER_IOC_MAP_ADD_TO_MAPPING and
 * PCIAER_IOC_MAP_DEL_FROM_MAPPING all pass a pointer to a pciaer_mapping
 * structure, and this contains a pointer member (unsigned short *dest_ae)
 * which means that the structure needs translating from the 32-bit world into
 * the 64-bit world.
 * All other ioctls are passed to the driver-generic CompatIoctl function.
 */
#ifdef CONFIG_COMPAT

#define IOC_COMPAT(C, SIZE) _IOC(_IOC_DIR(C), PCIAER_IOC_MAGIC, _IOC_NR(C), SIZE)

static long Map_compat_ioctl(struct file *fp, unsigned int cmd, unsigned long ulArgs)
{
	struct compat_mapping __user *pUserMapping;
	struct pciaer_board *pBoard;
	struct user_mapping KernMapping;
	long error;

	pUserMapping = (__user struct compat_mapping *)ulArgs;
	pBoard = (struct pciaer_board *)(fp->private_data);

	switch (cmd) {
	case IOC_COMPAT(PCIAER_IOC_SET_MAP_MAPPING, sizeof(struct compat_mapping)):
		error = CopyCompatMappingStruct(&KernMapping,
						pUserMapping,
						VERIFY_READ);
		if (error == 0L)
			error = SetMapping(pBoard, &KernMapping);
		break;

	case IOC_COMPAT(PCIAER_IOC_GET_MAP_MAPPING_COUNT, sizeof(struct compat_mapping)):
		error = CopyCompatMappingStruct(&KernMapping,
						pUserMapping,
						VERIFY_WRITE);
		if (error == 0L)
			error = GetMappingCount(pBoard, &KernMapping, &pUserMapping->dest_count);
		break;

	case IOC_COMPAT(PCIAER_IOC_GET_MAP_MAPPING, sizeof(struct compat_mapping)):
		error = CopyCompatMappingStruct(&KernMapping,
						pUserMapping,
						VERIFY_WRITE);
		if (error == 0L)
			error = GetMapping(pBoard, &KernMapping, &pUserMapping->dest_count);
		break;

	case IOC_COMPAT(PCIAER_IOC_MAP_ADD_TO_MAPPING, sizeof(struct compat_mapping)):
		/* @@@ */
	case IOC_COMPAT(PCIAER_IOC_MAP_DEL_FROM_MAPPING, sizeof(struct compat_mapping)):
		/* @@@ */
	default:
		error = Map_ioctl(fp, cmd, ulArgs);
		break;
	}
	return error;
}
#endif
/*****************************************************************************/

/*
 * Map_mmap
 * Implements the mmap call for the Mapper minor device.
 */

static int Map_mmap(struct file *fp, struct vm_area_struct *vma)
{
	unsigned long ulMemMapLength;
	struct pci_dev *pPciDev;
	unsigned long ulPhysAddr;

	pPciDev = ((struct pciaer_board *) (fp->private_data))->pPciDev;

	dev_dbg(&pPciDev->dev, "Map_mmap:\n");
	dev_dbg(&pPciDev->dev, "vm_start 0x%lx\n", vma->vm_start);
	dev_dbg(&pPciDev->dev, "vm_end   0x%lx\n", vma->vm_end);
	dev_dbg(&pPciDev->dev, "vm_pgoff 0x%lx\n", vma->vm_pgoff);
	dev_dbg(&pPciDev->dev, "vm_flags 0x%lx\n", vma->vm_flags);

	/*
	 * Parameter validation.
	 *
	 * First of all, we check the caller is not asking for more rights than
	 * we're prepared to grant, ie. PROT_WRITE or PROT_EXEC was supplied to
	 * the mmap call.
	 */

	if ((vma->vm_flags & VM_WRITE)
	    || (!read_implies_exec() && (vma->vm_flags & VM_EXEC)))
		return -EACCES;

	dev_dbg(&pPciDev->dev, "mmap prot OK\n");

	/*
	 * Unless overridden by the module parameter allow_mmap_write and we
	 * have the RAWIO capability, turn off VM_MAYWRITE bit to stop mprotect
	 * being used to gain direct write access.
	 */

	if (!allow_mmap_write || !capable(CAP_SYS_RAWIO))
		vma->vm_flags &= ~VM_MAYWRITE;

	/*
	 * Similarly prevent mprotect being used to gain execute access, insofar
	 * that this is separate from read access.
	 */

	if (!read_implies_exec())
		vma->vm_flags &= ~VM_MAYEXEC;

	/*
	 * If mmap was called with MAP_PRIVATE, return -EINVAL, since having a
	 * copy-on-write copy of the device's SRAM could be quite misleading.
	 */

	if (!(vma->vm_flags & VM_SHARED))
		return -EINVAL;

	dev_dbg(&pPciDev->dev, "mmap flags OK\n");

	/*
	 * For simplicity, we only accept requests with a size that exactly
	 * matches the size of the Mapper SRAM PCI region and with an offset of
	 * 0.
	 */

	ulMemMapLength = vma->vm_end - vma->vm_start;

	dev_dbg(&pPciDev->dev, "pci_resource_len 0x%lx\n",
		(unsigned long)pci_resource_len(pPciDev, MAPPERSRAM_BADR));

	if (ulMemMapLength != pci_resource_len(pPciDev, MAPPERSRAM_BADR))
		return -EINVAL;

	dev_dbg(&pPciDev->dev, "mmap length OK\n");

	if (vma->vm_pgoff != 0)
		return -EINVAL;

	dev_dbg(&pPciDev->dev, "mmap offset OK\n");

	/* At this point, we have a vm_area_struct that is OK */

	/* Turn on extra appropriate bits in flags */

	vma->vm_flags |= VM_IO | VM_DONTEXPAND | VM_RESERVED;
	dev_dbg(&pPciDev->dev, "vm_flags 0x%lx\n", vma->vm_flags);

	/* Perform memory mapping on physical address */

	ulPhysAddr = pci_resource_start(pPciDev, MAPPERSRAM_BADR);
	dev_dbg(&pPciDev->dev, "ulPhysAddr 0x%lx\n", ulPhysAddr);

	return remap_pfn_range(vma, vma->vm_start, ulPhysAddr >> PAGE_SHIFT,
			       ulMemMapLength, vma->vm_page_prot);
}

/*****************************************************************************/

/*
 * Map_open
 * Supplementary to PciAerMain_open, fail open with ENXIO if all the relevant
 * hardware is not available or EBUSY if another user is using the board.
 */

static int Map_open(struct inode *INode, struct file *fp)
{
	struct pciaer_board *pBoard;

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (S5920OpRegAddress(pBoard, 0) == NULL
	    || XilinxRegAddress(pBoard, 0) == NULL
	    || SramAddress(pBoard) == NULL)
		return -ENXIO;

	if (IncSingleUserLock(pBoard) != 0)
		return -EBUSY;

	INode->i_blkbits = ilog2(pBoard->uMapFifoDepth * sizeof(u32) / 2);

	return nonseekable_open(INode, fp);
}

/*****************************************************************************/

/*
 * Map_release
 * Re-enable the mapper if disabled, free buffer memory, decrement the single
 * user lock and call PciAer_release.
 */

static int Map_release(struct inode *INode, struct file *fp)
{
	struct pciaer_board *pBoard;

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (!(GetMapperOutState(pBoard) && PCIAER_IOC_MAP_OUT_ENABLED))
		EnableMapperOut(pBoard);

	if (pBoard->puhMapperDestListBuffer != NULL) {
		down_write(&pBoard->rwsMapMem);
		devm_kfree(&pBoard->pPciDev->dev, pBoard->puhMapperDestListBuffer);
		pBoard->puhMapperDestListBuffer = NULL;
		pBoard->uhMapperDestListBufferLen = 0;
		up_write(&pBoard->rwsMapMem);
	}

	DecSingleUserLock(pBoard);

	return PciAer_release(INode, fp);
}

/*****************************************************************************/

/*
 * MapFileOps
 * The dispatch table for accesses to the Mapper minor device. Only open,
 * release, mmap and ioctl are supported.
 */

const struct file_operations MapFileOps = {
	.owner = THIS_MODULE,
	.unlocked_ioctl = Map_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = Map_compat_ioctl,
#endif
	.mmap = Map_mmap,
	.open = Map_open,
	.release = Map_release,
};

/*****************************************************************************/

/* EOF */
