/* monitor.c

Monitor minor device code for the Linux character device driver
module for the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2000 & 2003-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

monitor.c $Id: 68434ad91f155bc73fa4544d29a4590385fc2d4a $
*/

#include <linux/version.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/ioctls.h>

#include "pciaerhw.h"
#include "internal.h"

/*****************************************************************************/

static inline unsigned int MonBufferOrder(struct pciaer_board *pBoard)
{
	return get_order(pBoard->uMonFifoDepth * sizeof(u32));
}

/*****************************************************************************/

void MonFifoHalfFullISR(struct pciaer_board *pBoard)
{
	atomic_inc(&pBoard->MonFifoHalfFullIntCount);
}

/*****************************************************************************/

void MonFifoFullISR(struct pciaer_board *pBoard)
{
	atomic_inc(&pBoard->MonFifoFullIntCount);
}

/*****************************************************************************/

/*
 * SetMonitorChannelSelect
 * Called by Mon_ioctl to implement PCIAER_IOC_SET_MON_CH_SEL.The parameter
 * ulMask should be a 4-bit mask containing a 1 for each channel which should be
 * monitored and a 0 for each channel which should be ignored. If any other bits
 * are set, the function returns -EINVAL, otherwise the function returns 0.
 */

static int SetMonitorChannelSelect(struct pciaer_board *pBoard,
				   unsigned long ulMask)
{
	void __iomem *pCfgRegA1;
	unsigned long ulHwLockFlags;
	u32 u32RegContents;

	if (ulMask & ~0xF)
		return -EINVAL;

	pCfgRegA1 = XilinxRegAddress(pBoard, CFG_REG_A1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	u32RegContents = ioread32(pCfgRegA1);
	u32RegContents &= ~CRA1_MON_IGN_CH_MASK;
	u32RegContents |=
	    (ulMask << CRA1_MON_IGN_CH_SHIFT) ^ CRA1_MON_IGN_CH_MASK;
	iowrite32(u32RegContents, pCfgRegA1);
	FlushPciWrites(pBoard);

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	return 0;
}

/*****************************************************************************/

/*
 * GetMonitorChannelSelect
 * Called by Mon_ioctl to implement PCIAER_IOC_GET_MON_CH_SEL. The return value
 * is a 4-bit mask containing a 1 for each channel which is being
 * monitored and a 0 for each channel which is being ignored.
 */

static int GetMonitorChannelSelect(struct pciaer_board *pBoard)
{
	void __iomem *pCfgRegA1;
	u32 u32RegContents;

	pCfgRegA1 = XilinxRegAddress(pBoard, CFG_REG_A1);

	u32RegContents = ioread32(pCfgRegA1);
	return ((u32RegContents & CRA1_MON_IGN_CH_MASK) ^ CRA1_MON_IGN_CH_MASK)
	    >> CRA1_MON_IGN_CH_SHIFT;
}

/*****************************************************************************/

/*
 * SetMonitorTimeLabelFlag
 * Called by Mon_ioctl to implement PCIAER_IOC_SET_MON_TIME_LBL_FLAG. If the
 * argument ulArg is zero, time labels are disabled, otherwise they are enabled.
 * There is no error return, since no value of the parameter is invalid and
 * nothing can go wrong (famous last words).
 */

static void SetMonitorTimeLabelFlag(struct pciaer_board *pBoard,
				    unsigned long ulArg)
{
	void __iomem *pCfgRegA1;
	unsigned long ulHwLockFlags;
	u32 u32RegContents;

	pCfgRegA1 = XilinxRegAddress(pBoard, CFG_REG_A1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	u32RegContents = ioread32(pCfgRegA1);
	if (ulArg == 0L)
		u32RegContents |= CRA1_DIS_MON_TIME_LBL;
	else
		u32RegContents &= ~CRA1_DIS_MON_TIME_LBL;
	iowrite32(u32RegContents, pCfgRegA1);
	FlushPciWrites(pBoard);

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);
}

/*****************************************************************************/

/*
 * GetMonitorTimeLabelFlag
 * Called by Mon_ioctl to implement PCIAER_IOC_GET_MON_TIME_LBL_FLAG. The
 * function returns true or false according to whether time labels are enabled
 * or disabled respectively.
 */

static bool GetMonitorTimeLabelFlag(struct pciaer_board *pBoard)
{
	void __iomem *pCfgRegA1;
	u32 u32RegContents;

	pCfgRegA1 = XilinxRegAddress(pBoard, CFG_REG_A1);

	u32RegContents = ioread32(pCfgRegA1);
	return !(u32RegContents & CRA1_DIS_MON_TIME_LBL);
}

/*****************************************************************************/

/*
 * GetMonitorStats
 * Fills in an internal_stats struct on behalf of debugfs.c and resets the
 * statistics if required.
 */

int GetMonitorStats(struct pciaer_board *pBoard, struct internal_stats *pStats,
		    bool fReset)
{
	unsigned long ulWords;
	unsigned long ulReads;
	int cFifoHalfFullInts;
	int cFifoFullInts;

	if (mutex_trylock(&pBoard->mtxMonitor) == 0)
		return -EAGAIN;

	ulWords = pBoard->ulMonWordsRead;
	ulReads = pBoard->ulMonReads;
	cFifoHalfFullInts = atomic_read(&pBoard->MonFifoHalfFullIntCount);
	cFifoFullInts = atomic_read(&pBoard->MonFifoFullIntCount);

	if (fReset) {
		pBoard->ulMonWordsRead = 0L;
		pBoard->ulMonReads = 0L;
		atomic_sub(cFifoHalfFullInts, &pBoard->MonFifoHalfFullIntCount);
		atomic_sub(cFifoFullInts, &pBoard->MonFifoFullIntCount);
	}

	mutex_unlock(&pBoard->mtxMonitor);

	if (pStats != NULL) {
		pStats->words_transferred = ulWords;
		pStats->user_transfers = ulReads;
		pStats->total_interrupts = cFifoHalfFullInts + cFifoFullInts;
		pStats->overflows_underflows = cFifoFullInts;
		pStats->memory_usage = PAGE_SIZE << MonBufferOrder(pBoard);
	}

	return 0;
}

/*****************************************************************************/

/*
 * GetMonitorDWordsAvailable
 * A helper function for Mon_read which returns the minimum number of DWords
 * available to be read from the Monitor FIFO.
 */

static size_t GetMonitorDWordsAvailable(const struct pciaer_board *pBoard)
{
	unsigned int uFifoFlags;
	size_t DWordsAvailable;

	uFifoFlags = ioread32(XilinxRegAddress(pBoard, STAT_REG_A1));
	DWordsAvailable = 0;
	switch (uFifoFlags & MON_FIFO_FLAGS_MASK) {
	case SRA1_MON_FIFO_HF_ | SRA1_MON_FIFO_FF_:
		/* Not half full, not full and not not empty, i.e. empty */
		DWordsAvailable = 0;
		break;
	case SRA1_MON_FIFO_EF_ | SRA1_MON_FIFO_HF_ | SRA1_MON_FIFO_FF_:
		/* Not empty, not half full and not full, i.e. between empty and half full */
		DWordsAvailable = 1;
		break;
	case SRA1_MON_FIFO_EF_ | SRA1_MON_FIFO_FF_:
		/* Not empty, not full and not not half full, i.e between half full and full */
		DWordsAvailable = pBoard->uMonFifoDepth / 2;
		break;
	case SRA1_MON_FIFO_EF_:
		/* Not empty, not not half full and not not full, i.e full */
		DWordsAvailable = pBoard->uMonFifoDepth;
		break;
	default:
		dev_err(&pBoard->pPciDev->dev,
			"Seemingly impossible combination of Monitor FIFO flags detected: SRA1 octal 0%o\n",
			uFifoFlags & 0xFFFF);
		DWordsAvailable = 0;
		break;
	}
	return DWordsAvailable;
}

/*****************************************************************************/

/*
 * Mon_ioctl
 * This function handles all ioctls on the Monitor minor device.
 */

static long Mon_ioctl(struct file *fp, unsigned int uCmd, unsigned long ulArgs)
{
	struct pciaer_board *pBoard;
	int nError;

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (uCmd == FIONREAD)
		return put_user(GetMonitorDWordsAvailable(pBoard) * sizeof(u32),
				(__user int *)ulArgs);

	nError = CheckIoctlPerm(fp, uCmd);
	if (nError != 0 && uCmd != PCIAER_IOC_RST_FIFO
	    && uCmd != PCIAER_IOC_RST_CNTR_GET_TIME)
		return nError;

	nError = -ENOTTY;

	switch (uCmd) {
	case PCIAER_IOC_GET_VERSION_INFO:
		nError =
		    GetVersionInfo(pBoard,
				   (__user struct pciaer_version_info *) ulArgs);
		break;

	case PCIAER_IOC_SET_CNTR_PERIOD:
		nError = SetCounterPeriod(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_CNTR_PERIOD:
		nError =
		    put_user(GetCounterPeriod(pBoard), (__user int *)ulArgs);
		break;

	case PCIAER_IOC_RST_CNTR_GET_TIME:
		nError =
		    ResetCounterGetTime(pBoard, (__user struct timeval *)ulArgs,
					fp->f_mode);
		break;

	case PCIAER_IOC_GET_CNTR_LAST_RST_TIME:
		nError =
		    GetLastCounterResetTime(pBoard,
					    (__user struct timeval *)ulArgs);
		break;

	case PCIAER_IOC_SET_ARB_CONFIG:
		nError = SetArbiterConfig(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_ARB_CONFIG:
		nError =
		    put_user(GetArbiterConfig(pBoard), (__user int *)ulArgs);
		break;

	case PCIAER_IOC_SET_MON_CH_SEL:
		nError = SetMonitorChannelSelect(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_MON_CH_SEL:
		nError =
		    put_user(GetMonitorChannelSelect(pBoard),
			     (__user int *)ulArgs);
		break;

	case PCIAER_IOC_SET_MON_TIME_LBL_FLAG:
		SetMonitorTimeLabelFlag(pBoard, ulArgs);
		nError = 0;
		break;

	case PCIAER_IOC_GET_MON_TIME_LBL_FLAG:
		nError =
		    put_user((int)GetMonitorTimeLabelFlag(pBoard),
			     (__user int *)ulArgs);
		break;

	case PCIAER_IOC_GET_MON_FIFO_FLAGS:
		/*
		 * Check that the flag bit positions defined for the IOCTL match
		 * those defined for the hardware.
		 */
		BUILD_BUG_ON(PCIAER_IOC_MON_EMPTY != SRA1_MON_FIFO_EF_);
		BUILD_BUG_ON(PCIAER_IOC_MON_HALF_FULL != SRA1_MON_FIFO_HF_);
		BUILD_BUG_ON(PCIAER_IOC_MON_FULL != SRA1_MON_FIFO_FF_);

		nError = put_user(~ioread32(XilinxRegAddress(pBoard, STAT_REG_A1)), (__user int *)ulArgs);	/* Inverts bits for +ve logic result */
		break;

	case PCIAER_IOC_RST_FIFO:
		if (fp->f_mode & FMODE_READ) {
			ResetFifo(pBoard, CRA1_MON_FIFO_MR_);
			nError = 0;
		} else
			nError = -EACCES;
		break;

	default:
		nError = -ENOTTY;
		break;
	}
	return nError;
}

/*****************************************************************************/

/*
 * Mon_poll
 * Mon_poll is called by the kernel when it deciding whether to put the current
 * process to sleep as a result of a user's poll or select call. If any of the
 * returned flags matches a state which the user is interested in, the process
 * will not sleep; otherwise it may (if the user's timeout is not reached and
 * there are no signals) sleep interruptibly on the queue referred to in the
 * call to poll_wait.
 * @@@ Currently there is no interrupt support implemented, so there can be no
 * wake-up, so there is no wait queue and the only way to get back to the
 * running state once sleeping as a result of a user's poll or select is either
 * to timeout or to receive a signal.
 */

static unsigned int Mon_poll(struct file *fp, struct poll_table_struct *ptsp)
{
	struct pciaer_board *pBoard;
	unsigned int uPollFlags;

	pBoard = (struct pciaer_board *) (fp->private_data);

	poll_wait(fp, NULL, ptsp);	/*
					 * @@@ The 2nd parameter will eventually
					 * be a pointer to the Monitor read wait
					 * queue.
					 */

	/* Always writeable without blocking, since write always return -EINVAL */

	uPollFlags = POLLOUT | POLLWRNORM;

	/* Iff the FIFO is not empty, then we're also readable without blocking */

	if (ioread32(XilinxRegAddress(pBoard, STAT_REG_A1)) & SRA1_MON_FIFO_EF_)
		uPollFlags |= (POLLIN | POLLRDNORM);

	return uPollFlags;
}

/*****************************************************************************/

/*
 * Mon_read
 */

static ssize_t Mon_read(struct file *fp, __user char *pchUserBuffer,
			size_t BytesRequested, loff_t *pOffset)
{
	size_t DWordsToRead;
	struct pciaer_board *pBoard;
	size_t DWordsAvailable;
	size_t BytesToRead;
	unsigned long ulHwLockFlags;
	unsigned long ulBytesNotCopied;

	if (unlikely(BytesRequested < sizeof(u32) && BytesRequested != 0))
		return -EINVAL;

	DWordsToRead = BytesRequested / sizeof(u32);

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (fp->f_flags & O_NONBLOCK) {
		if (mutex_trylock(&pBoard->mtxMonitor) == 0)
			return -EAGAIN;
	} else {
		if (mutex_lock_interruptible(&pBoard->mtxMonitor) != 0)
			return -ERESTARTSYS;
	}

	pBoard->ulMonReads++;

	if (unlikely(DWordsToRead == 0)) {
		mutex_unlock(&pBoard->mtxMonitor);
		return 0;
	}

	DWordsAvailable = GetMonitorDWordsAvailable(pBoard);

	if (DWordsAvailable == 0) {
		mutex_unlock(&pBoard->mtxMonitor);

		if (fp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		else
			return 0;	/* @@@ Should block here */
	}

	if (DWordsAvailable < DWordsToRead)
		DWordsToRead = DWordsAvailable;

	BytesToRead = DWordsToRead * sizeof(u32);

	read_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	memcpy_fromio(pBoard->pMonitorBuffer, FifoAddress(pBoard), BytesToRead);
	read_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	pBoard->ulMonWordsRead += DWordsToRead;

	ulBytesNotCopied =
	    __copy_to_user(pchUserBuffer, pBoard->pMonitorBuffer, BytesToRead);

	mutex_unlock(&pBoard->mtxMonitor);

	if (unlikely(ulBytesNotCopied != 0L))
		return -EFAULT;
	else
		return BytesToRead;
}

/*****************************************************************************/

/*
 * Mon_open
 * Supplementary to PciAerMain_open, fail open with ENXIO if all the relevant
 * hardware is not available, EOPNOTSUPP if the Monitor FIFO appears to be
 * 0-depth, EBUSY if another user is using the board, or ENOMEM if no buffer
 * can be allocated.
 */

static int Mon_open(struct inode *INode, struct file *fp)
{
	struct pciaer_board *pBoard;

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (S5920OpRegAddress(pBoard, 0) == NULL
	    || XilinxRegAddress(pBoard, 0) == NULL
	    || FifoAddress(pBoard) == NULL)
		return -ENXIO;

	if (pBoard->uMonFifoDepth == 0)
		return -EOPNOTSUPP;

	if (IncSingleUserLock(pBoard) != 0)
		return -EBUSY;

	if (fp->f_flags & O_NONBLOCK) {
		if (mutex_trylock(&pBoard->mtxMonitor) == 0) {
			DecSingleUserLock(pBoard);
			return -EAGAIN;
		}
	} else {
		if (mutex_lock_interruptible(&pBoard->mtxMonitor) != 0) {
			DecSingleUserLock(pBoard);
			return -ERESTARTSYS;
		}
	}

	if (pBoard->pMonitorBuffer == NULL) {
		pBoard->pMonitorBuffer =
		    (void *)__get_free_pages(GFP_KERNEL,
					     MonBufferOrder(pBoard));
		if (pBoard->pMonitorBuffer == NULL) {
			mutex_unlock(&pBoard->mtxMonitor);
			DecSingleUserLock(pBoard);
			return -ENOMEM;
		}
	}

	pBoard->nMonitorOpens++;

	mutex_unlock(&pBoard->mtxMonitor);

	INode->i_blkbits = ilog2(pBoard->uMonFifoDepth * sizeof(u32) / 2);

	return nonseekable_open(INode, fp);
}

/*****************************************************************************/

/*
 * Mon_release
 * Decrement the monitor open count, and if it's reached 0, free the monitor
 * buffer. Finally decrement the single user lock and call PciAer_release.
 */

static int Mon_release(struct inode *INode, struct file *fp)
{
	struct pciaer_board *pBoard;

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (fp->f_flags & O_NONBLOCK) {
		if (mutex_trylock(&pBoard->mtxMonitor) == 0)
			return -EAGAIN;
	} else {
		if (mutex_lock_interruptible(&pBoard->mtxMonitor) != 0)
			return -ERESTARTSYS;
	}

	if (--pBoard->nMonitorOpens == 0) {
		free_pages((unsigned long)pBoard->pMonitorBuffer,
			   MonBufferOrder(pBoard));
		pBoard->pMonitorBuffer = NULL;
	}

	mutex_unlock(&pBoard->mtxMonitor);

	DecSingleUserLock(pBoard);

	return PciAer_release(INode, fp);
}

/*****************************************************************************/

/*
 * MonFileOps
 * The dispatch table for accesses to the non-seekable, read-only Monitor minor
 * device.
 */

const struct file_operations MonFileOps = {
	.owner = THIS_MODULE,
	.read = Mon_read,
	.poll = Mon_poll,
	.unlocked_ioctl = Mon_ioctl,
	.open = Mon_open,
	.release = Mon_release,
};

/*****************************************************************************/

/* EOF */
