/* pciaer.h */

#ifndef PCIAER_H
#define PCIAER_H

/*
This header file represents the public interface to the Linux character
device driver module for the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2000, 2003-2007 & 2010-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

pciaer.h $Id: 05a2a1b8f04e62c7c219fde5e0b74007e92229bf $
*/

#ifdef __KERNEL__
#include <linux/ioctl.h>
#include <linux/types.h>
#else
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#endif

/* Maximum number of boards the driver can handle */

#define PCIAER_MAX_BOARDS	4

/*
 * Minor numbers.
 * The bottom 2 bits of the minor number indicate the sub-device (S5920 NVRAM,
 * mapper, monitor or sequencer); the remaining bits determine which board to
 * address.
 */

#define PCIAER_MINOR_TYPE_MASK	0x03
#define PCIAER_MINOR_NUM_MASK	0xfc
#define PCIAER_MINOR_NUM_SHIFT	2

#define PCIAER_MINOR_TYPE_NVR	0
#define PCIAER_MINOR_TYPE_MAP	1
#define PCIAER_MINOR_TYPE_MON	2
#define PCIAER_MINOR_TYPE_SEQ	3

/* Ioctl magic number for this device */

#define PCIAER_IOC_MAGIC	'='

/* Ioctls for monitor */

/*
 * The following two ioctls deal with which channels are actually monitored. The
 * 'SET' ioctl uses its argument directly to specify a 4-bit mask containing a 1
 * for each channel which should be monitored, and a 0 for each channel which
 * should not be monitored. The 'GET' ioctl fills the int pointed to by its
 * argument with such a bit mask according to the current status. The 'SET'
 * ioctl requires write access to the device; the 'GET' ioctl requires read
 * access.
 */

#define PCIAER_IOC_SET_MON_CH_SEL	_IO(PCIAER_IOC_MAGIC, 'C')
#define PCIAER_IOC_GET_MON_CH_SEL	_IOR(PCIAER_IOC_MAGIC, 'D', int)

/*
 * The following two ioctls deal with the Monitor's Time Label flag. If the
 * argument of the 'SET' ioctl is zero, time labels will be disabled, otherwise
 * they will be enabled. The 'GET' ioctl sets the int pointed to by its argument
 * to 0 or 1 according to the current status of the flag. The 'SET' ioctl
 * requires write access to the device; the 'GET' ioctl requires read access.
 */

#define PCIAER_IOC_SET_MON_TIME_LBL_FLAG	_IO(PCIAER_IOC_MAGIC, 'L')
#define PCIAER_IOC_GET_MON_TIME_LBL_FLAG	_IOR(PCIAER_IOC_MAGIC, 'M', int)

/*
 * The PCIAER_IOC_GET_MON_FIFO_FLAGS ioctl call fills the int pointed to by its
 * argument with a status word containing, among other bits, a positive logic
 * copy of the Monitor FIFO flag bits as defined below.
 * This call can be used to determine whether a subsequent read call
 * might be able to read a half or a full FIFO's worth of data. Just how much
 * that is depends on the depth of the FIFO. This ioctl requires read access to
 * the device.
*/

#define PCIAER_IOC_GET_MON_FIFO_FLAGS	_IOR(PCIAER_IOC_MAGIC, 'f', int)
#define PCIAER_IOC_MON_EMPTY		0x0001
#define PCIAER_IOC_MON_HALF_FULL	0x0002
#define PCIAER_IOC_MON_FULL		0x0004

/* Ioctls for sequencer */

/*
 * The PCIAER_IOC_GET_SEQ_FIFO_FLAGS ioctl call fills the int pointed to by its
 * argument with a status word containing, among other bits, a positive logic
 * copy of the Sequencer FIFO flag bits as defined below.
 * This ioctl requires read access to the device.
 */

#define PCIAER_IOC_GET_SEQ_FIFO_FLAGS	_IOR(PCIAER_IOC_MAGIC, 'g', int)
#define PCIAER_IOC_SEQ_EMPTY			0x0008
#define PCIAER_IOC_SEQ_HALF_FULL		0x0010
#define PCIAER_IOC_SEQ_FULL			0x0020

/*
 * The following two ioctls deal with connecting the sequencer output to one of
 * the arbiter input channels. The 'SET' ioctl uses its argument directly: an
 * argument of 0, 1, 2 or 3 means connect the sequencer to the given arbiter
 * input.The 'GET' ioctl fills the int pointed to by its argument with 0, 1, 2
 * or 3 according to the current status. The 'SET' ioctl requires write access
 * to the device; the 'GET' ioctl requires read access.
 * NB. These are now also callable on the mapper and have been moved to the
 * 'Common ioctls' section of the spec., as of spec. version 2.2.
 */

#define PCIAER_IOC_SET_SEQ_ARB_CHANNEL	_IO(PCIAER_IOC_MAGIC, 'A')
#define PCIAER_IOC_GET_SEQ_ARB_CHANNEL	_IOR(PCIAER_IOC_MAGIC, 'B', int)

/* Ioctls for mapper */

/*
 * The following five ioctls all use their argument to point to a
 * pciaer_mapping structure where dest_ae, when valid, points to an array of
 * at least dest_count unsigned shorts.
 */

struct pciaer_mapping {
	unsigned short source_ae;
	unsigned short dest_count;
	unsigned short *dest_ae;
};

#define PCIAER_IOC_SET_MAP_MAPPING		_IOW(PCIAER_IOC_MAGIC, 'E', struct pciaer_mapping)
#define PCIAER_IOC_GET_MAP_MAPPING_COUNT	_IOWR(PCIAER_IOC_MAGIC, 'F', struct pciaer_mapping)
#define PCIAER_IOC_GET_MAP_MAPPING		_IOWR(PCIAER_IOC_MAGIC, 'G', struct pciaer_mapping)
#define PCIAER_IOC_MAP_ADD_TO_MAPPING		_IOW(PCIAER_IOC_MAGIC, 'H', struct pciaer_mapping)
#define PCIAER_IOC_MAP_DEL_FROM_MAPPING		_IOW(PCIAER_IOC_MAGIC, 'I', struct pciaer_mapping)

/*
 * In the 'SET_MAP_MAPPING' case, all fields must be valid on entry, and the
 * call establishes a new mapping or replaces the existing mapping from the
 * given source_ae to the given list of dest_ae, or if dest_count is 0, deletes
 * any existing mapping for the given source_ae. This ioctl requires write
 * access to the device.
 * In the 'GET_MAP_MAPPING_COUNT' case, only the source_ae field must be valid
 * on entry and the driver fills in the dest_count field. This ioctl requires
 * read access to the device.
 * In the 'GET_MAP_MAPPING' case, the driver fills the dest_ae array with the
 * list of destinations for the given source_ae and replaces dest_count with
 * the actual number of destinations copied to dest_ae. If on entry dest_count
 * is less than the number of destination addresses there currently are for the
 * given source_ae, EINVAL is returned. This ioctl requires read access to the
 * device.
 * In the 'ADD_TO_MAP_MAPPING' and 'DEL_FROM_MAP_MAPPING' cases, all fields must
 * be valid on entry, and the given list of dest_ae will be added or deleted
 * respectively from any existing list for the given source_ae. These ioctls
 * require write access to the device.
 * Both PCIAER_IOC_SET_MAP_MAPPING and PCIAER_IOC_ADD_TO_MAP_MAPPING may fail
 * returning ENOSPC due to there being insufficient free contiguous Mapper
 * memory available, in which case PCIAER_IOC_MAP_COMPACT may help.
 * In all cases, the address-event FFFFh is reserved, and attempting to use
 * this will result in the error EINVAL.
 */

/*
 * PCIAER_IOC_FIND_NEXT_MAP_MAPPING replaces the int pointed to by its argument
 * with the next source address-event for which a mapping exists after the one
 * specified. This ioctl requires read access to the device.
 */

#define PCIAER_IOC_FIND_NEXT_MAP_MAPPING	_IOWR(PCIAER_IOC_MAGIC, 'J', int)

/*
 * PCIAER_IOC_GET_MAP_MAPPED_MAP fills the 8K of memory pointed to by its
 * argument with a bit vector in which a 0 represents a source address-event for
 * which a mapping does not exist and a 1 represents one for which a mapping
 * does exist. This ioctl requires read access to the device.
 */

#define PCIAER_IOC_GET_MAP_MAPPED_MAP		_IO(PCIAER_IOC_MAGIC, 'K')

/*
 * PCIAER_IOC_MAP_CLEAR clears the Mapper's memory to a state in which no
 * address-events are mapped. No argument is used. This ioctl requires write
 * access to the device.
 */

#define PCIAER_IOC_MAP_CLEAR			_IO(PCIAER_IOC_MAGIC, 'X')

/*
 * PCIAER_IOC_MAP_COMPACT forces maximal compaction of the Mapper's address
 * table. No argument is used. This ioctl requires write access to the device.
 */

#define PCIAER_IOC_MAP_COMPACT			_IO(PCIAER_IOC_MAGIC, 'Z')

/*
 * PCIAER_IOC_GET_MAP_FREE_SPACE fills the unsigned int pointed to by its
 * argument with the number of free words in the mapper SRAM. This gives an
 * indication of how many more destination addresses could be stored, but the
 * free space may be fragmented and need compaction before it is used. This
 * ioctl requires read access to the device.
 */

#define PCIAER_IOC_GET_MAP_FREE_SPACE		_IOR(PCIAER_IOC_MAGIC, '[', unsigned int)

/*
 * PCIAER_IOC_GET_MAP_FIFO_FLAGS fills the int pointed to by its argument with a
 * status word containing, among other bits, a positive logic copy of the Mapper
 * FIFO flag bits as defined below.
 * This ioctl requires read access to the device.
 */

#define PCIAER_IOC_GET_MAP_FIFO_FLAGS		_IOR(PCIAER_IOC_MAGIC, 'e', int)
#define PCIAER_IOC_MAP_EMPTY			0x0040
#define PCIAER_IOC_MAP_FULL			0x0080

/*
 * The following two ioctls deal with the configuration of the Mapper output.
 * The 'SET' ioctl uses its argument directly to specify one of the values
 * PCIAER_IOC_MAP_OUT_PASS_THRU, PCIAER_IOC_MAP_OUT_1_TO_1 or
 * PCIAER_IOC_MAP_OUT_1_TO_MANY. The 'GET' ioctl fills the int pointed to by its
 * argument with one of these values according to the current status. The 'SET'
 * ioctl requires write access to the device; the 'GET' ioctl requires read
 * access.
 */

#define PCIAER_IOC_SET_MAP_OUT_CONFIG		_IO(PCIAER_IOC_MAGIC, 'P')
#define PCIAER_IOC_GET_MAP_OUT_CONFIG		_IOR(PCIAER_IOC_MAGIC, 'Q', int)
#define PCIAER_IOC_MAP_OUT_PASS_THRU		1
#define PCIAER_IOC_MAP_OUT_1_TO_1		2
#define PCIAER_IOC_MAP_OUT_1_TO_MANY		3

/*
 * The following two ioctls deal with the configuration of the AER demultiplexer
 * on the Mapper output. The 'SET' ioctl uses its argument directly to specify
 * one of the values PCIAER_IOC_MAP_DEMUX_DISABLED, PCIAER_IOC_MAP_DEMUX_0_16,
 * PCIAER_IOC_MAP_DEMUX_1_15 or PCIAER_IOC_MAP_DEMUX_2_14. The names of these
 * constants reflect the way the 16 bits are split between channel number and
 * actual AE bits. The 'GET' ioctl fills the int pointed to by its argument with
 * one of these values according to the current status. The 'SET' ioctl requires
 * write access to the device; the 'GET' ioctl requires read access.
 */

#define PCIAER_IOC_SET_MAP_DEMUX_CONFIG		_IO(PCIAER_IOC_MAGIC, 'S')
#define PCIAER_IOC_GET_MAP_DEMUX_CONFIG		_IOR(PCIAER_IOC_MAGIC, 'T', int)
#define PCIAER_IOC_MAP_DEMUX_DISABLED		0
#define PCIAER_IOC_MAP_DEMUX_0_16		1
#define PCIAER_IOC_MAP_DEMUX_1_15		2
#define PCIAER_IOC_MAP_DEMUX_2_14		4

/*
 * The following two ioctls deal with the type of AER protocol used on the
 * Mapper output. The 'SET' ioctl uses its argument directly to specify one of
 * the values PCIAER_IOC_MAP_AER_P2P or PCIAER_IOC_MAP_AER_SCX indicating the
 * classic point to point, four phase handshake protocol or the shared bus, data
 * only driven on acknowledge protocol respectively. The 'GET' ioctl fills the
 * int pointed to by its argument with one of these values according to the
 * current status. The 'SET' ioctl requires write access to the device; the
 * 'GET' ioctl requires read access.
 */

#define PCIAER_IOC_SET_MAP_AER_PROTOCOL		_IO(PCIAER_IOC_MAGIC, 'U')
#define PCIAER_IOC_GET_MAP_AER_PROTOCOL		_IOR(PCIAER_IOC_MAGIC, 'W', int)
#define PCIAER_IOC_MAP_AER_P2P			0
#define PCIAER_IOC_MAP_AER_SCX			1

/*
 * The following two ioctls deal with which channels are processed by the
 * Mapper. The 'SET' ioctl uses its argument directly to specify a 4-bit mask
 * containing a 1 for each channel which should be processed, and a 0 for each
 * channel which should be ignored. The 'GET' ioctl fills the int pointed to by
 * its argument with such a bit mask according to the current status. The 'SET'
 * ioctl requires write access to the device; the 'GET' ioctl requires read
 * access.
 */

#define PCIAER_IOC_SET_MAP_CH_SEL	_IO(PCIAER_IOC_MAGIC, 'Y')
#define PCIAER_IOC_GET_MAP_CH_SEL	_IOR(PCIAER_IOC_MAGIC, 'y', int)

/*
 * The following three ioctls deal with enabling, disabling and getting the
 * current enabled/disabled state of the Mapper output respectively. The
 * ENABLE' ioctl takes no argument and requires write access to the device. The
 * 'DISABLE' ioctl uses its argument directly. If non-zero, the call will not
 * return until the mapper is idle. It requires write access to the device.
 * The 'GET' ioctl fills the int pointed to by its argument with a combination
 * of the bits PCIAER_IOC_MAP_OUT_ENABLED and PCIAER_IOC_MAP_OUT_BUSY, and
 * requires read access to the device.
 */

#define PCIAER_IOC_ENABLE_MAP_OUT	_IO(PCIAER_IOC_MAGIC, 'b')
#define PCIAER_IOC_DISABLE_MAP_OUT	_IO(PCIAER_IOC_MAGIC, 'c')
#define PCIAER_IOC_GET_MAP_OUT_STATE	_IOR(PCIAER_IOC_MAGIC, 'd', int)
#define PCIAER_IOC_MAP_OUT_ENABLED	0x0010
#define PCIAER_IOC_MAP_OUT_BUSY		0x0001

/* Common ioctls */

/*
 * Ioctl for getting version info. Fills the pciaer_version_info structure
 * pointed to by its argument with the version number of the driver, the
 * contents of the release registers of the two FPGAs and the contents of the
 * S5920's revision identification register (RID). The FPGA release and driver
 * version information is divided within the respective unsigned short into a
 * high byte containing a major version number and low byte containing a minor
 * revision number. This ioctl can be called on any of the minor devices and
 * requires read access to the device.
 */

struct pciaer_version_info {
	unsigned short driver_version;
	unsigned short fpga1_release;
	unsigned short fpga2_release;
	unsigned char s5920_revision_id;
};

#define PCIAER_IOC_GET_VERSION_INFO	_IOR(PCIAER_IOC_MAGIC, 'V', struct pciaer_version_info)

/*
 * Ioctls for AER Clock period selection. The following two ioctls deal with the
 * rate at which the interval counter is incremented. The 'SET' ioctl uses its
 * argument directly to specify the period in microseconds between counter
 * updates. The only valid values are 1, 10, 50 and 100. The 'GET' ioctl fills
 * the int pointed to by its argument with one of these values according to the
 * current status. These ioctls can be called on either monitor or sequencer.
 * The 'SET' ioctl requires write access to the device; the 'GET' ioctl requires
 * read access.
 */

#define PCIAER_IOC_SET_CNTR_PERIOD		_IO(PCIAER_IOC_MAGIC, 'k')
#define PCIAER_IOC_GET_CNTR_PERIOD		_IOR(PCIAER_IOC_MAGIC, 'l', int)

/*
 * The following two ioctls deal with resetting the counter to 0 and can be
 * called on either Monitor or Sequencer. The first resets the counter and fills
 * the timeval struct pointed to by its argument with the time at which it did
 * so (unless the pointer is NULL). This ioctl requires read and write access
 * to the device. The second ioctl simply fills the timeval struct pointed
 * to by its argument with the time at which the counter was last reset (without
 * causing a reset). This requires read access to the device.
 */

#define PCIAER_IOC_RST_CNTR_GET_TIME		_IOR(PCIAER_IOC_MAGIC, 'R', struct timeval)
#define PCIAER_IOC_GET_CNTR_LAST_RST_TIME	_IOR(PCIAER_IOC_MAGIC, '\\', struct timeval)

/*
 * Ioctls for arbiter configuration. The following two ioctls deal with how
 * many devices are multiplexed into the arbiter. The 'SET' ioctl uses its
 * argument directly to specify one of the values listed below. The names of
 * these constants reflect the way the 16 available bits are split between
 * device number and actual AE bits. The 'GET' ioctl fills the int pointed to
 * by its argument with one of these values according to the current status.
 * These ioctls can be called on either monitor or mapper. The 'SET' ioctl
 * requires write access to the device; the 'GET' ioctl requires read access.
 * NB. PCIAER_IOC_ARB_DISABLED is not supported by current hardware.
 */

#define PCIAER_IOC_SET_ARB_CONFIG		_IO(PCIAER_IOC_MAGIC, 'm')
#define PCIAER_IOC_GET_ARB_CONFIG		_IOR(PCIAER_IOC_MAGIC, 'n', int)
#define PCIAER_IOC_ARB_DISABLED			0
#define PCIAER_IOC_ARB_0_16			1
#define PCIAER_IOC_ARB_1_15			2
#define PCIAER_IOC_ARB_2_14			4

/*
 * The PCIAER_IOC_RST_FIFO ioctl call resets the FIFO associated with the
 * monitor, sequencer or mapper, depending on the device on which it is called.
 * No argument is used. This ioctl requires read access on the monitor, but
 * write access on the sequencer and mapper.
 */

#define PCIAER_IOC_RST_FIFO		_IO(PCIAER_IOC_MAGIC, 'h')

/* Bits in Monitor FIFO words, as defined in table 3 of the board spec. */

#define MONITOR_DWORD_TYPE_MASK		0x00030000
#define MONITOR_DWORD_DATA_MASK		0x0000ffff
#define MONITOR_DWORD_AER_ADDR		0x00000000
#define MONITOR_DWORD_TIME_HI		0x00010000
#define MONITOR_DWORD_TIME_LO		0x00020000
#define MONITOR_DWORD_ERROR		0x00030000

/* Bits in Sequencer FIFO words, as defined in table 4 of the board spec. */

#define SEQUENCER_DWORD_TYPE_MASK	0x00030000
#define SEQUENCER_DWORD_DATA_MASK	0x0000ffff
#define SEQUENCER_DWORD_END_SEQUENCE	0x00000000
#define SEQUENCER_DWORD_AER_ADDR	0x00010000
#define SEQUENCER_DWORD_DELAY		0x00020000
#define SEQUENCER_DWORD_WAIT_TIME	0x00030000

/*
 * Bits in Mapper SRAM words, as defined in table 5 of the board spec.
 *
 * Bits 23,22 | Bit 21 | Bits 20..0
 * -----------+--------+---------------------------------------------------
 *     00     |    x   | Pointer to start word in the spike table (21 bits)
 *     01     |    x   | Not used [mapper does nothing]
 *     10     |    x   | Not used [mapper does nothing]
 *     11     |    x   | Address [AE] of spike (only bits 15..0 are used)
 */

#define MAPPER_PTR_TYPE_MASK		0x00c00000
#define MAPPER_PTR_TYPE_INDIRECT	0x00000000
#define MAPPER_PTR_TYPE_NOT_USED_1	0x00400000
#define MAPPER_PTR_TYPE_NOT_USED_2	0x00800000
#define MAPPER_PTR_TYPE_DIRECT		0x00c00000
#define MAPPER_PTR_NOT_USED_MASK	0x00200000
#define MAPPER_PTR_INDIRECT_MASK	0x001fffff
#define MAPPER_PTR_DIRECT_MASK		0x0000ffff

/* For backward compatibility */

#define MAPPER_PTR_TYPE_NOT_USED	MAPPER_PTR_TYPE_NOT_USED_2

/*
 * Pattern defining the end of a series of address-events to be written out by
 * the mapper.
 */

#define MAPPER_END_LABEL	0xffff

/* Structure of Mapper SRAM, divided into so-called Pointer Table and Spike Table */

#define PCIAER_MAPPER_SRAM_LENGTH	(2 * 1024 * 1024)
#define PCIAER_MAPPER_PTR_TBL_LENGTH   	(64 * 1024)
#define PCIAER_MAPPER_SPIKE_TBL_LENGTH	(PCIAER_MAPPER_SRAM_LENGTH - PCIAER_MAPPER_PTR_TBL_LENGTH)

struct pciaer_mapper_sram {
	unsigned int pointer_table[PCIAER_MAPPER_PTR_TBL_LENGTH];
	unsigned int spike_table[PCIAER_MAPPER_SPIKE_TBL_LENGTH];
};

/* The following ioctls apply to the S5920 config NVRAM device: */

#define PCIAER_IOC_GET_NVR_CACHE_STATUS	_IOR(PCIAER_IOC_MAGIC, '3', int)
#define PCIAER_IOC_REVERT_NVR_CACHE	_IO(PCIAER_IOC_MAGIC, '4')

/* Bits defined in NVRAM cache status: */

#define PCIAER_IOC_NVR_CACHE_DIRTY	0x00000001

/* structs and typedefs for user space */

#ifndef __KERNEL__

struct pciaer_stats {
	size_t size;
	unsigned long words_transferred;
	unsigned long user_transfers;
	unsigned long total_interrupts;
	unsigned long overflows_underflows;
	unsigned long timeouts;
	unsigned long memory_usage;
};

struct pciaer_stats_2 {
	size_t size;
	unsigned long words_transferred;
	unsigned long user_transfers;
	unsigned long total_interrupts;
	unsigned long overflows_underflows;
	unsigned long timeouts;
	unsigned long memory_usage;
	struct timeval timestamp;
};

typedef struct pciaer_mapping       pciaer_mapping_t;
typedef struct pciaer_version_info  pciaer_version_info_t;
typedef struct pciaer_stats         pciaer_stats_t;
typedef struct pciaer_stats_2       pciaer_stats_t_2;
typedef struct pciaer_mapper_sram   pciaer_mapper_sram_t;

#endif

#endif

/* EOF */
