/* pciaerhw.h */
#ifndef PCIAERHW_H
#define PCIAERHW_H

/*
This header file defines register offsets and bitmasks relating to
the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2000, 2003, 2004, 2006, 2008 & 2010-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

pciaerhw.h $Id: 8845fc93efe765223f5e8fa49b62fea7cd6f4ed0 $
*/

/* PCI bus speed */

#define PCIAER_PCI_MHZ		33

/* PCI Vendor and Device IDs */

/* Currently just the values assigned to the AMCC S5920 */

#define PCIAER_PCI_VENDOR_ID	0x10E8
#define PCIAER_PCI_DEVICE_ID	0x5920

#define PCIAER_PCI_DEVICE_NAME	"Applied Micro Circuits Corp. S5920"

/* Base Address Registers and Memory regions */

/*
 * Maximum number of base addresses provided by S5920 and number of base address
 * registers used.
 */

#define MAX_NUM_BADR	5
#define NUM_USED_BADR	4

/* Base address register functional assignments */

#define S5920OPREGS_BADR	0
#define XILINXREGS_BADR		1
#define MONSEQFIFO_BADR		2
#define MAPPERSRAM_BADR		3
#define UNUSED_BADR		4

/* Sizes of each region */

#define S5920OPREGS_SIZE	128	/* 128 bytes (32 DWords) */
#define XILINXREGS_SIZE		128	/* 128 bytes (32 DWords) */
/*      MONSEQFIFO_SIZE		Depends on size of FIFOs fitted, see below */
#define MAPPERSRAM_SIZE		(2 * 1024 * 1024 * sizeof(u32))	/* 2M DWords */

/* Wait states required in each region */

#define XILINXREGS_WS	2
#define MONSEQFIFO_WS	1
#define MAPPERSRAM_WS	2

/* Reset delays */

/* Number of milliseconds for which SYSRST# should be active to reset board */

#define XILINX_RESET_DELAY_MS	200

/* Number of milliseconds needed to reset FIFO */

#define FIFO_RESET_DELAY_MS	 10

/*
 * Number of microseconds needed after reset of S5920
 * This only needs to be a couple of microseconds to cover a couple of board
 * (not PCI) clock cycles, according to Vittorio on 27.11.03.
 */

#define S5920_RESET_DELAY_US	2

/* Definitions of offsets of PCI-AER board registers */

/* FPGA1 registers */

#define STAT_REG_A1  0x00	/* Status Register A1 offset */
#define CFG_REG_A1   0x04	/* Configuration Register A1 offset */
#define CFG_REG_B1   0x08	/* Configuration Register B1 offset */
#define CFG_REG_C1   0x0C	/* Configuration Register C1 offset */
#define REL_REG_1    0x20	/* Release Register of FPGA1 offset */

/* FPGA2 registers */

#define STAT_REG_A2  0x30	/* Status Register A2 offset */
#define CFG_REG_A2   0x34	/* Configuration Register A2 offset */
#define REL_REG_2    0x3C	/* Release Register of FPGA2 offset */

/* Release registers for FPGA1 & FPGA2 */

#define RR_RELEASE	0x0000FFFF
#define RR_VERSION	0x0000FF00
#define RR_REVISION	0x000000FF

#define EXPECTED_FPGA1_RELEASE	0x4202
#define EXPECTED_FPGA2_RELEASE	0x4203

/* Status Register A1 */

/* FIFO flags */

#define SRA1_MON_FIFO_EF_      0x0001	/*  */
#define SRA1_MON_FIFO_HF_      0x0002	/*  */
#define SRA1_MON_FIFO_FF_      0x0004	/*  */
#define SRA1_SEQ_FIFO_EF_      0x0008	/*  */
#define SRA1_SEQ_FIFO_HF_      0x0010	/*  */
#define SRA1_SEQ_FIFO_FF_      0x0020	/*  */
#define SRA1_MAP_FIFO_EF_      0x0040	/*  */
#define SRA1_MAP_FIFO_FF_      0x0080	/*  */
#define SEQ_FIFO_FLAGS_MASK	(SRA1_SEQ_FIFO_EF_|SRA1_SEQ_FIFO_HF_|SRA1_SEQ_FIFO_FF_)
#define MON_FIFO_FLAGS_MASK	(SRA1_MON_FIFO_EF_|SRA1_MON_FIFO_HF_|SRA1_MON_FIFO_FF_)
#define MAP_FIFO_FLAGS_MASK	(SRA1_MAP_FIFO_EF_|SRA1_MAP_FIFO_FF_)

/* Interrupt flags */

#define SRA1_INT_FH_MON     0x0100	/* Monitor FIFO Half-full interrupt flag */
#define SRA1_INT_FF_MON     0x0200	/* Monitor FIFO Full interrupt flag */
#define SRA1_INT_FE_SEQ     0x0400	/* Sequencer FIFO Empty interrupt flag */
#define SRA1_INT_FH_SEQ     0x0800	/* Sequencer FIFO Half-full interrupt flag */
#define SRA1_INT_FF_MAP     0x1000	/* Mapper FIFO Full interrupt flag */
#define SRA1_INT_FLAGS_MASK	(SRA1_INT_FH_MON|SRA1_INT_FF_MON|SRA1_INT_FE_SEQ|SRA1_INT_FH_SEQ|SRA1_INT_FF_MAP)

/* Configuration Register A1  */

/* MONITOR control bits */

#define CRA1_MON_DISABLE      0x0001	/*  */
#define CRA1_MON_IGN_CH_MASK  0x001E
#define CRA1_MON_IGN_CH_SHIFT 1
#define CRA1_MON_IGN_CH_0     0x0002
#define CRA1_MON_IGN_CH_1     0x0004
#define CRA1_MON_IGN_CH_2     0x0008
#define CRA1_MON_IGN_CH_3     0x0010

#define CRA1_DIS_MON_TIME_LBL 0x0020	/*  */
#define CRA1_MON_FIFO_MR_     0x0040	/*  */
#define CRA1_MON_FIFO_RT_     0x0080	/*  */

/* SEQUENCER control bits */

#define CRA1_SEQ_DISABLE      0x0100	/*  */
#define CRA1_SEQ_CH_SEL_MASK  0x0600
#define CRA1_SEQ_CH_SEL_0     0x0000
#define CRA1_SEQ_CH_SEL_1     0x0200
#define CRA1_SEQ_CH_SEL_2     0x0400
#define CRA1_SEQ_CH_SEL_3     0x0600

#define CRA1_SEQ_FIFO_MR_     0x0800	/*  */
#define CRA1_SEQ_FIFO_RT_     0x1000	/*  */

/* MAPPER-Input control bits */

#define CRA1_MAP_IN_DISABLE   0x2000	/*  */
#define CRA1_MAP_IN_FIFO_MR_  0x4000

/* Configuration Register B1 */

#define CRB1_TIME_DISABLE     0x0001	/*  */
#define CRB1_TIME_RESET_      0x0002	/*  */
#define CRB1_AER_CLOCK_MASK   0x000C
#define CRB1_AER_CLOCK_1US    0x0000	/*  */
#define CRB1_AER_CLOCK_10US   0x0004	/*  */
#define CRB1_AER_CLOCK_50US   0x0008	/*  */
#define CRB1_AER_CLOCK_100US  0x000C	/*  */
#define CRB1_ARB_CFG_MASK     0x0030
#define CRB1_ARB_DISABLE      0x0000	/*  */
#define CRB1_ARB_1_16         0x0010	/*  */
#define CRB1_ARB_2_15         0x0020	/*  */
#define CRB1_ARB_4_14         0x0030	/*  */
#define CRB1_MAP_IGN_CH_MASK  0x03C0
#define CRB1_MAP_IGN_CH_SHIFT 6
#define CRB1_MAP_IGN_CH_0     0x0040
#define CRB1_MAP_IGN_CH_1     0x0080
#define CRB1_MAP_IGN_CH_2     0x0100
#define CRB1_MAP_IGN_CH_3     0x0200

/* Configuration Register C1 */

#define CRC1_DIS_INT_FH_MON	0x0001	/* Disable monitor FIFO Half-full interrupt */
#define CRC1_DIS_INT_FF_MON	0x0002	/* Disable monitor FIFO Full interrupt */
#define CRC1_DIS_INT_FE_SEQ	0x0004	/* Disable sequencer FIFO Empty interrupt */
#define CRC1_DIS_INT_FH_SEQ	0x0008	/* Disable sequencer FIFO Half-full interrupt */
#define CRC1_DIS_INT_FF_MAP	0x0010	/* Disable mapper FIFO Full interrupt */
#define CRC1_EN_ANTI_META	0x2000	/* Enable anti-metastability flip-flops */
#define CRC1_DIS_SM_DEBUG	0x4000	/* Disable LED flashing state machine debug feature */

/* Status Register A2 */

#define SRA2_MMU_IDLE		0x000001

/* Configuration Register A2 */

/* MAPPER-Output control */

#define CRA2_MAP_OUT_CFG_MASK		0x000003
#define CRA2_MAP_OUT_CFG_DISABLE	0x000000	/* Not used on current hardware */
#define CRA2_MAP_OUT_CFG_PASSTHRU  	0x000001
#define CRA2_MAP_OUT_CFG_1_TO_1    	0x000002
#define CRA2_MAP_OUT_CFG_1_TO_MANY  	0x000003

#define CRA2_AER_OUT_CFG_MASK		0x00000C
#define CRA2_AER_OUT_CFG_DISABLE    	0x000000	/* Not used on current hardware */
#define CRA2_AER_OUT_CFG_1_16       	0x000004
#define CRA2_AER_OUT_CFG_2_15       	0x000008
#define CRA2_AER_OUT_CFG_4_14       	0x00000C

#define CRA2_MAP_OUT_DISABLE   		0x000010

#define CRA2_AER_PROTOCOL_MASK		0x000020
#define CRA2_AER_PROTOCOL_P2P		0x000020	/* Point to point */
#define CRA2_AER_PROTOCOL_MULTI		0x000000	/* Multi-sender (SCX-like) */

/*
 * FIFO sizes
 *
 * A range of different depths of FIFO can be fitted as follows.
 * The size in bytes of memory occupied by the Monitor & Sequencer FIFOs is
 * always (depth * sizeof(u32)).
 */

#define IDT7203_DEPTH	 2048
#define IDT7204_DEPTH	 4096
#define IDT7205_DEPTH	 8192
#define IDT7206_DEPTH	16384
#define IDT7207_DEPTH	32768
#define IDT7208_DEPTH	65536

#endif
