/*
pcimodule.c

This file contains code for the Linux character device driver
module for the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2000, 2003-2006 & 2008-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

pcimodule.c $Id: 859c0222ee40af4323b7aae3d94050b4a71c3206 $
*/

#ifdef CONFIG_PCIAER_DEBUG
#define DEBUG
#endif

#include <linux/device.h>
#include <linux/version.h>
#include <linux/module.h>

#ifndef CONFIG_PCI
#error No PCI support - cannot build!
#endif

#include <linux/ctype.h>
#include <linux/interrupt.h>
#include <linux/oom.h>

#include "s5920.h"
#include "pciaerhw.h"
#include "internal.h"

#include "version.h"

/*****************************************************************************/

/*
 * abort_init_at can be used to abort the driver initialization process for
 * debugging purposes by setting it to the number given in the ABORT_INIT_POINT
 * macro at the point desired. 0 means don't abort.
 */

#ifdef CONFIG_PCIAER_DEBUG
INTEGER_MODULE_PARAMETER(abort_init_at, 0,
			 "Checkpoint at which to abort driver initialization for debug.");
#define ABORT_INIT_POINT(N) {if (abort_init_at != 0 && (N) >= abort_init_at) {pr_debug("%s: Initialization aborted at checkpoint %d\n", DriverName, (N)); return -EINTR; } }
#else
#define ABORT_INIT_POINT(N) {}
#endif

/*
 * nvr_only can be set true to force the driver to load with support only for
 * the S5920 NVRAM sub-device in case some configuration problem is otherwise
 * preventing the driver loading. The NVRAM configuration can then be changed,
 * the machine re-booted, and the driver can then hopefully be loaded normally.
 */

BOOLEAN_MODULE_PARAMETER(nvr_only,
			 "Force driver to load with only /dev/aernvr* usable.");

/*
 * fifos can be used to tell the driver that one or more of the FIFOs fitted to
 * the boards are not the same size as the PCI memory region for reading and
 * writing the monitor and sequencer FIFOs respectively. The string takes the
 * form
 *	[<board>.]<func>:<d>[,[<board>.]<func>:<d>[,...]]
 * where <board> is the board number, <func> is one of mon, seq or map and <d>
 * is either one of the possible FIFO part designations e.g. IDT7203 or depths
 * e.g. 2048 or 2K. If the <board>. element is not present, board 0 is assumed.
 * In the case of the monitor and sequencer FIFOs, the size given by the fifos
 * string is not allowed to be bigger than the size of the corresponding PCI
 * memory region.
 */

STRING_MODULE_PARAMETER(fifos, "Override FIFO depth assumptions.");

/*****************************************************************************/

/* Forward declarations of entry points */

int init_module(void);
void cleanup_module(void);
static int PciProbeCallback(struct pci_dev *pDev,
			    const struct pci_device_id *pDevId);
static void PciRemoveCallback(struct pci_dev *pDev);
static int no_suspend(struct pci_dev *pDev, pm_message_t state);

/*****************************************************************************/

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR
	("Adrian M. Whatley, amw@ini.phys.ethz.ch, Institute of Neuroinformatics (INI), Zurich");
MODULE_DESCRIPTION
	("Driver for the PCI-AER board developed by Vittorio Dante at the Istituto Superiore di Sanita (ISS) in Rome");
MODULE_VERSION(VER_STRING);

/*****************************************************************************/

/* Global data */

/* Array of supported device id's */

static struct pci_device_id SupportedDeviceTable[] = {
	{PCIAER_PCI_VENDOR_ID, PCIAER_PCI_DEVICE_ID, PCI_ANY_ID, PCI_ANY_ID, 0,
	 0, 0},
	{0, 0, 0, 0, 0, 0, 0}
};

/* PCI driver struct. */

static struct pci_driver Driver = {
	.name = (char *)DriverName,
	.id_table = SupportedDeviceTable,
	.probe = PciProbeCallback,
	.remove = PciRemoveCallback,
	.suspend = no_suspend,
};

/* Count of number of boards sucessfully initialized */

static int cBoardsInitialized;

/*
 * Array of pointers to information structures relating to each PCI-AER board
 * found in the system, or NULL if there is no corresponding board.
 */

static struct pciaer_board *BoardPtrs[PCIAER_MAX_BOARDS];

/* List of fifo parameter structures */

enum func_enum {MON, SEQ, MAP};

struct fifo_param {
	struct list_head list;
	unsigned int board_no;
	enum func_enum func;
	unsigned int depth;
};

static LIST_HEAD(fifo_param_list);

/*****************************************************************************/

/*
 * no_suspend
 * Suspend is not supported since (1) we're a PCI device that's not going to be
 * fitted to a laptop; and (2) suspending, assuming it were reasonably possible
 * to save & restore the PCI-AER state anyway, could potentially stop attached
 * AER hardware from running.
 * Output a helpful message and deny suspend by returning -ENOSYS as described
 * in Documentation/SubmittingDrivers.
 */

static int no_suspend(struct pci_dev *pDev, pm_message_t state)
{
	dev_info(&pDev->dev, "Unable to suspend\n");
	return -ENOSYS;
}

/*****************************************************************************/

/*
 * GetBoard
 */

struct pciaer_board *GetBoard(unsigned int iBoard)
{
	if (iBoard >= PCIAER_MAX_BOARDS)
		return NULL;

	return BoardPtrs[iBoard];
}

/*****************************************************************************/

/* Out-of-memory handler and notifier block */

/*static*/ int ModuleOomHandler(struct notifier_block *nb,
	unsigned long dummy, void *pvFreedPages)
{
	int iBoard;
	struct pciaer_board *pBoard;
	unsigned long *pulFreedPages;

	pulFreedPages = pvFreedPages;

	for (iBoard = 0; iBoard <= PCIAER_MAX_BOARDS; iBoard++) {
		pBoard = GetBoard(iBoard);
		if (pBoard != NULL) {
			pvFreedPages += MapperOomHandler(pBoard);
			pvFreedPages += S5920OomHandler(pBoard);
		}
	}

	return NOTIFY_OK;
}

static struct notifier_block OomNotifierBlock = {
	.notifier_call = ModuleOomHandler
};

static bool OomHandlerRegistered; /* = false; */

/*****************************************************************************/

/*
 * ReleasePciMemory
 * This function calls pcim_iounmap and pci_release_region for each valid base
 * address recorded in the board structure by GetPciMemory.
 */

static void ReleasePciMemory(struct pciaer_board *pBoard)
{
	int iBaseAddr;

	for (iBaseAddr = 0; iBaseAddr < NUM_USED_BADR; iBaseAddr++)
		if (pBoard->pBaseAddr[iBaseAddr] != NULL) {
			pcim_iounmap(pBoard->pPciDev, pBoard->pBaseAddr[iBaseAddr]);
			pci_release_region(pBoard->pPciDev, iBaseAddr);
			pBoard->pBaseAddr[iBaseAddr] = NULL;
		}
}

/*****************************************************************************/

/*
 * ReleaseBoard
 * Disables the hardware & releases memory associated with a board. Normally
 * called from PciRemoveCallback, but may also be called from PciProbeCallback
 * if initialisation of a board fails.
 */

static void ReleaseBoard(struct pciaer_board *pBoard)
{
	struct device *pDev;

	if (pBoard != NULL) {
		pDev = &pBoard->pPciDev->dev;

		/* dev_dbg(pDev, "Calling S5920NvrCleanup\n"); */
		S5920NvrCleanup(pBoard);

		/* dev_dbg(pDev, "Calling DisableHardware\n"); */
		DisableHardware(pBoard);

		if (pBoard->fIrqHandlerInstalled) {
			/* dev_dbg(pDev, "Calling devm_free_irq\n"); */
			devm_free_irq(pDev, pBoard->pPciDev->irq, pBoard);
		}
		/* dev_dbg(pDev, "Calling ReleasePciMemory\n"); */
		ReleasePciMemory(pBoard);

		BoardPtrs[pBoard->iBoard] = NULL;

		/* dev_dbg(pDev, "Calling devm_kfree\n"); */
		devm_kfree(pDev, pBoard);
	}
}

/*****************************************************************************/

/*
 * PciRemoveCallback
 * This is the function given in the remove member of our pci_driver struct to
 * be called back when a board is to be removed. In practice, since we do not
 * support hot-plugging, this only happens when we call pci_unregister_driver
 * from cleanup_module. All we do here is extract our board struct pointer put
 * into the kernel's device struct by PciProbeCallback during initialisation and
 * leave the real work to ReleaseBoard.
 */

static void PciRemoveCallback(struct pci_dev *pDev)
{
	if (pDev != NULL)
		ReleaseBoard((struct pciaer_board *) pci_get_drvdata(pDev));
}

/*****************************************************************************/

/*
 * cleanup_module
 * Called by the kernel prior to the module being removed, or from init_module
 * if no boards are found or there is a fatal error during initialiseation.
 * We call pci_unregister_driver which will cause PciRemoveCallback to be called
 * for each board to disable the hardware and free up memory, then
 * we unregister the driver.
 */

void cleanup_module(void)
{
	if (OomHandlerRegistered)
		unregister_oom_notifier(&OomNotifierBlock);

	CleanUpDebugFS();

	pci_unregister_driver(&Driver);

	UnregisterDriver();
}

/*****************************************************************************/

/*
 * InitBoardStruct
 * Helper function called from PciProbeCallback to fill in fields in a struct
 * pciaer_board in part from a struct pci_dev. It acts like a constructor for
 * struct pciaer_board. The supplied pointers must both be non-NULL.
 */

static void InitBoardStruct(struct pciaer_board *pBoard, struct pci_dev *pDev,
			    unsigned int iBoard)
{
	int nDevNameLen;
	int iBaseReg;

	/* Record the pointer to the struct pci_dev */

	pBoard->pPciDev = pDev;

	/* Remember the board index */

	pBoard->iBoard = iBoard;

	/* Construct a device name from the DriverName + iBoard */

	nDevNameLen =
	    snprintf(pBoard->DeviceName, MAX_DEVNAME_LEN, "%s%d", DriverName,
		     iBoard);
	if (nDevNameLen > MAX_DEVNAME_LEN)
		dev_warn(&pDev->dev, "Device name %s%d truncated to %s\n",
			 DriverName, iBoard, pBoard->DeviceName);

	/*
	 * First set all the base addresses and the 'ReadToFlush' address to
	 * NULL and the sizes of the FIFOs to zero in case something goes wrong.
	 */

	for (iBaseReg = 0; iBaseReg < NUM_USED_BADR; iBaseReg++)
		pBoard->pBaseAddr[iBaseReg] = NULL;

	pBoard->pReadToFlush = NULL;

	rwlock_init(&pBoard->HardwareLock);

	pBoard->fpga1_release = 0xFFFF;
	pBoard->fpga2_release = 0xFFFF;

	pBoard->uMonFifoDepth = 0;
	pBoard->uSeqFifoDepth = 0;
	pBoard->uMapFifoDepth = 0;

	pBoard->fIrqHandlerInstalled = false;

	spin_lock_init(&pBoard->OpenFlagsLock);
	pBoard->nOpens = 0;
	pBoard->uidCurrentUser = -1;

	atomic_set(&pBoard->CounterPeriod_us, 1);
	pBoard->tvLastCounterResetTime.tv_sec = 0;
	pBoard->tvLastCounterResetTime.tv_usec = 0;

	mutex_init(&pBoard->mtxMonitor);
	pBoard->nMonitorOpens = 0;
	pBoard->pMonitorBuffer = NULL;
	pBoard->ulMonWordsRead = 0L;
	pBoard->ulMonReads = 0L;
	atomic_set(&pBoard->MonFifoHalfFullIntCount, 0);
	atomic_set(&pBoard->MonFifoFullIntCount, 0);

	mutex_init(&pBoard->mtxSequencer);
	pBoard->nSequencerOpens = 0;
	pBoard->pSequencerBuffer = NULL;
	pBoard->fSeqEnabled = false;
	pBoard->ulSeqWordsWritten = 0L;
	pBoard->ulSeqWrites = 0L;
	atomic_set(&pBoard->SeqFifoHalfEmptyIntCount, 0);
	atomic_set(&pBoard->SeqFifoEmptyIntCount, 0);

	init_rwsem(&pBoard->rwsMapMem);
	pBoard->u32MapperNextFreeSpikeTblEntry = PCIAER_MAPPER_PTR_TBL_LENGTH;	/* @@@ Kludge - see struct pciaer_board comment */
	pBoard->puhMapperDestListBuffer = NULL;
	pBoard->uhMapperDestListBufferLen = 0;
	atomic_set(&pBoard->MapFifoFullIntCount, 0);

	mutex_init(&pBoard->mtxS5920NvrCache);
	pBoard->uS5920NvrSize = 0;
	pBoard->S5920NvrCacheBytes = NULL;
	pBoard->S5920NvrCacheFlags = NULL;

	pBoard->dDebugFSEntry = NULL;
}

/*****************************************************************************/

/*
 * GetPciMemory
 * This function tries to request and map the region associated with each of the
 * used base addresses on a board.
 * The void _iomem * which is returned by pcim_iomap and which can be used by
 * ioread32 / iowrite32 is stored in the corresponding element of the pBaseAddr
 * array in the board struct. If an error occurs with one region, the function
 * tries to carry on to do the others, but an error will finally be reported to
 * the caller. Otherwise if there is no error, the function returns 0.
 */

static int GetPciMemory(struct pciaer_board *pBoard)
{
	struct pci_dev *pDev;
	int iMaxBaseAddr;
	int iBaseAddr;
	unsigned long ulResourceStart;
	int nReqRegionErr;
	int nError;

	pDev = pBoard->pPciDev;
	nError = 0;

	iMaxBaseAddr = nvr_only ? 0 : (NUM_USED_BADR - 1);
	for (iBaseAddr = 0; iBaseAddr <= iMaxBaseAddr; iBaseAddr++) {
		ulResourceStart = pci_resource_start(pDev, iBaseAddr);

		dev_dbg(&pDev->dev,
			"resource %pr (size 0x%08lX)\n",
			 &pDev->resource[iBaseAddr],
			 (unsigned long)pci_resource_len(pDev, iBaseAddr));

		if (!(pci_resource_flags(pDev, iBaseAddr) & IORESOURCE_MEM)) {
			dev_warn(&pDev->dev,
				 "Board %d has non-memory mapped region\n",
				 pBoard->iBoard);
		}

		nReqRegionErr = pci_request_region_exclusive(pDev, iBaseAddr,
							     pBoard->DeviceName);
		if (nReqRegionErr != 0) {
			dev_warn(&pDev->dev,
				 "Requesting region at 0x%08lX on board %d failed\n",
				 ulResourceStart, pBoard->iBoard);
			nError = nReqRegionErr;
			continue;
		}

		pBoard->pBaseAddr[iBaseAddr] = pcim_iomap(pDev, iBaseAddr, 0);
		if (pBoard->pBaseAddr[iBaseAddr] == NULL) {
			dev_warn(&pDev->dev,
				 "Failed to map region at 0x%08lX on board %d\n",
				 ulResourceStart, pBoard->iBoard);
			nError = -ENXIO;
			pci_release_region(pDev, iBaseAddr);
			continue;
		}
	}

	return nError;
}

/*****************************************************************************/

static bool is_valid_fifo_depth(unsigned int depth)
{
	switch (depth) {
	case IDT7203_DEPTH:
	case IDT7204_DEPTH:
	case IDT7205_DEPTH:
	case IDT7206_DEPTH:
	case IDT7207_DEPTH:
	case IDT7208_DEPTH:
		return true;
	default:
		return false;
	}
}

/*****************************************************************************/

/*
 * SizeFifos
 * Record the sizes of the FIFOs.
 * To save having to size the FIFOs dynamically, we assume that the size of all
 * three FIFOs is accurately reflected in the size implied by the PCI base
 * address register for the FIFO region, unless overridden by an entry in the
 * fifo_param_list generated from the 'fifos' command line parameter. In the
 * case of the Monitor and Sequencer FIFOs, such an override is only valid if
 * it specifies a FIFO depth less than (or equal to) that implied by the size
 * of the PCI region, since we want to be able to read or write the whole FIFO
 * in one go.
 * NB. The FIFO sizes are given in terms of depths, i.e. numbers of words,
 * not sizes (number of bytes).
 */

static int SizeFifos(struct pciaer_board *pBoard, unsigned int board_no)
{
	unsigned int implicit_depth;
	struct device *d;
	struct fifo_param *p;

	implicit_depth =
	    pci_resource_len(pBoard->pPciDev, MONSEQFIFO_BADR) / sizeof(u32);

	d = &pBoard->pPciDev->dev;

	if (!is_valid_fifo_depth(implicit_depth)) {
		dev_err(d, "PCI resource implies invalid fifo depth\n");
		return -EOPNOTSUPP;
	}

	pBoard->uMonFifoDepth = implicit_depth;
	pBoard->uSeqFifoDepth = implicit_depth;
	pBoard->uMapFifoDepth = implicit_depth;

	list_for_each_entry(p, &fifo_param_list, list) {
		if (p->board_no == board_no) {
			if ((p->func == MON || p->func == SEQ) &&
			    p->depth > implicit_depth) {
				dev_err(d,
				    "Monitor or Sequencer FIFO depth override too big\n");
				return -EINVAL;
			}
			switch (p->func) {
			case MON:
				pBoard->uMonFifoDepth = p->depth;
				break;
			case SEQ:
				pBoard->uSeqFifoDepth = p->depth;
				break;
			case MAP:
				pBoard->uMapFifoDepth = p->depth;
				break;
			}
		}
	}

	dev_dbg(d, "Monitor fifo depth = %u words\n", pBoard->uMonFifoDepth);
	dev_dbg(d, "Sequencer fifo depth = %u words\n", pBoard->uSeqFifoDepth);
	dev_dbg(d, "Mapper fifo depth = %u words\n", pBoard->uMapFifoDepth);

	return 0;
}

/*****************************************************************************/

/*
 * PciProbeCallback
 * This is the function given in the probe member of our pci_driver struct to be
 * called back when a board is to be added. In practice, since we do not support
 * hot-plugging, this only happens when we call pci_module_init from init_module.
 */

static int PciProbeCallback(struct pci_dev *pDev,
			    const struct pci_device_id *pDevId)
{
	struct pciaer_board *pBoard;
	int nError;

	/*
	 * If we have already found and initialized the maximum number of boards
	 * we support, we cannot hold pointers to any more in our BoardPtrs
	 * array so there is nothing to do but say so and return an error.
	 */

	dev_info(&pDev->dev, "Found board\n");

	if (cBoardsInitialized == PCIAER_MAX_BOARDS) {
		dev_info(&pDev->dev,
			 "Maximum no. of boards already reached - this board will not be used\n");
		return -ENOMEM;
	}

	ABORT_INIT_POINT(5);

	/*
	 * Ensure PCI device is enabled and stays enabled for the sake of the
	 * Monitor, which must keep handshaking with any attached AER devices.
	 */

	if (pcim_enable_device(pDev) != 0) {
		dev_err(&pDev->dev, "pci_enable_device failed\n");
		return -EIO;
	}
	pcim_pin_device(pDev);

	ABORT_INIT_POINT(6);

	/* Allocate memory for our board specific struct */

	pBoard = devm_kzalloc(&pDev->dev, sizeof(struct pciaer_board), GFP_KERNEL);
	if (pBoard == NULL) {
		dev_err(&pDev->dev, "Memory allocation failure\n");
		return -ENOMEM;
	}

	ABORT_INIT_POINT(7);

	/*
	 * Call InitBoardStruct. This is a bit like a constructor for our board
	 * struct.
	 */

	InitBoardStruct(pBoard, pDev, cBoardsInitialized);

	ABORT_INIT_POINT(8);

	/*
	 * Get PCI memory addresses, and if that works, set up the 'ReadToFlush'
	 * address and size the FIFOs. For the 'ReadToFlush' address, I choose
	 * to read from one of the S5920 operation registers so that nothing
	 * beyond the S5920 need be disturbed. Reading from the OMB has no side
	 * effects.
	 * Otherwise, in the special case that the S5920 op-regs are accessible
	 * via IO space (NVRAM contents not valid), record the IO space address
	 * for use by the S5920 NVRAM minor device.
	 */

	nError = GetPciMemory(pBoard);
	if (nError == 0) {
		pBoard->pReadToFlush = S5920OpRegAddress(pBoard, S5920_OMB);
		if (!nvr_only)
			nError = SizeFifos(pBoard, cBoardsInitialized);
	}

	if (nError != 0) {
		dev_err(&pDev->dev, "Initialization failed with error %d\n",
			-nError);
		ReleaseBoard(pBoard);
		return nError;
	}

	ABORT_INIT_POINT(9);

	/* Initialize hardware */

	nError = InitHardware(pBoard);
	if (nError != 0) {
		dev_err(&pDev->dev,
			"Hardware initialization failed with error %d\n",
			-nError);
		ReleaseBoard(pBoard);
		return nError;
	}

	dev_info(&pDev->dev, "Hardware initialization successful\n");

	dev_info(&pDev->dev, "Assigned as board %u\n", pBoard->iBoard);

	/*
	 * Record the pointer to our board struct for this board in the driver
	 * data member of the kernel's device struct so that we can retrieve it
	 * again when PciRemoveCallback is called and save it in the next free
	 * member of our BoardPtrs array for all other purposes.
	 */

	pci_set_drvdata(pDev, pBoard);

	BoardPtrs[cBoardsInitialized++] = pBoard;

	/* Try to setup interrupts for the current board. */

	SetupInterrupts(pBoard);

	/* Enable monitor interrupts and mapper FIFO full interrupt */

	EnableInt(pBoard,
		  CRC1_DIS_INT_FH_MON | CRC1_DIS_INT_FF_MON |
		  CRC1_DIS_INT_FF_MAP);

	/* Setup debugfs entries */

	AddBoardToDebugFS(pBoard);

	return 0;
}

/*****************************************************************************/

/*
 * ParseFifosParam
 * Parse the 'fifos' module parameter into the fifo_param_list.
 * The 'fifos' string should have the following form:
 *	[<board>.]<func>:<d>[,[<board>.]<func>:<d>[,...]]
 * Return 0 if the parameter is valid and -EINVAL otherwise (-ENOMEM is also
 * theoretically possible).
 */

static int parse_fifos_param(void)
{
	const char *c;
	unsigned long long result;
	char *endptr;
	unsigned int board;
	enum func_enum func;
	unsigned int depth;
	struct fifo_param *p;

	if (fifos == NULL)
		return 0;

	c = fifos;
	do {
		/* If there is a digit, read the board number, otherwise
		assume 0 */

		if (isdigit(*c)) {
			result = simple_strtoul(c, &endptr, 10);
			board = (unsigned int)result;
			if (*endptr != '.' || board != result)
				return -EINVAL;
			c = endptr + 1;
		} else {
			board = 0;
		}

		/* Check which function */

		if (strncmp(c, "mon", 3) == 0)
			func = MON;
		else if (strncmp(c, "seq", 3) == 0)
			func = SEQ;
		else if (strncmp(c, "map", 3) == 0)
			func = MAP;
		else
			return -EINVAL;
		c += 3;

		/* Next character must be ':' */

		if (*c != ':')
			return -EINVAL;

		/* Now comes either a FIFO part designation of the form IDT720x
		 or a depth e.g. 2048 or 2K. */

		if (isalpha(*++c)) {

			if (strnicmp(c, "IDT720", 6) != 0)
				return -EINVAL;

			switch (*(c += 6)) {
			case '3':
				depth = IDT7203_DEPTH;
				break;
			case '4':
				depth = IDT7204_DEPTH;
				break;
			case '5':
				depth = IDT7205_DEPTH;
				break;
			case '6':
				depth = IDT7206_DEPTH;
				break;
			case '7':
				depth = IDT7207_DEPTH;
				break;
			case '8':
				depth = IDT7208_DEPTH;
				break;
			default:
				return -EINVAL;
			}
			c++;
		} else {
			result = memparse(c, &endptr);
			c = endptr;

			if (result > IDT7208_DEPTH)
				return -EINVAL;

			depth = (unsigned int)result;

			if (!is_valid_fifo_depth(depth))
				return -EINVAL;
		}

		/* Next character must either be a '\0', in which case we're
		done, or ',', in which case we go round the loop again. */

		if (*c != '\0' && *c != ',')
			return -EINVAL;

		/* Allocate & add list item to fifo_param_list */
		/* NB: we can't use devm_kzalloc here since we don't have a
		pointer to a device struct here. */

		p = kmalloc(sizeof(*p), GFP_KERNEL);
		if (p == NULL)
			return -ENOMEM;

		p->board_no = board;
		p->func = func;
		p->depth = depth;

		list_add_tail(&p->list, &fifo_param_list);

	} while (*c++ != '\0');

	return 0;
}

/*****************************************************************************/

static void del_fifo_params_list(void)
{
	struct list_head *entry;
	struct list_head *tmp;

	list_for_each_safe(entry, tmp, &fifo_param_list) {
		list_del(entry);
		kfree(entry);
	}
}

/*****************************************************************************/

/*
 * init_module
 * Called by kernel to initialize the module at module insertion time. Returns
 * -error in the event of a problem, and 0 otherwise.
 * Its tasks are to check the validity of the major device no. to be used (which
 * may have been changed by insmod from the default of 0 used for dynamic major
 * number allocation), to  register the driver and to find and initialise our
 * PCI device(s) by calling pci_module_init. This in turn will eventually call
 * our PciProbeCallback for each of our boards.
 */

int init_module(void)
{
	int iBoard;
	int nResult;

	ABORT_INIT_POINT(1);

	/*
	 * Ensure board count & unused entries in the BoardPtrs array are
	 * initialized.
	 */

	cBoardsInitialized = 0;
	for (iBoard = 0; iBoard < PCIAER_MAX_BOARDS; iBoard++)
		BoardPtrs[iBoard] = NULL;

	ABORT_INIT_POINT(2);

	nResult = parse_fifos_param();
	if (nResult != 0) {
		del_fifo_params_list();
		return nResult;
	}

	ABORT_INIT_POINT(3);

	/* Register driver */

	nResult = RegisterDriver();
	if (nResult != 0) {
		del_fifo_params_list();
		return nResult;
	}

	ABORT_INIT_POINT(4);

	InitDebugFS();

	/* Find PCI devices matching our vendor and device IDs and allocate and
	 * fill in the Board structures.
	 * pci_register_driver will call PciProbeCallback for each board
	 * matching the criteria in SupportedDeviceTable pointed to from the
	 * Driver structure.
	 */

	nResult = pci_register_driver(&Driver);

	/* The list of fifo parameters is no longer needed and can be deleted */

	del_fifo_params_list();

	/*
	 * If there was an error, or we ended up not having found any devices we
	 * say so and exit via cleanup_module.
	 */

	if (nResult != 0)
		pr_err("%s: pci_module_init returned error %d\n",
		       DriverName, -nResult);

	if (cBoardsInitialized == 0) {
		pr_info("%s: failed to find any PCI-AER boards\n", DriverName);
		nResult = -ENODEV;
	}

	if (nResult != 0) {
		cleanup_module();
		return nResult;
	}

	/* Otherwise we report how many boards we sucessfully initialized. */

	pr_info("%s: %d board%s initialized\n", DriverName,
	       cBoardsInitialized, cBoardsInitialized == 1 ? "" : "s");

	/*
	 * Register an out-of-memory notifier. This is not essential, so we
	 * don't mind if it fails.
	 */

	if (register_oom_notifier(&OomNotifierBlock) == 0)
		OomHandlerRegistered = true;

	return 0;
}

/*****************************************************************************/

/* EOF */
