/* s5920.h */
#ifndef S5920_H
#define S5920_H

/*
Defines constants specifically relating to AMCC's S5920 PCI device interface
chip.

Copyright (C) 2000, 2003, 2004 & 2010-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This file was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

s5920.h $Id: 3e854583cdc9b07bc416a1dfae5eaebfb767db37 $
*/

/*
 * Vendor & device ID should be defined in linux/pci.h, but if they aren't
 * (or linux/pci.h has not been included), define them here.
 */

#ifndef PCI_VENDOR_ID_AMCC
#define PCI_VENDOR_ID_AMCC		0x10e8
#endif
#ifndef PCI_DEVICE_ID_AMCC_S5920
#define PCI_DEVICE_ID_AMCC_S5920	0x5920
#endif

/* Operation registers, pp2-53 onwards */

/*
 * Outgoing Mailbox Register (OMB), p2-54
 * Read/Write, bytewise addressable.
 */

#define	S5920_OMB	0x0c
#define	S5920_OMB_0	0x0c
#define	S5920_OMB_1	0x0d
#define	S5920_OMB_2	0x0e
#define	S5920_OMB_3	0x0f

/*
 * Incoming Mailbox Register (IMB), p2-55
 * Read Only, bytewise addressable.
 */

#define S5920_IMB	0x1c
#define S5920_IMB_0	0x1c
#define S5920_IMB_1	0x1d
#define S5920_IMB_2	0x1e
#define S5920_IMB_3	0x1f

/* Mailbox Empty/Full Status Register (MBEF), p2-56 */

#define S5920_MBEF	0x34
#define S5920_MBEF_RESERVED		0x0FFF0FFF
#define S5920_MBEF_IMB_STATUS		0xF0000000
#define S5920_MBEF_IMB_STATUS_3		0x80000000
#define S5920_MBEF_IMB_STATUS_2		0x40000000
#define S5920_MBEF_IMB_STATUS_1		0x20000000
#define S5920_MBEF_IMB_STATUS_0		0x10000000
#define S5920_MBEF_OMB_STATUS		0x0000F000
#define S5920_MBEF_OMB_STATUS_3		0x00008000
#define S5920_MBEF_OMB_STATUS_2		0x00004000
#define S5920_MBEF_OMB_STATUS_1		0x00002000
#define S5920_MBEF_OMB_STATUS_0		0x00001000

/* Interrupt Control/Status Register (INTCSR), pp2-57/8 */

#define S5920_INTCSR	0x38
#define S5920_INTCSR_RESERVED			0xff3cccec
#define S5920_INTCSR_OMB_INT_BYTE_NO_MASK	0x00000003
#define S5920_INTCSR_OMB_INT_ENABLE		0x00000010
#define S5920_INTCSR_IMB_INT_BYTE_NO_MASK	0x00000300
#define S5920_INTCSR_IMB_INT_ENABLE		0x00001000
#define S5920_INTCSR_ADDINT_ENABLE		0x00002000
#define S5920_INTCSR_OMB_INT			0x00010000
#define S5920_INTCSR_IMB_INT			0x00020000
#define S5920_INTCSR_ADDINT_INT			0x00400000
#define S5920_INTCSR_INT_ASSERTED		0x00800000
#define S5920_INTCSR_OMB_INT_BYTE_NO_SHIFT		0
#define S5920_INTCSR_IMB_INT_BYTE_NO_SHIFT		8

/* Reset Control Register (RCR), pp2-59/60 */

#define S5920_RCR	0x3c
#define S5920_RCR_RESERVED		0x0400ffff
#define S5920_RCR_NVRAM_OP_ADDR_MASK	0x00ff0000
#define S5920_RCR_NVRAM_OP_DATA_MASK	0x00ff0000
#define S5920_RCR_ADDON_RESET		0x01000000
#define S5920_RCR_READ_FIFO_RESET	0x02000000
#define S5920_RCR_MB_FLAGS_RESET	0x08000000
#define S5920_RCR_NVRAM_ACCESS_FAILED	0x10000000
#define S5920_RCR_NVRAM_ACESSS_CONTROL	0xe0000000
#define S5920_RCR_NVRAM_INACTIVE	0x00000000
#define S5920_RCR_NVRAM_LOAD_LO_ADDR	0x80000000
#define S5920_RCR_NVRAM_LOAD_HI_ADDR	0xa0000000
#define S5920_RCR_NVRAM_WRITE_DATA	0xc0000000
#define S5920_RCR_NVRAM_READ_DATA	0xe0000000
#define S5920_RCR_NVRAM_BUSY		0x80000000
#define S5920_RCR_NVRAM_OP_ADDR_SHIFT		16
#define S5920_RCR_NVRAM_OP_DATA_SHIFT		16

/* Pass-Thru (sic) Configuration Register (PTCR), pp2-61/2 */

#define S5920_PTCR	0x60
#define S5920_PTCR_R1_MASK	0x000000ff
#define S5920_PTCR_R2_MASK	0x0000ff00
#define S5920_PTCR_R3_MASK	0x00ff0000
#define S5920_PTCR_R4_MASK	0xff000000
#define S5920_PTCR_R1_SHIFT		0
#define S5920_PTCR_R2_SHIFT		8
#define S5920_PTCR_R3_SHIFT		16
#define S5920_PTCR_R4_SHIFT		24

#define S5920_PTCR_ACTIVE_MODE_WS	0x07
#define S5920_PTCR_PREFETCH		0x18
#define S5920_PTCR_PREFETCH_DISABLED	0x00
#define S5920_PTCR_PREFETCH_SMALL	0x08
#define S5920_PTCR_PREFETCH_MEDIUM	0x10
#define S5920_PTCR_PREFETCH_LARGE	0x18
#define S5920_PTCR_WRITE_FIFO_DISABLED	0x20
#define S5920_PTCR_BIG_ENDIAN_CONV	0x40
#define S5920_PTCR_PTADR_DRIVEN		0x80

#define S5920_PTCR_R1_ACTIVE_MODE_WS		0x00000007
#define S5920_PTCR_R1_PREFETCH			0x00000018
#define S5920_PTCR_R1_PREFETCH_DISABLED		0x00000000
#define S5920_PTCR_R1_PREFETCH_SMALL		0x00000008
#define S5920_PTCR_R1_PREFETCH_MEDIUM		0x00000010
#define S5920_PTCR_R1_PREFETCH_LARGE		0x00000018
#define S5920_PTCR_R1_WRITE_FIFO_DISABLED	0x00000020
#define S5920_PTCR_R1_BIG_ENDIAN_CONV		0x00000040
#define S5920_PTCR_R1_PTADR_DRIVEN		0x00000080
#define S5920_PTCR_R2_ACTIVE_MODE_WS		0x00000700
#define S5920_PTCR_R2_PREFETCH			0x00001800
#define S5920_PTCR_R2_PREFETCH_DISABLED		0x00000000
#define S5920_PTCR_R2_PREFETCH_SMALL		0x00000800
#define S5920_PTCR_R2_PREFETCH_MEDIUM		0x00001000
#define S5920_PTCR_R2_PREFETCH_LARGE		0x00001800
#define S5920_PTCR_R2_WRITE_FIFO_DISABLED	0x00002000
#define S5920_PTCR_R2_BIG_ENDIAN_CONV		0x00004000
#define S5920_PTCR_R2_PTADR_DRIVEN		0x00008000
#define S5920_PTCR_R3_ACTIVE_MODE_WS		0x00070000
#define S5920_PTCR_R3_PREFETCH			0x00180000
#define S5920_PTCR_R3_PREFETCH_DISABLED		0x00000000
#define S5920_PTCR_R3_PREFETCH_SMALL		0x00080000
#define S5920_PTCR_R3_PREFETCH_MEDIUM		0x00100000
#define S5920_PTCR_R3_PREFETCH_LARGE		0x00180000
#define S5920_PTCR_R3_WRITE_FIFO_DISABLED	0x00200000
#define S5920_PTCR_R3_BIG_ENDIAN_CONV		0x00400000
#define S5920_PTCR_R3_PTADR_DRIVEN		0x00800000
#define S5920_PTCR_R4_ACTIVE_MODE_WS		0x07000000
#define S5920_PTCR_R4_PREFETCH			0x18000000
#define S5920_PTCR_R4_PREFETCH_DISABLED		0x00000000
#define S5920_PTCR_R4_PREFETCH_SMALL		0x08000000
#define S5920_PTCR_R4_PREFETCH_MEDIUM		0x10000000
#define S5920_PTCR_R4_PREFETCH_LARGE		0x18000000
#define S5920_PTCR_R4_WRITE_FIFO_DISABLED	0x20000000
#define S5920_PTCR_R4_BIG_ENDIAN_CONV		0x40000000
#define S5920_PTCR_R4_PTADR_DRIVEN		0x80000000

/* Minimum and maximum sizes for the NVRAM */

#define S5920_NVRAM_MIN_SIZE		128
#define S5920_NVRAM_MAX_SIZE		2048

/* S5920_NVR_SER_CLK_DIVISOR (see p2-73 of data book) */

#define S5920_NVR_SER_CLK_DIVISOR	293

/* S5920_NVR_SER_CLK_*_CYCLES (see p2-76) */

#define S5920_NVR_SER_CLK_WRITE_CYCLES	29
#define S5920_NVR_SER_CLK_READ_CYCLES	39

/* S5920_NVR_*_WRITE_TIME_MS (see p2-79) */

#define S5920_NVR_MIN_WRITE_TIME_MS	 5
#define S5920_NVR_MAX_WRITE_TIME_MS	10

/* NVRAM layout, p2-82 */

#define S5920_NVRAM_OFFSET_VENDOR_ID		0x40
#define S5920_NVRAM_OFFSET_DEVICE_ID		0x42
#define S5920_NVRAM_OFFSET_SPECIAL_MODES	0x45
#define S5920_NVRAM_OFFSET_REVISION_ID		0x48
#define S5920_NVRAM_OFFSET_CLASS_CODE		0x49
#define S5920_NVRAM_OFFSET_HEADER_TYPE		0x4e
#define S5920_NVRAM_OFFSET_SELF_TEST		0x4f
#define S5920_NVRAM_OFFSET_REQD_LSB_BADR0	0x50
#define S5920_NVRAM_OFFSET_REQD_VALID_1		0x51
#define S5920_NVRAM_OFFSET_REQD_VALID_2		0x52
#define S5920_NVRAM_OFFSET_REQD_VALID_3		0x53
#define S5920_NVRAM_OFFSET_BASE_ADDR_1		0x54
#define S5920_NVRAM_OFFSET_BASE_ADDR_2		0x58
#define S5920_NVRAM_OFFSET_BASE_ADDR_3		0x5c
#define S5920_NVRAM_OFFSET_BASE_ADDR_4		0x60
#define S5920_NVRAM_OFFSET_SVID			0x6c
#define S5920_NVRAM_OFFSET_SID			0x6e
#define S5920_NVRAM_OFFSET_EXP_ROM_BASE_ADDR	0x70
#define S5920_NVRAM_OFFSET_INT_LINE		0x7c
#define S5920_NVRAM_OFFSET_INT_PIN		0x7d

struct s5920_nvram {
	unsigned char first_0x40_bytes[0x40];
	unsigned short vendor_id;
	unsigned short device_id;
	unsigned char not_used_0x44;
	unsigned char special_modes;
	unsigned short not_used_0x46;
	unsigned char revision_id;
	unsigned char class_code[3];
	unsigned char not_used_0x4c;
	unsigned char not_used_0x4d;
	unsigned char header_type;
	unsigned char self_test;
	unsigned char reqd_lsb_base_addr_0;
	unsigned char reqd_valid_1;
	unsigned char reqd_valid_2;
	unsigned char reqd_valid_3;
	unsigned int base_addr_1;
	unsigned int base_addr_2;
	unsigned int base_addr_3;
	unsigned int base_addr_4;
	unsigned int not_used_0x64;
	unsigned int not_used_0x68;
	unsigned short svid;
	unsigned short sid;
	unsigned int exp_rom_base_addr;
	unsigned char not_used_0x74[8];
	unsigned char int_line;
	unsigned char int_pin;
	unsigned char not_used_0x7e;
	unsigned char not_used_0x7f;
};

/* Values for NVRAM byte 0x45, p2-129 */

#define S5920_SPECIAL_MODES_TARGET_LATENCY_ENB	0x01
#define S5920_SPECIAL_MODES_RETRY_FLUSH_ENB	0x02
#define S5920_SPECIAL_MODES_WRITE_FIFO_MODE	0x04
#define S5920_SPECIAL_MODES_RESERVED		0xf8

/* Values for NVRAM byte 0x4f */

#define S5920_NVRAM_SELF_TEST_DESIRED		0x80
#define S5920_NVRAM_SELF_TEST_NOT_DESIRED	0x00

/*
 * Values for NVRAM byte 0x50
 * This controls whether the S5920's PCI operation registers appear in IO
 * or memory space at startup.
 * NB. Specifying S5920_LSB_BADR0_MEM_BELOW_1MB is not PCI 2.2 compliant!
 */

#define S5920_LSB_BADR0_IO			0x81
#define S5920_LSB_BADR0_MEM			0x80
#define S5920_LSB_BADR0_MEM_BELOW_1MB		0x82

/*
 * The three signature bytes that must be present for the NVRAM to be
 * recognised as valid at startup (p2-74).
 */

#define S5920_NVRAM_REQD_VALID_1		0xff
#define S5920_NVRAM_REQD_VALID_2		0xe8
#define S5920_NVRAM_REQD_VALID_3		0x10

/*
 * Field in NVRAM base address register definitions that define the data bus
 * width for a Pass-Thru region (p2-128).
 */

#define S5920_BADR_PASSTHRU_REGION_WIDTH	0xc0000000
#define S5920_BADR_PASSTHRU_REGION_DISABLED	0x00000000
#define S5920_BADR_PASSTHRU_REGION_WIDTH_8	0x40000000
#define S5920_BADR_PASSTHRU_REGION_WIDTH_16	0x80000000
#define S5920_BADR_PASSTHRU_REGION_WIDTH_32	0xc0000000

#endif
