/* s5920nvram.c

PCI interface (S5920) configuration NVRAM minor device code for the
Linux character device driver module for the PCI-AER board developed by
Vittorio Dante of the Istituto Superiore di Sanita' in Rome.

Copyright (C) 2000, 2003-2005 & 2007-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

s5920nvram.c $Id: a7d12ec04e485bee2ff3c0f946b40ea041ac710a $
*/

#ifdef CONFIG_PCIAER_DEBUG
#define DEBUG
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/capability.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "s5920.h"
#include "pciaerhw.h"
#include "internal.h"

/* PC compatible expansion ROM offsets (Table 2, p2-81) */

#define ISA_EXP_ROM_SIG_BYTE1_OFFSET	0
#define ISA_EXP_ROM_SIG_BYTE2_OFFSET	1
#define ISA_EXP_ROM_LEN_512B_OFFSET	2
#define ISA_EXP_ROM_ENTRYPOINT_OFFSET	3
#define ISA_EXP_ROM_PCI_DATA_PTR_OFFSET	0x18

/* S5920 to NVRAM serial clock cycle time in microseconds (p2-73) */

#define NVR_SER_CLK_CYCLE_US		(S5920_NVR_SER_CLK_DIVISOR / PCIAER_PCI_MHZ + 1)

/* Flag bits used in S5920NvrCacheFlags */

#define NVR_CACHE_FLAG_PRESENT		1
#define NVR_CACHE_FLAG_DIRTY		2
#define NVR_CACHE_FLAG_READONLY		4

/*****************************************************************************/

/*
 * ReadS5920OpReg
 * Reads from the given op-regs register, whether IO or memory mapped, and
 * returns its contents.
 */

unsigned int ReadS5920OpReg(struct pciaer_board *pBoard, unsigned int uReg)
{
	return ioread32(S5920OpRegAddress(pBoard, uReg));
}

/*****************************************************************************/

/*
 * WriteS5920OpReg
 * Writes to the given op-regs register, whether IO or memory mapped.
 */

void WriteS5920OpReg(struct pciaer_board *pBoard, unsigned int uReg,
		     unsigned int uValue)
{
	iowrite32(uValue, S5920OpRegAddress(pBoard, uReg));
	FlushPciWrites(pBoard);
}

/*****************************************************************************/

/*
 * PollAndGetRCR
 * This function polls the S5920 RCR register up to rcr_max_polls times at
 * intervals given by the NVRAM serial clock cycle time, if necessary, for the
 * RCR NVRAM access mechanism to be not busy, and then returns the contents of
 * the register.
 */

static u32 PollAndGetRCR(struct pciaer_board *pBoard)
{
	static const unsigned int rcr_max_polls = 50;
	u32 u32RCRValue;
	unsigned int uBusyWaitCycles;

	for (u32RCRValue = ReadS5920OpReg(pBoard, S5920_RCR),
		uBusyWaitCycles = 0;
	     (u32RCRValue & S5920_RCR_NVRAM_BUSY) &&
		uBusyWaitCycles < rcr_max_polls;
	     u32RCRValue =
		ReadS5920OpReg(pBoard, S5920_RCR), uBusyWaitCycles++) {
			usleep_range(NVR_SER_CLK_CYCLE_US,
				     NVR_SER_CLK_CYCLE_US + 1);
	}

	return u32RCRValue;
}

/*****************************************************************************/

/*
 * PhysReadNvrByte
 * Physically read an NVRAM byte from NVRAM address uhAddress. If successful, 0
 * is returned and the byte read is placed into *pchByte. Otherwise, the return
 * code is a -ve errno.
 */

static int PhysReadNvrByte(struct pciaer_board *pBoard, unsigned short uhAddress,
			   unsigned char *pchByte)
{
	static const unsigned int read_cycles_delay_us =
		S5920_NVR_SER_CLK_READ_CYCLES * NVR_SER_CLK_CYCLE_US;
	u32 u32RCRValue;
	unsigned long ulHwLockFlags;

	/* Verify that the busy bit is not set. */

	u32RCRValue = PollAndGetRCR(pBoard);
	if (u32RCRValue & S5920_RCR_NVRAM_BUSY)
		return -EBUSY;

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	/* Write the low address byte */

	WriteS5920OpReg(pBoard, S5920_RCR,
			S5920_RCR_NVRAM_LOAD_LO_ADDR | (LOBYTE(uhAddress) <<
							S5920_RCR_NVRAM_OP_ADDR_SHIFT));

	/* Write the high address byte */

	WriteS5920OpReg(pBoard, S5920_RCR,
			S5920_RCR_NVRAM_LOAD_HI_ADDR | (HIBYTE(uhAddress) <<
							S5920_RCR_NVRAM_OP_ADDR_SHIFT));

	/* Initiate read operation */

	WriteS5920OpReg(pBoard, S5920_RCR, S5920_RCR_NVRAM_READ_DATA);

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	/* Read the data */

	usleep_range(read_cycles_delay_us, read_cycles_delay_us + 1);
	u32RCRValue = PollAndGetRCR(pBoard);
	if (u32RCRValue & S5920_RCR_NVRAM_BUSY)
		return -EBUSY;
	if (u32RCRValue & S5920_RCR_NVRAM_ACCESS_FAILED) {
		dev_err(&pBoard->pPciDev->dev,
			"S5920 nvRAM XFER_FAIL asserted after read at address 0x%03x\n",
			uhAddress);
		return -EIO;
	}

	*pchByte =
	    (unsigned char)((u32RCRValue & S5920_RCR_NVRAM_OP_DATA_MASK) >>
			    S5920_RCR_NVRAM_OP_DATA_SHIFT);
	return 0;
}

/*****************************************************************************/

/*
 * PhysWriteNvrByte
 * Physically write an NVRAM byte to NVRAM address uhAddress. If successful, 0
 * is returned. Otherwise, the return code is a -ve errno.
 */

static int PhysWriteNvrByte(struct pciaer_board *pBoard, unsigned short uhAddress,
			    unsigned char chByte)
{
	u32 u32RCRValue;
	unsigned long ulHwLockFlags;

	/* Verify that the busy bit is not set. */

	u32RCRValue = PollAndGetRCR(pBoard);
	if (u32RCRValue & S5920_RCR_NVRAM_BUSY)
		return -EBUSY;

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	/* Write the low address byte */

	WriteS5920OpReg(pBoard, S5920_RCR,
			S5920_RCR_NVRAM_LOAD_LO_ADDR | (LOBYTE(uhAddress) <<
							S5920_RCR_NVRAM_OP_ADDR_SHIFT));

	/* Write the high address byte */

	WriteS5920OpReg(pBoard, S5920_RCR,
			S5920_RCR_NVRAM_LOAD_HI_ADDR | (HIBYTE(uhAddress) <<
							S5920_RCR_NVRAM_OP_ADDR_SHIFT));

	/*
	 * Last check for CAP_SYS_ADMIN. If control arrived here correctly, this
	 * has already been checked with less drastic consequences if the
	 * condition is not satisfied.
	 */

	BUG_ON(!capable(CAP_SYS_ADMIN));

	/* Initiate write operation */

	WriteS5920OpReg(pBoard, S5920_RCR,
			S5920_RCR_NVRAM_WRITE_DATA | (chByte <<
						      S5920_RCR_NVRAM_OP_DATA_SHIFT));

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	/* Poll for completion of the write */

	msleep(1 + S5920_NVR_MIN_WRITE_TIME_MS);
	u32RCRValue = PollAndGetRCR(pBoard);
	if (u32RCRValue & S5920_RCR_NVRAM_BUSY) {
		msleep(S5920_NVR_MAX_WRITE_TIME_MS -
		       S5920_NVR_MIN_WRITE_TIME_MS);
		u32RCRValue = PollAndGetRCR(pBoard);
	}

	if (u32RCRValue & S5920_RCR_NVRAM_BUSY)
		return -EBUSY;
	if (u32RCRValue & S5920_RCR_NVRAM_ACCESS_FAILED) {
		dev_err(&pBoard->pPciDev->dev,
			"S5920 nvRAM XFER_FAIL asserted after write at address 0x%03x\n",
			uhAddress);
		return -EIO;
	}

	return 0;
}

/*****************************************************************************/

/*
 * RequiredByteValue
 * Returns the value required to be present in the given address or -1 if any
 * value can be permitted.
 */

static int RequiredByteValue(unsigned short uhAddress)
{
	switch (uhAddress) {
	case S5920_NVRAM_OFFSET_VENDOR_ID:
		return LOBYTE(PCI_VENDOR_ID_AMCC);
		break;
	case S5920_NVRAM_OFFSET_VENDOR_ID + 1:
		return HIBYTE(PCI_VENDOR_ID_AMCC);
		break;
	case S5920_NVRAM_OFFSET_DEVICE_ID:
		return LOBYTE(PCI_DEVICE_ID_AMCC_S5920);
		break;
	case S5920_NVRAM_OFFSET_DEVICE_ID + 1:
		return HIBYTE(PCI_DEVICE_ID_AMCC_S5920);
		break;
	case S5920_NVRAM_OFFSET_REQD_LSB_BADR0:
		return S5920_LSB_BADR0_MEM;
		break;
	case S5920_NVRAM_OFFSET_REQD_VALID_1:
		return S5920_NVRAM_REQD_VALID_1;
		break;
	case S5920_NVRAM_OFFSET_REQD_VALID_2:
		return S5920_NVRAM_REQD_VALID_2;
		break;
	case S5920_NVRAM_OFFSET_REQD_VALID_3:
		return S5920_NVRAM_REQD_VALID_3;
		break;
	default:
		break;
	}
	return -1;
}

/*****************************************************************************/

/*
 * GetS5920NvrCache
 * If the NVRAM cache has already been allocated and initialized, the function
 * returns true immediately. Otherwise it attempts to allocate memory for the
 * cache and to initialize the cache flags appropriately. If successful, it
 * initializes the pointers S5920NvrCacheBytes & S5920NvrCacheFlags in the board
 * structure and again returns true. If cache creation fails, it returns false.
 * GetS5920NvrCache must be called with mtxS5920NvramCache held.
 * NB: As currently implemented, one devm_kzalloc allocation serves for both the
 * cached bytes and contiguously for the corresponding cache flag bytes.
 * Therefore pBoard->S5920NvrCacheBytes holds a devm_kzalloc'd address, whereas
 * pBoard->S5920NvrCacheFlags holds a pointer to half way through the same
 * block, and only the former should be devm_kfree'd at the end of the cache's
 * lifetime.
 * Freeing the cache should be done by calling FreeS5920NvrCache.
 * It is intended that by calling this GetS5920NvrCache function the NVRAM cache
 * should be allocated and initialized at time of first use, but it need only be
 * freed at module cleanup time.
 */

static bool GetS5920NvrCache(struct pciaer_board *pBoard)
{
	unsigned int uCacheSize;
	unsigned char *pCache;
	unsigned char *pCacheFlags;
	int i;

	lockdep_assert_held(&pBoard->mtxS5920NvramCache);

	/* If cache already established, return true immediately */

	if (pBoard->S5920NvrCacheBytes != NULL)
		return true;

	/*
	 * Attempt allocation. NB: One allocation serves for both the cached
	 * bytes and contiguously for the corresponding cache flags.
	 * If allocation fails, return false.
	 */

	uCacheSize = pBoard->uS5920NvrSize;

	pCache = devm_kzalloc(&pBoard->pPciDev->dev, uCacheSize * 2, GFP_KERNEL);
	if (pCache == NULL) {
		dev_warn(&pBoard->pPciDev->dev,
			 "Failed to get memory for S5920 NVRAM cache\n");
		return false;
	}

	pCacheFlags = pCache + uCacheSize;

	/*
	 * All the cache flags (PRESENT, DIRTY & READONLY) are initially false
	 * thanks to the devm_kzalloc. 'Required' values (Table 1, p2-74) must
	 * be marked READONLY.
	 */

	for (i = 0; i < uCacheSize; i++)
		if (RequiredByteValue(i) >= 0)
			pCacheFlags[i] |= NVR_CACHE_FLAG_READONLY;

	/* Store appropriate pointers in the board structure. */

	pBoard->S5920NvrCacheBytes = pCache;
	pBoard->S5920NvrCacheFlags = pCacheFlags;

	return true;
}

/*****************************************************************************/

/*
 * ResizeCache
 * Called from Nvr_open if it discovers that the length byte in the NVRAM has
 * changed since the last Nvr_open and if there is an extant cache. The function
 * proceeds by allocating space for the newly sized cache and copying the
 * contents of the existing cache to new one before replacing the old with the
 * new. If the operation is successful, the function also updates the record of
 * the NVRAM size and hence cache size in the board structure.
 * ResizeCache must be called with mtxS5920NvramCache held.
 */

static void ResizeCache(struct pciaer_board *pBoard, unsigned int uNewSize)
{
	unsigned char *pNewCache;
	unsigned char *pNewCacheFlags;
	unsigned int uCacheEntriesToCopy;
	unsigned int i;

	lockdep_assert_held(&pBoard->mtxS5920NvramCache);

	if (pBoard->S5920NvrCacheBytes == NULL
	    || uNewSize == pBoard->uS5920NvrSize)
		return;		/* Nothing to do */

	/* Allocate new cache */

	pNewCache = devm_kzalloc(&pBoard->pPciDev->dev, uNewSize * 2, GFP_KERNEL);
	if (pNewCache == NULL) {
		dev_warn(&pBoard->pPciDev->dev,
			 "Failed to get memory to resize S5920 NVRAM cache\n");
		return;
	}

	pNewCacheFlags = pNewCache + uNewSize;

	/* Copy from old to new */

	uCacheEntriesToCopy = min(pBoard->uS5920NvrSize, uNewSize);
	for (i = 0; i < uCacheEntriesToCopy; i++) {
		pNewCache[i] = pBoard->S5920NvrCacheBytes[i];
		pNewCacheFlags[i] = pBoard->S5920NvrCacheFlags[i];
	}

	/*
	 * Thanks to devm_kzalloc, all of the flags in any remaining part of the
	 * new cache (if larger than the old) will be clear. This
	 * part cannot contain 'required' bytes that should be marked read-only,
	 * because such bytes occur only within the range of the smallest
	 * possible NVRAM/cache size (128 bytes).
	 */

	/* Free old cache & replace with new */

	devm_kfree(&pBoard->pPciDev->dev, pBoard->S5920NvrCacheBytes);
	pBoard->S5920NvrCacheBytes = pNewCache;
	pBoard->S5920NvrCacheFlags = pNewCacheFlags;

	/* Update NVRAM/cache size in board struct */

	pBoard->uS5920NvrSize = uNewSize;
}

/*****************************************************************************/

/*
 * FreeS5920NvrCache
 * Called at module cleanup time (or exceptionally by S5920OomHandler) to free
 * up any memory devm_kzalloc'd for the NVRAM cache by GetS5920NvrCache.
 * FreeS5920NvrCache must be called with mtxS5920NvramCache held.
 * As currently implemented, one devm_kzalloc allocation serves for both the
 * cached bytes and contiguously for the corresponding cache flag bytes.
 * Therefore pBoard->S5920NvrCacheBytes holds a devm_kzalloc'd address, whereas
 * pBoard->S5920NvrCacheFlags holds a pointer to half way through the same
 * block, and so only the former is devm_kfree'd.
 * When this function is called at module cleanup time, because the
 * module cannot be cleaned up until all devices are closed, there can be no
 * question of having dirty bytes in the cache, since any dirty bytes are
 * written back to the NVRAM at the time the device is closed, since Nvr_release
 * calls FlushCache.
 * S5920OomHandler only frees the cache if it finds it is not dirty.
 */

static void FreeS5920NvrCache(struct pciaer_board *pBoard)
{
	lockdep_assert_held(&pBoard->mtxS5920NvramCache);

	/* If there's an allocated NVRAM cache, free it. */

	if (pBoard->S5920NvrCacheBytes != NULL) {
		devm_kfree(&pBoard->pPciDev->dev, pBoard->S5920NvrCacheBytes);
		pBoard->S5920NvrCacheBytes = NULL;
		pBoard->S5920NvrCacheFlags = NULL;
	}
}

/*****************************************************************************/

/*
 * ReadCacheNvrByte
 * Read an NVRAM byte from the cache, physically from the NVRAM if the byte is
 * not cached, updating the cache along the way.
 * This function must only be called with mtxS5920NvramCache held and after
 * GetS5920NvrCache returns true, and uhAddress must be < pBoard->uS5920NvrSize.
 * If successful, 0 is returned and the byte read is placed into *pchByte.
 * Otherwise, the return code is a -ve errno.
 */

static int ReadCacheNvrByte(struct pciaer_board *pBoard, unsigned short uhAddress,
			    unsigned char *pchByte)
{
	lockdep_assert_held(&pBoard->mtxS5920NvramCache);

	/*
	 * If there is a cache, and the byte is present in the cache, read it
	 * from there, otherwise go to the physical NVRAM, and update the
	 * cache.
	 */

	if (pBoard->S5920NvrCacheFlags[uhAddress] & NVR_CACHE_FLAG_PRESENT) {
		*pchByte = pBoard->S5920NvrCacheBytes[uhAddress];
	} else {
		int nError;

		nError = PhysReadNvrByte(pBoard, uhAddress, pchByte);
		if (nError != 0)
			return nError;
		pBoard->S5920NvrCacheBytes[uhAddress] = *pchByte;
		pBoard->S5920NvrCacheFlags[uhAddress] |= NVR_CACHE_FLAG_PRESENT;
	}

	return 0;
}

/*****************************************************************************/

/*
 * WriteCacheNvrByte
 * Write an NVRAM byte to the cache.
 * This function must only be called with mtxS5920NvramCache held and after
 * GetS5920NvrCache returns true, and uhAddress must be < pBoard->uS5920NvrSize.
 * If an attempt is made to overwrite one of the required byte values, the
 * required value will be written, irrespective of the value supplied. If the
 * O_SYNC bit is set in the uFlags parameter, the write is 'written through' to
 * the hardware, otherwise only the contents of the cache are changed. If
 * successful, the function returns 0, otherwise the return code is a -ve errno.
 */

static int WriteCacheNvrByte(struct pciaer_board *pBoard, unsigned short uhAddress,
			     unsigned char chByte, unsigned int uFlags)
{
	int nError;

	lockdep_assert_held(&pBoard->mtxS5920NvramCache);

	if (pBoard->S5920NvrCacheFlags[uhAddress] & NVR_CACHE_FLAG_READONLY) {
		unsigned char chRequiredValue;

		chRequiredValue = RequiredByteValue(uhAddress);
		if (chByte != chRequiredValue) {
			dev_notice(&pBoard->pPciDev->dev,
				   "Writing required value 0x%02x instead of supplied value "
					"0x%02x at S5920 NVRAM address 0x%03x\n",
				   chRequiredValue, chByte, uhAddress);
			chByte = chRequiredValue;
		}
	}

	if (uFlags & O_SYNC) {
		nError = PhysWriteNvrByte(pBoard, uhAddress, chByte);
		if (nError != 0)
			return nError;
		pBoard->S5920NvrCacheBytes[uhAddress] = chByte;
		pBoard->S5920NvrCacheFlags[uhAddress] |= NVR_CACHE_FLAG_PRESENT;
		pBoard->S5920NvrCacheFlags[uhAddress] &= ~NVR_CACHE_FLAG_DIRTY;
	} else
	    if (!
		(pBoard->S5920NvrCacheFlags[uhAddress] & NVR_CACHE_FLAG_PRESENT)
		|| pBoard->S5920NvrCacheBytes[uhAddress] != chByte) {
		pBoard->S5920NvrCacheBytes[uhAddress] = chByte;
		pBoard->S5920NvrCacheFlags[uhAddress] |=
		    (NVR_CACHE_FLAG_PRESENT | NVR_CACHE_FLAG_DIRTY);
	}

	return 0;
}

/*****************************************************************************/

/*
 * IsCacheDirty
 * Returns true if the cache is dirty and false if it is not.
 * Must be called with mtxS5920NvramCache held.
 */

static bool IsCacheDirty(struct pciaer_board *pBoard)
{
	unsigned int i;
	bool bDirty;

	lockdep_assert_held(&pBoard->mtxS5920NvramCache);

	bDirty = false;

	for (i = 0; i < pBoard->uS5920NvrSize; i++) {
		if (pBoard->S5920NvrCacheFlags[i] & NVR_CACHE_FLAG_DIRTY) {
			bDirty = true;
			break;
		}
	}

	return bDirty;
}

/*****************************************************************************/

/*
 * GetNvrCacheStatus
 * Implements PCIAER_IOC_GET_NVR_CACHE_STATUS, setting PCIAER_IOC_NVR_CACHE_DIRTY
 * in the integer argument if the cache is dirty and clearing it to zero
 * otherwise. Must be called with mtxS5920NvramCache NOT held. Returns -ENOMEM
 * if no cache could be allocated.
 */

static int GetNvrCacheStatus(struct pciaer_board *pBoard, __user int *pnArg)
{
	int nStatus;

	if (!access_ok(VERIFY_WRITE, pnArg, sizeof(int)))
		return -EFAULT;

	if (mutex_lock_interruptible(&pBoard->mtxS5920NvrCache) != 0)
		return -ERESTARTSYS;

	if (!GetS5920NvrCache(pBoard)) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return -ENOMEM;
	}

	nStatus = IsCacheDirty(pBoard) ? PCIAER_IOC_NVR_CACHE_DIRTY : 0;

	mutex_unlock(&pBoard->mtxS5920NvrCache);

	return __put_user(nStatus, pnArg);
}

/*****************************************************************************/

/*
 * RevertNvrCache
 * Implements PCIAER_IOC_REVERT_NVR_CACHE by marking dirtied bytes as not
 * present and not dirty, thus undoing any pending writes.
 * Returns -ENOMEM if no cache could be allocated, otherwise 0.
 */

static int RevertNvrCache(struct pciaer_board *pBoard)
{
	int nError;
	int i;

	nError = TestSysAdminCapability();
	if (nError != 0)
		return nError;

	if (mutex_lock_interruptible(&pBoard->mtxS5920NvrCache) != 0)
		return -ERESTARTSYS;

	if (!GetS5920NvrCache(pBoard)) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return -ENOMEM;
	}

	for (i = 0; i < pBoard->uS5920NvrSize; i++)
		if (pBoard->S5920NvrCacheFlags[i] & NVR_CACHE_FLAG_DIRTY)
			pBoard->S5920NvrCacheFlags[i] &=
			    ~(NVR_CACHE_FLAG_DIRTY | NVR_CACHE_FLAG_PRESENT);

	mutex_unlock(&pBoard->mtxS5920NvrCache);

	return 0;
}

/*****************************************************************************/

/*
 * FlushCache
 * Flushes the cache, if there is one. May be called from Nvr_fsync or
 * Nvr_release. Must be called with mtxS5920NvramCache NOT held.
 * The start and end parameters dictate the range of bytes to be flushed.
 * If end is UINT_MAX, then the upper limit is the end of the cache.
 * Returns a -ve errno in case of error.
 */

static int FlushCache(struct pciaer_board *pBoard, unsigned int start, unsigned int end)
{
	unsigned int i;
	int nByteError;
	int nFirstError;

	if (mutex_lock_interruptible(&pBoard->mtxS5920NvrCache) != 0)
		return -ERESTARTSYS;

	if (pBoard->S5920NvrCacheBytes == NULL) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return 0;
	}

	if (end == UINT_MAX)
		end = pBoard->uS5920NvrSize - 1;

	if (start >= pBoard->uS5920NvrSize || end >= pBoard->uS5920NvrSize) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return -EINVAL;
	}

	/*
	 * Wherever there's a byte with it's dirty flag set, write that byte to
	 * NVRAM and clear the dirty flag. If any error occurs, save the first
	 * one to return to the caller.
	 */

	nFirstError = 0;
	for (i = start; i <= end; i++)
		if (pBoard->S5920NvrCacheFlags[i] & NVR_CACHE_FLAG_DIRTY) {
			nByteError =
			    PhysWriteNvrByte(pBoard, (unsigned short)i,
					     pBoard->S5920NvrCacheBytes[i]);
			if (nByteError == 0)
				pBoard->S5920NvrCacheFlags[i] &=
				    ~NVR_CACHE_FLAG_DIRTY;
			else {
				dev_err(&pBoard->pPciDev->dev,
					"Failed to flush byte value 0x%02x to S5920 NVRAM address "
						"0x%03x (error %d)\n",
					pBoard->S5920NvrCacheBytes[i], i,
					-nByteError);
				if (nFirstError == 0)
					nFirstError = nByteError;
			}
		}

	mutex_unlock(&pBoard->mtxS5920NvrCache);

	return nFirstError;
}

/*****************************************************************************/

/*
 * S5920OomHandler
 * If our cache is not being actively used and is clean, we can free it.
 * The number of pages freed returned is always 0 though, since our cache is
 * in devm_kzalloc'd memory, and not directly in pages.
 */

unsigned long S5920OomHandler(struct pciaer_board *pBoard)
{
	if (mutex_trylock(&pBoard->mtxS5920NvrCache)) {
		if (pBoard->S5920NvrCacheBytes != NULL && !IsCacheDirty(pBoard)) {
			FreeS5920NvrCache(pBoard);
			dev_dbg(&pBoard->pPciDev->dev, "S5920OomHandler freed cache\n");
		}
		mutex_unlock(&pBoard->mtxS5920NvrCache);
	}

	return 0UL;
}

/*****************************************************************************/

/*
 * Nvr_poll
 * The poll method always returns with the POLLIN bit set to show that the
 * device can be read from, even if the read would block, since only the act of
 * attempting the read can cause new data to become available. Whether or not
 * the POLLOUT bit is set to show that the device can be written to without
 * blocking depends only on whether the O_SYNC flag is set, since if it is not,
 * write always writes only to the cache, whereas if O_SYNC is set, write will
 * block while writing to the hardware. At end of file, POLLHUP is set.
 */

static unsigned int Nvr_poll(struct file *fp,
			     struct poll_table_struct *pPollTable)
{
	struct pciaer_board *pBoard;
	unsigned int uPollFlags;

	pBoard = (struct pciaer_board *) (fp->private_data);

	poll_wait(fp, NULL, pPollTable);

	uPollFlags = POLLIN | POLLRDNORM;

	if (!(fp->f_flags & O_SYNC))
		uPollFlags |= POLLOUT | POLLWRNORM;

	if (fp->f_pos >= pBoard->uS5920NvrSize)
		uPollFlags |= POLLHUP;

	return uPollFlags;
}

/*****************************************************************************/

/*
 * Nvr_llseek
 * Allows seeks relative to EOF, which is taken to be at pBoard->uS5920NvrSize.
 * Seeks to before the beginning of the file or beyond EOF (or invalid 'whence'
 * parameters) give rise to the error EINVAL.
 */

static loff_t Nvr_llseek(struct file *fp, loff_t Offset, int iWhence)
{
	struct pciaer_board *pBoard;
	loff_t NewOffset;

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (mutex_lock_interruptible(&pBoard->mtxS5920NvrCache) != 0)
		return -ERESTARTSYS;

	switch (iWhence) {
	case 0:		/* SEEK_SET: */
		NewOffset = Offset;
		break;
	case 1:		/* SEEK_CUR: */
		NewOffset = fp->f_pos + Offset;
		break;
	case 2:		/* SEEK_END: */
		NewOffset = pBoard->uS5920NvrSize + Offset;
		break;
	default:
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return -EINVAL;
		break;
	}

	if (NewOffset < 0 || NewOffset > pBoard->uS5920NvrSize) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return -EINVAL;
	}

	mutex_unlock(&pBoard->mtxS5920NvrCache);

	fp->f_pos = NewOffset;

	return NewOffset;
}

/*****************************************************************************/

/*
 * ReadDirect
 * This is a fall back for Nvr_read to call in the event that no cache could be
 * allocated. It's very simple and just reads one byte from the given offset in
 * NVRAM to the user's buffer, and advances the offset by one.
 */

static ssize_t ReadDirect(struct pciaer_board *pBoard,
			  __user char *pchUserBuffer, loff_t *pOffset)
{
	unsigned char chByte;
	int nError;

	nError = PhysReadNvrByte(pBoard, *pOffset, &chByte);
	if (nError != 0)
		return nError;

	nError = __put_user(chByte, pchUserBuffer);
	if (nError != 0)
		return nError;

	(*pOffset)++;
	return 1;
}

/*****************************************************************************/

/*
 * Nvr_read
 */

static ssize_t Nvr_read(struct file *fp, __user char *pchUserBuffer,
			size_t BytesRequested, loff_t *pOffset)
{
	struct pciaer_board *pBoard;
	loff_t EndOffset;
	unsigned short uhStartAddress;
	unsigned short uhEndAddress;
	unsigned short uhAddress;
	int nError;
	unsigned int uBytesAvailable;
	unsigned char chByte;
	unsigned int uBytesCopied;

	if (BytesRequested == 0)
		return 0;

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (mutex_lock_interruptible(&pBoard->mtxS5920NvrCache) != 0)
		return -ERESTARTSYS;

	if (*pOffset >= pBoard->uS5920NvrSize) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return 0;
	}

	if (!GetS5920NvrCache(pBoard)) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return ReadDirect(pBoard, pchUserBuffer, pOffset);
	}

	uhStartAddress = (unsigned short)*pOffset;
	EndOffset = *pOffset + BytesRequested;
	uhEndAddress =
	    (unsigned short)min_t(loff_t, EndOffset, pBoard->uS5920NvrSize);

	nError = 0;
	uBytesAvailable = 0;
	for (uhAddress = uhStartAddress; uhAddress < uhEndAddress; uhAddress++) {
		nError = ReadCacheNvrByte(pBoard, uhAddress, &chByte);
		if (nError != 0)
			break;
		uBytesAvailable++;
	}

	if (uBytesAvailable == 0) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return nError;
	}

	uBytesCopied =
	    uBytesAvailable - (unsigned int)__copy_to_user(pchUserBuffer,
							   &pBoard->
							   S5920NvrCacheBytes
							   [uhStartAddress],
							   uBytesAvailable);

	mutex_unlock(&pBoard->mtxS5920NvrCache);

	if (uBytesCopied == 0)
		return -EFAULT;

	*pOffset += uBytesCopied;
	return uBytesCopied;
}

/*****************************************************************************/

/*
 * WriteDirect
 * This is a fall back for Nvr_write to call in the event that no cache could be
 * allocated. It's very simple and just writes one byte to the given offset in
 * NVRAM from the user's buffer, and advances the offset by one. NB: Unlike
 * Nvr_write, WriteDirect returns -EINVAL if asked to write an incorrect
 * required byte value.
 */

static ssize_t WriteDirect(struct pciaer_board *pBoard,
			  const __user char *pchUserBuffer, loff_t *pOffset)
{
	int nError;
	unsigned char chByte;
	int nRequiredByte;

	nError = __get_user(chByte, pchUserBuffer);
	if (nError != 0)
		return nError;

	nRequiredByte = RequiredByteValue(*pOffset);
	if (nRequiredByte >= 0 && chByte != (unsigned char)nRequiredByte)
		return -EINVAL;

	nError = PhysWriteNvrByte(pBoard, *pOffset, chByte);
	if (nError != 0)
		return nError;

	(*pOffset)++;
	return 1;
}

/*****************************************************************************/

/*
 * Nvr_write
 */

static ssize_t Nvr_write(struct file *fp, const __user char *pchUserBuffer,
			 size_t BytesToWrite, loff_t *pOffset)
{
	struct pciaer_board *pBoard;
	struct device *pDev;
	unsigned char *pchKernBuffer;
	unsigned short uhStartAddress;
	unsigned short uhEndAddress;
	unsigned short uhAddress;
	int nError;
	unsigned int iByte;

	nError = TestSysAdminCapability();
	if (nError != 0)
		return nError;

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (mutex_lock_interruptible(&pBoard->mtxS5920NvrCache) != 0)
		return -ERESTARTSYS;

	if (BytesToWrite == 0 || *pOffset >= pBoard->uS5920NvrSize) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return 0;
	}

	if (!GetS5920NvrCache(pBoard)) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return WriteDirect(pBoard, pchUserBuffer, pOffset);
	}

	BytesToWrite =
	    min(BytesToWrite, (size_t) (pBoard->uS5920NvrSize - *pOffset));

	pDev = &pBoard->pPciDev->dev;
	pchKernBuffer = devm_kzalloc(pDev, BytesToWrite, GFP_KERNEL);
	if (pchKernBuffer == NULL) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		return -ENOMEM;
	}

	if (__copy_from_user(pchKernBuffer, pchUserBuffer, BytesToWrite) != 0) {
		mutex_unlock(&pBoard->mtxS5920NvrCache);
		devm_kfree(pDev, pchKernBuffer);
		return -EFAULT;
	}

	uhStartAddress = (unsigned short)*pOffset;
	uhEndAddress = (unsigned short)(*pOffset + BytesToWrite);
	nError = 0;
	iByte = 0;
	for (uhAddress = uhStartAddress; uhAddress < uhEndAddress; uhAddress++) {
		nError =
		    WriteCacheNvrByte(pBoard, uhAddress, pchKernBuffer[iByte],
				      fp->f_flags);
		if (nError != 0)
			break;
		iByte++;
	}

	mutex_unlock(&pBoard->mtxS5920NvrCache);
	devm_kfree(pDev, pchKernBuffer);

	if (iByte == 0)
		return nError;

	*pOffset += iByte;
	return iByte;
}

/*****************************************************************************/

/*
 * Nvr_ioctl
 * This function should handles all the ioctls on the S5920 NVRAM minor device.
 */

static long Nvr_ioctl(struct file *fp, unsigned int uCmd, unsigned long ulArgs)
{
	int nError;
	struct pciaer_board *pBoard;

	nError = CheckIoctlPerm(fp, uCmd);
	if (nError != 0)
		return nError;

	pBoard = (struct pciaer_board *) (fp->private_data);

	switch (uCmd) {
	case PCIAER_IOC_GET_VERSION_INFO:
		nError =
		    GetVersionInfo(pBoard,
				   (__user struct pciaer_version_info *) ulArgs);
		break;
	case PCIAER_IOC_GET_NVR_CACHE_STATUS:
		nError = GetNvrCacheStatus(pBoard, (__user int *)ulArgs);
		break;
	case PCIAER_IOC_REVERT_NVR_CACHE:
		nError = RevertNvrCache(pBoard);
		break;
	default:
		nError = -ENOTTY;
		break;
	}

	return nError;
}

/*****************************************************************************/

/*
 * Nvr_open
 * Supplementary to PciAerMain_open, fail open with EINVAL if non-blocking mode
 * is requested. Otherwise attempt to determine the active NVRAM size by looking
 * at the ROM length byte.
 */

static int Nvr_open(struct inode *INode, struct file *fp)
{
	struct pciaer_board *pBoard;
	char chLengthByte;
	unsigned int uAssumedNvrSize;

	if (fp->f_flags & O_NONBLOCK)
		return -EINVAL;

	pBoard = (struct pciaer_board *) (fp->private_data);

	uAssumedNvrSize = S5920_NVRAM_MIN_SIZE;

	if (mutex_lock_interruptible(&pBoard->mtxS5920NvrCache) != 0)
		return -ERESTARTSYS;

	if (PhysReadNvrByte(pBoard, ISA_EXP_ROM_LEN_512B_OFFSET, &chLengthByte)
	    == 0) {
		if (chLengthByte > 0
		    && chLengthByte * 512 <= S5920_NVRAM_MAX_SIZE)
			uAssumedNvrSize = chLengthByte * 512;
		if (pBoard->S5920NvrCacheBytes != NULL
		    && uAssumedNvrSize != pBoard->uS5920NvrSize)
			ResizeCache(pBoard, uAssumedNvrSize);
	}
	if (pBoard->S5920NvrCacheBytes == NULL)
		pBoard->uS5920NvrSize = uAssumedNvrSize;

	mutex_unlock(&pBoard->mtxS5920NvrCache);

	INode->i_blkbits = 0;

	return 0;
}

/*****************************************************************************/

/*
 * Nvr_release
 * Flushes the entire cache and calls PciAer_release.
 */

static int Nvr_release(struct inode *INode, struct file *fp)
{
	FlushCache((struct pciaer_board *)(fp->private_data), 0, UINT_MAX);
	return PciAer_release(INode, fp);
}

/*****************************************************************************/

/*
 * Nvr_fsync
 * Calls FlushCache after performing simple range checks on the start and end
 * parameters and passes them as unsigned ints to FlushCache. The S5920 NVRAM
 * is at most 64K so unsigned ints more than suffice as offsets for our
 * purposes. Note that we don't compare start and end with uS5920NvrSize here,
 * since we need to hold the mtxS5920NvramCache mutex to read that, and we
 * leave that to FlushCache.
 * The datasync parameter is ignored as we have no metadata to flush.
 */

static int Nvr_fsync(struct file *fp, loff_t start, loff_t end, int datasync)
{
	if (end == LLONG_MAX)
		end = UINT_MAX;

	if (start > UINT_MAX || end > UINT_MAX || start > end)
		return -EINVAL;

	return FlushCache((struct pciaer_board *)(fp->private_data),
			  (unsigned int)start, (unsigned int)end);
}

/*****************************************************************************/

/*
 * NvrFileOps
 * The dispatch table for accesses to the S5920 NVRAM minor device.
 */

const struct file_operations NvrFileOps = {
	.owner = THIS_MODULE,
	.llseek = Nvr_llseek,
	.read = Nvr_read,
	.write = Nvr_write,
	.poll = Nvr_poll,
	.unlocked_ioctl = Nvr_ioctl,
	.open = Nvr_open,
	.release = Nvr_release,
	.fsync = Nvr_fsync,
};

/*****************************************************************************/

/*
 * S5920NvrCleanup
 * Called from ReleaseBoard in pcimodule.c
 * Its only purpose at present is to call FreeS5920NvrCache.
 */

void S5920NvrCleanup(struct pciaer_board *pBoard)
{
	mutex_lock(&pBoard->mtxS5920NvrCache);
	FreeS5920NvrCache(pBoard);
	mutex_unlock(&pBoard->mtxS5920NvrCache);
}

/*****************************************************************************/

/* EOF */
