/* sequencer.c

Sequencer minor device code for the Linux character device driver
module for the PCI-AER board developed by Vittorio Dante of the
Istituto Superiore di Sanita' in Rome.

Copyright (C) 2000 & 2003-2012 Adrian Maurice Whatley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This code was written by:
Adrian M. Whatley (amw@ini.phys.ethz.ch)
Institut fuer Neuroinformatik (www.ini.uzh.ch)
ETH u. Universitaet Zuerich,
CH-8057 Zuerich, Switzerland

sequencer.c $Id: fda3ab450f4de8846229fbcf91b03be6727bfe90 $
*/

#ifdef CONFIG_PCIAER_DEBUG
#define DEBUG
#endif

#include <linux/device.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "pciaerhw.h"
#include "internal.h"

/*****************************************************************************/

static inline unsigned int SeqBufferOrder(struct pciaer_board *pBoard)
{
	return get_order(pBoard->uSeqFifoDepth * sizeof(u32));
}

/*****************************************************************************/

void SeqFifoHalfEmptyISR(struct pciaer_board *pBoard)
{
	atomic_inc(&pBoard->SeqFifoHalfEmptyIntCount);
}

/*****************************************************************************/

void SeqFifoEmptyISR(struct pciaer_board *pBoard)
{
	atomic_inc(&pBoard->SeqFifoEmptyIntCount);
}

/*****************************************************************************/

/*
 * GetSequencerStats
 * Fills in an internal_stats struct on behalf of debugfs.c and resets the
 * statistics if required.
 */

int GetSequencerStats(struct pciaer_board *pBoard, struct internal_stats *pStats,
		      bool fReset)
{
	unsigned long ulWords;
	unsigned long ulWrites;
	int cFifoHalfEmptyInts;
	int cFifoEmptyInts;

	if (mutex_trylock(&pBoard->mtxSequencer) == 0)
		return -EAGAIN;

	ulWords = pBoard->ulSeqWordsWritten;
	ulWrites = pBoard->ulSeqWrites;
	cFifoHalfEmptyInts = atomic_read(&pBoard->SeqFifoHalfEmptyIntCount);
	cFifoEmptyInts = atomic_read(&pBoard->SeqFifoEmptyIntCount);

	if (fReset) {
		pBoard->ulSeqWordsWritten = 0L;
		pBoard->ulSeqWrites = 0L;
		atomic_sub(cFifoHalfEmptyInts,
			   &pBoard->SeqFifoHalfEmptyIntCount);
		atomic_sub(cFifoEmptyInts, &pBoard->SeqFifoEmptyIntCount);
	}

	mutex_unlock(&pBoard->mtxSequencer);

	if (pStats != NULL) {
		pStats->words_transferred = ulWords;
		pStats->user_transfers = ulWrites;
		pStats->total_interrupts = cFifoHalfEmptyInts + cFifoEmptyInts;
		pStats->overflows_underflows = cFifoEmptyInts;
		pStats->memory_usage = PAGE_SIZE << SeqBufferOrder(pBoard);
	}

	return 0;
}

/*****************************************************************************/

/*
 * EnableSequencer
 * Enable sequencer, bring sequencer FIFO out of reset and record in board
 * struct that sequencer is enabled. Must be called with mtxSequencer held.
 */

static void EnableSequencer(struct pciaer_board *pBoard)
{
	void __iomem *pCfgRegA1;
	unsigned long ulHwLockFlags;
	u32 u32RegContents;

	pCfgRegA1 = XilinxRegAddress(pBoard, CFG_REG_A1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	u32RegContents = ioread32(pCfgRegA1);
	u32RegContents &= ~CRA1_SEQ_DISABLE;	/* Enable sequencer */
	u32RegContents |= CRA1_SEQ_FIFO_MR_;	/* Bring sequencer FIFO out of reset */
	iowrite32(u32RegContents, pCfgRegA1);
	FlushPciWrites(pBoard);

	pBoard->fSeqEnabled = true;

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	EnableInt(pBoard, CRC1_DIS_INT_FE_SEQ | CRC1_DIS_INT_FH_SEQ);
}

/*****************************************************************************/

/*
 * DisableSequencer
 * Disable sequencer, put sequencer FIFO in reset and record in board struct
 * that sequencer is disabled. Must be called with mtxSequencer held.
 */

static void DisableSequencer(struct pciaer_board *pBoard)
{
	void __iomem *pCfgRegA1;
	unsigned long ulHwLockFlags;
	u32 u32RegContents;

	DisableInt(pBoard, CRC1_DIS_INT_FE_SEQ | CRC1_DIS_INT_FH_SEQ);

	pCfgRegA1 = XilinxRegAddress(pBoard, CFG_REG_A1);

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);

	u32RegContents = ioread32(pCfgRegA1);
	u32RegContents |= CRA1_SEQ_DISABLE;	/* Disable sequencer */
	u32RegContents &= ~CRA1_SEQ_FIFO_MR_;	/* Put sequencer FIFO in reset */
	iowrite32(u32RegContents, pCfgRegA1);
	FlushPciWrites(pBoard);

	pBoard->fSeqEnabled = false;

	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	msleep(FIFO_RESET_DELAY_MS);	/* Ensure minimum FIFO reset time */
}

/*****************************************************************************/

/*
 * Seq_open
 * Supplementary to PciAerMain_open, fail open with ENXIO if all the relevant
 * hardware is not available, EOPNOTSUPP if the Sequencer FIFO appears to be
 * 0-depth, EBUSY if another user is using the board, or ENOMEM if no buffer can
 * be allocated.
 */

static int Seq_open(struct inode *INode, struct file *fp)
{
	struct pciaer_board *pBoard;

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (S5920OpRegAddress(pBoard, 0) == NULL
	    || XilinxRegAddress(pBoard, 0) == NULL
	    || FifoAddress(pBoard) == NULL)
		return -ENXIO;

	if (pBoard->uSeqFifoDepth == 0)
		return -EOPNOTSUPP;

	if (IncSingleUserLock(pBoard) != 0)
		return -EBUSY;

	if (fp->f_flags & O_NONBLOCK) {
		if (mutex_trylock(&pBoard->mtxSequencer) == 0) {
			DecSingleUserLock(pBoard);
			return -EAGAIN;
		}
	} else {
		if (mutex_lock_interruptible(&pBoard->mtxSequencer) != 0) {
			DecSingleUserLock(pBoard);
			return -ERESTARTSYS;
		}
	}

	if (pBoard->pSequencerBuffer == NULL) {
		pBoard->pSequencerBuffer =
		    (void *)__get_free_pages(GFP_KERNEL,
					     SeqBufferOrder(pBoard));
		if (pBoard->pSequencerBuffer == NULL) {
			mutex_unlock(&pBoard->mtxSequencer);
			DecSingleUserLock(pBoard);
			return -ENOMEM;
		}
	}
	pBoard->nSequencerOpens++;
	mutex_unlock(&pBoard->mtxSequencer);

	INode->i_blkbits = ilog2(pBoard->uSeqFifoDepth * sizeof(u32) / 2);

	return nonseekable_open(INode, fp);
}

/*****************************************************************************/

/*
 * Seq_release
 * Decrement the sequencer open count, and if it's reached 0, free the sequencer
 * buffer and turn off the sequencer hardware. Finally decrement the single user
 * lock and call PciAer_release.
 */

static int Seq_release(struct inode *INode, struct file *fp)
{
	struct pciaer_board *pBoard;

	pBoard = (struct pciaer_board *) (fp->private_data);

	if (fp->f_flags & O_NONBLOCK) {
		if (mutex_trylock(&pBoard->mtxSequencer) == 0)
			return -EAGAIN;
	} else {
		if (mutex_lock_interruptible(&pBoard->mtxSequencer) != 0)
			return -ERESTARTSYS;
	}

	if (--pBoard->nSequencerOpens == 0) {
		free_pages((unsigned long)pBoard->pSequencerBuffer,
			   SeqBufferOrder(pBoard));
		pBoard->pSequencerBuffer = NULL;
		if (pBoard->fSeqEnabled)
			DisableSequencer(pBoard);
	}

	mutex_unlock(&pBoard->mtxSequencer);

	DecSingleUserLock(pBoard);

	return PciAer_release(INode, fp);
}

/*****************************************************************************/

/*
 * Seq_ioctl
 * This function handles all ioctls on the Sequencer minor device.
 */

static long Seq_ioctl(struct file *fp, unsigned int uCmd, unsigned long ulArgs)
{
	struct pciaer_board *pBoard;
	int nError;

	nError = CheckIoctlPerm(fp, uCmd);
	if (nError != 0 && uCmd != PCIAER_IOC_RST_CNTR_GET_TIME)
		return nError;

	pBoard = (struct pciaer_board *) (fp->private_data);

	nError = -ENOTTY;

	switch (uCmd) {
	case PCIAER_IOC_GET_VERSION_INFO:
		nError =
		    GetVersionInfo(pBoard,
				   (__user struct pciaer_version_info *) ulArgs);
		break;

	case PCIAER_IOC_SET_CNTR_PERIOD:
		nError = SetCounterPeriod(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_CNTR_PERIOD:
		nError =
		    put_user(GetCounterPeriod(pBoard), (__user int *)ulArgs);
		break;

	case PCIAER_IOC_RST_CNTR_GET_TIME:
		nError =
		    ResetCounterGetTime(pBoard, (__user struct timeval *)ulArgs,
					fp->f_mode);
		break;

	case PCIAER_IOC_GET_CNTR_LAST_RST_TIME:
		nError =
		    GetLastCounterResetTime(pBoard,
					    (__user struct timeval *)ulArgs);
		break;

	case PCIAER_IOC_GET_SEQ_FIFO_FLAGS:
		/* Check that the flag bit positions defined for the IOCTL match
		   those defined for the hardware */
		BUILD_BUG_ON(PCIAER_IOC_SEQ_EMPTY != SRA1_SEQ_FIFO_EF_);
		BUILD_BUG_ON(PCIAER_IOC_SEQ_HALF_FULL != SRA1_SEQ_FIFO_HF_);
		BUILD_BUG_ON(PCIAER_IOC_SEQ_FULL != SRA1_SEQ_FIFO_FF_);

		nError = put_user(~ioread32(XilinxRegAddress(pBoard, STAT_REG_A1)), (__user int *)ulArgs);	/* Inverts bits for +ve logic result */
		break;

	case PCIAER_IOC_RST_FIFO:
		ResetFifo(pBoard, CRA1_SEQ_FIFO_MR_);
		nError = 0;
		break;

	case PCIAER_IOC_SET_SEQ_ARB_CHANNEL:
		nError = SetSequencerArbiterChannel(pBoard, ulArgs);
		break;

	case PCIAER_IOC_GET_SEQ_ARB_CHANNEL:
		nError =
		    put_user(GetSequencerArbiterChannel(pBoard),
			     (__user int *)ulArgs);
		break;

	default:
		nError = -ENOTTY;
		break;
	}
	return nError;
}

/*****************************************************************************/

/*
 * GetSequencerSpaceAvailable
 * A helper function for Seq_write which returns the minimum number of DWords
 * free space available in the Sequencer FIFO.
 */

static size_t GetSequencerSpaceAvailable(const struct pciaer_board *pBoard)
{
	unsigned int uFifoFlags;
	size_t SpaceAvailable;

	uFifoFlags = ioread32(XilinxRegAddress(pBoard, STAT_REG_A1));
	SpaceAvailable = 0;
	switch (uFifoFlags & SEQ_FIFO_FLAGS_MASK) {
	case SRA1_SEQ_FIFO_HF_ | SRA1_SEQ_FIFO_FF_:
		/* Not half full, not full and not not empty, i.e. empty */
		SpaceAvailable = pBoard->uSeqFifoDepth;
		break;
	case SRA1_SEQ_FIFO_EF_ | SRA1_SEQ_FIFO_HF_ | SRA1_SEQ_FIFO_FF_:
		/* Not empty, not half full and not full, i.e. between empty and half full */
		SpaceAvailable = pBoard->uSeqFifoDepth / 2;
		break;
	case SRA1_SEQ_FIFO_EF_ | SRA1_SEQ_FIFO_FF_:
		/* Not empty, not full and not not half full, i.e between half full and full */
		SpaceAvailable = 1;
		break;
	case SRA1_SEQ_FIFO_EF_:
		/* Not empty, not not half full and not not full, i.e full */
		SpaceAvailable = 0;
		break;
	default:
		dev_err(&pBoard->pPciDev->dev,
			"Seemingly impossible combination of Sequencer FIFO flags detected: SRA1 octal 0%o\n",
			uFifoFlags & 0xFFFF);
		SpaceAvailable = 0;
		break;
	}
	return SpaceAvailable;
}

/*****************************************************************************/

/*
 * Seq_write
 * Write to sequencer. The buffer should contain u32's formatted according to
 * table 2 of the draft board spec. It must therefore be an exact number of
 * DWORDS long, otherwise the call will be failed with EINVAL.
 */

static ssize_t Seq_write(struct file *fp, const __user char *pchUserBuffer,
			 size_t BytesToWrite, loff_t *pOffset)
{
	size_t DWordsToWrite;
	struct pciaer_board *pBoard;
	size_t DWordsFreeInFifo;
	unsigned long ulHwLockFlags;

	if (unlikely(BytesToWrite % sizeof(u32) != 0))
		return -EINVAL;

	pBoard = (struct pciaer_board *) (fp->private_data);

	DWordsToWrite = BytesToWrite / sizeof(u32);

	if (fp->f_flags & O_NONBLOCK) {
		if (mutex_trylock(&pBoard->mtxSequencer) == 0)
			return -EAGAIN;
	} else {
		if (mutex_lock_interruptible(&pBoard->mtxSequencer) != 0)
			return -ERESTARTSYS;
	}

	pBoard->ulSeqWrites++;

	if (unlikely(!pBoard->fSeqEnabled))
		EnableSequencer(pBoard);

	DWordsFreeInFifo = GetSequencerSpaceAvailable(pBoard);

	if (DWordsFreeInFifo == 0) {
		mutex_unlock(&pBoard->mtxSequencer);

		if (fp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		else
			return 0;	/* @@@ Should block here */
	}

	if (DWordsFreeInFifo < DWordsToWrite)
		DWordsToWrite = DWordsFreeInFifo;

	BytesToWrite = DWordsToWrite * sizeof(u32);

	if (unlikely
	    (__copy_from_user
	     (pBoard->pSequencerBuffer, pchUserBuffer, BytesToWrite) != 0)) {
		mutex_unlock(&pBoard->mtxSequencer);
		return -EFAULT;
	}

	write_lock_irqsave(&pBoard->HardwareLock, ulHwLockFlags);
	memcpy_toio(FifoAddress(pBoard), pBoard->pSequencerBuffer,
		    BytesToWrite);
	write_unlock_irqrestore(&pBoard->HardwareLock, ulHwLockFlags);

	pBoard->ulSeqWordsWritten += DWordsToWrite;

	mutex_unlock(&pBoard->mtxSequencer);

	return BytesToWrite;
}

/*****************************************************************************/

/*
 * Seq_poll
 * Seq_poll is called by the kernel when it deciding whether to put the current
 * process to sleep as a result of a user's poll or select call. If any of the
 * returned flags matches a state which the user is interested in, the process
 * will not sleep; otherwise it may (if the user's timeout is not reached and
 * there are no signals) sleep interruptibly on the queue referred to in the
 * call to poll_wait.
 * @@@ Currently there is no interrupt support implemented, so there can be no
 * wake-up, so there is no wait queue and the only way to get back to the
 * running state once sleeping as a result of a user's poll or select is either
 * to timeout or to receive a signal.
 */

static unsigned int Seq_poll(struct file *fp, struct poll_table_struct *ptsp)
{
	struct pciaer_board *pBoard;
	unsigned int uPollFlags;

	pBoard = (struct pciaer_board *) (fp->private_data);

	poll_wait(fp, NULL, ptsp);	/*
					 * @@@ The 2nd parameter will eventually
					 * be a pointer to the Sequencer write
					 * wait queue.
					 */

	/* Always readable without blocking, since read always returns -EINVAL */

	uPollFlags = (POLLIN | POLLRDNORM);

	/* Iff the FIFO is not full, then we're also writeable without blocking */

	if (ioread32(XilinxRegAddress(pBoard, STAT_REG_A1)) & SRA1_SEQ_FIFO_FF_)
		uPollFlags |= (POLLOUT | POLLWRNORM);

	return uPollFlags;
}

/*****************************************************************************/

/*
 * Seq_fsync
 */

static int Seq_fsync(struct file *fp, loff_t start, loff_t end, int datasync)
{
	struct pciaer_board *pBoard;

	pBoard = (struct pciaer_board *) (fp->private_data);

	dev_dbg(&pBoard->pPciDev->dev, "Sequencer fsync\n");

	/* @@@ Wait for sequencer FIFO to be empty */

	return 0;
}

/*****************************************************************************/

/*
 * SeqFileOps
 * The dispatch table for accesses to the non-seekable, write-only Sequencer
 * minor device.
 */

const struct file_operations SeqFileOps = {
	.owner = THIS_MODULE,
	.write = Seq_write,
	.poll = Seq_poll,
	.unlocked_ioctl = Seq_ioctl,
	.open = Seq_open,
	.release = Seq_release,
	.fsync = Seq_fsync,
};

/*****************************************************************************/

/* EOF */
