/* PCI-AER library, Mapper support. */

/* mapper.c $Id: 6d3aa6a624047d6a2bb4cab4e2f2e41dceef13e6 $ */

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "pciaer.h"
#include "pciaerlib.h"


#define BASE_DEVICE_PATH	"/dev/aermap"


int PciaerMapOpen( unsigned int iBoard, int flags, int *pnHandle )
{
  char achDevicePathname[ sizeof(BASE_DEVICE_PATH) + 2 ]; /* Length is length of BASE_DEVICE_PATH + one digit + \0 */
  int nOpenReturn;

  /* Open requires that iBoard is less than 10 so that we can guarantee its
     string representation doesn't overflow the achDevicePathname buffer. */

  if ( iBoard > 9 )
    return EINVAL;

  if ( pnHandle == NULL )
    return EINVAL;

  sprintf( achDevicePathname, BASE_DEVICE_PATH "%1d", iBoard );

  nOpenReturn = open( achDevicePathname, flags | O_NOCTTY );

  if ( nOpenReturn == -1 )
    return errno;

  *pnHandle = nOpenReturn;
  return 0;
}


const volatile pciaer_mapper_sram_t * PciaerMapGetMapperMemory( int handle );


int PciaerMapSetMapping( int handle, unsigned short source, unsigned short count, const unsigned short *pDestList )
{
    pciaer_mapping_t mapping;
    
    mapping.source_ae = source;
    mapping.dest_count = count;
    mapping.dest_ae = (unsigned short *)pDestList;  /* cast only to remove const */

    return ioctl( handle, PCIAER_IOC_SET_MAP_MAPPING, &mapping ) ? errno : 0;
}

int PciaerMapGetMapping(int handle, unsigned short source, unsigned short bufsize, unsigned short *pBuffer, unsigned short *pCount)
{
    pciaer_mapping_t mapping;
    
    mapping.source_ae = source;
    mapping.dest_count = bufsize;
    mapping.dest_ae = pBuffer;

    if ( ioctl( handle, PCIAER_IOC_GET_MAP_MAPPING, &mapping ) != 0 )
	return errno;

    *pCount = mapping.dest_count;
    return 0;
}

int PciaerMapGetMappingCount(int handle, unsigned short source, unsigned short *pCount)
{
    pciaer_mapping_t mapping;
    
    mapping.source_ae = source;

    if ( ioctl( handle, PCIAER_IOC_GET_MAP_MAPPING_COUNT, &mapping ) != 0 )
	return errno;

    *pCount = mapping.dest_count;
    return 0;
}

int PciaerMapAddToMapping(int handle, unsigned short source, unsigned short count, const unsigned short *pDestList)
{
    pciaer_mapping_t mapping;
    
    /*
    //    mapping.source_ae = source;
    //    mapping.dest_count = count;
    //    mapping.dest_ae = (unsigned short *)pDestList;  // cast only to remove const
    //    return ioctl( handle, PCIAER_IOC_MAP_ADD_TO_MAPPING, &mapping ) ? errno : 0;
    */

	int ret;
	
	if ( count == 0 )
	    return 0;

	if ( pDestList == NULL )
	    return EINVAL;

	/* first, allocate buffer with needed length */

	mapping.source_ae = source;
	if ( ioctl( handle, PCIAER_IOC_GET_MAP_MAPPING_COUNT, &mapping ) != 0 )
	    return errno;
	mapping.dest_count += count;
	mapping.dest_ae = (unsigned short *)malloc( mapping.dest_count * sizeof(unsigned short) );
	if ( mapping.dest_ae == NULL )
	    return ENOMEM;
	
	/* then get the old destinations */

	if ( ioctl( handle, PCIAER_IOC_GET_MAP_MAPPING, &mapping ) != 0 )
	    {
		ret = errno;
		free( mapping.dest_ae );
		return ret;
	}
	
	/* copy the new destinations */
	memcpy( mapping.dest_ae + mapping.dest_count, pDestList, count * sizeof(unsigned short) );
	mapping.dest_count += count;
	
	/* set the new mapping */
    ret = ioctl( handle, PCIAER_IOC_SET_MAP_MAPPING, &mapping ) ? errno : 0;
	
	/* free buffer */
	free( mapping.dest_ae );
	
	return ret;
}

int PciaerMapDeleteFromMapping(int handle, unsigned short source, unsigned short count, const unsigned short *pDestList)
{
    pciaer_mapping_t mapping;
   
    /*
    //    mapping.source_ae = source;
    //    mapping.dest_count = count;
    //    mapping.dest_ae = (unsigned short *)pDestList;  // cast only to remove const
    //    return ioctl( handle, PCIAER_IOC_MAP_DEL_FROM_MAPPING, &mapping ) ? errno : 0;
    */

	int ret;
	unsigned i1, i2;
	int new_count;

	if ( count == 0 )
	    return 0;

	if ( pDestList == NULL )
	    return EINVAL;

	/* first, allocate buffer with needed length */
	mapping.source_ae = source;
	if ( ioctl( handle, PCIAER_IOC_GET_MAP_MAPPING_COUNT, &mapping ) != 0 )
	    return errno;

	/* time saving: if there is no dest address, we cannot remove anything */
	if ( mapping.dest_count == 0 )
	    return 0;

	mapping.dest_ae = (unsigned short *)malloc( mapping.dest_count * sizeof(unsigned short) );
	if( mapping.dest_ae == NULL )
	    return ENOMEM;

	/* Get the current mapping */
	
	if ( ioctl( handle, PCIAER_IOC_GET_MAP_MAPPING, &mapping ) != 0 )
	{
	    free( mapping.dest_ae );
	    return errno;
	}

	/* For each element in pDestList, we go through the current list of
	   destinations in mapping.dest_ae and mark for deletion the first (and
	   only the first) matching destination address (if any) by overwriting
	   it with the MAPPER_END_LABEL marker, which is not a valid address.
	   All the elements so marked are then squashed out of the
	   mapping.dest_ae list afterwards. */
	
	new_count = mapping.dest_count;
	for (i1 = 0; i1 < count; i1++) {
		for (i2 = 0; i2 < mapping.dest_count; i2++) {
			if (pDestList[i1] == mapping.dest_ae[i2]) {
			        mapping.dest_ae[i2] = MAPPER_END_LABEL;
				new_count--;
				break;
			}
		}
	}

	/* If any addresses were clobbered with MAPPER_END_LABEL to mark
	   elements for deletion, then we now delete those elements, squashing
	   the destination list together, and set the new mapping back on the
	   device. */

	ret = 0;
	if ( new_count != mapping.dest_count )
	    {
		for (i1 = 0, i2 = 0; i2 < new_count; i1++)
		    {
			if ( mapping.dest_ae[i1] != MAPPER_END_LABEL )
			    mapping.dest_ae[i2++] = mapping.dest_ae[i1];
		    }

		mapping.dest_count = new_count;

		if ( ioctl( handle, PCIAER_IOC_SET_MAP_MAPPING, &mapping ) != 0 )
		    ret = errno;
	    }

	free( mapping.dest_ae );
	return ret;
}

int PciaerMapGetFifoFillCount( int handle, unsigned long *pc )
{
	pciaer_stats_t s;
	int retval;

	s.size = sizeof(pciaer_stats_t);
	retval = PciaerGetStatistics( handle, &s );
	if ( retval != 0 )
		return retval;
	*pc = s.overflows_underflows;
	return 0;
}

/* EOF */
