/* moncookwtl.c $Id: 11eb63981465bfeb27dd1aa34c1fe8123d5c7716 $ */

#include "pciaer.h"
#include "pciaerlib.h"

/******************************************************************************/

extern int PciaerGetCounterPeriodLocalCopy( void );

/******************************************************************************/

long CookWithTimeLabels(const int *pRaw, unsigned int nRaw, pciaer_monitor_read_ae_t *pCooked, unsigned int nToCook, unsigned int *pnUsedRaw, unsigned int *pnCooked)
{
  int iRaw;
  int nClockPeriod_us;

  enum { ExpectAEAddr = 0, ExpectTimeHi = 1, ExpectTimeLo = 2 } Expected; /* The numerical values in this enum are important. See comment before final return. */

  *pnUsedRaw = 0;
  *pnCooked = 0;

  Expected = ExpectAEAddr;

  /* Skip any leading TIME_HI or TIME_LO words to find the first AER_ADDR word, but return if an ERROR word is found */

  iRaw = -1;

  while ( ((pRaw[ ++iRaw ] & MONITOR_DWORD_TYPE_MASK) != MONITOR_DWORD_AER_ADDR) && (iRaw < nRaw) )
      if ( (pRaw[ iRaw ] & MONITOR_DWORD_TYPE_MASK) == MONITOR_DWORD_ERROR )
	  {	
	      *pnUsedRaw = iRaw + 1;	/* Include the ERROR word in the 'used' part. */
	      return (long)pRaw[ iRaw ] << (8 * sizeof(long) / 2);
	  }

  nClockPeriod_us = PciaerGetCounterPeriodLocalCopy();

  /* Start the serious processing */

  while ( iRaw < nRaw && *pnCooked <= nToCook )
      {
	  switch ( pRaw[ iRaw ] & MONITOR_DWORD_TYPE_MASK )
	      {
	      case MONITOR_DWORD_AER_ADDR:
		  if ( Expected == ExpectAEAddr )
		      {
			  /* Normal correct case. We got an AER_ADDR when we were expecting one. */

			  pCooked[ *pnCooked ].ae = pRaw[ iRaw++ ] & MONITOR_DWORD_DATA_MASK;
			  Expected = ExpectTimeHi;
		      }
		  else if ( Expected == ExpectTimeHi )
		      {
			  /* Special case for dealing with consecutive AER_ADDRs, i.e. if we get another AER_ADDR when we were expecting a TIME_HI.
			     The assumption is that if there is a long line of consecutive AER_ADDR's followed
			     by a TIME_HI TIME_LO pair then we can use this last timestamp for all of them. */

			  unsigned int nConsecAddrs;
			  int nPutativeTimeHi;
			  int nPutativeTimeLo;
			  unsigned int uTimestamp;
			  int i;

			  /* First count how many consecutive AER_ADDRs we have in the buffer including and beyond the one indexed by iRaw. */
			  
			  nConsecAddrs = 1;
			  while ( (iRaw + nConsecAddrs < nRaw) && ((pRaw[ iRaw + nConsecAddrs ] & MONITOR_DWORD_TYPE_MASK) == MONITOR_DWORD_AER_ADDR) )
			      nConsecAddrs++;

			  if ( iRaw + nConsecAddrs >= nRaw - 1 )
			      {
				  /* The whole of the remainder of the raw buffer is filled with AER_ADDRs without there being room enough at the end
				     for two words of timestamp info. Therefore, return indicating that everything before these AER_ADDRs was used,
				     but don't consider the AER_ADDRs themselves used because if the function is retried with more raw data, we might
				     get a timestamp. */

				  *pnUsedRaw = iRaw - 1;
				  return 0L;
			      }

			  nPutativeTimeHi = pRaw[ iRaw + nConsecAddrs ];
			  nPutativeTimeLo = pRaw[ iRaw + nConsecAddrs + 1 ];

			  if (    ((nPutativeTimeHi & MONITOR_DWORD_TYPE_MASK) != MONITOR_DWORD_TIME_HI)
			       || ((nPutativeTimeLo & MONITOR_DWORD_TYPE_MASK) != MONITOR_DWORD_TIME_LO) )
			      {
				  /* We got a bunch of addresses but they were followed by something other than the required TIME_HI TIME_LO pair.
				     Count all the AER_ADDRs as used, and return an error.
				     We could analyse this situation further to decide between having got an ERROR word or a wrong timestamp word
				     but it's probably not worth looking at this 2nd order error condition. */
				  
				  *pnUsedRaw = iRaw + nConsecAddrs;
				  return PCOLERR_MISSING_TIME_LBLS;
			      }

			  /* Calculate the timestamp to be applied, correcting for the internal PCI-AER clock period. */

			  uTimestamp = ((nPutativeTimeHi & MONITOR_DWORD_DATA_MASK) << (8 * sizeof(pCooked->time_us) / 2)) | (nPutativeTimeLo & MONITOR_DWORD_DATA_MASK);
			  uTimestamp *= nClockPeriod_us;

			  /* Go through the consecutive events, including also now the one before
			     the one indexed by iRaw and produce corresponding cooked output. */

			  for ( i = 0; i < nConsecAddrs; i++ )
			      {
				  pCooked[ *pnCooked ].ae = pRaw[ iRaw + i - 1 ] & MONITOR_DWORD_DATA_MASK;
				  pCooked[ *pnCooked ].time_us = uTimestamp;
				  (*pnCooked)++;
			      }

			  /* Update iRaw to index the next word after the ones that provided
			     our timestamp and go back to expecting the AER_ADDR. */

			  iRaw += nConsecAddrs + 2;
			  Expected = ExpectAEAddr;
		      }
		  else
		      {
			  /* We got an AER_ADDR after a TIME_HI.
			     We leave *pnUsedRaw set so that the preceding
			     TIME_HI is considered used, but not the current
			     AER_ADDR which is hopefully the beginning of a
			     valid AER_ADDR, TIME_HI, TIME_LO sequence. */

			  *pnUsedRaw = iRaw;
			  return PCOLERR_MISSING_TIME_LO;
		      }
		  break;
	      
	      case MONITOR_DWORD_TIME_HI:
		  if ( Expected == ExpectTimeHi )
		      {
			  /* Normal correct case. We got a TIME_HI when we were expecting one. */

			  pCooked[ *pnCooked ].time_us = (pRaw[ iRaw++ ] & MONITOR_DWORD_DATA_MASK) << (8 * sizeof(pCooked->time_us) / 2);
			  Expected = ExpectTimeLo;
		      }
		  else
		      {
			  /* We got a TIME_HI when we weren't expecting one (previous word was also a TIME_* word)
			     We set *pnUsedRaw so that the erroneous TIME_HI is included in the 'used'. */

			  *pnUsedRaw = iRaw + 1;
			  return PCOLERR_UNEXPECTED_TIME_HI;
		      }
		  break;
	      
	      case MONITOR_DWORD_TIME_LO:
		  if ( Expected == ExpectTimeLo )
		      {
			  /* Normal correct case. We got a TIME_LO when we were expecting one. */

			  pCooked[ *pnCooked ].time_us |= (pRaw[ iRaw++ ] & MONITOR_DWORD_DATA_MASK);
			  pCooked[ *pnCooked ].time_us *= nClockPeriod_us; /* Correct for internal PCI-AER clock period */
			  (*pnCooked)++;
			  Expected = ExpectAEAddr;
		      }
		  else
		      {
			  /* We got a TIME_LO when we weren't expecting one (previous word was either AER_ADDR or TIME_LO).
			     We set *pnUsedRaw so that the erroneous TIME_LO is included in the 'used'. */

			  *pnUsedRaw = iRaw + 1;
			  return PCOLERR_UNEXPECTED_TIME_LO;
		      }
		  break;
	      
	      case MONITOR_DWORD_ERROR:
		  /* This is always an error condition irrespective of what word was expected.
		     We set *pnUsedRaw so that ERROR word is included in the 'used'. */

		  *pnUsedRaw = iRaw + 1;

		  /* The contents of the hardware error word are returned in the high order part of the return value. */

		  return (long)pRaw[ iRaw ] << (8 * sizeof(long) / 2);
		  break;
	      }
      }

  *pnUsedRaw = iRaw - Expected; /* Subtracting the numerical value of Expected from iRaw allows for the fact that the
				   last one or two raw words that were read might not have contributed to a completed
				   cooked output data structure if the raw buffer a pRaw did not end with the sequence
				   AER_ADDR, TIME_HI, TIME_LO, and in that case should not count as having been used. */
                                  
  return 0L;
}

/******************************************************************************/

/* EOF */
