/* pciinfo.c $Id$ */

#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysfs/libsysfs.h>

#include "pciaer.h"
#include "pciaerlib.h"
#include "s5920.h"

#define base_class(C)   (((unsigned char *)&(C))[2])
#define sub_class(C)    (((unsigned char *)&(C))[1])
#define prog_if(C)      (((unsigned char *)&(C))[0])


static int read_attribute( struct sysfs_attribute *a, const char *name, unsigned long *result )
{
	char *endptr;

	if ( strcmp( a->name, name ) == 0 )
	{
		*result = strtoul( a->value, &endptr, 0 );
		if ( *a->value == '\0' || isgraph( *endptr ) )
			return EIO;
		return 0;
	}
	else
		return -1;
}

int PciaerGetPciInfo(int handle, pciaer_pci_info_t *ppi)
{
	int error;
	struct sysfs_bus *b;
	struct sysfs_driver *drv;
	struct dlist *drvlinks;
	char *d;
	struct sysfs_device *dev = NULL;
	struct dlist *attrs;
	struct sysfs_attribute *a;
	unsigned long value;
	unsigned int fields;
	unsigned long class;

	b = sysfs_open_bus( "pci" );
	if ( b == NULL )
	{
		error = errno;
		goto exit;
	}

	drv = sysfs_get_bus_driver( b, "pciaer" );
	if ( drv == NULL )
	{
		error = errno;
		goto exit;
	}

	drvlinks = sysfs_open_link_list( drv->path );
	if ( drvlinks == NULL )
	{
		error = ENODEV;
		goto close_bus;
	}

	/* Currently assumes the first link which is a valid (openable) device name is the one that we want. */

	dlist_for_each_data( drvlinks, d, char )
	{
		dev = sysfs_open_device( "pci", d );
		if ( dev )
			break;
	}
	if ( !dev )
	{
		error = ENODEV;
		goto close_list;
	}

	/* If ppi->slot_name is non-NULL, we can do nothing other than assume the buffer is big enough :-( */

	if ( ppi->slot_name != NULL )
		strcpy( ppi->slot_name, dev->name );

	if ( sscanf( dev->name, "%*x:%2x:%2hhx.%hhx", &ppi->bus, &ppi->slot, &ppi->func ) != 3 )
	{
		error = EIO;
		goto close_dev;
	}

	attrs = sysfs_get_device_attributes( dev );
	if ( attrs == NULL )
	{
		error = errno;
		goto close_dev;
	}

	fields = 0;
	dlist_for_each_data( attrs, a, struct sysfs_attribute )
	{
		error = read_attribute( a, "vendor", &value );
		if ( error != -1 )
		{
			if ( error == 0 && value <= USHRT_MAX )
			{
				fields++;
				ppi->vendor = (unsigned short)value;
				continue;
			}
			else
				goto close_dev;
		}
		error = read_attribute( a, "device", &value );
		if ( error != -1 )
		{
			if ( error == 0 && value <= USHRT_MAX )
			{
				fields++;
				ppi->device = (unsigned short)value;
				continue;
			}
			else
				goto close_dev;
		}
		error = read_attribute( a, "subsystem_vendor", &value );
		if ( error != -1 )
		{
			if ( error == 0 && value <= USHRT_MAX )
			{
				fields++;
				ppi->subsys_vendor = (unsigned short)value;
				continue;
			}
			else
				goto close_dev;
		}
		error = read_attribute( a, "subsystem_device", &value );
		if ( error != -1 )
		{
			if ( error == 0 && value <= USHRT_MAX )
			{
				fields++;
				ppi->subsys_device = (unsigned short)value;
				continue;
			}
			else
				goto close_dev;
		}
		error = read_attribute( a, "class", &class );
		if ( error != -1 )
		{
			if ( error == 0 && class <= 0xFFFFFF )
			{
				fields++;
				continue;
			}
			else
				goto close_dev;
		}
		error = read_attribute( a, "irq", &value );
		if ( error != -1 )
		{
			if ( error == 0 && value <= UCHAR_MAX )
			{
				fields++;
				ppi->irq = (unsigned char)value;
				continue;
			}
			else
				goto close_dev;
		}
	}
	if ( fields != 6 )
	{
		error = EIO;
		goto close_dev;
	}

	/* If ppi->device_name is non-NULL, we can do nothing other than assume the buffer is big enough :-( */

	if ( ppi->device_name != NULL )
	{
		if ( ppi->vendor == PCI_VENDOR_ID_AMCC && ppi->device == PCI_DEVICE_ID_AMCC_S5920 )
			strcpy( ppi->device_name, "Applied Micro Circuits Corp. S5920" );
		else
			strcpy( ppi->device_name, "" );
	}

	ppi->base_class = base_class( class );
	ppi->sub_class = sub_class( class );
	ppi->prog_if = prog_if( class );

	error = 0;

close_dev:
	sysfs_close_device( dev );
close_list:
	sysfs_close_list( drvlinks );
close_bus:
	sysfs_close_bus( b );
exit:
	return error;
}
