/* seqwrite.c $Id: eef7de2bddd10566d053100258c688b6022c29d5 $ */

#include <stdlib.h>

#include "pciaer.h"
#include "pciaerlib.h"

/******************************************************************************/

/* This is the minimum length of the raw write buffer for PciaerSeqWrite */
#define SEQ_WRITE_BUFFER_MIN_ELEMENTS   128

/******************************************************************************/

int PciaerSeqWrite( int handle, const pciaer_sequencer_write_ae_t *p, unsigned int nEventsToWrite, unsigned int *pnEventsWritten )
{
    unsigned int nRawBufferElements;	/* Number of words currently allocated to the write buffer */
    unsigned int *pBufRaw;		/* Pointer to a buffer in which the raw data is prepared */
    unsigned int iBufferWord;		/* Index of raw word being written from buffer */
    unsigned int nTotalEventsWritten;	/* Running total of the no. of words written updated after each PciaerSeqWriteRaw call */
    unsigned int nEventsInBuffer;       /* Number of events in the raw buffer */
    unsigned int nWordsInBuffer;        /* Number of words used in the raw buffer */
    unsigned int nWordsWrittenPerCall;	/* No. of words reported as being written by each PciaerSeqWriteRaw call */
    unsigned int bNonBlockingExit;      /* BOOL: if true, we're in non-blocking mode and should exit */
    int write_return = EINVAL; 		/* Return value from PciaerSeqWriteRaw (error or zero) subsequently used as our return value */
    int prepare_return;         	/* Return value from PrepareRawWriteBuffer */

    /* Our estimate of the required buffer size is two sequencer commands per event
       This could be reduced as necessary */
    nRawBufferElements = 2 * nEventsToWrite;
    
    /* We should specify a minimum buffer size */
    if (nRawBufferElements < SEQ_WRITE_BUFFER_MIN_ELEMENTS) {
        nRawBufferElements = SEQ_WRITE_BUFFER_MIN_ELEMENTS;
    }

    /* Allocate buffer in which the raw data is to be prepared */

    pBufRaw = (unsigned int *) malloc( nRawBufferElements * sizeof(unsigned int) );
    if ( pBufRaw == NULL )
	{
	    *pnEventsWritten = 0;
	    return ENOMEM;
	}

    /* -- Write the events to the device (supports NON-BLOCKING) */
    nTotalEventsWritten = 0;
    bNonBlockingExit = 0;
    
    while ((nTotalEventsWritten < nEventsToWrite) & !bNonBlockingExit) {
        bNonBlockingExit = 0;
    
        /* -- Convert the user-supplied events into a raw buffer */
        prepare_return = PrepareRawWriteBuffer(p + nTotalEventsWritten, nEventsToWrite - nTotalEventsWritten, pBufRaw, nRawBufferElements, &nEventsInBuffer, &nWordsInBuffer);
        if (prepare_return != 0) {
            /* Error */
            free (pBufRaw);
            pBufRaw = NULL;
            return prepare_return;
        }
        
        /* -- Check that we could fit at least one event into the buffer */
        if (nEventsInBuffer == 0) {
            /* Currently this is an error -- in future we could reallocate the buffer */
            free (pBufRaw);
            pBufRaw = NULL;
            return ENOMEM;
        }

        /* -- Write the raw buffer to the device    (always BLOCKING) */
        iBufferWord = 0;
        while (iBufferWord < nWordsInBuffer) {
           write_return = PciaerSeqWriteRaw(handle, pBufRaw + iBufferWord, nWordsInBuffer - iBufferWord, &nWordsWrittenPerCall);
           
           /* Check the return value */
           if (write_return != 0) {
	       /* If the 'error' is EAGAIN, this shows that we're in non-blocking mode, and should not
		  continue to loop over events, but it's not an error as such. */
		      
		      if (write_return == EAGAIN) {
		          bNonBlockingExit = 1;
		          
		      } else {		          
		          /* BUG: In this case, we may have written more events than we say
			     we have, but at least we won't have written less.  In any case, any error here is
			     so severe that it's probably tough luck anyway. */
		          
		          free(pBufRaw);
		          pBufRaw = NULL;
		          return write_return;
		      }
		  }
		  
		  iBufferWord += nWordsWrittenPerCall;
        }
        
        nTotalEventsWritten += nEventsInBuffer;
        *pnEventsWritten = nTotalEventsWritten;
    }

    free(pBufRaw);
    pBufRaw = NULL;
    return write_return;
}

/******************************************************************************/

/* EOF */
