/* pciinfo.c $Id: 6fd388c0fe668ac2fa73d042f418980aa93f9456 $ */

#define _GNU_SOURCE

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pciaer.h"
#include "pciaerlib.h"


// handle_to_path
// Given a handle to an open file, return the path to that file in pathbuf by
// looking up the handle in procfs. bufsiz gives the available size of pathbuf.
// Returns 0 or an errno value in case of error.

static int handle_to_path( int handle, char *pathbuf, size_t bufsiz )
{
	const int maxprocpathlen = 40;
	char procpath[ maxprocpathlen ];
	int retval;

	retval = snprintf( procpath, maxprocpathlen, "/proc/%d/fd/%d", getpid(), handle );
	if ( retval >= maxprocpathlen )
		return ENAMETOOLONG;
	if ( retval < 0 )
		return retval;

	retval = readlink( procpath, pathbuf, bufsiz );
	if ( retval < 0 )
		return errno;
	if ( retval == bufsiz )
		return ENAMETOOLONG;
	pathbuf[ retval ] = '\0';
	return 0;
}


// parse_devpath
// Given a path to a device node of the form .../aerxxxn, set *devno to n, and
// put the xxx part (mon, seq or map) into the buffer at minor, which must have
// room for at least 4 characters (includes terminating \0).
// Returns 0 or an errno value in case of error.

static int parse_devpath( const char *devpath, unsigned int *devno, char *minor )
{
	char *device;
	int consumed;
	int retval;

	device = basename( devpath );
	retval = sscanf( device, "aer%3c%u%n", minor, devno, &consumed );
	if ( retval < 2 || consumed != strlen( device ) )
		return ENODEV;
	minor[3] = '\0';
	return 0;
}


// dev_from_handle
// Given a handle to an open device node of the form .../aerxxxn, set *devno to
// n, and put the xxx part (mon, seq or map) into the buffer at minor, which
// must have room for at least 4 characters (includes terminating \0).
// Returns 0 or an errno value in case of error.

static int dev_from_handle( int handle, unsigned int *devno, char *minor )
{
	const int maxdevpathlen = 24;
	int retval;
	char devpath[ maxdevpathlen ];

	retval = handle_to_path( handle, devpath, maxdevpathlen );
	if ( retval != 0 )
		return retval;

	retval = parse_devpath( devpath, devno, minor );
	if ( retval != 0 )
		return retval;

	return 0;
}


// read_and_parse_stats
// Given a handle to one of the debugfs stats files, read the statistics
// info into a buffer and then parse this into a pciaer_stats_t[_2] structure.
// The actual type of the structure is assumed based on the size indicated by
// the first element, stats->size. If certain fields are not listed in the data
// read from the stats file, the corresponding field in the structure will not
// be changed.
// Returns 0 or an errno value in case of error.

static int read_and_parse_stats( int stats_handle, pciaer_stats_t_2 *stats )
{
	const int nfields = 7;
	const int maxlinelen = 44;
	const int bufsiz = nfields * maxlinelen;
	char buf[ bufsiz ];
	int retval;
	char *field;
	char *decimal;
	char *saveptr;
	char *endconv;
	unsigned long stat;

	retval = read( stats_handle, buf, bufsiz );
	if ( retval == -1 )
		return errno;

	for ( decimal = strtok_r( buf, " ", &saveptr ), field = strtok_r( NULL, "\n", &saveptr );
	      decimal != NULL && field != NULL;
	      decimal = strtok_r( NULL, " ", &saveptr ), field = strtok_r( NULL, "\n", &saveptr ) )
	{
		if ( strcmp( field, "timestamp" ) == 0 )
		{
			if ( stats->size == sizeof(pciaer_stats_t_2) )
			{
				retval = sscanf( decimal, "%lu.%lu",
						 &stats->timestamp.tv_sec,
						 &stats->timestamp.tv_usec );
				if ( retval != 2 )
					return EIO;
			}
			continue;
		}
		stat = strtoul( decimal, &endconv, 10 );
		if ( *endconv != '\0' )
			return EIO;
		if ( strcmp( field, "words_transferred" ) == 0 )
		{
			stats->words_transferred = stat;
			continue;
		}
		if ( strcmp( field, "user_transfers" ) == 0 )
		{
			stats->user_transfers = stat;
			continue;
		}
		if ( strcmp( field, "total_interrupts" ) == 0 )
		{
			stats->total_interrupts = stat;
			continue;
		}
		if ( strcmp( field, "overflows_underflows" ) == 0 )
		{
			stats->overflows_underflows = stat;
			continue;
		}
		if ( strcmp( field, "timeouts" ) == 0 )
		{
			stats->timeouts = stat;
			continue;
		}
		if ( strcmp( field, "memory_usage" ) == 0 )
		{
			stats->memory_usage = stat;
			continue;
		}
	}

	return 0;
}


// reset_stats
// Resets the statistics of the stats file given by the supplied handle by
// writing a single byte to the file.
// Returns 0 or an errno value in case of error.

static int reset_stats( int stats_handle )
{
	return (write( stats_handle, "", 1 ) == 1) ? 0 : errno;
}


// get_reset_statistics
// Gets and/or resets the statistics for the device on which the supplied handle
// is open filling in the structure at s and resetting the statistics if reset
// is true.
// Returns 0 or an errno value in case of error.

static int get_reset_statistics( int handle, pciaer_stats_t_2 *s, int reset )
{
	const int maxstatspathlen = 52;
	int retval;
	int flags;
	unsigned int devno;
	char minor[ 4 ];
	char path[ maxstatspathlen ];
	int stats_handle;

	if ( s != NULL && s->size != sizeof(pciaer_stats_t_2) )
		s->size = sizeof(pciaer_stats_t);

	if ( !reset && s != NULL )
		flags = O_RDONLY;
	else if ( reset && s == NULL )
		flags = O_WRONLY;
	else if ( reset && s != NULL )
		flags = O_RDWR;
	else
		return EINVAL;

	retval = dev_from_handle( handle, &devno, minor );
	if ( retval != 0 )
		return retval;

	retval = snprintf( path, maxstatspathlen, "/sys/kernel/debug/pciaer/pciaer%u/%sstats", devno, minor );
	if ( retval >= maxstatspathlen )
		return -ENAMETOOLONG;
	if ( retval < 0 )
		return retval;

	retval = open( path, flags );
	if ( retval < 0 )
		return -retval;

	stats_handle = retval;

	if ( s != NULL )
		retval = read_and_parse_stats( stats_handle, s );

	if ( retval == 0 && reset )
		retval = reset_stats( stats_handle );

	close( stats_handle );
	return retval;
}


int PciaerGetStatistics( int handle, pciaer_stats_t *pstats )
{
	return get_reset_statistics( handle, (pciaer_stats_t_2 *)pstats, 0 );
}


int PciaerResetStatistics( int handle, pciaer_stats_t *pstats )
{
	return get_reset_statistics( handle, (pciaer_stats_t_2 *)pstats, 1 );
}

/* EOF */
