/*           Code for configuring the PCI-AER board                  */

/* pciaercfg.c $Id: c9b8b67aaf39247b0f691d919824dde1ce748056 $ */

#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "pciaer.h"
#include "pciaerlib.h"

#define MONFIFOFLAGS (PCIAER_IOC_MON_EMPTY|PCIAER_IOC_MON_HALF_FULL|PCIAER_IOC_MON_FULL)
#define SEQFIFOFLAGS (PCIAER_IOC_SEQ_EMPTY|PCIAER_IOC_SEQ_HALF_FULL|PCIAER_IOC_SEQ_FULL)

#define MONITOR   1
#define SEQUENCER 2
#define MAPPER    3

struct setting
{
  int getioctl;
  int setioctl;
  int min;
  int max;
  const char *setting;
  const char *desc;
//  const char **map;
};

//const static char *muxmap[] = {NULL, "0_16", "1_15", NULL, "2_14"};
//const static char *boolmap[] = {"disabled", "enabled"};
//const static char *mapmode[] = {NULL, "Pass-Through", "One-to-One", "One-to-Many"};
//const static char *protomap[] = {"P2P", "SCX"};

const static struct setting validsetting[]= {
  {PCIAER_IOC_GET_VERSION_INFO, 0, 0, 0,
          "version", "      Display PCI-AER board version information" },

  {PCIAER_IOC_GET_CNTR_PERIOD, PCIAER_IOC_SET_CNTR_PERIOD, 1, 100,
          "cntprd", "[N]    Get/set counter period in microseconds" },

  {PCIAER_IOC_RST_CNTR_GET_TIME, 0, 0, 0,
          "cntrst", "       Reset Counter" },

  {PCIAER_IOC_GET_ARB_CONFIG,  PCIAER_IOC_SET_ARB_CONFIG, 1, 4,
          "arbcfg", "[N]    Get/set arbiter configuration (1:0_16, 2:1_15, 4:2_14)" }, //muxmap},

  {PCIAER_IOC_GET_MON_CH_SEL,  PCIAER_IOC_SET_MON_CH_SEL, 0, 15,
          "monchm", "[MASK] Get/set monitor channel mask" },

  {PCIAER_IOC_GET_MON_TIME_LBL_FLAG, PCIAER_IOC_SET_MON_TIME_LBL_FLAG, 0, 1,
          "montml", "[BOOL] Get/set monitor time label flag (0|1 disables|enables labels)" }, //boolmap},

  {PCIAER_IOC_GET_MON_FIFO_FLAGS, 0, 0, 0,
          "monffl", "       Get monitor FIFO flags" },

  {PCIAER_IOC_GET_SEQ_ARB_CHANNEL, PCIAER_IOC_SET_SEQ_ARB_CHANNEL, 0, 3,
          "seqchn", "[N]    Get/set sequencer channel number (0-3)" },

  {PCIAER_IOC_GET_SEQ_FIFO_FLAGS, 0, 0, 0,
          "seqffl", "       Get sequencer FIFO flags" },

  {PCIAER_IOC_GET_MAP_OUT_CONFIG, PCIAER_IOC_SET_MAP_OUT_CONFIG, 1, 3,
          "mapmod", "[N]    Get/set mapper mode (1:Pass-Through, 2:One-to-One, 3:One-to-Many" }, //mapmode},

  {PCIAER_IOC_MAP_CLEAR, 0, 0, 0,
          "mapclr", "       Clear mapper" },

  {PCIAER_IOC_GET_MAP_CH_SEL,     PCIAER_IOC_SET_MAP_CH_SEL, 0, 15,
          "mapchm", "[MASK] Get/set Mapper channel mask" },

  {PCIAER_IOC_GET_MAP_DEMUX_CONFIG, PCIAER_IOC_SET_MAP_DEMUX_CONFIG, 1, 4,
          "mapdmx", "[N]    Get/set Mapper demux config (1:0_16, 2:1_15, 4:2_14)" }, //muxmap},

  {PCIAER_IOC_GET_MON_FIFO_FLAGS, 0, 0, 0,
          "mapffl", "       Get mapper FIFO flags" },

  {PCIAER_IOC_GET_MAP_AER_PROTOCOL, PCIAER_IOC_SET_MAP_AER_PROTOCOL, 0, 1,
          "mappcl", "[N]    Get/set mapper protocol (0:P2P, 1:SCX)" }//, protomap }
};

static int totalsettings = sizeof(validsetting) / sizeof(struct setting);

static int board = 0;

static int monhandle = -1;
static int seqhandle = -1;
static int maphandle = -1;

/*********************************************************************/

static void help( void )
{
    int i;

    printf( "Usage: pciaercfg -h\n"
             "   or: pciaercfg [-b BOARD] [SETTING]...\n"
             "   or: pciaercfg [-b BOARD] -a\n"
             "   or: pciaercfg [-b BOARD] -g\n"
             "   or: pciaercfg -v\n"
             "\n"
             "  -h          print this help and exit\n"
             "  -b BOARD    use board number BOARD\n"
             "  -a          print all current settings in human-readable form\n"
             "  -g          print all current settings in a pciaercfg-readable form\n"
             "  -v          print CVS revision information for pciaercfg\n"
             "\n"
             "SETTINGs:\n");
    for( i = 0; i < totalsettings; i++ ) 
    printf( "   %s %s\n", validsetting[i].setting, validsetting[i].desc );
}

/*********************************************************************/

/* Perhaps contrary to the name, this function does not always exit. If the
supplied error number is 0, the function returns immediately without doing
anything. */

static void print_error_and_exit( const char *message, int error_no )
{
    if ( error_no == 0 )
        return;
    fprintf( stderr, "%s: %s", message, strerror( error_no ) );
    exit( error_no );
}

/*********************************************************************/

static int findsetting(char *s)
{
    int i;
    int result = -1;

    for(i=0;i<totalsettings;i++)
    {
        if(strcmp(s,validsetting[i].setting)==0)
        {
            result = i;
            break;
        }
    }
    return result;
}

/*********************************************************************/

static void opendev(int board, int dev)
{
    int open_err = 0;
    char *msg;

    switch (dev) {
    case MONITOR:
        msg = "PciaerMonOpen";
        if ( monhandle == -1 )
            open_err = PciaerMonOpen( board, O_RDWR, &monhandle );
        break;
    case SEQUENCER:
        msg = "PciaerSeqOpen";
        if ( seqhandle == -1 )
            open_err = PciaerSeqOpen( board, O_RDWR, &seqhandle );
        break;
    case MAPPER:
        msg = "PciaerMapOpen";
        if ( maphandle == -1 )
            open_err = PciaerMapOpen( board, O_RDWR, &maphandle );
        break;
    default:
        open_err = EINVAL;
        msg = "opendev";
    }
    print_error_and_exit( msg, open_err );
}

/*********************************************************************/

static void version( void )
{
    pciaer_version_info_t pvi;
    int err;

    opendev( board, MONITOR );
    err = PciaerGetVersionInfo( monhandle, &pvi );
    print_error_and_exit( "PciaerGetVersionInfo", err );
    printf("PCI-AER Driver Version: %d.%d\n",pvi.driver_version>>8,pvi.driver_version&0xFF);
    printf("PCI-AER FPGA1 Release: %04x\n",pvi.fpga1_release);
    printf("PCI-AER FPGA2 Release: %04x\n",pvi.fpga2_release);
    printf("PCI-AER s5920 Revision: %d\n",pvi.s5920_revision_id);
}

/*********************************************************************/

static int getarbcfg( int verbose )
{
    int cfg;
    int err;

    opendev( board, MONITOR );
    err = PciaerGetArbConfig( monhandle, &cfg );
    print_error_and_exit( "PciaerGetArbConfig", err );
    if ( verbose )
    {
        switch(cfg) {
        case PCIAER_IOC_ARB_DISABLED:
            printf("Arbiter Configuration: Disabled\n");
            break;
        case PCIAER_IOC_ARB_0_16:
            printf("Arbiter Configuration: 1 device with 16 address bits\n");
            break;
        case PCIAER_IOC_ARB_1_15:
            printf("Arbiter Configuration: 2 devices with 15 address bits each\n");
            break;
        case PCIAER_IOC_ARB_2_14:
            printf("Arbiter Configuration: 4 devices with 14 address bits each\n");
            break;
        default:
            break;
        }
    }
    return cfg;
}

/*********************************************************************/

static int getmapmode( int verbose )
{
    int cfg;
    int err;

    opendev( board, MAPPER );
    err = PciaerMapGetOutputConfig( maphandle, &cfg );
    print_error_and_exit( "PciaerMapGetOutputConfig", err );
    if ( verbose )
    {
        switch(cfg) {
        case PCIAER_IOC_MAP_OUT_PASS_THRU:
            printf("Mapper Output Mode: Pass-Through\n");
            break;
        case PCIAER_IOC_MAP_OUT_1_TO_1:
            printf("Mapper Output Mode: One-to-One\n");
            break;
        case PCIAER_IOC_MAP_OUT_1_TO_MANY:
            printf("Mapper Output Mode: One-to-Many\n");
            break;
        default:
            break;
        }
    }
    return cfg;
}

/*********************************************************************/

static int getmapdemuxcfg( int verbose )
{
    int cfg;
    int err;

    opendev( board, MAPPER );
    err = PciaerMapGetDemuxConfig( maphandle, &cfg );
    print_error_and_exit( "PciaerMapGetDemuxConfig", err );
    if ( verbose )
    {
        switch(cfg) {
        case PCIAER_IOC_MAP_DEMUX_DISABLED:
            printf("Mapper Output Demux: Disabled\n");
            break;
        case PCIAER_IOC_MAP_DEMUX_0_16:
            printf("Mapper Output Demux: 1 device with 16 address bits\n");
            break;
        case PCIAER_IOC_MAP_DEMUX_1_15:
            printf("Mapper Output Demux: 2 devices with 15 address bits each\n");
            break;
        case PCIAER_IOC_MAP_DEMUX_2_14:
            printf("Mapper Output Demux: 4 devices with 14 address bits each\n");
            break;
        default:
            break;
        }
    }
    return cfg;
}

/*********************************************************************/

static void setmapproto( int cfg )
{
    int err;

    opendev( board, MAPPER );
    switch(cfg) {
    case 0:
        cfg=PCIAER_IOC_MAP_AER_P2P;
        break;
    case 1:
        cfg=PCIAER_IOC_MAP_AER_SCX;
        break;
    default:
        abort();
    }
    err = PciaerMapSetProtocol(maphandle, cfg);
    print_error_and_exit( "PciaerMapSetProtocol", err );
}

/*********************************************************************/

static int getmapproto( int verbose )
{
    int cfg;
    int err;

    opendev( board, MAPPER );
    err = PciaerMapGetProtocol(maphandle, &cfg);
    print_error_and_exit( "PciaerMapGetProtocol", err );
    if ( verbose )
    {
        switch(cfg) {
        case PCIAER_IOC_MAP_AER_P2P:
            printf("Mapper Output Protocol: P2P\n");
            break;
        case PCIAER_IOC_MAP_AER_SCX:
            printf("Mapper Output Protocol: SCX\n");
            break;
        default:
            break;
        }
    }
    return cfg;
}

/*********************************************************************/

static int do_setting( int ioctlno, int cfg, int verbose )
{
    int err;

    switch(ioctlno)
    {
    case PCIAER_IOC_GET_VERSION_INFO:
        version();
        break;

    case PCIAER_IOC_SET_CNTR_PERIOD:            /* Set Counter Period */
        opendev( board, MONITOR );
        err = PciaerSetCounterPeriod( monhandle, cfg );
        print_error_and_exit( "PciaerSetCounterPeriod", err );
        break;

    case PCIAER_IOC_GET_CNTR_PERIOD:         /* Get Counter Period */
        opendev( board, MONITOR );
        err = PciaerGetCounterPeriod(monhandle, &cfg);
        print_error_and_exit( "PciaerGetCounterPeriod", err );
        if ( verbose )
            printf("Counter Period (us): %d\n",cfg);
        break;

    case PCIAER_IOC_RST_CNTR_GET_TIME:          /* Reset Counter */
        opendev( board, MONITOR );
        err = PciaerResetCounter(monhandle);
        print_error_and_exit( "PciaerResetCounter", err );
        break;

    case PCIAER_IOC_SET_ARB_CONFIG:         /* Set Arbiter Configuration */
        opendev( board, MONITOR );
        err = PciaerSetArbConfig(monhandle, cfg);
        print_error_and_exit( "PciaerSetArbConfig", err );
        break;

    case PCIAER_IOC_GET_ARB_CONFIG:         /* Get Arbiter Configuration */
        cfg = getarbcfg( verbose );
        break;

    case PCIAER_IOC_SET_MON_CH_SEL:         /* Set Monitor Channel Mask */
        opendev( board, MONITOR );
        err = PciaerMonSetChannelSel(monhandle, cfg);
        print_error_and_exit( "PciaerMonSetChannelSel", err );
        if(cfg==0)
            fprintf( stderr, "Warning: monitor disabled\n" );
        break;

    case PCIAER_IOC_GET_MON_CH_SEL:         /* Get Monitor Channel Mask */
        opendev( board, MONITOR );
        err = PciaerMonGetChannelSel(monhandle, &cfg);
        print_error_and_exit( "PciaerMonGetChannelSel", err );
        if ( verbose )
            printf("Monitor Channel Enable Mask: 0x%X\n",cfg);
        break;

    case PCIAER_IOC_SET_MON_TIME_LBL_FLAG:          /* Set Monitor Time Label Flag */
        opendev( board, MONITOR );
        err = PciaerMonSetTimeLabelFlag(monhandle,cfg);
        print_error_and_exit( "PciaerMonSetTimeLabelFlag", err );
        break;

    case PCIAER_IOC_GET_MON_TIME_LBL_FLAG:          /* Get Monitor Time Label Flag */
        opendev( board, MONITOR );
        err = PciaerMonGetTimeLabelFlag(monhandle,&cfg);
        print_error_and_exit( "PciaerMonGetTimeLabelFlag", err );
        if ( verbose )
            printf("Monitor Time Label Flag: %d\n",cfg);
        break;

    case PCIAER_IOC_GET_MON_FIFO_FLAGS:         /* Get Monitor FIFO Flags */
        opendev( board, MONITOR );
        err = PciaerMonGetFifoFlags(monhandle,&cfg);
        print_error_and_exit( "PciaerMonGetFifoFlags", err );
        if ( verbose )
            printf("Monitor FIFO Flags (octal): %02o\n",cfg & MONFIFOFLAGS);
        break;

    case PCIAER_IOC_SET_SEQ_ARB_CHANNEL:            /* Set Sequencer Arbiter Channel */
        opendev( board, SEQUENCER );
        err = PciaerSetSeqArbChannel(seqhandle, cfg);
        print_error_and_exit( "PciaerSetSeqArbChannel", err );
        break;

    case PCIAER_IOC_GET_SEQ_ARB_CHANNEL:            /* Get Sequencer Arbiter Channel */
        opendev( board, SEQUENCER );
        err = PciaerGetSeqArbChannel(seqhandle, &cfg);
        print_error_and_exit( "PciaerGetSeqArbChannel", err );
        if ( verbose )
            printf("Sequencer Channel: %d\n",cfg);
        break;

    case PCIAER_IOC_GET_SEQ_FIFO_FLAGS:         /* Get Sequencer FIFO Flags */
        opendev( board, SEQUENCER );
        err = PciaerSeqGetFifoFlags(seqhandle,&cfg);
        print_error_and_exit( "PciaerSeqGetFifoFlags", err );
        if ( verbose )
            printf("Sequencer FIFO Flags (octal): %02o\n",cfg & SEQFIFOFLAGS);
        break;

    case PCIAER_IOC_SET_MAP_OUT_CONFIG:         /* Set Mapper Mode */
        opendev( board, MAPPER );
        err = PciaerMapSetOutputConfig(maphandle, cfg);
        print_error_and_exit( "PciaerMapSetOutputConfig", err );
        break;

    case PCIAER_IOC_GET_MAP_OUT_CONFIG:         /* Get Mapper Mode */
        cfg = getmapmode( verbose );
        break;

    case PCIAER_IOC_MAP_CLEAR:          /* Clear Mapper */
        opendev( board, MAPPER );
        err = PciaerMapClearAllMappings(maphandle);
        print_error_and_exit( "PciaerMapClearAllMappings", err );
        break;

    case PCIAER_IOC_SET_MAP_CH_SEL:         /* Set Mapper Channel Mask */
        opendev( board, MAPPER );
        err = PciaerMapSetChannelSel(maphandle, cfg);
        print_error_and_exit( "PciaerMapSetChannelSel", err );
        break;

    case PCIAER_IOC_GET_MAP_CH_SEL:         /* Get Mapper Channel Mask */
        opendev( board, MAPPER );
        err = PciaerMapGetChannelSel(maphandle, &cfg);
        print_error_and_exit( "PciaerMapGetChannelSel", err );
        if ( verbose )
            printf("Mapper Channel Output Mask: 0x0%X\n",cfg);
        break;

    case PCIAER_IOC_SET_MAP_DEMUX_CONFIG:           /* Set Mapper Demux Configuration */
        opendev( board, MAPPER );
        err = PciaerMapSetDemuxConfig(maphandle, cfg);
        print_error_and_exit( "PciaerMapSetDemuxConfig", err );
        break;

    case PCIAER_IOC_GET_MAP_DEMUX_CONFIG:           /* Get Mapper Demux Configuration */
        cfg = getmapdemuxcfg( verbose );
        break;

    case PCIAER_IOC_GET_MAP_FIFO_FLAGS:         /* Get Mapper FIFO Flags */
        opendev( board, MAPPER );
        err = PciaerMapGetFifoFlags(maphandle,&cfg);
        print_error_and_exit( "PciaerMapGetFifoFlags", err );
        if ( verbose )
            printf("Mapper FIFO Flags (octal): %03o\n",cfg);
        break;

    case PCIAER_IOC_SET_MAP_AER_PROTOCOL:           /* Set Mapper output Protocol Configuration */
        setmapproto( cfg );
        break;

    case PCIAER_IOC_GET_MAP_AER_PROTOCOL:           /* Set Get Mapper output Protocol Configuration */
        cfg = getmapproto( verbose );
        break;

    default:
        fprintf( stderr, "Unknown ioctl: 0x%x\n", ioctlno );
        abort();
    }

    return cfg;
}

/*********************************************************************/

static void all_human_readable( void )
{
    int i;
    int cfg = 0;

    printf("PCI-AER Board #%d\n",board);
    for( i = 0; i < totalsettings; i++ )
        if ( validsetting[i].setioctl != 0 )
            do_setting( validsetting[i].getioctl, cfg, 1 );
}

/*********************************************************************/

static void all_machine_readable( void )
{
    int i;
    int cfg;

    for ( i = 0; i < totalsettings; i++ )
        if ( validsetting[i].setioctl != 0 )
        {
            cfg = do_setting( validsetting[i].getioctl, cfg, 0 );
            printf( "%s %d ", validsetting[i].setting, cfg);
        }
    printf( "\n" );
}

/*********************************************************************/

// Parse command line options and return argc index of first non-option
// argument.

static int options( int argc, char *argv[] )
{
    int arg = 0;
    int i;
    char c;
    char *next;

top:
    while( ++arg < argc && argv[arg][0] == '-' )
    {
        for ( i = 1; (c = argv[arg][i]); i++ )
        {
            switch (c)
            {
            case 'h':
                help();
                exit(0);

            case 'b':
                next = &argv[arg][++i];
                if ( isdigit( *next ) )
                    board = atoi( next );
                else if ( *next == '\0' )
                {
                    board = strtol( argv[++arg], &next, 0);
                    if ( *next != '\0' )
                        print_error_and_exit( "Board number required", EINVAL );
                    goto top;
                }
                else
                    print_error_and_exit( "Board number required", EINVAL );
                break;

            case 'a':
                all_human_readable();
                exit( 0 );

            case 'g':
                all_machine_readable();
                exit( 0 );

            case 'v':
                printf( "%s: CVS $Revision: 1.3.2.5 $\n", __FILE__ );
                exit( 0 );

            default:
                fprintf( stderr, "%s -%c: unknown command line switch\n", argv[0], c );
                fprintf( stderr, "Type pciaercfg -h for help\n" );
                exit( EINVAL );
            }
        }
    }
    return arg;
}

/*********************************************************************/

int main( int argc, char *argv[] )
{
    int arg;
    int id;
    int ioctlno;
    int cfg;
    char *endptr;

    if( argc == 1 )
    {
        fprintf( stderr, "PCI-AER Board Configuration Program\n"
                          "Type pciaercfg -h for help\n");
        exit( 0 );
    }

    arg = options( argc, argv );

    printf("PCI-AER Board #%d\n",board);

    //Parse settings
    for ( ; arg < argc; arg++ )
    {
        id = findsetting( argv[arg] );
        if ( id == -1 )
        {
            printf("Unknown setting:%s\n",argv[arg]);
            continue;
        }

        ioctlno = validsetting[id].getioctl;

        if ( validsetting[id].setioctl != 0 && arg+1 < argc )
        {
            cfg = strtol( argv[arg+1], &endptr, 0 );
            if ( *endptr == '\0' )
            {
                if ( cfg < validsetting[id].min || cfg > validsetting[id].max )
                    print_error_and_exit( argv[arg], EINVAL );
                ioctlno = validsetting[id].setioctl;
                arg++;
            }
        }

        do_setting( ioctlno, cfg, 1 );
    }

    if( monhandle != -1 ) PciaerMonClose( monhandle );
    if( seqhandle != -1 ) PciaerSeqClose( seqhandle );
    if( maphandle != -1 ) PciaerMapClose( maphandle );

    return 0;
}
