// cmdline.cc $Id: 4f5b691ef4a60da43864e864f1e152eb50f271c4 $

// See comments in cmdline.h

// Derived from my private source code.

#include <algorithm>
#include <cctype>
#include <iostream>
#include "cmdline.h"


CCmdLine::CCmdLine( int argc, const char * const argv[], int flags ) : 
	m_argc( argc ),
	m_argv( argv ),
	m_flags( flags ),
	m_chSw( flags & flagUseOSSwCh ? chOSSw : chDefSw ),
	m_pOptRecords( NULL ),
	m_iFirstNonOptArg( 0 )
{
}


const char *CCmdLine::Parse( const char *optstring, BOOL opterr )
{
        using std::cerr;
        using std::endl;

	// Check characters in Optstring

	const char *pchFail = CheckOptString( optstring );

	if ( pchFail != NULL )
		return pchFail;

	// optstring OK - store in member variable & convert to all upper case if case is to be ignored.
		
	m_OptString = optstring;
		
	if ( m_flags & flagIgnoreOptCase )
		transform( m_OptString.begin(), m_OptString.end(), m_OptString.begin(), ::toupper );
		
	// Create an array of unions of BOOLs and const char*s, one union per optstring character and
	// initialize to FALSE / NULL.
		
	m_pOptRecords = new UOptRecord[ m_OptString.length() ];
	for ( unsigned int iOpt = 0; iOpt < m_OptString.length(); iOpt++ )
		if ( optstring[ iOpt + 1 ] != ':' )
			m_pOptRecords[ iOpt ].fPresent = FALSE;
		else
			m_pOptRecords[ iOpt ].pchOptArg = NULL;
	
	// Scan the argv looking for options. If found, record presence or pointer to option argument in
	// the array of unions.

	int iArg;
	for ( iArg = 1; iArg < m_argc; iArg++ )
	{
		const char *pchArg = m_argv[ iArg ];
			
		if ( *pchArg != m_chSw )
			break; // No more options

		const char *pchNext = pchArg + 1;
		
		if ( *pchNext == '\0' )
			break; // This is not an option, but just a single '-', an argument meaning stdin.
		
		if ( *pchNext++ == m_chSw )
		{
			// deal with "--"
			
			if ( *pchNext == '\0' )
			{
				// "--" marks end of option list
				
				iArg++;
				break;
			}
			else
			{
				// "--x" is some kind of unknown option
				
				if ( opterr )
					cerr << m_argv[ 0 ] << ": Invalid option " << pchArg << endl;
				if ( pchFail == NULL )
					pchFail = pchArg;
				continue;
			}
		}
		
		while ( *++pchArg )
		{
			int iOpt = m_OptString.find( ( m_flags & flagIgnoreOptCase ) ? (char)toupper( *pchArg ) : *pchArg );
			if ( iOpt != -1 && *pchArg != ':' )
			{
				if ( optstring[ iOpt + 1 ] != ':' )
				{
					m_pOptRecords[ iOpt ].fPresent = TRUE;
				}
				else
				{
					// deal with opt-arg
									
					if ( *(pchArg - 1) == m_chSw || (m_flags & flagAllowArgOptGroup) )
					{
						pchNext = pchArg + 1;
	
						if ( *pchNext == ':' && (m_flags & flagAllowOptArgColon) )
							pchNext++;
	
						if ( *pchNext == '\0' && iArg + 1 < m_argc )
						{
							m_pOptRecords[ iOpt ].pchOptArg = m_argv[ ++iArg ];
							break; // skip to next arg
						}
						else if ( *pchNext != '\0' && (m_flags & flagAllowOptArgNoSpace) )
						{
							m_pOptRecords[ iOpt ].pchOptArg = pchNext;
							break; // skip to next arg
						}
						else
						{
							if ( opterr )
								cerr << m_argv[ 0 ] << ": Option requires an argument -" << *pchArg << endl;
							if ( pchFail == NULL )
								pchFail = pchArg;
							if ( *(pchArg + 1) == ':' && (m_flags & flagAllowOptArgColon) )
								pchArg++;
						}
					}
					else
					{
						if ( opterr )
							cerr << m_argv[ 0 ] << ": Option requiring argument may not be grouped -" << *pchArg << endl;
						if ( pchFail == NULL )
							pchFail = pchArg;
					}
				}
			}
			else
			{
				// deal with unknown options.

				if ( opterr )
					cerr << m_argv[ 0 ] << ": Invalid option -" << *pchArg << endl;
				if ( pchFail == NULL )
					pchFail = pchArg;
			}
		}
	}
	
	m_iFirstNonOptArg = (iArg != m_argc) ? iArg : 0;
	return pchFail;
}


int CCmdLine::GetNonOptArgCount() const
{
	if ( m_iFirstNonOptArg == 0 )
		return 0;

	return m_argc - m_iFirstNonOptArg;
}


const char *CCmdLine::Arg( int i ) const
{
  // As a special case, Arg( 0 ) always returns argv[ 0 ], otherwise option arguments are not
  // counted and Arg( 1 ) will return the first non-option argument etc.

  if ( i != 0 )
    {
      if ( m_iFirstNonOptArg == 0 )
	return NULL;
      else	
	i += m_iFirstNonOptArg - 1;
    }

  return (i < m_argc) ? m_argv[ i ] : NULL;
}
	

BOOL CCmdLine::Opt( char chOpt ) const
{
	if ( m_OptString.empty() )
		return FALSE;
		
	int iOpt = m_OptString.find( ( m_flags & flagIgnoreOptCase ) ? (char)toupper( chOpt ) : chOpt );
	
	if ( iOpt == -1 )
		return FALSE;
	
	if ( m_OptString[ iOpt + 1 ] != ':' )
		return m_pOptRecords[ iOpt ].fPresent;
	else
		return (m_pOptRecords[ iOpt ].pchOptArg != NULL) ? TRUE : FALSE;
}


BOOL CCmdLine::Opt( char chOpt, string& OptArg ) const
{   
	OptArg.erase();
	
	if ( m_OptString.empty() )
		return FALSE;
		
	int iOpt = m_OptString.find( ( m_flags & flagIgnoreOptCase ) ? (char)toupper( chOpt ) : chOpt );

	if ( iOpt == -1 )
		return FALSE;
	
	if ( m_OptString[ iOpt + 1 ] == ':' )
	{
	  const char *pchArg = m_pOptRecords[ iOpt ].pchOptArg;
	  if ( pchArg != 0 )
	    OptArg = pchArg;

	  return !OptArg.empty();
	}

	return m_pOptRecords[ iOpt ].fPresent;
}


const char *CCmdLine::CheckOptString( const char *pchOptString )
{
	for ( const char *pch = pchOptString; *pch != '\0'; pch++ )
		if ( !isalnum( *pch ) && ( *pch != ':' || pch == pchOptString || *(pch - 1) == ':' ) )
			return pch;

	return NULL;
}

