// cmdline.h $Id: 4d51f1b84cb6588b4731055f11b6aeefaaeae167 $

// CCmdLine is a class that handles command lines supplied in traditional argc,argv form.

// After calling the constructor, Parse should be called with the parameter optstring containing
// the option letters to be recognized. If an option letter in the string is followed by a colon,
// the option is expected to have an option-argument. (optstring is like the third parameter to
// the getopt(3) function.) If Parse encounters options not listed in  optstring, it will return a
// pointer to the first failing option and (unless it was called with its parameter opterr = FALSE)
// print an error message on stderr. Parse can also fail before examining the command line and
// return a pointer into optstring if optstring contains non-alphanumeric option characters or
// colons in the wrong places. If Parse succeeds, it returns a null-pointer. Note that Parse
// operates in a POSIXly correct manner, interpreting any argument beyond the first non-option
// argument as a non-option argument even if it begins with the switch character. Thus all options
// on the command line must preceed all non-option arguments.
//
// Various non-standard variations on the standard command line theme are controlled by flags on the
// constructor - see below.
// 
// GNU extensions such as the use of a double colon in the optstring to indicate optional
// option-arguments are not yet supported, but might be added by the use of an additional
// constructor flag.
//
// Once Parse has been called, either of the Opt methods can be called to determine whether a given
// option letter was present on the command line, the 2 parameter method returns any option-argument
// via its second parameter. Also Arg(i) can be called to obtain a pointer to the i'th true argument
// beyond the options and option-arguments. As a special case Arg(0) always retrieves argv[0].

// Derived from my private source code.

#include <string>

#ifndef BOOL
#define BOOL bool
#define TRUE true
#define FALSE false
#endif

using std::string;

const char chDefSw = '-';	// Default switch character.

const char chOSSw		// OS switch character.
#ifdef MSDOS
	= '/';
#else
	= chDefSw;
#endif

const char chSwPlus = '+';


class CCmdLine
{
public:

	enum Flags
	{
		flagAllowArgOptGroup	= 0x0001,	// Allow eg. -abf filename as well as -ab -f filename
		flagAllowOptArgNoSpace	= 0x0002,	// Allow eg. -ffilename as well as -f filename
		flagAllowOptArgColon	= 0x0004,	// Allow eg. -f:filename as well as -f filename
		//flagAllowPlus		= 0x0008,	// Allow eg. +x as well as -x (not with flagUseOSOptCh)
		flagIgnoreOptCase	= 0x0010,	// eg. -F filename is equivalent to -f filename
		flagUseOSSwCh		= 0x0020	// ie. on DOS, look for /x rather than -x (not with flagAllowPlus)
	};

	// NB. flagAllowPlus is not yet supported.
	
	CCmdLine( int argc, const char * const argv[], int flags = 0 );
	const char *Parse( const char *optstring, BOOL opterr = TRUE );
	int GetFirstNonOptArg( void ) const { return m_iFirstNonOptArg; }
	int GetNonOptArgCount() const;
	const char *Arg( int i ) const;
	BOOL Opt( char chOpt ) const;
	BOOL Opt( char chOpt, string& OptArg ) const;

	~CCmdLine() { delete [] m_pOptRecords; }

private:

	static const char *CheckOptString( const char *pchOptString );

	const int m_argc;			// Copy of argc supplied to constructor.
	const char * const * const m_argv;	// Copy of argv supplied to constructor.
	const int m_flags;			// See above enumeration.
	const char m_chSw;			// The chosen switch character.
	string m_OptString;			// String containing list of option letters to be recognized.
	union UOptRecord { BOOL fPresent; const char *pchOptArg; } *m_pOptRecords;	// Result of Parse, one entry per character in m_OptString
	int m_iFirstNonOptArg;			// Index in argv of first non-option arg.
};
