// file.h $Id: a03f194ae7628552e42166cf7b4d26e859c6cc1d $

// Was revision 1.2 in the expo repository Floor/common module.

#ifndef FILE_H
#define FILE_H

#include <cerrno>
#include <string>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/file.h>

using std::string;

class File
{
 public:

  // Constructor & destructor

  inline File() : m_nDescriptor( -1 ) {}
  inline virtual ~File() { Close(); }


  // Copy constructor and op= (dup, and dup or dup2 respectively)

  inline File( const File& f )
    { m_nDescriptor = dup( f.m_nDescriptor ); }

  virtual File& operator=( const File& f );


  // Construct from open file descriptor

  inline File( int fd ) : m_nDescriptor( fd ) {}


  // Operations

  virtual int Open( const char *pName, int flags = O_RDWR );

  inline virtual int Open( const string sName, int flags = O_RDWR )
    { return Open( sName.c_str(), flags ); }

  virtual int Create( const char *pName, int flags = O_RDWR, mode_t mode = 0666 );

  virtual int Create( const string sName, int flags = O_RDWR, mode_t mode = 0666 )
    { return Create( sName.c_str(), flags, mode ); }

  virtual int Close();
  virtual int Seek( off_t Offset, int iWhence, off_t& Result ) const;
  virtual int Read( void *pv, size_t BytesToRead, size_t& BytesRead ) const;
  virtual int Write( const void *pv, size_t BytesToWrite, size_t& BytesWritten ) const;

  virtual bool CloseOnExec() const;

  inline virtual void CloseOnExec( bool f ) const
    { fcntl( m_nDescriptor, F_SETFD, long(f ? FD_CLOEXEC : 0) ); }

  virtual int GetFlags( int& f ) const;
  virtual int SetFlags( int f ) const;
  virtual bool IsReadable() const;
  virtual bool IsWritable() const;
  virtual bool IsReadWritable() const;
  virtual bool Append() const;
  virtual void Append( bool fAppend ) const;
  virtual bool NonBlocking() const;
  virtual void NonBlocking( bool fNonBlock ) const;
  virtual bool SyncWrite() const;
  virtual void SyncWrite( bool fSync ) const;
  virtual bool Async() const;
  virtual void Async( bool fAsync ) const;
  virtual int GetOwnerProcessOrGroup( pid_t& pid ) const;
  virtual int SetOwnerProcess( pid_t pid ) const;

  inline virtual int SetOwnerProcessGroup( pid_t pgid ) const
    { return SetOwnerProcess( -pgid ); }
  
  virtual int Ioctl( int iRequest ) const;
  virtual int Ioctl( int iRequest, void *pvArg ) const;

  virtual int Stat( struct stat& StatStruct ) const;
  virtual int ChangeMode( mode_t mode ) const;

  virtual int ChangeOwnerAndGroup( uid_t uidOwner, gid_t gidGroup ) const;

  inline virtual int ChangeOwner( uid_t uidOwner ) const
    { return ChangeOwnerAndGroup( uidOwner, gid_t(-1) ); }

  virtual int ChangeGroup( gid_t gidGroup ) const
    { return ChangeOwnerAndGroup( uid_t(-1), gidGroup ); }

  virtual int Truncate( off_t Length = 0 ) const;
  virtual int Sync() const;
  virtual int DataSync() const;

  virtual int GetRecordLock( const struct flock& FLockStructInput, struct flock& FLockStructOutput ) const;
  virtual int SetRecordLock( const struct flock& FLockStruct, bool fWait = false ) const;

  virtual bool IsRangeReadLockable( off_t Offset, int iWhence, size_t Length, pid_t& pid ) const;

  inline virtual bool IsRangeReadLockable( off_t Offset, int iWhence, size_t Length ) const
    { pid_t pid; return IsRangeReadLockable( Offset, iWhence, Length, pid ); }
  
  virtual bool IsRangeWriteLockable( off_t Offset, int iWhence, size_t Length, pid_t& pid ) const;

  inline virtual bool IsRangeWriteLockable( off_t Offset, int iWhence, size_t Length ) const
    { pid_t pid; return IsRangeWriteLockable( Offset, iWhence, Length, pid ); }

  virtual int ReadLockRange( off_t Offset, int iWhence, size_t Length ) const;
  virtual int WriteLockRange( off_t Offset, int iWhence, size_t Length ) const;
  virtual int ReadLockRangeWait( off_t Offset, int iWhence, size_t Length ) const;
  virtual int WriteLockRangeWait( off_t Offset, int iWhence, size_t Length ) const;
  virtual int UnlockRange( off_t Offset, int iWhence, size_t Length ) const;

  inline virtual int LockShared() const
    { return FLock( LOCK_SH ); }
  inline virtual int LockExclusive() const
    { return FLock( LOCK_EX ); }
  inline virtual int LockSharedNonBlocking() const
    { return FLock( LOCK_SH | LOCK_NB ); }
  inline virtual int LockExclusiveNonBlocking() const
    { return FLock( LOCK_EX | LOCK_NB ); }
  inline virtual int Unlock() const
    { return FLock( LOCK_UN ); }


  // Status

  inline virtual bool IsOpen() const
    { return m_nDescriptor != -1; }


  // Member access

  inline virtual int FileDescriptor() const
    { return m_nDescriptor; }


 protected:

  // Seek, Read & Write that return exactly the multiplexed result and error returns
  // of the underlying OS lseek, read & write calls.

  inline virtual off_t Seek( off_t Offset, int iWhence ) const
    { return lseek( m_nDescriptor, Offset, iWhence ); }

  inline virtual ssize_t Read( void *pv, size_t n ) const
    { return read( m_nDescriptor, pv, n ); }

  inline virtual ssize_t Write( const void *pv, size_t n ) const
    { return write( m_nDescriptor, pv, n ); }


  inline virtual int FLock( int nOp ) const
    { return flock( m_nDescriptor, nOp ) == 0 ? 0 : errno; }


  // Member variables

  int m_nDescriptor;

};

#endif
