/* mks5920bootimg.c $Id: 5844484901e69390eecd79d82ebb8f34e90010ea $

Program to make an AMCC S5920 boot image, according to Table 2, page 2-82 of
the data book. The values that actually end up in the boot image are determined
by the contents of an included file, s5920bootimg.c, which can draw upon
constants defined in pci.h and/or s5920.h.
*/


#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>

/* Some constants & macros that it would have been nice if they were in pci.h */

#define PCI_INTERRUPT_LINE_UNK	0xff
#define PCI_INTERRUPT_LINE_NC	0xff
#define PCI_INTERRUPT_PIN_NONE	0
#define PCI_INTERRUPT_PIN_INTA	1
#define PCI_INTERRUPT_PIN_INTB	2
#define PCI_INTERRUPT_PIN_INTC	3
#define PCI_INTERRUPT_PIN_INTD	4

#define base_class(C)	(((unsigned char *)&(C))[ 2 ])
#define sub_class(C)	(((unsigned char *)&(C))[ 1 ])
#define prog_if(C)	(((unsigned char *)&(C))[ 0 ])


#include "s5920.h"


/* Macro to extract sub-class from 2-byte combined class & sub-class constants */

#define PCI_SUB_CLASS(C) ((C)&0xFF)


/* Stucture that will contain the S5920 boot image */

struct s5920_nvram bootimg;


/* Function called to initialise the boot image.
Lines which actually set the contents of the boot image structure to something 
other than the defaults are in the separate include file s5920bootimg.c, and
normally only that file is changed to produce different boot image contents */

void InitBootImage( void )
{
    /* Clear whole structure to zero */

    memset( &bootimg, 0, sizeof(bootimg) );

    /* Provide some essential and/or sensible defaults */

    bootimg.vendor_id = PCI_VENDOR_ID_AMCC;
    bootimg.device_id = PCI_DEVICE_ID_AMCC_S5920;

    bootimg.special_modes = S5920_SPECIAL_MODES_TARGET_LATENCY_ENB;

    base_class(bootimg.class_code) = PCI_CLASS_OTHERS;

    bootimg.header_type = PCI_HEADER_TYPE_NORMAL;

    bootimg.self_test = S5920_NVRAM_SELF_TEST_NOT_DESIRED;

    bootimg.reqd_lsb_base_addr_0 = S5920_LSB_BADR0_IO;

    bootimg.reqd_valid_1 = S5920_NVRAM_REQD_VALID_1;
    bootimg.reqd_valid_2 = S5920_NVRAM_REQD_VALID_2;
    bootimg.reqd_valid_3 = S5920_NVRAM_REQD_VALID_3;

    bootimg.base_addr_1 = S5920_BADR_PASSTHRU_REGION_DISABLED;
    bootimg.base_addr_2 = S5920_BADR_PASSTHRU_REGION_DISABLED;
    bootimg.base_addr_3 = S5920_BADR_PASSTHRU_REGION_DISABLED;
    bootimg.base_addr_4 = S5920_BADR_PASSTHRU_REGION_DISABLED;

    bootimg.int_line = PCI_INTERRUPT_LINE_NC;
    bootimg.int_pin = PCI_INTERRUPT_PIN_NONE;

    /* Include the file in which device specific values can be set */

#include "s5920bootimg.c"

}


/* Function to write the bootimg to a named file.
   If it succeeds it returns 0; if it fails it returns an errno value. */

int WriteBootImageToFile( char *Filename )
{
    int fd;
    ssize_t WriteReturn;

    fd = creat( Filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
    if ( fd == -1 )
	return errno;
    
    WriteReturn = write( fd, &bootimg, sizeof(bootimg) );
    if ( WriteReturn == -1 )
	return errno;

    close( fd );

    if ( WriteReturn != sizeof(bootimg) )
	return EINTR;

    return 0;
}


/* Initialise the boot image, check the command line for the presence of a
   filename to write to, and write the image to that file. */

int main( int argc, char *argv[] )
{
    InitBootImage();

    switch ( argc )
	{
	case 0:
	    errno = ENOEXEC;
	    break;
	case 1:
	    errno = EINVAL;
	    break;
	case 2:
	    errno = WriteBootImageToFile( argv[ 1 ] );
	    break;
	default:
	    errno = E2BIG;
	    break;
	}

    if ( errno != 0 )
	perror( argv[ 0 ] );

    return errno;
}
