// mks5920romimg.cc $Id: a55aec0171ca2f2f716d245e8d34060c10be72cf $

#include <iostream>
#include <limits>
#include <sstream>
#include <string>
#include <cstdlib>

using namespace std;

#include "cmdline.h"
#include "file.h"
#include "s5920.h"
#include "PcExpRom.h"
#include "PciDS.h"

typedef struct s5920_nvram s5920_nvram_t;


void SafeAssignString( string& s, const char *pch )
{
    if ( pch == 0 )
	s.empty();
    else
	s = pch;
}


unsigned char ByteSum( unsigned char *p, size_t nBytes )
{
    unsigned char uchSum = 0;
    for ( size_t i = 0; i < nBytes; ++i )
	uchSum += p[ i ];
    return uchSum;
}


int main( int argc, char *argv[] )
{
    CCmdLine CommandLine( argc, argv, CCmdLine::flagAllowArgOptGroup | CCmdLine::flagAllowOptArgNoSpace );

    if ( CommandLine.Parse( "i:o:v:pr:ns:" ) != NULL )
	return 1;

    string InputFilename;
    string OutputFilename;

    int nRequiredArgs = 2;

    if ( !CommandLine.Opt( 'i' ) && !CommandLine.Opt( 'o' ) )
	{
	    SafeAssignString( InputFilename, CommandLine.Arg( 1 ) );
	    SafeAssignString( OutputFilename, CommandLine.Arg( 2 ) );
	}

    if ( CommandLine.Opt( 'i' ) && !CommandLine.Opt( 'o' ) )
	{
	    nRequiredArgs = 1;
	    CommandLine.Opt( 'i', InputFilename );
	    SafeAssignString( OutputFilename, CommandLine.Arg( 1 ) );
	}
    
    if ( !CommandLine.Opt( 'i' ) && CommandLine.Opt( 'o' ) )
	{
	    nRequiredArgs = 1;
	    SafeAssignString( InputFilename, CommandLine.Arg( 1 ) );
	    CommandLine.Opt( 'o', OutputFilename );
	}

    if ( CommandLine.Opt( 'i' ) && CommandLine.Opt( 'o' ) )
	{
	    nRequiredArgs = 0;
	    CommandLine.Opt( 'i', InputFilename );
	    CommandLine.Opt( 'o', OutputFilename );
	}

    if ( CommandLine.GetNonOptArgCount() != nRequiredArgs )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Wrong number of command line arguments" << endl;
	    return 1;
	}

    if ( InputFilename.empty() || OutputFilename.empty() )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Input and/or output file not specified" << endl;
	    return 1;
	}

    string VpdFilename;
    CommandLine.Opt( 'v', VpdFilename );

    bool IncludePciDS = CommandLine.Opt( 'v' ) || CommandLine.Opt( 'p' );

    unsigned short uhRevision = 0;
    if ( CommandLine.Opt( 'r' ) )
	{
	    if ( !IncludePciDS )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Option -r requires -v or -p" << endl;
		    return 1;
		}

	    string sRevision;
	    CommandLine.Opt( 'r', sRevision );

	    char *pchEnd;
	    unsigned long ulRev = strtoul( sRevision.c_str(), &pchEnd, 0 );
	    if ( *pchEnd != '\0' || ulRev > numeric_limits<unsigned short>::max() )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Bad -r value" << endl;
		    return 1;
		}
	    uhRevision = ulRev;
	}

    bool SigRequired = !CommandLine.Opt( 'n' );

    unsigned short uhRomSize = 0;	// Use 0 to mean no size specified
    if ( CommandLine.Opt( 's' ) )
	{
	    string sSize;
	    CommandLine.Opt( 's', sSize );

	    char *pchEnd;
	    unsigned long ulSize = strtoul( sSize.c_str(), &pchEnd, 0 );
	    if ( *pchEnd != '\0' || ulSize > numeric_limits<unsigned short>::max() || ulSize == 0L || (ulSize % PC_EXP_ROM_LENGTH_MULTIPLE) != 0 )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Bad -s value" << endl;
		    return 1;
		}
	    uhRomSize = ulSize;
	}

#ifdef DEBUG
    cout << InputFilename << " -> " << OutputFilename << endl;
    if ( uhRomSize != 0 )
	cout << "ROM size = " << uhRomSize << " bytes\n";
    if ( IncludePciDS )
	cout << "PCIDS rev. " << uhRevision << endl;
    if ( !VpdFilename.empty() )
	cout << "VPD " << VpdFilename << endl;
    if ( !SigRequired )
	cout << "No signature required\n";
#endif
    
    File InputFile;
    File OutputFile;
    File VpdFile;

    if ( InputFile.Open( InputFilename, O_RDONLY ) != 0 )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Could not open input file " << InputFilename << endl;
	    return 1;
	}

    struct stat StatStruct;

    if ( InputFile.Stat( StatStruct ) != 0 )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Could not stat input file " << InputFilename << endl;
	    return 1;
	}
    if ( StatStruct.st_size != sizeof(s5920_nvram_t) )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Input file " << InputFilename << " is not the correct size" << endl;
	    return 1;
	}

    s5920_nvram_t BootImage;
    size_t BytesRead;
    if ( InputFile.Read( &BootImage, sizeof(s5920_nvram_t), BytesRead ) != 0  || BytesRead != sizeof(s5920_nvram_t) )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Error reading input file " << InputFilename << endl;
	    return 1;
	}

    if ( BootImage.reqd_valid_1 != S5920_NVRAM_REQD_VALID_1
	 || BootImage.reqd_valid_2 != S5920_NVRAM_REQD_VALID_2
	 || BootImage.reqd_valid_3 != S5920_NVRAM_REQD_VALID_3 )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Input file " << InputFilename << " does not contain a valid S5920 boot image" << endl;
	    return 1;
	}

    if ( OutputFile.Create( OutputFilename, O_WRONLY | O_TRUNC ) != 0 )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Could not open output file " << OutputFilename << endl;
	    return 1;
	}

    unsigned short uhVpdSize = 0;
    unsigned char *Vpd = 0;
    if ( !VpdFilename.empty() )
	{
	    if ( VpdFile.Open( VpdFilename, O_RDONLY ) != 0 )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Could not open VPD file " << VpdFilename << endl;
		    return 1;
		}
	    if ( VpdFile.Stat( StatStruct ) != 0 )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Could not stat VPD file " << InputFilename << endl;
		    return 1;
		}
	    if ( size_t( StatStruct.st_size ) > numeric_limits<unsigned short>::max() - sizeof(s5920_nvram_t) )
		{
		    cerr << CommandLine.Arg( 0 ) << ": VPD file " << InputFilename << " is too big" << endl;
		    return 1;
		}
	    uhVpdSize = StatStruct.st_size;
	    Vpd = new unsigned char [ uhVpdSize ];
	    if ( VpdFile.Read( Vpd, uhVpdSize, BytesRead ) != 0 || BytesRead != uhVpdSize )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Error reading VPD file " << VpdFilename << endl;
		    return 1;
		}
	}


    // Create ROM header

    PcExpansionRomHeader Header;


    // Determine where in ROM the various items will go and hence the length of ROM required

    unsigned short uhMinRomSize = sizeof(s5920_nvram_t);
    const unsigned short uhHoleSize = 0x40 - sizeof(pc_exp_rom_hdr_t); // Size of hole between end of header and start of boot image

    // 3 enum variables for the possible dispositions of the various items that can take up space

    enum { DispAbsent, DispHdrRsrvd, DispHole, DispBootImg, DispEnd } dispPciDS, dispVpd, dispRetf;

    // Decide disposition of VPD if present
    // Try reserved area in header, hole, and failing that, after the boot image

    if ( Vpd == 0 )
	dispVpd = DispAbsent;
    else if ( uhVpdSize < Header.GetFreeSpace() )
	dispVpd = DispHdrRsrvd;
    else if ( uhVpdSize < uhHoleSize && (sizeof(pci_data_struct_t) + uhVpdSize < uhHoleSize || uhVpdSize > sizeof(pci_data_struct_t)) )
	dispVpd = DispHole;
    else
	{
	    dispVpd = DispEnd;
	    uhMinRomSize += uhVpdSize;
	}

    // Decide disposition of PCI DS if present
    // Try hole, and failing that, after the boot image

    if ( !IncludePciDS )
	dispPciDS = DispAbsent;
    else if ( sizeof(pci_data_struct_t) < uhHoleSize && (sizeof(pci_data_struct_t) + uhVpdSize < uhHoleSize || sizeof(pci_data_struct_t) >= uhVpdSize) )
	dispPciDS = DispHole;
    else
	{
	    dispPciDS = DispEnd;
	    uhMinRomSize += sizeof(pci_data_struct_t);
	}

    // Decide disposition of RETF if present
    // Try reserved area in header, hole, and failing that, use the penultimate (reserved) byte of the boot image

    if ( !SigRequired )
	dispRetf = DispAbsent;
    else if ( (dispVpd != DispHdrRsrvd) || uhVpdSize < Header.GetFreeSpace() )
	dispRetf = DispHdrRsrvd;
    else if ( (dispVpd == DispHole) * uhVpdSize + (dispPciDS == DispHole) * sizeof(pci_data_struct_t) < uhHoleSize )
	dispRetf = DispHole;
    else
	dispRetf = DispBootImg;

    // Decide disposition of Checksum if present
    // Must be at the end, but this can be the last (reserved) byte of the boot image if nothing else will be after that
    // therefore we only need to reserve an extra byte for the checksum if the PCI DS or the VPD is at the end of the ROM.

    if ( SigRequired && ( dispPciDS == DispEnd || dispVpd == DispEnd ) )
	uhMinRomSize += 1;

    // Calculate the size of ROM necessary and if a size was given on the command line, check that it will fit

    uhMinRomSize = (uhMinRomSize + PC_EXP_ROM_LENGTH_MULTIPLE - 1) & ~(PC_EXP_ROM_LENGTH_MULTIPLE - 1);

    if ( uhRomSize == 0 )
	uhRomSize = uhMinRomSize;
    else if ( uhMinRomSize > uhRomSize )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Required contents will not fit into a ROM of " << uhRomSize << " bytes" << endl;
	    return 1;
	}


    // Init ROM header length

    Header.SetLength( uhRomSize );


    // Allocate memory for ROM image

    unsigned char *RomImage = new unsigned char [ uhRomSize ];
    

    // Copy boot image into ROM image

    memcpy( RomImage, &BootImage, sizeof(s5920_nvram_t) );


    // Create PCI-DS & init PCI-DS pointer in header if required

    PciDataStructure *pPciDS = 0;
    if ( IncludePciDS )
	{
	    pPciDS = new PciDataStructure( BootImage.vendor_id, BootImage.device_id, uhRomSize, uhRevision, PCIDS_CODE_TYPE_INTEL_X86, BootImage.class_code );
	    switch ( dispPciDS )
		{
		case DispHole:
		    Header.SetPciDSOffset( sizeof(pc_exp_rom_hdr_t) );
		    break;
		case DispEnd:
		    Header.SetPciDSOffset( sizeof(s5920_nvram_t) );
		    break;
		default:
		    cerr << CommandLine.Arg( 0 ) << ": Internal error " << __LINE__ << endl;
		    return 1;
		    break;
		}

	    // If vpd is reqd, copy into ROM image (or header) and set VPD pointer in PCI-DS

	    switch ( dispVpd )
		{
		    unsigned short uhVpdOffset;

		case DispAbsent:
		    break;
		case DispHdrRsrvd:
		    uhVpdOffset = Header.CopyToFreeSpace( Vpd, uhVpdSize );
		    pPciDS->SetVPDOffset( uhVpdOffset );
		    break;
		case DispHole:
		    uhVpdOffset = dispPciDS != DispHole ? sizeof(pc_exp_rom_hdr_t) : sizeof(pc_exp_rom_hdr_t) + sizeof(pci_data_struct_t);
		    memcpy( RomImage + uhVpdOffset, Vpd, uhVpdSize );
		    pPciDS->SetVPDOffset( uhVpdOffset );
		    break;
		case DispEnd:
		    uhVpdOffset = dispPciDS != DispEnd ? sizeof(s5920_nvram_t) : sizeof(s5920_nvram_t) + sizeof(pci_data_struct_t);
		    memcpy( RomImage + uhVpdOffset, Vpd, uhVpdSize );
		    pPciDS->SetVPDOffset( uhVpdOffset );
		    break;
		default:
		    cerr << CommandLine.Arg( 0 ) << ": Internal error " << __LINE__ << endl;
		    return 1;
		    break;
		}
	    
	    delete [] Vpd;
	    Vpd = 0;

	    // Write PCI-DS into ROM image

	    pPciDS->Write( RomImage + Header.GetPciDSOffset() );

	    delete pPciDS;
	    pPciDS = 0;
	}


    if ( SigRequired )
	{
	    // Insert RETF into ROM image (or header) if entry point to be set

	    const unsigned char RETF = PC_EXP_ROM_FAR_RET;
	    unsigned short uhEntryPoint;

	    switch ( dispRetf )
		{
		case DispHdrRsrvd:
		    uhEntryPoint = Header.CopyToFreeSpace( &RETF, 1 );
		    break;
		case DispHole:
		    uhEntryPoint = sizeof(pc_exp_rom_hdr_t);
		    if ( dispPciDS == DispHole )
			uhEntryPoint += sizeof(pci_data_struct_t);
		    if ( dispVpd == DispHole )
			uhEntryPoint += uhVpdSize;
		    RomImage[ uhEntryPoint ] = RETF;
		    break;
		case DispBootImg:
		    uhEntryPoint = 0x7e;
		    RomImage[ uhEntryPoint ] = RETF;
		    break;
		default:
		    cerr << CommandLine.Arg( 0 ) << ": Internal error " << __LINE__ << endl;
		    return 1;
		    break;
		}

	    // Sign & set entry point in header

	    Header.Sign();
	    Header.SetEntryPoint( uhEntryPoint );
	}


    // Write ROM header into ROM image

    Header.Write( RomImage );


    // Calc checksum and write into ROM image if reqd.

    if ( SigRequired )
	RomImage[ uhRomSize - 1 ] = -ByteSum( RomImage, uhRomSize - 1 );


    // Write ROM image to output file

    size_t BytesWritten;
    if ( OutputFile.Write( RomImage, uhRomSize, BytesWritten ) != 0 || BytesWritten != uhRomSize )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Error writing output file " << OutputFilename << endl;
	    return 1;
	}
    
    delete [] RomImage;

    return 0;
}
