// vpd.cc $Id: 4444381f36bdd8c129c4e6c622a4673fff4709fe $

#include <iostream>
#include <fstream>
#include <cctype>
#include <climits>

using namespace std;

#include "cmdline.h"
#include "VpdClass.h"
#include "PcExpRom.h"
#include "PciDS.h"


static bool StringIsAscii( const string& String )
{
    const char *p = String.c_str();
    while ( *p != '\0' )
	if ( !isascii( *p++ ) )
	    return false;
    return true;
}


static bool ParseEqualityLikeString( const string& input, string& lhs, string& rhs )
{
    string::size_type iEq = input.find( '=' );
    if ( iEq == string::npos )
	return false;
    lhs = input.substr( 0, iEq );
    rhs = input.substr( iEq + 1 );
    return true;
}


static bool ReadTextVpd( istream& is, CVitalProductData& vpd )
{
    string Line;
    unsigned int iLineNumber = 0;
    bool TooLateForIdentifier = false;
    string Keyword;
    string Data;
    while ( getline( is, Line ) )
	{
	    ++iLineNumber;
	    if ( Line.empty() || Line[ 0 ] == '#' )
		continue;
	    if ( !StringIsAscii( Line ) )
		{
		    cerr << "Non-ASCII character at line " << iLineNumber << endl;
		    return false;
		}
	    if ( !ParseEqualityLikeString( Line, Keyword, Data ) )
		{
		    if ( TooLateForIdentifier )
			{
			    cerr << "Missing '=' at line " << iLineNumber << endl;
			    return false;
			}
		    if ( Line.length() > USHRT_MAX )
			{
			    cerr << "Identifier string too long at line " << iLineNumber << endl;
			    return false;
			}
		    vpd.SetIdentifier( Line );
		    TooLateForIdentifier = true;
		}
	    else
		{
		    if ( Keyword.length() != 2 )
			{
			    cerr << "Invalid length keyword '" << Keyword << "' at line " << iLineNumber << endl;
			    return false;
			}
		    if ( Data.length() > UCHAR_MAX )
			{
			    cerr << "Keyword data too long for keyword '" << Keyword << "' at line " << iLineNumber << endl;
			    return false;
			}
		    vpd.SetKeyword( Keyword, Data );
		    TooLateForIdentifier = true;
		}
	}

    return true;
}


static void PrintKeywordPair( const string& Keyword, const string& Data )
{
    cout << Keyword << '=' << Data << endl;
}


static int Read( const CCmdLine& CommandLine )
{
    string Filename;
    CommandLine.Opt( 'r', Filename );
    ifstream InputFile( Filename.c_str() );
    if ( !InputFile )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Can't open " << Filename << endl;
	    return 1;
	}

    if ( CommandLine.Opt( 'R' ) || InputFile.peek() != VpdDescriptor::IDStringTag )
	{
	    PcExpansionRomHeader RomHdr;
	    InputFile >> RomHdr;
	    if ( !CommandLine.Opt( 'R' ) && !RomHdr.SignatureOK() )
		{
		    cerr << CommandLine.Arg( 0 ) << ": " << Filename << " not recognised as VPD or ROM image" << endl;
		    return 1;
		}
	    InputFile.seekg( RomHdr.GetPciDSOffset() );
	    PciDataStructure PciDS;
	    InputFile >> PciDS;
	    if ( !PciDS.IsValid() )
		{
		    cerr << CommandLine.Arg( 0 ) << ": " << Filename << " contains neither VPD nor a ROM image with a valid PCI Data Structure" << endl;
		    return 1;
		}
	    InputFile.seekg( PciDS.GetVPDOffset() );
	}

    CVitalProductData Vpd;

    InputFile >> Vpd;

    cout << Vpd.GetIdentifier() << endl;
    Vpd.ForEachKeyword( PrintKeywordPair );

    return 0;
}


static int Write( const CCmdLine& CommandLine )
{
    string Filename;
    CommandLine.Opt( 'w', Filename );
    ofstream OutputFile( Filename.c_str() );
    if ( !OutputFile )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Can't open " << Filename << endl;
	    return 1;
	}
    
    string Identifier;
    string Keyword;
    string Data;
    unsigned int iLastArg = CommandLine.GetNonOptArgCount();
    for ( unsigned int iArg = 1; iArg <= iLastArg; ++iArg )
	{
	    const string &Argument( CommandLine.Arg( iArg ) );
	    if ( !StringIsAscii( Argument ) )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Non-ASCII character in command line argument" << endl;
		    return 1;
		}
	    if ( !ParseEqualityLikeString( Argument, Keyword, Data ) )
		{
		    if ( iArg != 1 )
			{
			    cerr << CommandLine.Arg( 0 ) << ": Missing '=' in command line keyword '" << Argument << "'" << endl;
			    return 1;
			}
		    if ( Argument.length() > USHRT_MAX )
			{
			    cerr << CommandLine.Arg( 0 ) << ": Identifier string on command line too long" << endl;
			    return 1;
			}
		}
	    else
		{
		    if ( Keyword.length() != 2 )
			{
			    cerr << CommandLine.Arg( 0 ) << ": Invalid length keyword '" << Keyword << "' on command line " << endl;
			    return false;
			}
		    if ( Data.length() > UCHAR_MAX )
			{
			    cerr << CommandLine.Arg( 0 ) << ": Keyword data too long for keyword '" << Keyword << "' on command line " << endl;
			    return false;
			}
		}
	}

    CVitalProductData Vpd;

    if ( !CommandLine.Opt( 'n' ) )
	ReadTextVpd( cin, Vpd );

    for ( unsigned int iArg = 1; iArg <= iLastArg; ++iArg )
	{
	    const string &Argument( CommandLine.Arg( iArg ) );
	    if ( !ParseEqualityLikeString( Argument, Keyword, Data ) )
		Vpd.SetIdentifier( Argument );
	    else
		Vpd.SetKeyword( Keyword, Data );
	}

    if ( !Vpd.HasIdentifier() )
	{
	    cerr << CommandLine.Arg( 0 ) << ": VPD identifier string has not been defined" << endl;
	    return false;
	}

    OutputFile << Vpd;

    return 0;
}


int main( int argc, char *argv[] )
{
    CCmdLine CommandLine( argc, argv, CCmdLine::flagAllowArgOptGroup | CCmdLine::flagAllowOptArgNoSpace );

    if ( CommandLine.Parse( "r:w:Rn" ) != NULL )
	return 1;

    if ( !CommandLine.Opt( 'r' ) && !CommandLine.Opt( 'w' ) )
	{
	    cerr << CommandLine.Arg( 0 ) << ": Must specify -r or -w option" << endl;
	    return 1;
	}

    if ( CommandLine.Opt( 'r' ) )
	{
	    if ( CommandLine.Opt( 'w' ) )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Option -w not permitted in combination with option -r" << endl;
		    return 1;
		}
	    if ( CommandLine.Opt( 'n' ) )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Option -n not permitted in combination with option -r" << endl;
		    return 1;
		}

	    return Read( CommandLine );
	}

    if ( CommandLine.Opt( 'w' ) )
	{
	    if ( CommandLine.Opt( 'r' ) )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Option -r not permitted in combination with option -w" << endl;
		    return 1;
		}
	    if ( CommandLine.Opt( 'R' ) )
		{
		    cerr << CommandLine.Arg( 0 ) << ": Option -R not permitted in combination with option -w" << endl;
		    return 1;
		}

	    return Write( CommandLine );
	}
}
