/*
Hardware test of PCI-AER board.

boardtest.c $Id: 4d2ed566040b9ce15b5230a70b14a041cc55db02 $
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>

#include "pciaer.h"
#include "s5920.h"
#include "pciaerhw.h"

#define SRAM_ADDR_BITS	0x07FFFF
#define SRAM_BANK_BITS	0x180000
#define SRAM_BANK_SHIFT 19

typedef struct {
	char *pchRegName;
	unsigned int uMask;
	unsigned int uExpected;
} reg_read_test_t;

typedef struct {
	char *pchRegName;
	unsigned int uMask;
	char *pchAltRegName;
} reg_walk_test_t;

typedef struct {
	char *pchName;
	off_t offset;
	unsigned int uMask;
} sram_walk_test_t;

static reg_read_test_t S5920OpRegsRead[] = {
		{ "s5920/mbef",		S5920_MBEF_RESERVED, 	0 },
		{ "s5920/intcsr",	S5920_INTCSR_RESERVED,	0x00000C0C },
		{ "s5920/rcr",		S5920_RCR_RESERVED, 	0 }
};

static reg_read_test_t XilinxRegsRead[] = {
		{ "xilinx/rr1",		0x0000FFFF,				EXPECTED_FPGA1_RELEASE },
		{ "xilinx/rr2",		0x0000FFFF,				EXPECTED_FPGA2_RELEASE }
};

static reg_walk_test_t XilinxRegsWalk[] = {
		{ "xilinx/cra1", 0x0000FFFF, "xilinx/rr1" },
		{ "xilinx/crb1", 0x0000FFFF, "xilinx/rr1" },
		{ "xilinx/crc1", 0x0000FFFF, "xilinx/rr1" },
		{ "xilinx/cra2", 0x00FFFFFF, "xilinx/rr2" },
};

static int nLoopOnTest;
static int nTest;
static int nIteration;
static int nStartAtTest;

static int LoopOnTest()
{
    if ( nTest < nStartAtTest )
        return 0;

    if ( nIteration++ == 0 || nTest == nLoopOnTest )
    {
        printf( "TEST %d:\n", nTest );
        return 1;
    }
    else
        return 0;
}

static int NextTest()
{
    nIteration = 0;
    return ++nTest;
}

static int Error( unsigned int uLine, char *pchMessage )
{
    int e;

    e = errno;
    fprintf( stderr, "(%u):", uLine );
    perror( pchMessage );
    return e;
}

static void PrintVersion( int hNvr )
{
    pciaer_version_info_t v;

    printf( "Version info: " );

    if ( ioctl( hNvr, PCIAER_IOC_GET_VERSION_INFO, &v ) != 0 )
        exit( Error( __LINE__, "PCIAER_IOC_GET_VERSION_INFO" ) );

    printf( "{0x%04hX;0x%04hX;0x%04hX;0x%02X}\n",
	    v.driver_version,
	    v.fpga1_release,
	    v.fpga2_release,
	    (int)v.s5920_revision_id
	  );
}

static char *GetRegPath( char *pchRegName )
{
    static const char DebugFSPath[ ] = "/sys/kernel/debug/pciaer/pciaer0/";
    char *pchRegPath;

    pchRegPath = malloc( strlen( DebugFSPath ) + strlen( pchRegName ) + 1 );

    if ( pchRegPath == NULL )
        exit( Error( __LINE__, "malloc" ) );

    strcpy( pchRegPath, DebugFSPath );
    strcat( pchRegPath, pchRegName );
    return pchRegPath;
}

static unsigned int ReadReg( char *pchRegName )
{
    char *pchPath;
    FILE *file;
    char HexBuffer[ 12 ];

    pchPath = GetRegPath( pchRegName );
    file = fopen( pchPath, "r" );
    if ( file == NULL )
        exit( Error( __LINE__, "fopen" ) );

    free( pchPath );

    fgets( HexBuffer, sizeof(HexBuffer), file );

    fclose( file );

    return (unsigned int)strtoul( HexBuffer, NULL, 16 );
}

static void WriteReg( char *pchRegName, unsigned int uValue )
{
    char *pchPath;
    FILE *file;

    pchPath = GetRegPath( pchRegName );
    file = fopen( pchPath, "a" );
    if ( file == NULL )
        exit( Error( __LINE__, "fopen" ) );

    free( pchPath );

    fprintf( file, "%x", uValue );

    fclose( file );
}

static void ReadRegTest( reg_read_test_t *pTestList, int nToTest )
{
    int i;
    reg_read_test_t *pT;
    unsigned int uValue;

    for ( i = 0; i < nToTest; i++ )
    {
        NextTest();
        while ( LoopOnTest() )
        {
            pT = &pTestList[ i ];
            printf( "Reading %s...\n", pT->pchRegName );
            uValue = ReadReg( pT->pchRegName );
            printf( " read 0x%08X\t\t\t\t\t\t%s\n", uValue, ((uValue
                    & pT->uMask) == pT->uExpected) ? "OK" : "FAIL" );
        }
    }
}

static void WalkRegTest( int iZeroOrOne, reg_walk_test_t *pTest )
{
    int fail;
    unsigned int uWalkingOne;
    unsigned int uPattern;
    unsigned int uValueRead;

    NextTest();
    while ( LoopOnTest() )
    {
        printf( "Walking %d test of %s...\n", iZeroOrOne, pTest->pchRegName );
        fail = 0;
        for ( uWalkingOne = 1; (uWalkingOne & pTest->uMask) != 0; uWalkingOne
                <<= 1 )
        {
            uPattern = iZeroOrOne ? uWalkingOne : ~uWalkingOne;
            WriteReg( pTest->pchRegName, uPattern );
            WriteReg( pTest->pchAltRegName, ~uPattern );
            uValueRead = ReadReg( pTest->pchRegName );
            if ( (uValueRead & pTest->uMask) != (uPattern & pTest->uMask) )
            {
                fail = 1;
                printf( " wrote 0x%08X; read 0x%08X\n", uPattern, uValueRead );
            }
        }
        printf( " walking %d test of %s\t\t\t%s\n", iZeroOrOne,
                pTest->pchRegName, fail ? "FAIL" : "OK" );
    }
}

volatile unsigned int dummy;

static void WriteSRam( pciaer_mapper_sram_t *p, off_t RegOffset, unsigned int uValue )
{
    *(volatile unsigned int *)(((unsigned char *)p) + RegOffset) = uValue;
    dummy = *(volatile unsigned int *)p;
}

static unsigned int ReadSRam( pciaer_mapper_sram_t *p, off_t RegOffset )
{
    return *(volatile unsigned int *)(((unsigned char *)p) + RegOffset);
}

static void WalkSRamDataTest( pciaer_mapper_sram_t *pSRam, int iZeroOrOne, sram_walk_test_t *pTest )
{
    int fail;
    unsigned int uWalkingOne;
    unsigned int uPattern;
    unsigned int uValueRead;

    NextTest();
    while ( LoopOnTest() )
    {
        printf( "Walking %d test of %s...\n", iZeroOrOne, pTest->pchName );
        fail = 0;
        for ( uWalkingOne = 1; (uWalkingOne & pTest->uMask) != 0; uWalkingOne
                <<= 1 )
        {
            uPattern = iZeroOrOne ? uWalkingOne : ~uWalkingOne;
            WriteSRam( pSRam, pTest->offset, uPattern );
            WriteSRam( pSRam, (pTest->offset == 0) ? 0x3FFFC : 0, ~uPattern );
            uValueRead = ReadSRam( pSRam, pTest->offset );
            if ( (uValueRead & pTest->uMask) != (uPattern & pTest->uMask) )
            {
                fail = 1;
                printf( " wrote 0x%08X; read 0x%08X\n", uPattern, uValueRead );
            }
        }
        printf( " walking %d test of %s\t\t\t%s\n", iZeroOrOne, pTest->pchName,
                fail ? "FAIL" : "OK" );
    }

}

static void WalkSRamAddrTest( pciaer_mapper_sram_t *pSRam, int iBank, unsigned int uMask, int iZeroOrOne )
{
    int fail;
    unsigned int uWalkingOne;
    unsigned int uSRamWordAddress;
    unsigned int uPCByteAddress;
    unsigned int uPattern;
    unsigned int uValueRead;

    NextTest();
    uPattern = 0xAAAAAA;
    while( LoopOnTest() )
	{
	    printf( "Walking %d test of SRAM bank %d address bus...\n", iZeroOrOne, iBank );
	    fail = 0;
	    for ( uWalkingOne = 1; uWalkingOne < SRAM_ADDR_BITS + 1; uWalkingOne <<= 1 )
		{
		    uSRamWordAddress = ((iZeroOrOne ? uWalkingOne : ~uWalkingOne) & SRAM_ADDR_BITS);
		    uSRamWordAddress |= iBank << SRAM_BANK_SHIFT;
		    uPCByteAddress = uSRamWordAddress << 2;
		    WriteSRam( pSRam, uPCByteAddress, uPattern );
		    WriteSRam( pSRam, 0, ~uPattern );
		    uValueRead = ReadSRam( pSRam, uPCByteAddress );
		    if ( (uValueRead & uMask ) != (uPattern & uMask) )
			{
			    fail = 1;
			    printf( " wrote 0x%08X to SRAM bank %d address 0x%05X; read 0x%08X\n", uPattern, iBank, uSRamWordAddress & SRAM_ADDR_BITS, uValueRead );
			}
		}
	    printf( " walking %d test of SRAM bank %d address bus\t\t\t%s\n", iZeroOrOne, iBank, fail ? "FAIL" : "OK" );
	}
}

static void TestS5920OpRegs( void )
{
    reg_walk_test_t S5920OmbWalkTest = { "s5920/omb", ~0, "s5920/imb" };

    ReadRegTest( S5920OpRegsRead, sizeof(S5920OpRegsRead)
            / sizeof(reg_read_test_t) );

    WalkRegTest( 1, &S5920OmbWalkTest );
    WalkRegTest( 0, &S5920OmbWalkTest );
}

static void TestXilinxRegs( void )
{
    int i;

    ReadRegTest( XilinxRegsRead, sizeof(XilinxRegsRead)
            / sizeof(reg_read_test_t) );

    for ( i = 0; i < sizeof(XilinxRegsWalk) / sizeof(reg_walk_test_t); i++ )
    {
        WalkRegTest( 1, &XilinxRegsWalk[ i ] );
        WalkRegTest( 0, &XilinxRegsWalk[ i ] );
    }
}

static void TestSRam( void )
{
    int hMapper;
    pciaer_mapper_sram_t *p;

    hMapper = open( "/dev/aermap0", O_RDWR );
    if ( hMapper == -1 )
        exit( Error( __LINE__, "open /dev/aermap0" ) );

    // Driver only allows write access to Mapper SRAM iff module parameter
    // allow_mmap_write=1 is set, and then only if write access is acquired via
    // mprotect.

    p = mmap( 0, sizeof(pciaer_mapper_sram_t), PROT_NONE, MAP_SHARED, hMapper, 0 );
    if ( p == MAP_FAILED )
        exit( Error( __LINE__, "mmap") );

    if ( mprotect( p, sizeof(pciaer_mapper_sram_t), PROT_READ | PROT_WRITE ) != 0 )
    {
        if ( errno == EACCES )
        {
            fprintf( stderr, "mprotect failed to grant PROT_WRITE. "
                    "Load module with allow_mmap_write=1\n" );
            exit( EACCES );
        }
        else
            exit( Error( __LINE__, "mprotect") );
    }

    sram_walk_test_t SRamWalkTest[ 4 ] = {
		{ "SRAM bank 0 address 0", 0x000000 << 2, 0x00FFFFFF },
		{ "SRAM bank 1 address 0", 0x080000 << 2, 0x0000FFFF },
		{ "SRAM bank 2 address 0", 0x100000 << 2, 0x0000FFFF },
		{ "SRAM bank 3 address 0", 0x180000 << 2, 0x0000FFFF } };
    int iBank;

    WriteReg( "xilinx/cra2", CRA2_MAP_OUT_DISABLE );
    sleep( 1 );

    for ( iBank = 0; iBank <= 3; iBank++ )
    {
        WalkSRamDataTest( p, 1, &SRamWalkTest[ iBank ] );
        WalkSRamDataTest( p, 0, &SRamWalkTest[ iBank ] );
        WalkSRamAddrTest( p, iBank, SRamWalkTest[ iBank ].uMask, 1 );
        WalkSRamAddrTest( p, iBank, SRamWalkTest[ iBank ].uMask, 0 );
    }

    if ( munmap( p, sizeof(pciaer_mapper_sram_t) ) != 0 )
        exit( Error( __LINE__, "munmap" ) );

    close( hMapper );
}

static void TestFifos()
{
}

int main( int argc, char *argv[] )
{
    int hNvr;

    if ( argc >= 2 )
        nStartAtTest = atoi( argv[ 1 ] );
    else
        nStartAtTest = 0;

    if ( argc >= 3 )
        nLoopOnTest = atoi( argv[ 2 ] );
    else
        nLoopOnTest = -1;

    hNvr = open( "/dev/aernvr0", O_RDONLY );
    if ( hNvr == -1 )
        exit( Error( __LINE__, "open /dev/aernvr0" ) );

    nTest = -1;
    NextTest();
    while ( LoopOnTest() )
        PrintVersion( hNvr );

    close( hNvr );

    TestS5920OpRegs();
    TestXilinxRegs();
    TestSRam();
    TestFifos();

    return 0;
}
