/*
mapch2ch s d
Sets up a mapping from each possible AE on channel s to the same AE on channel d.
(The channels are numbered 0 to 3.)
The arbiter is set into 4 channel mode.
The mapper output is also set for 4 channel mode, SCX protocol.
*/


#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "pciaer.h"
#include "pciaerlib.h"


int main( int argc, char *argv[] )
{
  int src_channel = 1;
  int dst_channel = 0;
  int n_address_bits = 14;
  int arb_config = PCIAER_IOC_ARB_2_14;
  int demux_config = PCIAER_IOC_MAP_DEMUX_2_14;
  int protocol = PCIAER_IOC_MAP_AER_SCX;
  int map_channels;
  int error;
  int handle;
  pciaer_version_info_t v;
  unsigned int max_ae;
  unsigned int ae;
  unsigned short src_ae;
  unsigned short dst_ae;

  map_channels = 1 << src_channel;
  max_ae = (1 << n_address_bits) - 1;

  // Open Mapper

  error = PciaerMapOpen( 0, O_RDWR, &handle );
  if ( error != 0 )
      {
	  perror( "PciaerMapOpen" );
	  exit( 1 );
      }

  error = PciaerGetVersionInfo( handle, &v );
  if ( error != 0 )
      {
	  perror( "PciaerGetVersionInfo" );
	  exit( 1 );
      }
  printf( "Version info: {0x%04hX;0x%04hX;0x%04hX;0x%02X}\n",
	  v.driver_version,
	  v.fpga1_release,
	  v.fpga2_release,
	  (int)v.s5920_revision_id
	);

  // Turn off all input to mapper while we re-configure

  error = PciaerMapSetChannelSel( handle, 0 );
  if ( error != 0 )
      {
	  perror( "PciaerMapSetChannelSel" );
	  exit( 1 );
      }

  // Clear mapper contents

  printf( "Clearing Mapper...\n" );
  error = PciaerMapClearAllMappings( handle );
  if ( error != 0 )
      {
	  perror( "PciaerMapClearAllMappings" );
	  exit( 1 );
      }
  printf( "... done!\n" );

  // Set arbiter configuration

  error = PciaerSetArbConfig( handle, arb_config );
  if ( error != 0 )
      {
	  perror( "PciaerSetArbConfig" );
	  exit( 1 );
      }
  printf( "Arbiter config set to %d\n", arb_config );

  // Set mapper mode to 1-to-1

  error = PciaerMapSetOutputConfig( handle, PCIAER_IOC_MAP_OUT_1_TO_MANY );
  if ( error != 0 )
      {
	  perror( "PciaerMapSetOutputConfig" );
	  exit( 1 );
      }
  printf( "Mapper mode set to 1-to-1" );

  // Set mapper demux config

  error = PciaerMapSetDemuxConfig( handle, demux_config );
  if ( error != 0 )
    {
      perror( "PciaerMapSetDemuxConfig" );
      exit( 1 );
    }
  printf( "Demux config set to %d\n", demux_config );

  // Set mapper output protocol

  error = PciaerMapSetProtocol( handle, protocol );

  if ( ioctl( handle, PCIAER_IOC_GET_MAP_AER_PROTOCOL, &protocol ) != 0 )
    {
      perror( "PCIAER_IOC_GET_MAP_AER_PROTOCOL" );
      exit( 1 );
    }
  printf( "Protocol set to %d\n", protocol );

  // Establish mappings

  src_channel <<= n_address_bits;
  dst_channel <<= n_address_bits;
  for ( ae = 0; ae <= max_ae; ae++ )
      {
	  src_ae = ae | src_channel;
	  dst_ae = ae | dst_channel;
	  error = PciaerMapSetMapping( handle, src_ae, 1, &dst_ae );
	  if ( error != 0 )
	      {
		  perror( "PCIAER_IOC_SET_MAP_MAPPING" );
		  exit( 1 );
	      }
	  printf( "Mapping %04x->%04x\n", src_ae, dst_ae );
      }

  // Turn input on again for channel to be mapped
  
  error = PciaerMapSetChannelSel( handle, map_channels );
  if ( error != 0 )
      {
	  perror( "PciaerMapSetChannelSel" );
	  exit( 1 );
      }
  printf( "Mapper channel selection bitmask set to %x\n", map_channels );

  // All done

  PciaerMapClose( handle );
    
  return 0;
}
