/*
Dumps PCI-AER S5920 NVRAM bytes.

readnvram.c $Id: f7ae2dc07f87d10d5bb37b648b605b003a8da8c6 $
*/

#define _GNU_SOURCE

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/stat.h>

#include "pciaer.h"
#include "pciaerhw.h"


void Seek( int handle, off_t offset, int whence )
{
	int e;

	if ( lseek( handle, offset, whence ) == (off_t)-1 )
	{
		e = errno;
		perror( "lseek" );
		exit( e );
	}
}


unsigned char ReadByte( int handle, unsigned short uhOffset )
{
	unsigned char data;
	ssize_t read_return;
	int e;

	Seek( handle, uhOffset, SEEK_SET );
	read_return = read( handle, &data, 1 );
	if ( read_return == -1 )
	{
	    e = errno;
	    perror( "read" );
	    exit( e );
	}
	if ( read_return != 1 )
	{
	    fprintf( stderr, "read returned %zd\n", read_return );
	    exit( 101 );
	}

	return data;
}


void WriteByte( int handle, unsigned short uhOffset, unsigned char ucByte )
{
	int e;
	ssize_t write_return;

	Seek( handle, uhOffset, SEEK_SET );
	write_return = write( handle, &ucByte, 1 );
	if ( write_return == -1 )
	{
	    e = errno;
	    perror( "write" );
	    exit( e );
	}
	if ( write_return != 1 )
	{
	    fprintf( stderr, "write returned %zd\n", write_return );
	    exit( 102 );
	}
}


void PrintDriverVersion( int handle )
{
    pciaer_version_info_t v;
    int e;

    if ( ioctl( handle, PCIAER_IOC_GET_VERSION_INFO, &v ) != 0 )
	{
	    e = errno;
	    perror( "PCIAER_IOC_GET_VERSION_INFO" );
	    exit( e );
	}

    printf( "Driver version info: {0x%04hX;0x%04hX;0x%04hX;0x%02X}\n\n",
	    v.driver_version,
	    v.fpga1_release,
	    v.fpga2_release,
	    (int)v.s5920_revision_id
	  );
}


int GetCacheStatus( int handle )
{
	int status;
	int e;

	if ( ioctl( handle, PCIAER_IOC_GET_NVR_CACHE_STATUS, &status ) != 0 )
	{
		e = errno;
		perror( "PCIAER_IOC_GET_NVR_CACHE_STATUS" );
		exit( e );
	}
	return status;
}


void RevertCache( int handle )
{
	int e;

	if ( ioctl( handle, PCIAER_IOC_REVERT_NVR_CACHE ) != 0 )
	{
		e = errno;
		perror( "PCIAER_IOC_REVERT_NVR_CACHE" );
		exit( e );
	}
}


unsigned int GetNVRamSize( int handle )
{
    off_t InitialOffset;
    off_t NVRamSize;
    int e;
    
    InitialOffset = lseek( handle, 0, SEEK_CUR );
    if ( InitialOffset == (off_t)-1 )
	{
	    e = errno;
	    perror( "lseek SEEK_CUR" );
	    exit( e );
	}

    NVRamSize = lseek( handle, 0, SEEK_END );
    if ( NVRamSize == (off_t)-1 )
	{
	    e = errno;
	    perror( "lseek SEEK_END" );
	    exit( e );
	}

    Seek( handle, InitialOffset, SEEK_SET );
    
    return (unsigned int)NVRamSize;
}


unsigned long GetPreferredIOBlockSize( handle )
{
  struct stat stat_buffer;
  if ( fstat( handle, &stat_buffer ) != 0 )
      {
	  perror( "fstat" );
	  exit( errno );
      }
  return stat_buffer.st_blksize;
}


void TestDirtyRevertStatus( int handle, unsigned char *pBuffer, unsigned int uNVRamSize )
{
	static const unsigned short uhTestLoc = 7;
	unsigned char ucTestByte;
	int fail = 0;

	if ( GetCacheStatus( handle ) != 0 )
	{
		fail = 1;
		puts( "Cache not clean before write!" );
	}

	ucTestByte = ReadByte( handle, uhTestLoc );

	if ( GetCacheStatus( handle ) != 0 )
	{
		fail = 1;
		puts( "Cache not clean before write!" );
	}

	WriteByte( handle, uhTestLoc, ~ucTestByte );

	if ( GetCacheStatus( handle ) != PCIAER_IOC_NVR_CACHE_DIRTY )
	{
		fail = 1;
		puts( "Cache not dirty after write!" );
	}

	RevertCache( handle );

	if ( GetCacheStatus( handle ) != 0 )
	{
		fail = 1;
		puts( "Cache not clean after revert!" );
	}

	if ( ReadByte( handle, uhTestLoc ) != ucTestByte )
	{
		fail = 1;
		puts( "Cache not reverted to initial value!" );
	}

	if ( !fail )
		puts( "Passed cache tests." );
}


int main( int argc, char *argv[] )
{
  int flags;
  int handle;
  int e;
  unsigned int uNVRamSize;
  unsigned long ulPrefIoBlkSize;
  unsigned char *pBuffer;
  int nReadReturn;
  int nBytesRead;
  unsigned short uhAddress;

  flags = (geteuid() == 0) ? O_RDWR : O_RDONLY;
  if (argc == 2 && strcmp( argv[1], "-s" ) == 0)
  {
	flags |= O_SYNC;
	puts( "Using O_SYNC\n" );
  }

  // Open the NVRAM

  handle = open( "/dev/aernvr0", flags );
  if ( handle == -1 )
    {
      e = errno;
      perror( "open" );
      exit ( e );
    }

  PrintDriverVersion( handle );

  uNVRamSize = GetNVRamSize( handle );
  ulPrefIoBlkSize = GetPreferredIOBlockSize( handle );

  printf( "NVRAM size: %u\n", uNVRamSize );
  printf( "Preferred IO block size: %lu\n\n", ulPrefIoBlkSize );

  pBuffer = malloc( uNVRamSize );

  for ( nBytesRead = 0; nBytesRead != uNVRamSize; nBytesRead += nReadReturn )
  {
	nReadReturn = read( handle, pBuffer + nBytesRead, uNVRamSize - nBytesRead );
	if ( nReadReturn <= 0 )
		break;
  }
  if ( nReadReturn < 0 )
      {
	  e = errno;
	  perror( "read" );
	  fprintf( stderr, "read returned %d\n", nReadReturn );
	  exit( e );
      }
  
  for ( uhAddress = 0; uhAddress < uNVRamSize; uhAddress++ )
      printf( "%03Xh : %02Xh\n", uhAddress, pBuffer[ uhAddress ] );

  if ( (flags & (O_RDONLY|O_WRONLY|O_RDWR)) == O_RDWR )
	TestDirtyRevertStatus( handle, pBuffer, uNVRamSize );

  free( pBuffer );
  close( handle );
  return 0;
}
