/*
Stress test code for Mapper memory management part of PCI-AER driver, pciaer.o

stressmap.c $Id: de5a1ab89db35fde7a4a8e87d2af40ea7784ff3c $
*/

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include "pciaer.h"

#ifndef max
#define max(A,B) ((A)>(B)?(A):(B))
#endif

#ifdef NOBEEP
#define BEEP ""
#else
#define BEEP "\a"
#endif

#define MAX_DEST_AE 128

#define DO_CLEAR
#define DO_WRITE
#define READ_BY_IOCTL
//#define READ_BY_MMAP
//#define PRINT_STATS
//#define PRINT_IOCTL_LETTERS
//#define DUMP_EVERY_WRITE

#if defined(READ_BY_IOCTL) && defined(READ_BY_MMAP)
#error READ_BY_IOCTL and READ_BY_MMAP are contradictory
#endif


typedef struct { unsigned long ulReads; unsigned long ulErrors; } stats_t;


static int signalled = 0;

void signal_handler( const int signum )
{
  signalled = 1;
}


#if defined(DO_CLEAR)
void Clear( const int handle )
{
  printf( "Clear\n" );
  if ( ioctl( handle, PCIAER_IOC_MAP_CLEAR ) != 0 )
    {
      perror( "PCIAER_IOC_MAP_CLEAR" );
      exit( 1 );
    }
}

#if defined(READ_BY_IOCTL) || defined(READ_BY_MMAP)
void TestClear( const int handle, const volatile unsigned int *p )
{
  unsigned short i;
  int fError;
#if defined(READ_BY_IOCTL)
  pciaer_mapping_t Mapping;
#endif
#if defined(READ_BY_MMAP)
  unsigned int uPointerTableEntry;
#endif

  for ( i = 0; i < MAPPER_END_LABEL; i++ )
    {
	fError = 0;
#if defined(READ_BY_IOCTL)
	Mapping.source_ae = i;
	if ( ioctl( handle, PCIAER_IOC_GET_MAP_MAPPING_COUNT, &Mapping ) != 0 )
	    {
		perror( "PCIAER_IOC_GET_MAP_MAPPING_COUNT" );
		break;
	    }
	if ( Mapping.dest_count != 0 )
	    fError = 1;
#endif
#if defined(READ_BY_MMAP)
	uPointerTableEntry = p[ i ] & 0x00ffffff;
	if ( uPointerTableEntry != MAPPER_PTR_TYPE_NOT_USED_1 && uPointerTableEntry != MAPPER_PTR_TYPE_NOT_USED_2 )
	    fError = 1;
#endif
	if ( fError != 0 )
	    printf( BEEP "Pointer table entry at 0x%04X is not null\n", i );
    }
}
#endif
#endif


void PrepareRandomMapping( pciaer_mapping_t *pMapping )
{
    static unsigned short serial = 0;

    int i;

    pMapping->source_ae = (int)(((float)rand() * (MAPPER_END_LABEL - 1)) / RAND_MAX);
    pMapping->dest_count = (int)(((float)rand() * MAX_DEST_AE) / RAND_MAX);
    for ( i = 0; i < pMapping->dest_count; i++ )
	{
	    pMapping->dest_ae[ i ] = serial++;
	    if ( serial == MAPPER_END_LABEL )
		serial = 0;
	}
}


void DumpMapping( FILE *pFile, const pciaer_mapping_t *pMapping )
{
  int i;

  fprintf( pFile, "0x%04hX[%hu]:", pMapping->source_ae, pMapping->dest_count );
  if ( pMapping->dest_count == 0 )
    fprintf( pFile, "NULL" );
  else
    {
      for ( i = 0; i < pMapping->dest_count; i++ )
	fprintf( pFile, "0x%04hX;", pMapping->dest_ae[ i ] );
      fprintf( pFile, "\b" ); // \b should remove trailing semicolon
    }
}


#if defined(READ_BY_MMAP)
void ReportDirectReadMapError( const int iLine, const pciaer_mapping_t *pMapping, const volatile unsigned int *p )
{
  int i;
  unsigned int uPointerTableEntry;
  char chType;

  printf( BEEP "\nLine %d: Written ", iLine );
  DumpMapping( stdout, pMapping );
  printf( "\n" );

  uPointerTableEntry = p[ pMapping->source_ae ];

  switch ( uPointerTableEntry  & MAPPER_PTR_TYPE_MASK )
    {
    case MAPPER_PTR_TYPE_INDIRECT:
      chType = 'I';
      break;
    case MAPPER_PTR_TYPE_NOT_USED_1:
      chType = '1';
      break;
    case MAPPER_PTR_TYPE_NOT_USED_2:
      chType = '2';
      break;
    case MAPPER_PTR_TYPE_DIRECT:
      chType = 'D';
      break;
    }
  printf( "Read [0x%04X]:0x%08X %c ", pMapping->source_ae, uPointerTableEntry, chType );
/*   if ( chType == 'D' ) */
/*     printf( " 0x%04X", uPointerTableEntry & MAPPER_PTR_DIRECT_MASK ); */
  if ( chType == 'I' )
    {
      i = uPointerTableEntry & MAPPER_PTR_INDIRECT_MASK;
      do
	{
	  uPointerTableEntry = p[ i++ ];
	  printf( "0x%08X;", uPointerTableEntry );
	}
      while ( (uPointerTableEntry & 0xffff) != MAPPER_END_LABEL );
    }
  printf( "\n" );
}


int VerifyMappingByDirectRead( const pciaer_mapping_t *pMappingWritten, const volatile unsigned int *pMappedMemory, stats_t *pStats )
{
    unsigned int uPointerTableEntry;
    const volatile unsigned int *pList;
    unsigned short i;

    uPointerTableEntry = pMappedMemory[ pMappingWritten->source_ae ];

    switch ( pMappingWritten->dest_count )
	{
	case 0:
	    if ( (uPointerTableEntry & MAPPER_PTR_TYPE_MASK) != MAPPER_PTR_TYPE_NOT_USED_1
		 && (uPointerTableEntry & MAPPER_PTR_TYPE_MASK) != MAPPER_PTR_TYPE_NOT_USED_2 )
		{
		    ReportDirectReadMapError( __LINE__, pMappingWritten, pMappedMemory );
		    pStats->ulErrors++;
		}
	    break;
	case 1:
	    if ( (uPointerTableEntry & MAPPER_PTR_TYPE_MASK) != MAPPER_PTR_TYPE_DIRECT
		 || (uPointerTableEntry & MAPPER_PTR_DIRECT_MASK) != pMappingWritten->dest_ae[ 0 ] )
		{
		    ReportDirectReadMapError( __LINE__, pMappingWritten, pMappedMemory );
		    pStats->ulErrors++;
		}
	    break;
	default:
	    if ( (uPointerTableEntry & MAPPER_PTR_TYPE_MASK) != MAPPER_PTR_TYPE_INDIRECT )
		{
		    ReportDirectReadMapError( __LINE__, pMappingWritten, pMappedMemory );
		    pStats->ulErrors++;
		}
	    pMappingWritten->dest_ae[ pMappingWritten->dest_count ] = MAPPER_END_LABEL;
	    pList = pMappedMemory + (uPointerTableEntry & MAPPER_PTR_INDIRECT_MASK);
	    for ( i = 0; i <= pMappingWritten->dest_count; i++ )
		{
		    uPointerTableEntry = pList[ i ];
		    pStats->ulReads++;
		    if ( (uPointerTableEntry & 0xffff) != pMappingWritten->dest_ae[ i ] )
			{
			    ReportDirectReadMapError( __LINE__, pMappingWritten, pMappedMemory );
			    pStats->ulErrors++;
			    printf( "Failing SRAM location = 0x%06X\n", (&pList[ i ] - pMappedMemory) >> 2 );
			    // Try re-reading
			    uPointerTableEntry = pList[ i ];
			    printf( "Re-read 0x%08X;", uPointerTableEntry );
			    uPointerTableEntry = pList[ i + 1 ];
			    printf( "0x%08X\n", uPointerTableEntry );
			    uPointerTableEntry = pList[ i ];
			    printf( "Re-read 0x%08X;", uPointerTableEntry );
			    uPointerTableEntry = pList[ i + 1 ];
			    printf( "0x%08X\n", uPointerTableEntry );
			    break;
			}
		}
	    break;
	}
    return 0;
}
#endif


#if defined(READ_BY_IOCTL)
int VerifyMappingByIoctl( const pciaer_mapping_t *pMappingWritten, const int handle, stats_t *pStats )
{
    pciaer_mapping_t MappingRead;
    int nErr = 0;

    MappingRead.source_ae = pMappingWritten->source_ae;
#ifdef PRINT_IOCTL_LETTERS
    fprintf( stderr, "c" );
#endif
    if ( ioctl( handle, PCIAER_IOC_GET_MAP_MAPPING_COUNT, &MappingRead ) != 0 )
	{
	    nErr = errno;
	    perror( "PCIAER_IOC_GET_MAP_MAPPING_COUNT" );
	    return nErr;
	}
    if ( MappingRead.dest_count != pMappingWritten->dest_count )
	{
	    printf( BEEP "\nCount written: %u; Count read: %u\n", pMappingWritten->dest_count, MappingRead.dest_count );
	    pStats->ulErrors++;
	}
    if ( MappingRead.dest_count != 0 )	// was MappingRead.dest_count
	{
	    MappingRead.dest_count = max(pMappingWritten->dest_count,MappingRead.dest_count);
	    MappingRead.dest_ae = malloc( MappingRead.dest_count * sizeof( short ) );
#ifdef PRINT_IOCTL_LETTERS
	    fprintf( stderr, "g" );
#endif
	    if ( ioctl( handle, PCIAER_IOC_GET_MAP_MAPPING, &MappingRead ) != 0 )
		{
		    nErr = errno;
		    perror( "PCIAER_IOC_GET_MAP_MAPPING" );
		    DumpMapping( stdout, &MappingRead );
		    //		    free( MappingRead.dest_ae );
		    //		    return nErr;
		}
	    else
		{
		    pStats->ulReads += MappingRead.dest_count;
		    if ( memcmp( pMappingWritten->dest_ae, MappingRead.dest_ae, MappingRead.dest_count * sizeof( short ) ) != 0 )
			{
			    printf( BEEP "\nWritten " );
			    DumpMapping( stdout, pMappingWritten );
			    printf(      "\nRead    " );
			    DumpMapping( stdout, &MappingRead );
			    printf( "\n" );
			    pStats->ulErrors++;
			}
		}
	    free( MappingRead.dest_ae );
	}

    return 0;
}
#endif


int main( int argc, char *argv[] )
{
  int handle;
  pciaer_version_info_t v;
  int state;
  pciaer_mapping_t mappingW;
  volatile unsigned int *pMappedMemory = NULL;
  int nErr;
  unsigned short DestAEListW[ MAX_DEST_AE + 1 ];
  stats_t Stats;
  unsigned int uFreeSpace;

  handle = open( "/dev/aermap0", O_RDWR );
  if ( handle == -1 )
    {
      perror( "open" );
      exit( 1 );
    }

  if ( ioctl( handle, PCIAER_IOC_GET_VERSION_INFO, &v ) != 0 )
    {
      perror( "PCIAER_IOC_GET_VERSION_INFO" );
      exit( 1 );
    }
  printf( "Version info: {0x%04hX;0x%04hX;0x%04hX;0x%02X}\n",
	  v.driver_version,
	  v.fpga1_release,
	  v.fpga2_release,
	  (int)v.s5920_revision_id
	);

#if defined(READ_BY_MMAP)
  printf( "Disabling Mapper out\n" );
  if ( ioctl( handle, PCIAER_IOC_DISABLE_MAP_OUT, 1 ) != 0 )
    {
      perror( "PCIAER_IOC_DISABLE_MAP_OUT" );
      exit( 1 );
    }
#endif

  if ( ioctl( handle, PCIAER_IOC_GET_MAP_OUT_STATE, &state ) != 0 )
    {
      perror( "PCIAER_IOC_GET_MAP_OUT_STATE" );
      exit( 1 );
    }
  printf( "Mapper out state now 0x%08X\n", state );
  
#if defined(READ_BY_MMAP)
  pMappedMemory = mmap( 0, sizeof(pciaer_mapper_sram_t), PROT_READ, MAP_SHARED, handle, 0 );
  if ( pMappedMemory == MAP_FAILED )
    {
      perror( "mmap" );
      exit( 1 );
    }

  // If we wanted to write directly to the SRAM as well as read from it, we'd
  // need to do the following:

/*   if ( mprotect( p, sizeof(pciaer_mapper_sram_t), PROT_READ | PROT_WRITE ) != 0 ) */
/*     { */
/*       perror( "mprotect" ); */
/*       exit( 1 ); */
/*     } */
#endif

  printf( "Testing using %s read method.\n",
	  pMappedMemory != NULL ? "direct" : "ioctl" );
  
  srand( getpid() );

  mappingW.dest_ae = DestAEListW;

  Stats.ulReads = 0L;
  Stats.ulErrors = 0L;

  while ( !signalled )
    {
#if defined(DO_CLEAR)
      Clear( handle );
      printf( "Done Clear\n" );
#if defined(READ_BY_IOCTL) || defined(READ_BY_MMAP)
      TestClear( handle, pMappedMemory );
      printf( "Done TestClear\n" );
#endif
#endif

      nErr = 0;
#if defined(DO_WRITE)
      while ( !signalled )
	{
	  PrepareRandomMapping( &mappingW );

#ifdef DUMP_EVERY_WRITE
	    DumpMapping( stderr, &mappingW );
	    fprintf( stderr, "\n" );
#endif

	  /* Write */

#ifdef PRINT_IOCTL_LETTERS
	  fprintf( stderr, "w" );
#endif
	  if ( ioctl( handle, PCIAER_IOC_SET_MAP_MAPPING, &mappingW ) != 0 )
	    {
	      nErr = errno;
	      perror( "PCIAER_IOC_SET_MAP_MAPPING" );
	      break;
	    }

	  /* Read back */

#if defined(READ_BY_IOCTL) || defined(READ_BY_MMAP)
#if defined(READ_BY_IOCTL)
	  nErr = VerifyMappingByIoctl( &mappingW, handle, &Stats );
#endif
#if defined(READ_BY_MMAP)
	  nErr = VerifyMappingByDirectRead( &mappingW, pMappedMemory, &Stats );
#endif
	  if ( nErr != 0 )
	      break;

#ifdef PRINT_STATS
	  printf( "%f%% failures (%lu/%lu)\n", (float)ulErrors/(float)ulReads*100.0, ulErrors, ulReads );
#endif
#endif
	}
      if ( !signalled && nErr != ENOSPC )
        exit( nErr );
      if ( nErr == ENOSPC )
	  {
	      if ( ioctl( handle, PCIAER_IOC_GET_MAP_FREE_SPACE, &uFreeSpace ) != 0 )
		  {
		      nErr = errno;
		      perror( "PCIAER_IOC_GET_MAP_FREE_SPACE" );
		      exit( nErr );
		  }
	      printf( "%u words free\n", uFreeSpace );
	  }
#endif
    }

#if defined(READ_BY_MMAP)
  if ( pMappedMemory != 0 && munmap( (void *)pMappedMemory, sizeof(pciaer_mapper_sram_t) ) != 0 )
    {
      perror( "munmap" );
      exit( 1 );
    }
#endif
  
  close( handle );
    
  return 0;
}
