/*
Test code for Monitor part of PCI-AER driver.

testmon.cc $Id: 7314c4807a53b67edfe1c3e7f6a36d98a6bc225a $
*/

#include <iostream>
#include <csignal>
#include <cstdio>
#include <cstdlib>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include "pciaer.h"
#include "../s5920nvram-support/cmdline.h"
#include "PciaerStats.h"
#include "common.h"

#define CHANNEL_BITMASK 0xF
#define ARB_CONFIG	PCIAER_IOC_ARB_0_16

int main( int argc, char *argv[] )
{
  CCmdLine command_line( argc, argv, CCmdLine::flagAllowArgOptGroup | CCmdLine::flagAllowOptArgNoSpace );
  int handle;
  int error;
  unsigned int fifo_depth;
  int channels;
  int time_lbl_flag;
  int arb_config;
  int fifo_flags;
  unsigned int *pBuffer;
  int read_return;
  unsigned int i;

  if ( command_line.Parse( "c:m:qtF" ) != NULL )
      exit( EINVAL );

  int arbiter_config = GetNumOpt( command_line, 'c', ARB_CONFIG );
  switch ( arbiter_config )
      {
      case 1:
	  arbiter_config = PCIAER_IOC_ARB_0_16;
	  break;
      case 2:
	  arbiter_config = PCIAER_IOC_ARB_1_15;
	  break;
      case 4:
	  arbiter_config = PCIAER_IOC_ARB_2_14;
	  break;
      default:
	  fprintf( stderr, "Invalid arbiter configuration option\n" );
	  exit( EINVAL );
	  break;
      }

  const unsigned int channel_mask = GetNumOpt( command_line, 'm', CHANNEL_BITMASK );
  if ( channel_mask > 0xF )
      {
	  fprintf( stderr, "Invalid channel mask option\n" );
	  exit( EINVAL );
      }

  const bool quiet = command_line.Opt( 'q' );
  const bool with_time_labels = !command_line.Opt( 't' );
  const bool test_fionread = command_line.Opt( 'F' );

  signal( SIGINT, signal_handler );

  handle = open( "/dev/aermon0", O_RDWR );
  if ( handle == -1 )
    {
      error = errno;
      perror( "open" );
      exit( error );
    }

  print_version( handle );

  test_lseek( handle );

  TestCounter( handle );

  fifo_depth = test_fifo_depth( handle );

  pBuffer = static_cast<unsigned int *>(malloc( fifo_depth * sizeof(unsigned int) ));

  if ( ioctl( handle, PCIAER_IOC_GET_MON_CH_SEL, &channels ) != 0 )
    {
      error = errno;
      perror( "PCIAER_IOC_GET_MON_CH_SEL" );
      exit( error );
    }
  printf( "Channel selection bitmask was 0x%08X\n", channels );
  if ( ioctl( handle, PCIAER_IOC_SET_MON_CH_SEL, channel_mask ) != 0 )
    {
      error = errno;
      perror( "PCIAER_IOC_SET_MON_CH_SEL" );
      exit( error );
    }
  if ( ioctl( handle, PCIAER_IOC_GET_MON_CH_SEL, &channels ) != 0 )
    {
      error = errno;
      perror( "PCIAER_IOC_GET_MON_CH_SEL" );
      exit( error );
    }
  printf( "Channel selection bitmask is now 0x%08X\n", channels );

  if ( ioctl( handle, PCIAER_IOC_GET_MON_TIME_LBL_FLAG, &time_lbl_flag ) != 0 )
      {
	  error = errno;
	  perror( "PCIAER_IOC_GET_MON_TIME_LBL_FLAG" );
	  exit( error );
      }
  printf( "Time labels were %s\n", time_lbl_flag ? "enabled" : "disabled" );
  if ( ioctl( handle, PCIAER_IOC_SET_MON_TIME_LBL_FLAG, with_time_labels ) != 0 )
      {
	  error = errno;
	  perror( "PCIAER_IOC_SET_MON_TIME_LBL_FLAG" );
	  exit( error );
      }
  if ( ioctl( handle, PCIAER_IOC_GET_MON_TIME_LBL_FLAG, &time_lbl_flag ) != 0 )
      {
	  error = errno;
	  perror( "PCIAER_IOC_GET_MON_TIME_LBL_FLAG" );
	  exit( error );
      }
  printf( "Time labels are now %s\n", time_lbl_flag ? "enabled" : "disabled" );

  if ( ioctl( handle, PCIAER_IOC_GET_ARB_CONFIG, &arb_config ) != 0 )
      {
	  error = errno;
	  perror( "PCIAER_IOC_GET_ARB_CONFIG" );
	  exit( error );
      }
  printf( "Arb config was %d\n", arb_config );
  if ( ioctl( handle, PCIAER_IOC_SET_ARB_CONFIG, arbiter_config ) != 0 )
      {
	  error = errno;
	  perror( "PCIAER_IOC_SET_ARB_CONFIG" );
	  exit( error );
      }
  if ( ioctl( handle, PCIAER_IOC_GET_ARB_CONFIG, &arb_config ) != 0 )
      {
	  error = errno;
	  perror( "PCIAER_IOC_GET_ARB_CONFIG" );
	  exit( error );
      }
  printf( "Arb config is now %d\n", arb_config );

  if ( ioctl( handle, PCIAER_IOC_RST_FIFO ) != 0 ) // Reset FIFO
    {
      error = errno;
      perror( "PCIAER_IOC_RST_FIFO" );
      exit( error );
    }

  if ( ioctl( handle, PCIAER_IOC_GET_MON_FIFO_FLAGS, &fifo_flags ) != 0 )
    {
      error = errno;
      perror( "PCIAER_IOC_GET_MON_FIFO_FLAGS" );
      exit( error );
    }
  printf( "Fifo flags = 0x%08X\n", fifo_flags );

  PciaerStats stats( handle );
  error = stats.Reset();
  if ( error != 0 )
      {
	  perror( "PciaerResetStatistics" );
	  exit( error );
      }

  printf( "Running...\n" );

  int fionread_bytes;
  unsigned long bytes_read = 0;
  while ( !is_signalled() )
    {
      if ( test_fionread )
	  {
	      if ( ioctl( handle, FIONREAD, &fionread_bytes ) != 0 )
		  {
		      error = errno;
		      perror( "FIONREAD" );
		      exit( error );
		  }
	      if ( !quiet )
		  printf( "FIONREAD: %d words available\n", fionread_bytes / (int)sizeof(unsigned int) );
	  }
      read_return = read( handle, pBuffer, fifo_depth * sizeof(unsigned int) );
      if ( !quiet )
	  {
	      //      printf( "Read returned %d\n", read_return );
	      for ( i = 0; i < read_return / sizeof(unsigned int); i++ )
		  printf( "%08X ", pBuffer[ i ] );
	      if ( read_return > 0 )
		  printf( "\n" );
	  }
      if ( read_return == -1 )
	  perror( "read" );
      else
	  bytes_read += read_return;
    }

  printf( "Stopped\nRead %lu words\n", bytes_read / sizeof(unsigned int) );

  free( pBuffer );

  error = stats.Refresh();
  if ( error != 0 )
      {
	  perror( "PciaerGetStatistics" );
	  exit( error );
      }

  std::cout << stats;

  close( handle );

  return 0;
}
