/*
Test code for Sequencer part of PCI-AER driver.

testseq.cc $Id: 30f542e77c86f44fe2f527ff2b7ab491971cf958 $
*/

#include <iostream>
#include <csignal>
#include <cstdio>
#include <cstdlib>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include "pciaer.h"
#include "../s5920nvram-support/cmdline.h"
#include "PciaerStats.h"
#include "common.h"

#define MON_FIFO_FLAGS	(PCIAER_IOC_MON_EMPTY | PCIAER_IOC_MON_HALF_FULL | PCIAER_IOC_MON_FULL)
#define SEQ_FIFO_FLAGS	(PCIAER_IOC_SEQ_EMPTY | PCIAER_IOC_SEQ_HALF_FULL | PCIAER_IOC_SEQ_FULL)
#define MAP_FIFO_FLAGS	(PCIAER_IOC_MAP_EMPTY | PCIAER_IOC_MAP_FULL)
#define FIFO_FLAGS_MASK (MON_FIFO_FLAGS | SEQ_FIFO_FLAGS | MAP_FIFO_FLAGS)

#define BUFFER_SIZE	(1 << 14)
#define ISI_MICROSECONDS       2

#define ARBITER_CHANNEL        0

#define DEMUX_CONFIG	       PCIAER_IOC_MAP_DEMUX_0_16

typedef struct {
  unsigned int ae;
  unsigned int isi;
} event;

void PrintFifoFlags( int handle )
{
  int fifo_flags;
  int error;

  if ( ioctl( handle, PCIAER_IOC_GET_SEQ_FIFO_FLAGS, &fifo_flags ) != 0 )
    {
      error = errno;
      perror( "PCIAER_IOC_GET_SEQ_FIFO_FLAGS" );
      exit( error );
    }
  fifo_flags &= FIFO_FLAGS_MASK;
  printf( "+ve logic fifo flags = octal %03o\n", fifo_flags );
}

int main( int argc, char *argv[] )
{
  CCmdLine command_line( argc, argv, CCmdLine::flagAllowArgOptGroup | CCmdLine::flagAllowOptArgNoSpace );
  int handle;
  int error;
  int map_handle;
  int current_arb_channel;
  event *pEventBuffer;
  int write_return;
  int i;
  unsigned long bytes_written;

  if ( command_line.Parse( "a:d:i:o:q" ) != NULL )
      exit( EINVAL );

  const int arb_channel = GetNumOpt( command_line, 'a', ARBITER_CHANNEL );
  const int demux = GetNumOpt( command_line, 'd', DEMUX_CONFIG );
  const int isi_us = GetNumOpt( command_line, 'i', ISI_MICROSECONDS );
  const int op_channel = GetNumOpt( command_line, 'o', 0 );
  const bool quiet = command_line.Opt( 'q' );


  signal( SIGINT, signal_handler );


  handle = open( "/dev/aerseq0", O_RDWR );
  if ( handle == -1 )
    {
      error = errno;
      perror( "open /dev/aerseq0" );
      exit( error );
    }

  print_version( handle );

  test_lseek( handle );

  //Open the Mapper
  map_handle = open( "/dev/aermap0", O_RDWR );
  if ( map_handle == -1 )
	{
	  error = errno;
	  perror( "open /dev/aermap0" );
	  exit( error );
	}

  if ( ioctl( map_handle, PCIAER_IOC_SET_ARB_CONFIG, PCIAER_IOC_ARB_0_16 ) != 0 )
	{
	  error = errno;
	  perror( "PCIAER_IOC_SET_ARB_CONFIG" );
	  exit( error );
	}

  if ( ioctl( map_handle, PCIAER_IOC_SET_MAP_CH_SEL, 0xF ) != 0 )
	{
	  error = errno;
	  perror( "PCIAER_IOC_SET_MAP_CH_SEL" );
	  exit( error );
	}

  if ( ioctl( map_handle, PCIAER_IOC_SET_MAP_OUT_CONFIG, PCIAER_IOC_MAP_OUT_PASS_THRU ) != 0 )
	{
	  error = errno;
	  perror( "PCIAER_IOC_SET_MAP_OUT_CONFIG" );
	  exit( error );
	}

  if ( ioctl( map_handle, PCIAER_IOC_SET_MAP_DEMUX_CONFIG, demux ) != 0 )
	{
	  error = errno;
	  perror( "PCIAER_IOC_SET_MAP_DEMUX_CONFIG" );
	  exit( error );
	}
  printf( "Demux configured for %d channels\n", demux );

  if ( ioctl( map_handle, PCIAER_IOC_SET_MAP_AER_PROTOCOL, PCIAER_IOC_MAP_AER_SCX ) != 0 )
	{
	  error = errno;
	  perror( "PCIAER_IOC_SET_MAP_AER_PROTOCOL" );
	  exit( error );
	}

  close( map_handle );

  TestCounter( handle );

  test_fifo_depth( handle );

  if ( ioctl( handle, PCIAER_IOC_GET_SEQ_ARB_CHANNEL, &current_arb_channel ) != 0 )
    {
      error = errno;
      perror( "PCIAER_IOC_GET_SEQ_ARB_CHANNEL" );
      exit( error );
    }

  printf( "Arb channel was %d, ", current_arb_channel );
  if ( ioctl( handle, PCIAER_IOC_SET_SEQ_ARB_CHANNEL, arb_channel ) != 0 )
    {
      error = errno;
      perror( "PCIAER_IOC_SET_SEQ_ARB_CHANNEL" );
      exit( error );
    }
  if ( ioctl( handle, PCIAER_IOC_GET_SEQ_ARB_CHANNEL, &current_arb_channel ) != 0 )
    {
      error = errno;
      perror( "PCIAER_IOC_GET_SEQ_ARB_CHANNEL" );
      exit( error );
    }
  printf( "now %d\n", current_arb_channel );

  PrintFifoFlags( handle );

  PciaerStats stats( handle );
  error = stats.Reset();
  if ( error != 0 )
      {
	  perror( "PciaerResetStatistics" );
	  exit( error );
      }

  pEventBuffer = (event *)malloc( BUFFER_SIZE * sizeof(event) );
  
  for ( i = 0; i < BUFFER_SIZE; i++ )
    {
      pEventBuffer[ i ].ae = (op_channel << 14) | i | SEQUENCER_DWORD_AER_ADDR;
      pEventBuffer[ i ].isi = isi_us | SEQUENCER_DWORD_DELAY;
    }

  printf( "ISI = %dus\nRunning...\n", isi_us );

  i = 0;
  bytes_written = 0;
  while ( !is_signalled() )
    {
      write_return = write( handle, (char *)pEventBuffer + i, BUFFER_SIZE * sizeof(event) - i );
      if ( !quiet )
	  printf( "write returned %d\n", write_return );
      if ( write_return < 0 )
	  perror( "write" );
      else
	  {
	      i = (i + write_return) % (BUFFER_SIZE * sizeof(event));
	      bytes_written += write_return;
	  }
      if ( !quiet )
	  PrintFifoFlags( handle );
    }

  free( pEventBuffer );

  printf( "Wrote %lu events\n", bytes_written / sizeof(event) );

  fsync( handle );

  error = stats.Refresh();
  if ( error != 0 )
      {
	  perror( "PciaerGetStatistics" );
	  exit( error );
      }

  std::cout << stats;

  close( handle );

  return 0;
}
